! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

! Non-method subroutines contained:
!  subroutine extractRawXYBounds(rawxy_str, rawxy_sel, status)
!  subroutine validateEnergyInt(emin, emax, status)
!  subroutine interpret_str_gti(str, gti, status)
!  subroutine apply_filter_eventfile(task, eventfile_out, status)
!  subroutine update_filter_settings_eventfile(task, eventFile_out, status)
!  subroutine update_filter_gti_settings(Evtool, eventFile, status)
!  subroutine update_filter_telid_settings(Evtool, EventFile, status)
!  subroutine update_filter_energy_settings(Evtool, EventFile, status)
!  subroutine update_filter_flag_settings(Evtool, EventFile, status)
!  subroutine update_filter_pattern_settings(Evtool, EventFile, status)
!  subroutine update_filter_rawxy_settings(Evtool, EventFile, status)

module evtool2_mod

    use erolib_general_mod
    use erolib_sort_mod
    use erolib_str_mod
    use erolib_ffits_mod
    use erolib2_region_mod
    use erolib_file_mod
    use erolib_extensions_mod
    use erolib_gti_mod
    use erolib_rawxy_mod
    use erolib_events_mod
    use erolib_eventfile_mod
    use erolib_image2_mod
    use time_mod
    use param

    implicit none

logical, private, parameter :: doTest=.false.

    private :: extractRawXYBounds,validateEnergyInt,interpret_str_gti
    private :: update_filter_gti_settings,update_filter_telid_settings
    private :: update_filter_energy_settings,update_filter_flag_settings
    private :: update_filter_pattern_settings,update_filter_rawxy_settings

    type tEvtool
        integer :: nfile = 0
        character(buffsize), dimension(:), allocatable :: infile
        integer :: delete_outfile = 0
        character(buffsize) :: outfile = ""
        logical :: clobber = .false.
        logical :: memset = .false. ! the memory setting to read in all files at once.
        real(r8) :: overlap = 0.0

        logical :: write_events = .false.
        logical :: write_image = .false.

        integer :: nenergy = 0
        integer :: filter_energy = 0
        real(dp), dimension(:), allocatable :: emin, emax
        integer :: filter_telid = 0
        integer :: ninst = 0
        integer, dimension(:), allocatable :: ccdnr
        character(buffsize), dimension(:), allocatable :: instrum
        integer :: filter_region = 0
        character(buffsize) :: region = ""
        type(tRegions) :: regionObj
        integer :: filter_gti = 0
        integer :: filter_type_gti = 0
        character(buffsize) :: gti_str = ""
        integer :: filter_flag = 0
        character(buffsize) :: flag = ""
        logical :: flag_invert = .false.
        integer :: flag_hex = 0
        integer :: filter_pattern = 0
        integer :: pattern = 0
        integer :: filter_image = 0

        character(buffsize) :: image_centre_str = ""
        character(buffsize) :: image_dimensions_str = ""
        logical :: auto_centre_image = .false.
        logical :: auto_dimension_image = .false.
!        integer, dimension(:), allocatable :: resolution, rebin
!        real(dp), dimension(:), allocatable :: center_position
        integer, dimension(:), allocatable :: rebin
        integer, dimension(2) :: resolution
        real(r8), dimension(2) :: center_position

        character(buffsize) :: rawxy_str = ""
        type(tRawXYSelection) :: rawxy_sel
        logical :: filter_rawxy = .false.

        type(tGTI) :: externalgti
        logical :: write_externalgti = .false.

        ! temporary
        character(buffsize) :: external_skyfield = ""

    contains
        procedure :: read_params => read_params_tEvtool
        procedure :: print_params => print_params_tEvtool
        procedure :: validate => validate_tEvtool
        procedure :: free => free_tEvtool
        procedure :: init => init_tEvtool
    end type tEvtool

    character(*), parameter :: gti_ext = "GTI"
    character(*), parameter :: qualgti_ext = "QUALGTI"
    character(*), parameter :: flaregti_ext = "FLAREGTI"
    character(*), parameter :: external_gti_ext = "STDGTI"
    integer, parameter :: canonical_rebin=80

contains

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> init the Evtool type structure with preset of allocated arrays
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine init_tEvtool(this, status, nfile, ninst, nenergy)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in), optional :: nfile, ninst, nenergy
        class(tEvtool) :: this

        character(*), parameter :: function = "init_tEvtool"

        if(status > 100 .or. status == s_errRep) return

        if(present(nfile)) then
            this%nfile = nfile
        else
            this%nfile = 100
        end if

        if(present(ninst)) then
            this%ninst = ninst
        else
            this%ninst = 7
        end if

        if(present(nfile)) then
            this%nenergy = nenergy
        else
            this%nenergy = 9
        end if

        call alloc(this%infile, this%nfile, status)
        call alloc(this%emin, this%nenergy, status)
        call alloc(this%emax, this%nenergy, status)
        call alloc(this%ccdnr, this%ninst, status)
!        call alloc(this%resolution, 2, status)
!        call alloc(this%center_position, 2, status)
        call alloc(this%rebin, 2, status)

    end subroutine init_tEvtool

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> free all allocated arrays in the Evtool type
    !
    !
    !
    !> @param[in] this
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tEvtool(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tEvtool) :: this

        character(*), parameter :: function = "free_tEvtool"

        if(status > 100 .or. status == s_errRep) return

!        call dealloc(this%resolution, status)
        call dealloc(this%rebin, status)
!        call dealloc(this%center_position, status)
        call dealloc(this%instrum, status)
        call dealloc(this%ccdnr, status)
        call dealloc(this%emin, status)
        call dealloc(this%emax, status)
        call dealloc(this%infile, status)

        call this%rawxy_sel%mask%free(status)

    end subroutine free_tEvtool

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> read the parameters of the Evtool
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_params_tEvtool(this, status)
        implicit none

        class(tEvtool) :: this
        integer, intent(inout) :: status

        integer :: icount,ierr
        integer :: ninfiles = 100
        character(buffsize) :: filelist_file
        integer, dimension(4) :: tempRawXYSel
        integer :: ii

        character(*), parameter :: function = "read_params_tEvtool"

        if(status > 100 .or. status == s_errRep) return

        call rgeta ("eventfiles", ninfiles, this%infile, icount, ierr)
        if(ierr /= 0) call wrstop(function, "reading infile parameter")
        this%infile = this%infile(:icount)

        ! check whether infile is actually a file containing the infiles
        if(icount == 1) then
            if(this%infile(1)(1:1) == "@") then
                filelist_file = this%infile(1)(2:)
                call read_filelist(filelist_file, this%infile, status)
            end if
        end if

        call rget ("outfile", this%outfile, ierr)
        if(ierr /= 0) call wrstop(function, "reading outfile parameter")

        call rget ("clobber", this%clobber, ierr)
        if(ierr /= 0) call wrstop(function, "reading clobber parameter")

        call rget("events", this%write_events, ierr)
        if(ierr /= 0) call wrstop(function, "reading events parameter")

        call rget("image", this%write_image, ierr)
        if(ierr /= 0) call wrstop(function, "reading image parameter")

!        call rgeta("size", 2, this%resolution, icount, ierr)
!        if(ierr /= 0) call wrstop(function, "reading image resolution parameter")
!        if(icount == 1) this%resolution(2) = this%resolution(1)
!        if(icount == 0) this%resolution = this%resolution(:icount)
        call rget("size", this%image_dimensions_str, ierr)

!        call rgeta("center_position", 2, this%center_position, icount, ierr)
!        if(ierr /= 0) call wrstop(function, "reading image center_position parameter")
!        if(icount == 1) this%center_position(2) = this%center_position(1)
!        if(icount == 0) this%center_position = this%center_position(:icount)
        call rget("center_position", this%image_centre_str, ierr)

        call rgeta("rebin", 2, this%rebin, icount, ierr)
        if(ierr /= 0) call wrstop(function, "reading image rebin parameter")
        if(icount == 1) this%rebin(2) = this%rebin(1)
        if(icount == 0) this%rebin = this%rebin(:icount)

        ! max number of nenergy will be set to 9 due to std_header settings
        ! should be updated if std_header will be updated
        !    call rgeti ("nenergy", this%nenergy, ierr)
        !    if(ierr /= 0) call wrstop(function, "reading nenergy parameter")

        call rgeta ("emin", this%nenergy, this%emin, icount, ierr)
        if(ierr == 1) ierr = 0
        if(ierr /= 0) call wrstop(function, "reading emin parameter")
        this%emin = this%emin(:icount) ! changed conversion to be done only in the events extension conversion
        !    this%emin = this%emin(:icount) * 1000.0_r8 ! keV to eV conversion

        call rgeta ("emax", this%nenergy, this%emax, icount, ierr)
        if(ierr == 1) ierr = 0
        if(ierr /= 0) call wrstop(function, "reading emax parameter")
        this%emax = this%emax(:icount) ! changed conversion to be done only in the events extension conversion
        !    this%emax = this%emax(:icount) * 1000.0_r8 ! keV to eV conversion

        ! update the nenergy keyword
        this%nenergy = icount

        call rgeta("telid", 7, this%ccdnr, icount, ierr)
        if(ierr == 1) ierr = 0
        if(ierr /= 0) call wrstop(function, "reading telid parameter")
        this%ccdnr = this%ccdnr(:icount)

        call rget ("region", this%region, ierr)
        if(ierr == 1) then
            this%region = ""
        else if (ierr /= 0) then
            call wrstop(function, "reading region parameter")
        end if

        call rget ("gti", this%gti_str, ierr)
        if(ierr == 1) then
            this%gti_str = ""
        else if (ierr /= 0) then
            call wrstop(function, "reading gti parameter")
        end if

        call rget ("flag", this%flag, ierr)
        if(ierr == 1) then
            this%flag = ""
        else if (ierr /= 0) then
            call wrstop(function, "reading flag parameter")
        end if

        call rget ("flag_invert", this%flag_invert, ierr)
        if(ierr /= 0) call wrstop(function, "reading flag_invert parameter")

        call rget ("pattern", this%pattern, ierr)
        if(ierr /= 0) call wrstop(function, "reading pattern parameter")

        call rget ("rawxy", this%rawxy_str, ierr)
        if(ierr == 1) then
            this%rawxy_str = ""
        else if(ierr /= 0) then
          call wrstop(function, "reading rawxy parameter")
        end if

        this%rawxy_sel%ccdnr_defined = .true. ! default
        call rget ("rawxy_telid", this%rawxy_sel%ccdnr, ierr)
        if(this%rawxy_sel%ccdnr < 1 .or. this%rawxy_sel%ccdnr > ntel) then
            this%rawxy_sel%ccdnr = 0
            this%rawxy_sel%ccdnr_defined = .false.
        else if (ierr /= 0) then
            call wrstop(function, "reading rawxy_ccdnr parameter")
        end if

        call rget ("rawxy_invert", this%rawxy_sel%invert, ierr)
        if(ierr /= 0) call wrstop(function, "reading rawxy_invert parameter")

        call rget ("memset", this%memset, ierr)
        if(ierr /= 0) call wrstop(function, "reading memset parameter")

        call rget ("overlap", this%overlap, ierr)
        if(ierr /= 0) call wrstop(function, "reading overlap parameter")

        call rget ("skyfield", this%external_skyfield, ierr)
        if(ierr == 1) then
            this%external_skyfield = ""
        else if (ierr /= 0) then
            call wrstop(function, "reading skyfield parameter")
        end if

        ! adjust all read parameters
        this%infile  = adjustl(this%infile)
        this%outfile = adjustl(this%outfile)
        this%region  = adjustl(this%region)
        this%gti_str = adjustl(this%gti_str)
        this%rawxy_str = adjustl(this%rawxy_str)
!*** check other str params.

    end subroutine read_params_tEvtool

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> print the parameters of the Evtool
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine print_params_tEvtool(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tEvtool) :: this
        integer :: ii

        character(*), parameter :: function = "print_params_tEvtool"
        character(buffsize) :: entryStr, ccdnrStr, invertStr

        if(status > 100 .or. status == s_errRep) return

        call wrlog(function, "printing parameters")
        print *, "EVENTFILES: - @FILELISTS are already read in"
        do ii = 1, size(this%infile)
            print *, "*" // trim(this%infile(ii)) // "*"
        end do

        print *, "OUTFILE:"
        print *, "*" // trim(this%outfile) // "*"

        print *, "CLOBBER:"
        print * , this%clobber

        print *, "WRITE EVENTS:"
        print * , this%write_events

        print *, "WRITE IMAGE:"
        print * , this%write_image

        print *, "RESOLUTION"
        print *, this%resolution

        print *, "CENTER POSITION"
        print *, this%center_position

        print *, "REBIN"
        print *, this%rebin

        print *, "EMIN"
        print *, this%emin

        print *, "EMAX"
        print *, this%emax

        print *, "TELID"
        print *, this%ccdnr

        print *, "REGION"
        print *, "*" // trim(this%region) // "*"

        print *, "GTI"
        print *, "*" // trim(this%gti_str) // "*"

        print *, "FLAG"
        print *, "*" // trim(this%flag) // "*"

        print *, "PATTERN"
        print *, this%pattern

        print *, "RAWXY"
        call this%rawxy_sel%get_entry_strs(entryStr, ccdnrStr, invertStr)
        print trim(entryStr)

        print *, "RAWXY_TELID"
        print *, trim(ccdnrStr)

        print *, "RAWXY_INVERT"
        print *, trim(invertStr)

    end subroutine print_params_tEvtool

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> validate the input parameters of Evtool
    !
    !
    !
    !> @param[in] this
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine validate_tEvtool(this, status)
        implicit none

        class(tEvtool) :: this
        integer, intent(inout) :: status

        character(*), parameter :: function = "validate_tEvtool"
        integer :: exists, ii
        type(tFitsFile) :: external_GTI_file
        character(buffsize) :: msg

        if(status > 100 .or. status == s_errRep) return

        do ii = 1, size(this%infile)
            if(trim(this%infile(ii)) == "") call wrerr(function, s_err, "one of the infiles is blank")

            call ffits_file_exists(this%infile(ii), exists, status)
            if (exists == 0) then
                call wrerr(function, s_err, "infile " // trim(this%infile(ii)) // " does not exist")
                status = s_errRep
    return
            end if
        end do

        if(trim(this%outfile) /= "") then
            this%delete_outfile = false
            call ffits_file_exists(this%outfile, exists, status)
            if (exists /= 0) then
                if(this%clobber .eqv. .false.) then
                    call wrerr(function, s_err, "outfile " // trim(this%outfile) // " already exists")
                    status = s_errRep
    return
                else
                    this%delete_outfile = true
                end if
            end if
        else
            call wrerr(function, s_err, "outfile is blank")
        end if

        this%filter_image = false
        if(this%write_image .eqv. .true.) then
          call extract_pair_dbl(this%image_centre_str, this%auto_centre_image, this%center_position)
          call extract_pair_int(this%image_dimensions_str, this%auto_dimension_image, this%resolution)

          if(.not.this%auto_dimension_image .and.(this%resolution(1) < 1 .or. this%resolution(2) < 1)) then
            call wrerr(function, s_err, "zero or negative resolution selected")
            status = s_errRep
    return
          end if

          if(this%rebin(1) < 1 .or. this%rebin(2) < 1) then
            call wrerr(function, s_err, "zero or negative rebin selected")
            status = s_errRep
    return
          end if

          if(this%rebin(1) /= canonical_rebin .or. this%rebin(2) /= canonical_rebin) then
            msg = int2str(canonical_rebin)
            call wrerr(function, warn, "If pixel size (parameter --rebin) is not left at "&
              //trim(msg)//", reliable source detection is not guaranteed.")
          end if

          this%filter_image = true
        end if

        if(this%nenergy < 0 .and. this%nenergy > 9) then
            call wrerr(function, s_err, "max number of energy intervals"&
              //" cannot be less than zero or more than 9")
            status = s_errRep
    return
        end if

        if(size(this%emin) /= size(this%emax)) then
            call wrerr(function, s_err, "number of given energy boundaries"&
              //" (emin / emax) is inconsistent")
            status = s_errRep
        end if

        if(size(this%emin) > this%nenergy .or. size(this%emax) > this%nenergy) then
            call wrerr(function, s_err, "max number of energy intervals is"&
              //" lower than the max number of energy bands provided")
            status = s_errRep
        end if

        call validateEnergyInt(this%emin, this%emax, status)

        this%filter_energy = false
        if(size(this%emin) > 0) this%filter_energy = true
!*** 20190915 IMS: seems never to be preset to false, so I added this just above as the default.

        ! FIX for HB evtool call

        if(size(this%ccdnr) > 0) then
            this%filter_telid = true

            if(minval(this%ccdnr) < 1) then
                call wrerr(function, s_err, "telescope id is illegal, cannot be less than 1")
                status = s_errRep
            end if

            if(maxval(this%ccdnr) > 7) then
                call wrerr(function, s_err, "telescope id is illegal, cannot be greater than 7")
                status = s_errRep
            end if

        else if (size(this%ccdnr) > 7) then
            call wrerr(function, s_err, "more telescope ids selected than max number of instruments (7)")
            status = s_errRep
        else
            this%filter_telid = false
        end if

        if(trim(this%region) /= "") then
            call getRegionsFromStr(this%region, this%regionObj, status)
            if(status == ok) this%filter_region = true
        else
            call alloc_regRows(this%regionObj%regRows, 0, status)
        end if

        this%filter_gti = false ! default
        if(trim(this%gti_str) /= "") then
            ! determine the input format of the gti
            ! type is per telescope
            if(this%gti_str == gti_ext) then
                this%filter_type_gti = gti_type
            else if(this%gti_str == qualgti_ext) then
                this%filter_type_gti = qualgti_type
            ! type is global
            else if(this%gti_str == flaregti_ext) then
                this%filter_type_gti = flaregti_type
            else
                this%filter_type_gti = externalgti_type
                ! assume type is a FITS file
                call ffits_file_exists(this%gti_str, exists, status)
                if (exists == 1) then
                  call external_GTI_file%info%setFileName(this%gti_str)
                  call this%externalgti%init(iEXTERNAL, status, external_GTI_file%info)
                  this%externalgti%info%tableName = this%externalgti%extNameRoot
                  call this%externalgti%setUseage(0, 1, 0, 0)
                  call external_GTI_file%open(readonly, status)
                  call this%externalgti%read(status)
                  call this%externalgti%sanityCheck(status, "External GTI:")
                  if(.not. this%externalgti%containsData()) then
                    call wrerr(function, s_err, external_gti_ext&
                      //" in file "//trim(this%gti_str)//" could not be read")
                      status = s_errRep
    return
                  end if
                  call external_GTI_file%close(status)
                else
                    ! interpret string as two values indicating tstart and tstop
                    call interpret_str_gti(this%gti_str, this%externalgti, status)

                end if

            end if
            if(status == ok) this%filter_gti = true
        end if

        if(trim(this%flag) /= "") then
          this%flag_hex = str2hex(this%flag, status)
          if(status == ok) this%filter_flag = true
        end if

        ! single = 1, double = 2, triple = 4, quadruple = 8
        this%filter_pattern = false ! default
        if(this%pattern > 0 .and. this%pattern < 16) then
          this%filter_pattern = true
        end if

        this%filter_rawxy = .false. ! default
        if(trim(this%rawxy_str) /= "") then
          ! Determine which of the two acceptable strings it is - either (mask) file name or 'xLo xHi yLo yHi'.
          !
          call ffits_file_exists(this%rawxy_str, exists, status)
          if (exists == 1) then
            this%rawxy_sel%mask%filename = this%rawxy_str
            call this%rawxy_sel%mask%open(readonly, status)
            call this%rawxy_sel%mask%read(status)
            call this%rawxy_sel%mask%close(status)
            this%rawxy_sel%mask%containsData = .true.
          else
            call extractRawXYBounds(this%rawxy_str, this%rawxy_sel, status)
          end if

          if(status == ok) this%filter_rawxy = .true.
        end if

        if(this%rawxy_sel%ccdnr_defined) then
          if(this%rawxy_sel%ccdnr < 1 .or. this%rawxy_sel%ccdnr > ntel) then
            call wrerr(function, s_err, "RAWXY ccd number is out of range.")
            status = s_errRep
    return
          end if
        end if

    end subroutine validate_tEvtool

    !---------------------------------------------------------------------------
    !> @author
    !> Ian M Stewart
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> blah
    !
    !
    !
    !> @param[in] emin
    !> @param[out] emax
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine extractRawXYBounds(rawxy_str, rawxy_sel, status)
        implicit none

        character(*),             intent(in)    :: rawxy_str
        type(tRawXYSelection),    intent(inout) :: rawxy_sel
        integer,                  intent(inout) :: status

        integer :: ii
        character(buffsize), allocatable, dimension(:) :: listOfStrs
        integer :: icount,readstat

        character(*), parameter :: function = "extractRawXYBounds"

        if(status > 100 .or. status == s_errRep) return

        ! Divide the rawxy string into its space-separated substrings:
        !
        call strtok(rawxy_str, ' ', listOfStrs)
        icount = size(listOfStrs)

        if(icount == 4) then
          ! Read the 4 values into the relevant struct elements:
          !
          ii = 1
          read(listOfStrs(ii),*,iostat=readstat) rawxy_sel%axes(1)%lo%value
          if (readstat>0) then
            call wrerr(function, s_err, "Cannot read " // trim(listOfStrs(ii)))
          status = s_errRep
    return
          end if

          ii = 2
          read(listOfStrs(ii),*,iostat=readstat) rawxy_sel%axes(1)%hi%value
          if (readstat>0) then
            call wrerr(function, s_err, "Cannot read " // trim(listOfStrs(ii)))
          status = s_errRep
    return
          end if

          ii = 3
          read(listOfStrs(ii),*,iostat=readstat) rawxy_sel%axes(2)%lo%value
          if (readstat>0) then
            call wrerr(function, s_err, "Cannot read " // trim(listOfStrs(ii)))
          status = s_errRep
    return
          end if

          ii = 4
          read(listOfStrs(ii),*,iostat=readstat) rawxy_sel%axes(2)%hi%value
          if (readstat>0) then
            call wrerr(function, s_err, "Cannot read " // trim(listOfStrs(ii)))
          status = s_errRep
    return
          end if

          do ii = 1, 2
            if(rawxy_sel%axes(ii)%lo%value < 1)&
              rawxy_sel%axes(ii)%lo%value = 1
            if(rawxy_sel%axes(ii)%hi%value > n_ccd_pixels)&
              rawxy_sel%axes(ii)%hi%value = n_ccd_pixels

            rawxy_sel%axes(ii)%lo%is_defined = .true.
            rawxy_sel%axes(ii)%hi%is_defined = .true.
          end do

          ! Fill the mask.
          !
          allocate(rawxy_sel%mask%values(n_ccd_pixels,n_ccd_pixels))
          rawxy_sel%mask%values = .false.
          rawxy_sel%mask%values(rawxy_sel%axes(1)%lo%value:rawxy_sel%axes(1)%hi%value&
                           ,rawxy_sel%axes(2)%lo%value:rawxy_sel%axes(2)%hi%value) = .true.

        else
            call wrerr(function, warn, "Invalid rawxy argument "//trim(rawxy_str)&
              //". It is neither the name of an existing file or a list of 4 substrings.")
            do ii = 1, 2
              rawxy_sel%axes(ii)%lo%is_defined = .false.
              rawxy_sel%axes(ii)%lo%value = 0.0
              rawxy_sel%axes(ii)%hi%is_defined = .false.
              rawxy_sel%axes(ii)%hi%value = 0.0
            end do
        end if

        call dealloc(listOfStrs, status)

        do ii=1, 2
          if(rawxy_sel%axes(ii)%lo%is_defined .and. rawxy_sel%axes(ii)%hi%is_defined) then
            if(rawxy_sel%axes(ii)%lo%value > rawxy_sel%axes(ii)%hi%value) then
              call wrerr(function, s_err, "RAWXY filter: lo may not be greater than hi.")
              status = s_errRep
    return
            end if
          end if
        end do

    end subroutine extractRawXYBounds

  !.....................................................................
  subroutine extract_pair_dbl(image_centre_str, auto_centre_image, center_position)
    implicit none

    character(*),           intent(in)  :: image_centre_str
    logical,                intent(out) :: auto_centre_image
    real(r8), dimension(2), intent(out) :: center_position

    character(*), parameter :: functionName = "extract_pair_dbl"
    integer :: ii,ierr
    character(buffsize), allocatable, dimension(:) :: listOfStrs
    integer :: icount,readstat


!    if(status > 100 .or. status == s_errRep) return

    if(trim(image_centre_str)=='auto' .or. trim(image_centre_str)=='AUTO') then
      auto_centre_image = .true.
      center_position = 0.0d0

    else
      auto_centre_image = .false.

      ! Divide the string into its space-separated substrings:
      !
      call strtok(image_centre_str, ' ', listOfStrs)
      icount = size(listOfStrs)

      if(icount == 2) then
        ! Read the 2 values into the relevant struct elements:
        !
        do ii = 1,2
          read(listOfStrs(ii),*,iostat=readstat) center_position(ii)
          if (readstat>0) then
            call wrerr(functionName, s_err, "Cannot read " // trim(listOfStrs(ii)))
!            status = s_errRep
    return
          end if
        end do

      else if(icount == 1) then
        ii = 1
        read(listOfStrs(ii),*,iostat=readstat) center_position(ii)
        if (readstat>0) then
          call wrerr(functionName, s_err, "Cannot read " // trim(listOfStrs(ii)))
!          status = s_errRep
    return
        end if

        center_position(2) = center_position(1)

      else
        call wrerr(functionName, s_err, "Invalid 'center_position' parameter "&
          //trim(image_centre_str)//". It must either equal 'auto' or be a list of 2 numbers.")
      end if

      call dealloc(listOfStrs, ierr)
    end if

  end subroutine extract_pair_dbl

  !.....................................................................
  subroutine extract_pair_int(image_dimensions_str, auto_dimension_image, resolution)
    implicit none

    character(*),          intent(in)  :: image_dimensions_str
    logical,               intent(out) :: auto_dimension_image
    integer, dimension(2), intent(out) :: resolution

    character(*), parameter :: functionName = "extract_pair_int"
    integer :: ii,ierr
    character(buffsize), allocatable, dimension(:) :: listOfStrs
    integer :: icount,readstat


!    if(status > 100 .or. status == s_errRep) return

    if(trim(image_dimensions_str)=='auto' .or. trim(image_dimensions_str)=='AUTO') then
      auto_dimension_image = .true.
      resolution = 0.0d0

    else
      auto_dimension_image = .false.

      ! Divide the string into its space-separated substrings:
      !
      call strtok(image_dimensions_str, ' ', listOfStrs)
      icount = size(listOfStrs)

      if(icount == 2) then
        ! Read the 2 values into the relevant struct elements:
        !
        do ii = 1,2
          read(listOfStrs(ii),*,iostat=readstat) resolution(ii)
          if (readstat>0) then
            call wrerr(functionName, s_err, "Cannot read " // trim(listOfStrs(ii)))
!            status = s_errRep
    return
          end if
        end do

      else if(icount == 1) then
        ii = 1
        read(listOfStrs(ii),*,iostat=readstat) resolution(ii)
        if (readstat>0) then
          call wrerr(functionName, s_err, "Cannot read " // trim(listOfStrs(ii)))
!          status = s_errRep
    return
        end if

        resolution(2) = resolution(1)

      else
        call wrerr(functionName, s_err, "Invalid 'size' parameter "&
          //trim(image_dimensions_str)//". It must either equal 'auto' or be a list of 2 numbers.")
      end if

      call dealloc(listOfStrs, ierr)
    end if

  end subroutine extract_pair_int

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> validate energy intervalls
    !
    !
    !
    !> @param[in] emin
    !> @param[out] emax
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine validateEnergyInt(emin, emax, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii

        real(dp), dimension(:), intent(in) :: emin, emax

        character(*), parameter :: function = "validateEnergyInt"

        if(status > 100 .or. status == s_errRep) return

        if(size(emin) /= size(emax)) then
            call wrerr(function, s_err, "number of min/max energy entries is different")
            status = s_errRep
            return
        end if

        do ii = 1, size(emin)
            if(emin(ii) > emax(ii)) then
                call wrerr(function, s_err, "selected min energy > max energy")
                status = s_errRep
                return
            end if
            if(emin(ii) < 0.0_dp .or. emax(ii) < 0.0_dp) then
                call wrerr(function, s_err, "selected min and/or max energy is smaller than 0 eV")
                status = s_errRep
                return
            end if
        end do

    end subroutine validateEnergyInt


    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> interprete string as gti
    !
    !
    !
    !> @param[in] file
    !> @param[out]
    !> @param[inout] status, filelist
    !---------------------------------------------------------------------------

    subroutine interpret_str_gti(str, gti, status)
        implicit none

        integer, intent(inout) :: status
        character(*), intent(in) :: str
        type(tGTI), intent(out) :: gti

        character(buffsize), dimension(:), allocatable :: gti_str

        character(*), parameter :: function = "interpret_str_gti"

        if(status > 100 .or. status == s_errRep) return

        call strtok(str, " ", gti_str)
        if(size(gti_str) /= 2) then
            call wrerr(function, s_err, "could not interprete " // trim(str) // " as external gti")
            status = s_errRep
            return
        end if

        call gti%init(iTemp, status)
        call gti%setUseage(0, 0, 0, 1) ! set for internal useage.
        call gti%alloc(status, 1)
        gti%cols(r8GTI_START)%r8data(1) = str2double(gti_str(1), status)
        gti%cols(r8GTI_STOP )%r8data(1) = str2double(gti_str(2), status)
!        gti%header%keywords(r8GH_TSTART)%r8Value = gti%cols(r8GTI_START)%r8data(1)
!        gti%header%keywords(r8GH_TSTOP )%r8Value = gti%cols(r8GTI_STOP )%r8data(1)
        call gti%header%keywords(r8GH_TSTART)%setValue(gti%cols(r8GTI_START)%r8data(1), status)
        call gti%header%keywords(r8GH_TSTOP )%setValue(gti%cols(r8GTI_STOP )%r8data(1), status)
!        ! when set to negative mjdref, the mjdref value is ignored when gti is applied
!        gti%use_mjdref = .false.
!        gti%use_timezero = .false.
        call gti%sanityCheck(status, "User input GTI:")

    end subroutine interpret_str_gti

  !.....................................................................
  subroutine apply_filter_eventfile(task, eventfile_out, status)
    ! Filters the event list but also then constructs an image, if required.
    !
    implicit none

    type(tEvtool),    intent(in)    :: task
    type(tEventFile), intent(inout) :: eventfile_out
    integer,          intent(inout) :: status

    integer :: nenergy
    character(*), parameter :: function = "apply_filter_eventfile"

    if(status > 100 .or. status == s_errRep) return

    call eventfile_out%sort_events(status)

    if(task%filter_telid == true .and. status == ok) then
      call wrchatty(function, "telid filter")
      call eventfile_out%events%apply_telid(eventfile_out%ccdnr_infile, status)
    end if

    if(task%filter_energy == true .and. status == ok) then
      call wrchatty(function, "energy filter")
      nenergy = eventfile_out%std_header%nenergy
      ! only apply energy unit conversion on std_header here, since std_header is keV while energy column are in eV
      call eventfile_out%events%apply_energy(&
          eventfile_out%std_header%eminx(:nenergy)*1000.0_r8&
        , eventfile_out%std_header%emaxx(:nenergy)*1000.0_r8, status)
    end if

    if(task%filter_flag == true .and. status == ok) then
      call wrchatty(function, "flag filter")
!      call eventfile_out%events%apply_flag(eventfile_out%std_header%flagsel, status)
      call eventfile_out%events%apply_flag(eventfile_out%std_header%flagsel, status, invert=task%flag_invert)
    end if

    if(task%filter_pattern == true .and. status == ok) then
      call wrchatty(function, "pattern filter")
      call eventfile_out%events%apply_pattern(eventfile_out%std_header%pat_sel, status)
    end if

    if(task%filter_region == true .and. status == ok) then
      call wrchatty(function, "region filter")
      call eventfile_out%apply_filter_region(task%regionObj, status)

!!!      eventfile_out%regionObj%reg_file%info = eventfile_out%info

      ! if its a "physical filter a region gti cannot be derived -> no conversion from rawx/y to ra/dec
      ! it is enough to check first region, since only one type of wcs is allowed
      if(eventfile_out%regionObj%regRows(1)%wcs /= "physical") then
        call eventfile_out%derive_region_gti(status)

        call eventfile_out%apply_filter_tm_GTIs(iREGIONGTI, status, task%overlap, do_events=.false.)
      end if
    end if

    if(task%filter_gti == true .and. status == ok) then
      call wrchatty(function, "gti filter")

      if(task%filter_type_gti == gti_type) then
        call wrchatty(function, "- internal gti filter")
        call eventfile_out%apply_filter_tm_GTIs(iGTI,      status, task%overlap, do_events=.true.)
!        eventfile_out%std_header%tstart&
!          = eventfile_out%gtiList(iGTI)%gtiPerTM(1)%header%keywords(r8GH_TSTART)%r8Value
!        eventfile_out%std_header%tstop&
!          = eventfile_out%gtiList(iGTI)%gtiPerTM(1)%header%keywords(r8GH_TSTOP )%r8Value
        call eventfile_out%gtiList(iGTI)%gtiPerTM(1)%header%keywords(r8GH_TSTART)%setValue(eventfile_out%std_header%tstart, status)
        call eventfile_out%gtiList(iGTI)%gtiPerTM(1)%header%keywords(r8GH_TSTOP )%setValue(eventfile_out%std_header%tstop,  status)

      else if(task%filter_type_gti == qualgti_type) then
        call wrchatty(function, "- qual gti filter")
        call eventfile_out%apply_filter_tm_GTIs(iQUALGTI,  status, task%overlap, do_events=.true.)
!        eventfile_out%std_header%tstart&
!          = eventfile_out%gtiList(iQUALGTI)%gtiPerTM(1)%header%keywords(r8GH_TSTART)%r8Value
!        eventfile_out%std_header%tstop&
!          = eventfile_out%gtiList(iQUALGTI)%gtiPerTM(1)%header%keywords(r8GH_TSTOP )%r8Value
        call eventfile_out%gtiList(iQUALGTI)%gtiPerTM(1)%header%keywords(r8GH_TSTART)%setValue(eventfile_out%std_header%tstart, status)
        call eventfile_out%gtiList(iQUALGTI)%gtiPerTM(1)%header%keywords(r8GH_TSTOP )%setValue(eventfile_out%std_header%tstop,  status)

      else if(task%filter_type_gti == flaregti_type) then
        call wrchatty(function, "- flare gti filter")
        call eventfile_out%apply_filter_tm_GTIs(iFLAREGTI, status, task%overlap, do_events=.true.)
!        eventfile_out%std_header%tstart&
!          = eventfile_out%gtiList(iFLAREGTI)%gtiPerTM(1)%header%keywords(r8GH_TSTART)%r8Value
!        eventfile_out%std_header%tstop&
!          = eventfile_out%gtiList(iFLAREGTI)%gtiPerTM(1)%header%keywords(r8GH_TSTOP )%r8Value
        call eventfile_out%gtiList(iFLAREGTI)%gtiPerTM(1)%header%keywords(r8GH_TSTART)%setValue(eventfile_out%std_header%tstart, status)
        call eventfile_out%gtiList(iFLAREGTI)%gtiPerTM(1)%header%keywords(r8GH_TSTOP )%setValue(eventfile_out%std_header%tstop,  status)

      else if(task%filter_type_gti == externalgti_type) then
        call wrchatty(function, "- external gti filter")
        call eventfile_out%apply_filter_nontm_GTI(task%externalgti, status, task%overlap)
        ! Note that we can't call apply_filter_tm_GTIs() with GTI type iEXTERNAL because this call runs over all the TMs.
!        eventfile_out%std_header%tstart = task%externalgti%header%keywords(r8GH_TSTART)%r8Value
!        eventfile_out%std_header%tstop  = task%externalgti%header%keywords(r8GH_TSTOP )%r8Value
        call task%externalgti%header%keywords(r8GH_TSTART)%setValue(eventfile_out%std_header%tstart, status)
        call task%externalgti%header%keywords(r8GH_TSTOP )%setValue(eventfile_out%std_header%tstop,  status)

      else
call wrstop(function, "Unrecognized GTI filter type.")
      end if

      call task%externalgti%destroy(status)

      call wrchatty(function, "Set new DATE-OBS, DATE-END keyword values.")

      call sec2date(eventfile_out%std_header%tstart, eventfile_out%std_header%mjdref&
        , eventfile_out%std_header%date_obs, status)
      call sec2date(eventfile_out%std_header%tstop,  eventfile_out%std_header%mjdref&
        , eventfile_out%std_header%date_end, status)
    end if

    if(task%filter_rawxy .and. status == ok) then
      call wrchatty(function, "RAWX/Y filter")
      call eventfile_out%events%apply_rawxy(task%rawxy_sel, status)
    end if

    if(task%filter_image == true .and. status == ok) then
      if(.not. eventfile_out%events%wcs_containsData) then
        call wrerr(function, c_err, "No X/Y WCS was read from the events header, can't make image.")
  return
      end if

      call wrlog(function, "Creating image.")
      call eventfile_out%image%createFromEvents(eventfile_out%events&
        , task%resolution, task%center_position, task%rebin, task%auto_centre_image&
        , task%auto_dimension_image, status)
      eventfile_out%image_containsData = .true.
    end if

  end subroutine apply_filter_eventfile

  !.....................................................................
  subroutine update_filter_settings_eventfile(task, eventFile_out, status)
    implicit none

    type(tEvtool),    intent(in)    :: task
    type(tEventFile), intent(inout) :: eventFile_out
    integer,          intent(inout) :: status

    character(*), parameter :: function = "update_filter_settings_eventfile"

    if(status > 100 .or. status == s_errRep) return

    call update_filter_telid_settings(  task, eventFile_out, status)
    call update_filter_energy_settings( task, eventFile_out, status)
    call update_filter_flag_settings(   task, eventFile_out, status)
    call update_filter_pattern_settings(task, eventFile_out, status)
    call update_filter_rawxy_settings(  task, eventFile_out, status)

    !        call update_filter_region_settings(task, eventFile, status)
    call update_filter_gti_settings(    task, eventFile_out, status)

  end subroutine update_filter_settings_eventfile

  !.....................................................................
  subroutine update_filter_gti_settings(Evtool, eventFile, status)
    implicit none

    type(tEvtool),    intent(in)    :: Evtool
    type(tEventFile), intent(inout) :: eventFile
    integer,          intent(inout) :: status

    character(*), parameter :: function = "update_filter_gti_settings"

    if(status > 100 .or. status == s_errRep) return

    if(Evtool%filter_gti == true .and. .not. eventFile%filtered_gti) then
      eventFile%filtered_gti = .true.
      if(Evtool%filter_type_gti == gti_type)      eventFile%std_header%gti_sel = gti_ext
      if(Evtool%filter_type_gti == qualgti_type)  eventFile%std_header%gti_sel = qualgti_ext
      if(Evtool%filter_type_gti == flaregti_type) eventFile%std_header%gti_sel = flaregti_ext
      if(Evtool%filter_type_gti == externalgti_type) then
        eventFile%std_header%gti_sel = external_gti_ext
      end if
    end if

  end subroutine update_filter_gti_settings

  !.....................................................................
  subroutine update_filter_telid_settings(Evtool, EventFile, status)
    implicit none

    type(tEvtool),    intent(in)    :: Evtool
    type(tEventFile), intent(inout) :: EventFile
    integer,          intent(inout) :: status

    integer, dimension(:), allocatable :: intersect_array, complement_array
    integer :: ii
    character(*), parameter :: function = "update_filter_telid_settings"

    if(status > 100 .or. status == s_errRep) return

    if(Evtool%filter_telid == true .and. .not. EventFile%filtered_telid) then
      call wrerr(function, s_err, "EventFile out has no input ccdnr, this should never happen")
      status = s_errRep
  return
    end if

    if(Evtool%filter_telid == true .and. EventFile%filtered_telid) then
      intersect_array = intersect_int_order(EventFile%ccdnr_infile, Evtool%ccdnr, status)
      complement_array = complement_int_order(Evtool%ccdnr, EventFile%ccdnr_infile, status)
      EventFile%ccdnr_infile = EventFile%ccdnr_infile(intersect_array)

      if(EventFile%ninst < size(Evtool%ccdnr)) then
        do ii = 1, size(complement_array)
          call wrerr(function, warn, "Selected TELID " // trim(int2str(Evtool%ccdnr(complement_array(ii)), status)) // &
                  " is not present in the infile/s")
        end do
      end if

      EventFile%instrum = EventFile%instrum(intersect_array)
      EventFile%ninst = size(intersect_array)

      !update stdheader
      EventFile%std_header%ninst = EventFile%ninst
      EventFile%std_header%instrumx(:EventFile%ninst) = EventFile%instrum

      if(EventFile%std_header%ninst == 1) then
        EventFile%std_header%instrume = EventFile%std_header%instrumx(1)
      else
        EventFile%std_header%instrume = "merged"
      end if

      call dealloc(intersect_array, status)
      call dealloc(complement_array, status)

    end if
  end subroutine update_filter_telid_settings

  !.....................................................................
  subroutine update_filter_energy_settings(Evtool, EventFile, status)
    implicit none

    type(tEvtool),    intent(in)    :: Evtool
    type(tEventFile), intent(inout) :: EventFile
    integer,          intent(inout) :: status

    real(r8), dimension(:), allocatable :: emin, emax

    character(*), parameter :: function = "update_filter_energy_settings"

    if(status > 100 .or. status == s_errRep) return

    if(Evtool%filter_energy == true .and. .not. EventFile%filtered_energy) then
      EventFile%filtered_energy = .true.
      EventFile%std_header%nenergy = Evtool%nenergy
      EventFile%std_header%eminx(:Evtool%nenergy) = Evtool%emin
      EventFile%std_header%emaxx(:Evtool%nenergy) = Evtool%emax
  return
    end if

    if(Evtool%filter_energy == true .and. EventFile%filtered_energy) then
      call merge_and_r8(Evtool%emin, Evtool%emax, Evtool%nenergy&
        , EventFile%std_header%eminx, EventFile%std_header%emaxx&
        , EventFile%std_header%nenergy, emin, emax, status)
      if(size(emin) > 9) then
        call wrerr(function, s_err, "the number of energy intervals after user applied energy settings exceed the max of 9 energy intervals")
        status = s_errRep
  return
      end if
      EventFile%std_header%nenergy = size(emin)
      EventFile%std_header%eminx(:size(emin)) = emin
      EventFile%std_header%emaxx(:size(emin)) = emax
      call dealloc(emin, status)
      call dealloc(emax, status)
    end if

  end subroutine update_filter_energy_settings

  !.....................................................................
  subroutine update_filter_flag_settings(Evtool, EventFile, status)
    implicit none

    type(tEvtool),    intent(in)    :: Evtool
    type(tEventFile), intent(inout) :: EventFile
    integer,          intent(inout) :: status

    integer :: flag
    character(*), parameter :: function = "update_filter_flag_settings"

    if(status > 100 .or. status == s_errRep) return

    if(Evtool%filter_flag == true .and. .not.EventFile%filtered_flag) then
      EventFile%filtered_flag = .true.
      EventFile%std_header%flagsel = Evtool%flag_hex
      if(Evtool%flag_invert) then
        EventFile%std_header%c_flagsel = "Flag selection: "//trim(Evtool%flag)//" (inverted)"
      else
        EventFile%std_header%c_flagsel = "Flag selection: "//trim(Evtool%flag)
      end if
  return
    end if

    if(Evtool%filter_flag == true .and. EventFile%filtered_flag) then
      flag = ior(EventFile%std_header%flagsel, Evtool%flag_hex)
      EventFile%std_header%flagsel = flag
    end if

  end subroutine update_filter_flag_settings

  !.....................................................................
  subroutine update_filter_pattern_settings(Evtool, EventFile, status)
    implicit none

    type(tEvtool),    intent(in)    :: Evtool
    type(tEventFile), intent(inout) :: EventFile
    integer,          intent(inout) :: status

    integer :: pattern
    character(*), parameter :: function = "update_filter_pattern_settings"

    if(status > 100 .or. status == s_errRep) return

    if(Evtool%filter_pattern == true .and. .not. EventFile%filtered_pattern) then
      EventFile%filtered_pattern = .true.
      EventFile%std_header%pat_sel = Evtool%pattern
  return
    end if

    if(Evtool%filter_pattern == true .and. EventFile%filtered_pattern) then
      pattern = iand(EventFile%std_header%pat_sel, Evtool%pattern)
      EventFile%std_header%pat_sel = pattern
    end if

  end subroutine update_filter_pattern_settings

  !.....................................................................
  subroutine update_filter_rawxy_settings(Evtool, EventFile, status)
    implicit none

    type(tEvtool),    intent(in)    :: Evtool
    type(tEventFile), intent(inout) :: EventFile
    integer,          intent(inout) :: status

    character(*), parameter :: function = "update_filter_rawxy_settings"

    if(status > 100 .or. status == s_errRep) return

    if(Evtool%filter_rawxy .and. .not. EventFile%filtered_rawxy) then
      EventFile%filtered_rawxy = .true.
      call EventFile%record_rawxy_sel(Evtool%rawxy_sel, status, .true.)
      return
    end if

    if(Evtool%filter_rawxy .and. EventFile%filtered_rawxy) then
      call EventFile%record_rawxy_sel(Evtool%rawxy_sel, status)
    end if

  end subroutine update_filter_rawxy_settings

end module evtool2_mod
