! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module evtool_Badpix_mod

    use erolib_str_mod
    use erolib_ffits_mod
    use erolib_general_mod
    use erolib_sort_mod
    use erolib_math_mod
    use erolib_tbintable_mod
    implicit none

    private

    character(*), parameter :: badpix_ext = "BADPIX"
    character(*), parameter :: rawx_col = "RAWX"
    character(*), parameter :: rawy_col = "RAWY"
    character(*), parameter :: badflag_col = "BADFLAG"
    character(*), parameter :: timemin_col = "TIMEMIN"
    character(*), parameter :: timemax_col = "TIMEMAX"
    character(*), parameter :: phamax_col = "PHAMAX"
    character(*), parameter :: phamin_col = "PHAMIN"
    character(*), parameter :: phamed_col = "PHAMED"
    character(*), parameter :: type_col = "TYPE"
    character(*), parameter :: yextent_col = "YEXTENT"

    type tBadpix
        real(r8) :: mjdref = 0.0
        integer :: ccdnr = 0
        integer :: nelements = 0
        character(buffsize) :: extname = ""
        real (r8), allocatable, dimension(:) :: timemin, timemax
        real (r4), allocatable, dimension(:) :: phamed
        integer (i2), allocatable, dimension(:) :: rawx, rawy, yextent, type, badflag, phamin, phamax
    contains
        procedure, non_overridable :: read => read_tBadpix
        procedure, non_overridable :: write => write_tBadpix
        procedure, non_overridable :: sort => sort_tBadpix
        procedure, non_overridable :: free => free_tBadpix
        procedure, non_overridable :: init => init_tBadpix
        procedure, non_overridable :: detect_same_row => detect_same_row_tBadpix
        procedure, non_overridable :: fail_on_duplicates => fail_on_duplicates_tBadpix
        procedure, non_overridable :: delete_duplicates => delete_duplicates_tBadpix
        procedure, non_overridable :: apply_index_cols => apply_index_cols_tBadpix
        procedure, non_overridable :: create_extension => create_extension_tBadpix
    end type tBadpix

    public tBadpix, badpix_ext, alloc_badpix, dealloc_badpix, merge_badpix

contains

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> dealloacte badpix structures
    !
    !
    !
    !> @param[inout] badpix
    !> @param[inout] status
    !---------------------------------------------------------------------------
    subroutine dealloc_badpix(badpix, status)

        integer :: stat
        integer, intent(inout) :: status

        type(tBadpix), allocatable, dimension(:) :: badpix

        character(*), parameter :: function = "dealloc_badpix"

        if(allocated(badpix)) then
            deallocate(badpix, stat=stat)
            call FError(function, stat, "dealloc badpix type", status)
        end if

    end subroutine dealloc_badpix

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> alloacte n badpix structures
    !
    !
    !
    !> @param[inout] badpix
    !> @param[in] n
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine alloc_badpix(badpix, n, status)
        implicit none

        integer :: stat
        integer, intent(in) :: n
        integer, intent(inout) :: status

        type(tBadpix), allocatable, dimension(:) :: badpix

        character(buffsize) :: errmsg
        character(*), parameter :: function = "alloc_badpix"

        call dealloc_badpix(badpix, status)
        allocate(badpix(n), stat=stat, errmsg=errmsg)
        call FError(function, stat, errmsg, status)

    end subroutine alloc_badpix

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> create the merged badpix extensions for all infiles
    !
    !
    !
    !> @param[in] EvtenFile_out
    !> @param[out] files
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine merge_badpix(badpix_in, badpix_out, status)
        implicit none

        integer, intent(inout) :: status
        integer(i4) :: ii, nelements, a, b

        type(tBadpix), intent(inout) :: badpix_out
        !        type(tBadpix), intent(inout), dimension(:) :: badpix_in
        type(tBadpix), intent(in), dimension(:) :: badpix_in

        character(*), parameter :: function = "merge_badpix"

        if(status > 100 .or. status == s_errRep) return

        nelements = 0
        do ii = 1, size(badpix_in)
            nelements = nelements + badpix_in(ii)%nelements
        end do
        ! reallocate memory for columns
        call badpix_out%init(nelements, status)
        badpix_out%nelements = nelements
        badpix_out%ccdnr = badpix_in(1)%ccdnr
        badpix_out%extname = badpix_in(1)%extname
        badpix_out%mjdref = badpix_in(1)%mjdref

        if(nelements == 0) return

        ! merge all columns
        a = 1
        do ii = 1, size(badpix_in)
            b = a + badpix_in(ii)%nelements - 1
            badpix_out%timemin(a:b) = badpix_in(ii)%timemin
            badpix_out%timemax(a:b) = badpix_in(ii)%timemax
            badpix_out%phamed(a:b) = badpix_in(ii)%phamed
            badpix_out%rawx(a:b) = badpix_in(ii)%rawx
            badpix_out%rawy(a:b) = badpix_in(ii)%rawy
            badpix_out%yextent(a:b) = badpix_in(ii)%yextent
            badpix_out%type(a:b) = badpix_in(ii)%type
            badpix_out%badflag(a:b) = badpix_in(ii)%badflag
            badpix_out%phamin(a:b) = badpix_in(ii)%phamin
            badpix_out%phamax(a:b) = badpix_in(ii)%phamax

            call badpix_in(ii)%free(status)
            if(status == s_errRep) return
            a = b + 1
        end do

    end subroutine merge_badpix

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> sort the type entries according to one array
    !
    !
    !
    !> @param[in] this
    !> @param[in] array
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine sort_tBadpix(this, array, status, ascend)
        implicit none

        integer, intent(inout) :: status
        integer, allocatable, dimension(:) :: order
        integer, optional :: ascend

        class(tBadpix) :: this

        class(*), intent(in), dimension(:) :: array

        integer(i2), allocatable, dimension(:) :: tmparray_i2
        real(r4), allocatable, dimension(:) :: tmparray_r4
        real(r8), allocatable, dimension(:) :: tmparray_r8

        character(*), parameter :: function = "sort_tBadpix"

        if(status > 100 .or. status == s_errRep) return

        if(size(this%timemin) /= size(array)) then
            call wrerr(function // " tBadpix type", s_err, "length of sorting array does not conform to length of type")
            status = s_errRep
        end if

        select type (array)
            type is (integer(i2))
            call alloc(tmparray_i2, size(array), status, array)
            call sort(tmparray_i2, status, order)
            if(present(ascend)) then
                if(ascend == false) call reverse_array(order)
            end if

            type is (real(r4))
            call alloc(tmparray_r4, size(array), status, array)
            call sort(tmparray_r4, status, order)
            if(present(ascend)) then
                if(ascend == false) call reverse_array(order)
            end if

            type is (real(r8))
            call alloc(tmparray_r8, size(array), status, array)

            call sort(tmparray_r8, status, order)
            if(present(ascend)) then
                if(ascend == false) call reverse_array(order)
            end if

        end select

        this%timemin = this%timemin(order)
        this%timemax = this%timemax(order)
        this%phamed = this%phamed(order)
        this%yextent = this%yextent(order)
        this%type = this%type(order)
        this%badflag = this%badflag(order)
        this%phamin = this%phamin(order)
        this%phamax = this%phamax(order)
        this%rawx = this%rawx(order)
        this%rawy = this%rawy(order)

        call dealloc(order, status)
        call dealloc(tmparray_i2, status)
        call dealloc(tmparray_r4, status)
        call dealloc(tmparray_r8, status)

    end subroutine sort_tBadpix

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> read the values from the Badpix extension
    !
    !
    !
    !> @param[in] this (tBadpix)
    !> @param[in] unit, ccdnr
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_tBadpix(this, unit, ccdnr, success, status, mjdref)
        implicit none

        class(tBadpix) :: this

        integer :: nelements, rowsize, frow, nrows, lrow, felem
        integer :: rawx_colnum, rawy_colnum, yextent_colnum, type_colnum, badflag_colnum, phamin_colnum, phamax_colnum
        integer :: timemin_colnum, timemax_colnum, phamed_colnum
        integer, intent(in) :: ccdnr
        integer, intent(inout) :: unit, status
        integer, intent(out) :: success
        real(r8), intent(in), optional :: mjdref

        character(buffsize)  :: telid_str
        character(*), parameter :: function = "read_tBadpix"

        logical :: anyf

        if(status > 100 .or. status == s_errRep) return

        ! convert int to string
        telid_str = int2str(ccdnr, status)
        if(status == s_errRep) return

        success = false
        status = ignore
        call ffits_movnam_hdu(unit, BINARY_TBL, badpix_ext // trim(telid_str), status)
        if(status == ignore) then
            status = ok
        else
            status = ok
            return
        end if

        call ffits_get_num_rows(unit, nelements, status)
        call ffits_get_rowsize(unit, rowsize, status)

        call this%init(nelements, status)

        this%ccdnr = ccdnr
        this%nelements = nelements
        this%extname = badpix_ext // trim(telid_str)

        ! if no mjdref in the extension is found take the optional mjdref
        status = ignore
        call ffits_read_key(unit, "MJDREF", this%mjdref, status)
        if(status == ignore) then
            status = ok
        else
            status = ok
            if(present(mjdref)) this%mjdref = mjdref
        end if

        call ffits_get_colnum(unit, .false., timemin_col, timemin_colnum, status)
        call ffits_get_colnum(unit, .false., timemax_col, timemax_colnum, status)
        call ffits_get_colnum(unit, .false., phamed_col, phamed_colnum, status)
        call ffits_get_colnum(unit, .false., rawx_col, rawx_colnum, status)
        call ffits_get_colnum(unit, .false., rawy_col, rawy_colnum, status)
        call ffits_get_colnum(unit, .false., yextent_col, yextent_colnum, status)
        call ffits_get_colnum(unit, .false., type_col, type_colnum, status)
        call ffits_get_colnum(unit, .false., badflag_col, badflag_colnum, status)
        call ffits_get_colnum(unit, .false., phamin_col, phamin_colnum, status)
        call ffits_get_colnum(unit, .false., phamax_col, phamax_colnum, status)

        anyf = .false.
        felem = 1

        if(nelements == 0 .or. status == s_errRep) then
            if(status == ok) success = true
            return
        end if

        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1
            call ffits_read_col(unit, timemin_colnum, frow, felem, nrows, 0, this%timemin(frow:lrow), status, anyf)
            call ffits_read_col(unit, timemax_colnum, frow, felem, nrows, 0, this%timemax(frow:lrow), status, anyf)
            call ffits_read_col(unit, phamed_colnum, frow, felem, nrows, 0, this%phamed(frow:lrow), status, anyf)
            call ffits_read_col(unit, rawx_colnum, frow, felem, nrows, 0, this%rawx(frow:lrow), status, anyf)
            call ffits_read_col(unit, rawy_colnum, frow, felem, nrows, 0, this%rawy(frow:lrow), status, anyf)
            call ffits_read_col(unit, yextent_colnum, frow, felem, nrows, 0, this%yextent(frow:lrow), status, anyf)
            call ffits_read_col(unit, type_colnum, frow, felem, nrows, 0, this%type(frow:lrow), status, anyf)
            call ffits_read_col(unit, badflag_colnum, frow, felem, nrows, 0, this%badflag(frow:lrow), status, anyf)
            call ffits_read_col(unit, phamin_colnum, frow, felem, nrows, 0, this%phamin(frow:lrow), status, anyf)
            call ffits_read_col(unit, phamax_colnum, frow, felem, nrows, 0, this%phamax(frow:lrow), status, anyf)
        end do

        if(status == ok) success = true

    end subroutine read_tBadpix

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> write in the columns of an badpix extension
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_tBadpix(this, unit, status)
        implicit none

        integer :: nelements, rowsize, frow, nrows, lrow, felem
        integer :: rawx_colnum, rawy_colnum, yextent_colnum, type_colnum, badflag_colnum, phamin_colnum, phamax_colnum
        integer :: timemin_colnum, timemax_colnum, phamed_colnum
        integer, intent(inout) :: status
        integer, intent(in) :: unit

        class(tBadpix) :: this

        character(*), parameter :: function = "write_tBadpix"

        if(status > 100 .or. status == s_errRep) return

        ! move to the correct extension
        status = ignore
        call ffits_movnam_hdu(unit, BINARY_TBL, this%extname, status)
        if(status /= ignore) then
            status = ok
            call this%create_extension(unit, status)
        else
            status = ok
        end if

        call ffits_get_rowsize(unit, rowsize, status)

        nelements = this%nelements

        call ffits_get_colnum(unit, .false., timemin_col, timemin_colnum, status)
        call ffits_get_colnum(unit, .false., timemax_col, timemax_colnum, status)
        call ffits_get_colnum(unit, .false., phamed_col, phamed_colnum, status)
        call ffits_get_colnum(unit, .false., rawx_col, rawx_colnum, status)
        call ffits_get_colnum(unit, .false., rawy_col, rawy_colnum, status)
        call ffits_get_colnum(unit, .false., yextent_col, yextent_colnum, status)
        call ffits_get_colnum(unit, .false., type_col, type_colnum, status)
        call ffits_get_colnum(unit, .false., badflag_col, badflag_colnum, status)
        call ffits_get_colnum(unit, .false., phamin_col, phamin_colnum, status)
        call ffits_get_colnum(unit, .false., phamax_col, phamax_colnum, status)

        felem = 1

        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1
            call ffits_write_col(unit, timemin_colnum, frow, felem, nrows, this%timemin(frow:lrow), status)
            call ffits_write_col(unit, timemax_colnum, frow, felem, nrows, this%timemax(frow:lrow), status)
            call ffits_write_col(unit, phamed_colnum, frow, felem, nrows, this%phamed(frow:lrow), status)
            call ffits_write_col(unit, rawx_colnum, frow, felem, nrows, this%rawx(frow:lrow), status)
            call ffits_write_col(unit, rawy_colnum, frow, felem, nrows, this%rawy(frow:lrow), status)
            call ffits_write_col(unit, yextent_colnum, frow, felem, nrows, this%yextent(frow:lrow), status)
            call ffits_write_col(unit, type_colnum, frow, felem, nrows, this%type(frow:lrow), status)
            call ffits_write_col(unit, badflag_colnum, frow, felem, nrows, this%badflag(frow:lrow), status)
            call ffits_write_col(unit, phamin_colnum, frow, felem, nrows, this%phamin(frow:lrow), status)
            call ffits_write_col(unit, phamax_colnum, frow, felem, nrows, this%phamax(frow:lrow), status)
        end do

        call ffits_update_key(unit, "MJDREF", this%mjdref, "[d] updated by evtool", status)

    end subroutine write_tBadpix

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> create a basic badpix extension in the file
    !
    !
    !
    !> @param[in] this
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine create_extension_tBadpix(this, unit, status)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) :: unit

        character(*), parameter :: function = "create_extension_tBadpix"
        character(fcolnamelength), dimension(10) :: ttype
        character(fcolnamelength), dimension(10) :: tform
        character(fcolnamelength), dimension(10) :: tunit

        class(tBadpix) :: this
        type(tBinTable) :: table

        if(status > 100 .or. status == s_errRep) return

        ttype(1) = rawx_col
        ttype(2) = rawy_col
        ttype(3) = yextent_col
        ttype(4) = type_col
        ttype(5) = badflag_col
        ttype(6) = timemin_col
        ttype(7) = timemax_col
        ttype(8) = phamin_col
        ttype(9) = phamax_col
        ttype(10) = phamed_col

        tform = ["I","I","I","I","I","D","D","I","I","E"]
        tunit(:) = ""

        call table%init(ttype, tform, tunit, status)
        call table%insertFits(unit, this%extname, status);

    end subroutine create_extension_tBadpix

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> reduce all entries in the columns to contain only a subset of the complete
    !> by providing an array list with the desired indices
    !
    !
    !> @param[in] this
    !> @param[in] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine apply_index_cols_tBadpix(this, order, status)
        implicit none

        integer, intent(inout) :: status

        class(tBadpix) :: this

        integer, dimension(:), intent(in) :: order
        integer :: n

        character(*), parameter :: function = "apply_index_cols_tBadpix"

        if(status > 100 .or. status == s_errRep) return

        n = size(order)
        if(n == 0) then
            call wrchatty(function, trim(this%extname) // " was reduced to zero entries")
            call alloc(this%timemin, n, status)
            call alloc(this%timemax, n, status)
            call alloc(this%phamed, n, status)
            call alloc(this%yextent, n, status)
            call alloc(this%type, n, status)
            call alloc(this%badflag, n, status)
            call alloc(this%phamin, n, status)
            call alloc(this%phamax, n, status)
            call alloc(this%rawx, n, status)
            call alloc(this%rawy, n, status)
     return
        end if

        this%timemin = this%timemin(order)
        this%timemax = this%timemax(order)
        this%phamed = this%phamed(order)
        this%rawx = this%rawx(order)
        this%rawy = this%rawy(order)
        this%yextent = this%yextent(order)
        this%type = this%type(order)
        this%badflag = this%badflag(order)
        this%phamin = this%phamin(order)
        this%phamax = this%phamax(order)
        this%nelements = n

    end subroutine apply_index_cols_tBadpix

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> delete duplicate entries in the tBadpix type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

  subroutine detect_same_row_tBadpix(this, ri1, ri2, colIsSame, status)
    implicit none

    class(tBadpix) :: this
    integer,                intent(in)    :: ri1, ri2
    logical, dimension(10), intent(out)   :: colIsSame
    integer,                intent(inout) :: status

    integer :: ci
    character(*), parameter :: function = "detect_same_row_tBadpix"

    colIsSame = .false.

    ci = 1
    if(this%rawx(ri1) == this%rawx(ri2))&
      colIsSame(ci) = .true.

    ci = ci + 1
    if(this%rawy(ri1) == this%rawy(ri2))&
      colIsSame(ci) = .true.

    ci = ci + 1
    if(this%yextent(ri1) == this%yextent(ri2))&
      colIsSame(ci) = .true.

    ci = ci + 1
    if(this%type(ri1) == this%type(ri2))&
      colIsSame(ci) = .true.

    ci = ci + 1
    if(this%badflag(ri1) == this%badflag(ri2))&
      colIsSame(ci) = .true.

    ci = ci + 1
    if(this%phamin(ri1) == this%phamin(ri2))&
      colIsSame(ci) = .true.

    ci = ci + 1
    if(this%phamax(ri1) == this%phamax(ri2))&
      colIsSame(ci) = .true.

    ci = ci + 1
    if(abs(this%phamed(ri1) - this%phamed(ri2)) < epsilon(this%phamed(1)))&
      colIsSame(ci) = .true.

    ci = ci + 1
    if(abs(this%timemin(ri1) - this%timemin(ri2)) < epsilon(this%timemin(1)))&
      colIsSame(ci) = .true.

    ci = ci + 1
    if(abs(this%timemax(ri1) - this%timemax(ri2)) < epsilon(this%timemax(1)))&
      colIsSame(ci) = .true.

  end subroutine detect_same_row_tBadpix

  subroutine fail_on_duplicates_tBadpix(this, status)
    implicit none

    class(tBadpix) :: this
    integer, intent(inout) :: status

    integer, dimension(:), allocatable :: order
    logical, dimension(this%nelements) :: mask
    integer :: ii
    logical, dimension(10) :: colIsSame
    real(r8), dimension(:), allocatable :: timemin, timemax
    character(buffsize) :: errStr='',errStrBase
    logical :: identityFound
    character(*), parameter :: function = "fail_on_duplicates_tBadpix"

    if(status > 100 .or. status == s_errRep) return

    call this%sort(this%timemax, status)
    call this%sort(this%timemin, status)
    call this%sort(this%phamed, status)
    call this%sort(this%yextent, status)
    call this%sort(this%type, status)
    call this%sort(this%badflag, status)
    call this%sort(this%phamin, status)
    call this%sort(this%phamax, status)
    call this%sort(this%rawy, status)
    call this%sort(this%rawx, status)

    if(this%nelements < 2) return

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]

    identityFound = .false. ! default
    mask = .true.
    do ii = 2, this%nelements
      call this%detect_same_row(ii, ii-1, colIsSame, status)

      ! bad pixel found at the same position
      if(colIsSame(1) .and. colIsSame(2)) then
        write(errStrBase, *) "Bad pix at rawx " //trim(int2str(this%rawx(ii)))// " rawy "&
          //trim(int2str(this%rawy(ii)))

        ! check whether all other settings beside time intervals are identical
        if(colIsSame(3)) then
          write(errStr, *) trim(errStrBase)//" - different yextent values"
          call wrerr(function, s_err, errStr)
          identityFound = .true.
        end if

        if(colIsSame(4)) then
          write(errStr, *) trim(errStrBase)//" - different type values"
          call wrerr(function, s_err, errStr)
          identityFound = .true.
        end if

        if(colIsSame(5)) then
          write(errStr, *) trim(errStrBase)//" - different badflag values"
          call wrerr(function, s_err, errStr)
          identityFound = .true.
        end if

        if(colIsSame(6)) then
          write(errStr, *) trim(errStrBase)//" - different phamin values"
          call wrerr(function, s_err, errStr)
          identityFound = .true.
        end if

        if(colIsSame(7)) then
          write(errStr, *) trim(errStrBase)//" - different phamax values"
          call wrerr(function, s_err, errStr)
          identityFound = .true.
        end if

        if(colIsSame(8)) then
          write(errStr, *) trim(errStrBase)//" - different phamed values"
          call wrerr(function, s_err, errStr)
          identityFound = .true.
        end if

        if(identityFound) then
          status = s_errRep
  return
        end if

        ! if all other settings are fine merge timemin timemax

        call merge_or_r8([this%timemin(ii-1), this%timemin(ii)], [this%timemax(ii-1), this%timemax(ii)], 2, &
          [this%timemin(ii-1), this%timemin(ii)], [this%timemax(ii-1), this%timemax(ii)], 2, timemin, timemax, status)

        if(size(timemin) > 2 .and. size(timemax) > 2 ) then
          call wrerr(function, s_err, "merging times here should not have more intervals")
          status = s_errRep
  return
        end if

        if(size(timemin) == 1 .and. size(timemax) == 1 ) then
          mask(ii) = .false.
          this%timemin(ii-1) = timemin(1)
          this%timemax(ii-1) = timemax(1)
        end if

      end if
    end do

    order = pack(order, mask)
    call this%apply_index_cols(order, status)
    call dealloc(order, status)

  end subroutine fail_on_duplicates_tBadpix

  subroutine delete_duplicates_tBadpix(this, status)
    implicit none

    class(tBadpix) :: this
    integer, intent(inout) :: status

    integer, dimension(:), allocatable :: order
    logical, dimension(this%nelements) :: mask
    integer :: ii
    logical, dimension(10) :: colIsSame
    real(r8), dimension(:), allocatable :: timemin, timemax
!    character(buffsize) :: errStr='',errStrBase,nelements_dup_str
    character(buffsize) :: nelements_dup_str
    character(*), parameter :: function = "delete_duplicates_tBadpix"

    if(status > 100 .or. status == s_errRep) return

    call this%sort(this%timemax, status)
    call this%sort(this%timemin, status)
    call this%sort(this%phamed, status)
    call this%sort(this%yextent, status)
    call this%sort(this%type, status)
    call this%sort(this%badflag, status)
    call this%sort(this%phamin, status)
    call this%sort(this%phamax, status)
    call this%sort(this%rawy, status)
    call this%sort(this%rawx, status)

    if(this%nelements < 2) return

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]

    mask = .true.
    do ii = 2, this%nelements
      call this%detect_same_row(ii, ii-1, colIsSame, status)

      if(all(colIsSame))&
        mask(ii) = .false.
    end do

    order = pack(order, mask)

    if(this%nelements > size(order)) then
      nelements_dup_str = int2str(this%nelements - size(order), status)
      call wrerr(function, warn, "Number of duplicate BADPIX rows removed: " // trim(nelements_dup_str))
      call this%apply_index_cols(order, status)
    end if

    call dealloc(order, status)

  end subroutine delete_duplicates_tBadpix

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> free the content of tBadpix type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tBadpix(this, status)
        implicit none

        integer, intent(inout) :: status

        class(tBadpix) :: this

        character(*), parameter :: function = "free_tBadpix"

        call dealloc(this%timemin, status)
        call dealloc(this%timemax, status)
        call dealloc(this%phamed, status)
        call dealloc(this%yextent, status)
        call dealloc(this%type, status)
        call dealloc(this%badflag, status)
        call dealloc(this%phamin, status)
        call dealloc(this%phamax, status)
        call dealloc(this%rawx, status)
        call dealloc(this%rawy, status)

    end subroutine free_tBadpix

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> alloc the content of tBadpix type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine init_tBadpix(this, n, status)
        implicit none

        integer, intent(in) :: n
        integer, intent(inout) :: status

        class(tBadpix) :: this

        character(*), parameter :: function = "init_tBadpix"

        call alloc(this%timemin, n, status)
        call alloc(this%timemax, n, status)
        call alloc(this%phamed, n, status)
        call alloc(this%yextent, n, status)
        call alloc(this%type, n, status)
        call alloc(this%badflag, n, status)
        call alloc(this%phamin, n, status)
        call alloc(this%phamax, n, status)
        call alloc(this%rawx, n, status)
        call alloc(this%rawy, n, status)

        this%nelements = n
        this%ccdnr = -1
        this%extname = ""
        this%mjdref = 0_r8

    end subroutine init_tBadpix

end module evtool_Badpix_mod
