! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module evtool_corrat_mod

  use erolib_str_mod
  use erolib_ffits_mod
  use erolib_general_mod
  use erolib_sort_mod
  use erolib_math_mod
  use erolib_region_mod
  use evtool_gti_mod
  use erolib_tbintable_mod
  use srctool_generic , only : coord_type, COORDSYS_CODE_ICRS
  use srctool_generic_regions , only : region_type, srctool_regions_test_if_coord_in_region

  implicit none

  private

  real(r8), parameter :: d_gap = 10.0_r8

  character(*), parameter :: corratt_ext = "CORRATT"

  character(*), parameter :: ra_col = "RA"
  character(*), parameter :: dec_col = "DEC"
  character(*), parameter :: time_col = "TIME"
  character(*), parameter :: roll_col =  "ROLL"

  character(*), parameter :: region_gti_ext = "REGGTI"

  type tCorrat
    real(r8) :: mjdref = 0.0
    integer :: nelements = 0
    integer :: interpol = 0
    integer :: ccdnr = 0
    character(buffsize) :: extname = ""
    real (dp), allocatable, dimension(:) :: time, ra, dec, roll
  contains
    procedure, non_overridable :: read => read_tCorrat
    procedure, non_overridable :: write => write_tCorrat
    procedure, non_overridable :: create_extension => create_extension_tCorrat
    procedure, non_overridable :: sort => sort_tCorrat
    procedure, non_overridable :: interpolate => interpolate_tCorrat
    procedure, non_overridable :: free => free_tCorrat
    procedure, non_overridable :: init => init_tCorrat
    procedure, non_overridable :: delete_duplicates => delete_duplicates_tCorrat
    procedure, non_overridable :: derive_region_GTI => derive_region_GTI_tCorrat
    procedure, non_overridable :: derive_region_GTI_old => derive_region_GTI_tCorrat_old
    procedure, non_overridable :: apply_index_cols => apply_index_cols_tCorrat
    procedure, non_overridable :: apply_gti => apply_gti_tCorrat
  end type tCorrat

  public :: tCorrat, corratt_ext, alloc_corrat, dealloc_corrat, merge_corrat

contains
  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> dealloacte corrat structures
  !
  !
  !
  !> @param[inout] corrat
  !> @param[inout] status
  !---------------------------------------------------------------------------
  subroutine dealloc_corrat(corrat, status)

    integer :: stat
    integer, intent(inout) :: status

    type(tCorrat), allocatable, dimension(:) :: corrat

    character(*), parameter :: function = "dealloc_corrat"

    if(allocated(corrat)) then
      deallocate(corrat, stat=stat)
      call FError(function, stat, "dealloc corrat type", status)
    end if

  end subroutine dealloc_corrat

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloacte n corrat structures
  !
  !
  !
  !> @param[inout] corrat
  !> @param[in] n
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine alloc_corrat(corrat, n, status)
    implicit none

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    type(tCorrat), allocatable, dimension(:) :: corrat

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_corrat"

    call dealloc_corrat(corrat, status)
    allocate(corrat(n), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

  end subroutine alloc_corrat

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create the merged corrat extensions for all infiles
  !
  !
  !
  !> @param[in] EvtenFile_out
  !> @param[out] files
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine merge_Corrat(corrat_in, corrat_out, status)
    implicit none

    integer, intent(inout) :: status
    integer(i4) :: ii, nelements, a, b

    type(tCorrat), intent(out) :: corrat_out
    !        type(tCorrat), intent(inout), dimension(:) :: corrat_in
    type(tCorrat), intent(in), dimension(:) :: corrat_in

    character(*), parameter :: function = "merge_Corrat"

    if(status > 100 .or. status == s_errRep) return

    if(size(corrat_in) == 1) then
      corrat_out = corrat_in(1)
      return
    end if

    nelements = 0
    do ii = 1, size(corrat_in)
      nelements = nelements + corrat_in(ii)%nelements
    end do

    ! reallocate memory for columns
    call corrat_out%init(nelements, status)

    corrat_out%nelements = nelements
    corrat_out%ccdnr = corrat_in(1)%ccdnr
    corrat_out%extname = corrat_in(1)%extname
    corrat_out%mjdref = corrat_in(1)%mjdref

    if(nelements == 0) return

    ! merge all columns
    a = 1
    do ii = 1, size(corrat_in)
      b = a + corrat_in(ii)%nelements - 1
      corrat_out%time(a:b) = corrat_in(ii)%time
      corrat_out%ra(a:b) = corrat_in(ii)%ra
      corrat_out%dec(a:b) = corrat_in(ii)%dec
      corrat_out%roll(a:b) = corrat_in(ii)%roll
      if(status == s_errRep) return
      a = b + 1
    end do

  end subroutine merge_Corrat

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> derive the GTI of corrat extension when filtered with a region
  !
  !
  !
  !> @param[in] this
  !> @param[out] region
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine derive_region_GTI_tCorrat_old(this, regtable, gti, status)
    implicit none

    integer, intent(inout) :: status
    integer, dimension(:), allocatable :: mask, order
    integer :: nelements
    integer :: ii

    class(tCorrat) :: this

    type(tCorrat) :: iCorrat

    type(region_type) :: srctool_region
    type(coord_type) :: srctool_coord
    type(tRegTable), intent(in) :: regtable
    type(tGTI), intent(out) :: gti

    character(*), parameter :: function = "derive_region_GTI_tCorrat_old"

    if(status > 100 .or. status == s_errRep) return

    ! only if a region is found the region gti should be valid
    if(allocated(regtable%Region)) then
      call regtable%get_srctool_encap_circle_region(srctool_region, status)

      call this%interpolate(iCorrat, frametime, status)

      nelements = size(iCorrat%ra)

      call alloc(order, nelements, status)
      call alloc(mask, nelements, status)

      order = [(ii,ii=1,nelements,1)]
      mask = false

      do ii = 1, nelements
        srctool_coord%long = iCorrat%ra(ii)
        srctool_coord%lat = iCorrat%dec(ii)
        srctool_coord%coordsys_code = COORDSYS_CODE_ICRS
        call srctool_regions_test_if_coord_in_region(srctool_coord, srctool_region, mask(ii), status)
        if(status /= ok) status = s_errRep
      end do

      order = pack(order, mask == true)
      call create_GTI_from_order(iCorrat%time, order, gti, status)
        ! otherwise nothing should change
        ! TODO replace this with GTI init / alloc functions and manually set the elemnts to min and max of time
    else
      nelements = size(this%ra)
      call alloc(order, nelements, status)
      call alloc(mask, nelements, status)
      order = [(ii,ii=1,nelements,1)]
      call create_GTI_from_order(this%time, order, gti, status)
    end if

    gti%mjdref = this%mjdref
    gti%ccdnr = this%ccdnr
    gti%extname = region_gti_ext // trim(int2str(gti%ccdnr, status))

    call dealloc(order, status)
    call dealloc(mask, status)


  end subroutine derive_region_GTI_tCorrat_old

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> derive the GTI of corrat extension when filtered with a region
  !
  !
  !
  !> @param[in] this, regtable, overlap
  !> @param[out] gti
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine derive_region_GTI_tCorrat(this, regtable, gti, overlap, status)
    implicit none

    class(tCorrat) :: this
    type(tRegTable), intent(in) :: regtable
    type(tGTI), intent(out) :: gti
    real(r8), intent(in) :: overlap
    integer, intent(inout) :: status

    character(*), parameter :: function = "derive_region_GTI_tCorrat"
    integer :: nContSeq,ii,jj,dummyI,firstInsideI,lastInsideI,nContSeqInReg,nGTIs
    integer, dimension(this%nelements) :: csStartIs,csStopIs,csirStartIs,csirStopIs
    logical, dimension(this%nelements) :: isInside
    type(region_type) :: srctool_region
    type(coord_type) :: srctool_coord
    character(buffsize) :: comment
    logical :: presentlyInsideReg

    if(status > 100 .or. status == s_errRep) return

    ! Only if a region is found is the gti valid.
    !
    if(.not.allocated(regtable%Region) .or. this%nelements<1) then
      call gti%init(0, status)
  return
    end if

    ! Find which CORRATT times fall inside or outside the region:
    !
    call regtable%get_srctool_encap_circle_region(srctool_region, status)
    srctool_coord%coordsys_code = COORDSYS_CODE_ICRS
    do ii = 1, this%nelements
      srctool_coord%long = this%ra( ii)
      srctool_coord%lat  = this%dec(ii)
      call srctool_regions_test_if_coord_in_region(srctool_coord, srctool_region, dummyI, status)
      isInside(ii) = (dummyI/=0)
    end do

    ! Determine the start and stop indices of time sequences within which successive times are separated by less than d_gap:
    !
    csStartIs(1) = 1
    nContSeq = 1
    do ii = 2, this%nelements
      if( abs(this%time(ii) - this%time(ii-1)) > d_gap) then
        csStopIs(nContSeq) = ii - 1
        nContSeq = nContSeq + 1
        csStartIs(nContSeq) = ii
      end if
    end do
    csStopIs(nContSeq) = this%nelements

    ! Now we want to divide these sequences further. We excise time gaps > d_gap out of GTIs, but essentially then we want to AND the result with GTIS for the attitude falling inside the region. We'll do that here by setting up new start and stop indices for entering or leaving respectively a period in which the attitude falls inside the region.
    !
    nContSeqInReg = 0
    do ii=1, nContSeq
      presentlyInsideReg = .false.

      do jj=csStartIs(ii),csStopIs(ii)
        if(isInside(jj)) then
          if(.not.presentlyInsideReg) then
            nContSeqInReg = nContSeqInReg + 1
            csirStartIs(nContSeqInReg) = jj
            presentlyInsideReg = .true.
          end if
        else ! att(jj) is outside the region.
          if(presentlyInsideReg) then
            csirStopIs(nContSeqInReg) = jj - 1 ! because att(jj-1) was the last att still inside the region.
            presentlyInsideReg = .false.
          end if
        end if
      end do

      if(presentlyInsideReg) then
        csirStopIs(nContSeqInReg) = csStopIs(ii)
      end if
    end do

    ! Calculate the number of GTIs. It would be nContSeqInReg, but we discard sequences which only have 1 member.
    !
    nGTIs = 0
    do ii=1, nContSeqInReg
      if(csirStopIs(ii)>csirStartIs(ii)) nGTIs = nGTIs + 1
    end do

    if(nGTIs<1) then
      call gti%init(0, status)
  return
    end if

    call gti%init(nGTIs, status)
    gti%mjdref = this%mjdref
    gti%ccdnr = this%ccdnr
    gti%extname = region_gti_ext // trim(int2str(gti%ccdnr, status))

    ! Now set the GTIs.
    !
    jj = 0
    do ii=1, nContSeqInReg
      if(csirStopIs(ii)>csirStartIs(ii)) then
        jj = jj + 1
        gti%start(jj) = this%time(csirStartIs(ii)) - overlap
        gti%stop( jj) = this%time(csirStopIs( ii)) + overlap
        if(csirStartIs(ii)>1) then
          if( abs(this%time(csirStartIs(ii)) - this%time(csirStartIs(ii)-1)) > d_gap) then
            comment = "TM"//trim(int2str(this%ccdnr,status))//": not enough corrat information before derived region gti -"
            comment = trim(comment) // " " // double2str(this%time(csirStartIs(ii)), status)
            call wrerr(function, warn, comment)
          end if
        end if
      end if
    end do

  end subroutine derive_region_GTI_tCorrat

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> reduce all entries in the columns to contain only a subset of the complete
  !> by providing an array list with the desired indices
  !
  !
  !> @param[in] this
  !> @param[in] order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_index_cols_tCorrat(this, order, status)
    implicit none

    integer, intent(inout) :: status

    class(tCorrat) :: this

    integer, dimension(:), intent(in) :: order
    integer :: n

    character(*), parameter :: function = "apply_index_cols_tCorrat"

    if(status > 100 .or. status == s_errRep) return

    n = size(order)
    if(n < 2) then
      !      call wrerr(function, warn, trim(this%extname) // " would be reduced to below two entries - but at least two entries in " // trim(this%extname) // &
      !        " are required for further processing - no filter will be applied")
      call wrchatty(function, trim(this%extname) // " would be reduced to below two entries - but at least two entries in " // trim(this%extname) // &
        " are required for further processing - no filter will be applied")
      !      call wrerr(function, warn, trim(this%extname) // " is reduced to zero entries")
      !      call alloc(this%time, n, status)
      !      call alloc(this%deadc, n, status)
      !      this%nelements = n
      return
    end if

    this%time = this%time(order)
    this%ra = this%ra(order)
    this%dec = this%dec(order)
    this%roll = this%roll(order)
    this%nelements = n

  end subroutine apply_index_cols_tCorrat

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a GTI filter to the corrat type
  !
  !
  !
  !> @param[in] this
  !> @param[in] GTI
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_gti_tCorrat(this, GTI, status)
    implicit none

    class(tCorrat) :: this
    type(tGTI), intent(in)    :: GTI
    integer,    intent(inout) :: status

    character(*), parameter :: function = "apply_gti_tCorrat"
    integer :: ii
    integer, dimension(this%nelements) :: mask
    integer, dimension(:), allocatable :: order
    type(tGTI) :: iGTI

    if(status > 100 .or. status == s_errRep) return

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]

    iGTI = GTI
    ! internally correct the GTI
    if(iGTI%use_mjdref == true) then

      if(this%mjdref /= iGTI%mjdref) then

        iGTI%start = (iGTI%mjdref - this%mjdref ) * 3600.0_r8 * 24.0_r8 + iGTI%start
        iGTI%stop = ( iGTI%mjdref - this%mjdref ) * 3600.0_r8 * 24.0_r8 + iGTI%stop
        iGTI%mjdref = this%mjdref
      end if

    end if

    mask = get_gti_mask(this%time, iGTI, status)

    order = pack(order, mask == true)

    call this%apply_index_cols(order, status)
    call dealloc(order, status)
    call iGTI%free(status)

  end subroutine apply_gti_tCorrat

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete duplicate entries in the tcorrat type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_duplicates_tCorrat(this, status)
    implicit none

    class(tCorrat) :: this

    integer, intent(inout) :: status
    integer, dimension(:), allocatable :: order
    integer, dimension(this%nelements) :: mask
    integer :: ii

    character(*), parameter :: function = "delete_duplicates_tCorrat"

    if(status > 100 .or. status == s_errRep) return

    call this%sort(this%roll, status)
    call this%sort(this%dec, status)
    call this%sort(this%ra, status)
    call this%sort(this%time, status)

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]

    if(this%nelements == 1) return

    mask = true
    do ii = 2, this%nelements
      if(abs(this%time(ii) - this%time(ii-1)) < tiny(this%time(1))) then
        if(abs(this%ra(ii) - this%ra(ii-1)) < tiny(this%ra(1))) then
          if(abs(this%dec(ii) - this%dec(ii-1)) < tiny(this%dec(1))) then
            if(abs(this%roll(ii) - this%roll(ii-1)) < tiny(this%roll(1))) then
              mask(ii) = false
            end if
          end if
        end if
      end if
    end do

    order = pack(order, mask == true)
    call this%apply_index_cols(order, status)
    call dealloc(order, status)

  end subroutine delete_duplicates_tCorrat

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> return an interpolated corrat type - with a given time resolution
  !> time gaps of more than 300s are ignored and not interpolated
  !
  !
  !> @param[in] this
  !> @param[out] time_res, iCorrat
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine interpolate_tCorrat(this, iCorrat, time_res, status)
    implicit none

    integer, intent(inout) :: status
    integer :: nbin, n, ii, ngaps

    real(r8), intent(in) :: time_res
    real(r8) :: tmin, tmax
    real(r8), dimension(:), allocatable :: ra, dec, time, roll

    class(tCorrat) :: this
    type(tCorrat), intent(inout) :: iCorrat

    character(*), parameter :: function = "interpolate_tCorrat"

    if(status > 100 .or. status == s_errRep) return

    ! the number of given values
    n = size(this%time)

    ! search for jumps in the time
    ngaps = 1

    do ii = 1, n-1
        
      if(this%time(ii + 1) - this%time(ii) > d_gap .or. ii == n - 1) then
            
        tmin = this%time(ngaps)
        tmax = this%time(ii)
        ! the number of time bins to interpolate to
        nbin = int((tmax-tmin) / time_res)
        nbin = nbin + 1

        call alloc (time, nbin, status)
        call alloc (ra, nbin, status)
        call alloc (dec, nbin, status)
        call alloc (roll, nbin, status)

        call array_loop(roll, 360.0_dp, status);        ! might change to a different interpolation of roll angle if needed

        time = [ (ii, ii=0, nbin-1)] * time_res + tmin
                
        ! latitude - longitude
        call interpol_gc_double_sorted(time, dec, ra,&
          this%time(ngaps:ii), this%dec(ngaps:ii), this%ra(ngaps:ii), status)
                    
        roll = interpol_linear_double_sorted(time, this%time(ngaps:ii), this%roll(ngaps:ii), status)
        roll = mod_double(roll, 360.0_dp)

        if(allocated(iCorrat%ra)) then
          iCorrat%ra = [iCorrat%ra, ra]
        else
          iCorrat%ra = ra
        end if

        if(allocated(iCorrat%dec)) then
          iCorrat%dec = [iCorrat%dec, dec]
        else
          iCorrat%dec = dec
        end if
        if(allocated(iCorrat%roll)) then
          iCorrat%roll = [iCorrat%roll, roll]
        else
          iCorrat%roll = roll
        end if
        if(allocated(iCorrat%time)) then
          iCorrat%time = [iCorrat%time, time]
        else
          iCorrat%time = time
        end if

        ngaps = ii + 1

      end if

    end do

    iCorrat%interpol = true

  end subroutine interpolate_tCorrat

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort the type entries according to one array
  !
  !
  !
  !> @param[in] this
  !> @param[in] array
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_tCorrat(this, array, status, ascend)
    implicit none

    integer, intent(inout) :: status
    integer, allocatable, dimension(:) :: order
    integer, optional :: ascend

    class(tCorrat) :: this

    real(dp), intent(in), dimension(:) :: array
    real(dp), allocatable, dimension(:) :: tmparray

    character(*), parameter :: function = "sort_tCorrat"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%time) /= size(array)) then
      call wrerr(function // " tCorrat type", s_err, "length of sorting array does not conform to length of type")
      status = s_errRep
    end if

    tmparray = array

    call sort(tmparray, status, order)
    if(present(ascend)) then
      if(ascend == false) call reverse_array(order)
    end if

    this%time = this%time(order)
    this%ra = this%ra(order)
    this%dec = this%dec(order)
    this%roll = this%roll(order)

    call dealloc(order, status)
    call dealloc(tmparray, status)

  end subroutine sort_tCorrat

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the values from the corrat extension
  !
  !
  !
  !> @param[in] this (tCorrat)
  !> @param[in] unit, ccdnr
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_tCorrat(this, unit, ccdnr, success, status, mjdref)
    implicit none

    class(tCorrat) :: this

    integer :: nelements, rowsize, frow, nrows, lrow, felem
    integer :: time_colnum, roll_colnum, ra_colnum, dec_colnum
    integer, intent(in) :: ccdnr
    integer, intent(inout) :: unit, status
    integer, intent(out) :: success

    real(r8), optional, intent(in) :: mjdref

    character(buffsize)  :: telid_str
    character(*), parameter :: function = "read_tCorrat"

    logical :: anyf

    if(status > 100 .or. status == s_errRep) return

    ! convert int to string
    telid_str = int2str(ccdnr, status)
    if(status == s_errRep) return

    success = false
    status = ignore
    call ffits_movnam_hdu(unit, BINARY_TBL, corratt_ext // trim(telid_str), status)
    if(status == ignore) then
      status = ok
    else
      status = ok
      return
    end if

    call ffits_get_num_rows(unit, nelements, status)
    call ffits_get_rowsize(unit, rowsize, status)

    call this%init(nelements, status)

    this%ccdnr = ccdnr
    this%nelements = nelements
    this%extname = corratt_ext // trim(telid_str)

    ! if no mjdref in the extension is found take the optional mjdref
    status = ignore
    call ffits_read_key(unit, "MJDREF", this%mjdref, status)
    if(status == ignore) then
      status = ok
    else
      status = ok
      if(present(mjdref)) this%mjdref = mjdref
    end if

    call ffits_get_colnum(unit, .false., time_col, time_colnum, status)
    call ffits_get_colnum(unit, .false., roll_col, roll_colnum, status)
    call ffits_get_colnum(unit, .false., ra_col, ra_colnum, status)
    call ffits_get_colnum(unit, .false., dec_col, dec_colnum, status)

    anyf = .false.
    felem = 1

    if(nelements == 0 .or. status == s_errRep) then
      if(status == ok) success = true
      return
    end if

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1
      call ffits_read_col(unit, time_colnum, frow, felem, nrows, 0, this%time(frow:lrow), status, anyf)
      call ffits_read_col(unit, roll_colnum, frow, felem, nrows, 0, this%roll(frow:lrow), status, anyf)
      call ffits_read_col(unit, ra_colnum, frow, felem, nrows, 0, this%ra(frow:lrow), status, anyf)
      call ffits_read_col(unit, dec_colnum, frow, felem, nrows, 0, this%dec(frow:lrow), status, anyf)
    end do

    if(status == ok) success = true

  end subroutine read_tCorrat

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write in the columns of an eventfile
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_tCorrat(this, unit, status)
    implicit none

    integer :: nelements, rowsize, frow, nrows, lrow, felem
    integer :: time_colnum, roll_colnum, ra_colnum, dec_colnum
    integer, intent(inout) :: status
    integer, intent(in) :: unit

    class(tCorrat) :: this

    character(*), parameter :: function = "write_tCorrat"

    if(status > 100 .or. status == s_errRep) return

    ! move to the correct extension
    status = ignore
    call ffits_movnam_hdu(unit, BINARY_TBL, this%extname, status)
    if(status /= ignore) then
      status = ok
      call this%create_extension(unit, status)
    else
      status = ok
    end if

    call ffits_get_rowsize(unit, rowsize, status)

    nelements = this%nelements

    call ffits_get_colnum(unit, .false., time_col, time_colnum, status)
    call ffits_get_colnum(unit, .false., roll_col, roll_colnum, status)
    call ffits_get_colnum(unit, .false., ra_col, ra_colnum, status)
    call ffits_get_colnum(unit, .false., dec_col, dec_colnum, status)

    felem = 1

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1
      call ffits_write_col(unit, time_colnum, frow, felem, nrows, this%time(frow:lrow), status)
      call ffits_write_col(unit, roll_colnum, frow, felem, nrows, this%roll(frow:lrow), status)
      call ffits_write_col(unit, ra_colnum, frow, felem, nrows, this%ra(frow:lrow), status)
      call ffits_write_col(unit, dec_colnum, frow, felem, nrows, this%dec(frow:lrow), status)
    end do

    call ffits_update_key(unit, "MJDREF", this%mjdref, "[d] updated by evtool", status)

  end subroutine write_tCorrat

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create a basic corratt extension in the file
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine create_extension_tCorrat(this, unit, status)
    implicit none

    integer, intent(inout) :: status
    integer, intent(in) :: unit

    character(*), parameter :: function = "create_extension_tCorrat"
    character(fcolnamelength), dimension(4) :: ttype
    character(fcolnamelength), dimension(4) :: tform
    character(fcolnamelength), dimension(4) :: tunit

    class(tCorrat) :: this
    type(tBinTable) :: table

    if(status > 100 .or. status == s_errRep) return

    ttype(1) = time_col
    ttype(2) = ra_col
    ttype(3) = dec_col
    ttype(4) = roll_col
    tform(1) = "D"
    tform(2) = "D"
    tform(3) = "D"
    tform(4) = "D"
    tunit(1) = ""
    tunit(2) = "deg"
    tunit(3) = "deg"
    tunit(4) = "deg"

    call table%init(ttype, tform, tunit, status)
    call table%insertFits(unit, this%extname, status)

  end subroutine create_extension_tCorrat

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the content of tCorrat type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_tCorrat(this, status)
    implicit none

    integer, intent(inout) :: status

    class(tCorrat) :: this

    character(*), parameter :: function = "free_tCorrat"

    call dealloc(this%time, status)
    call dealloc(this%ra, status)
    call dealloc(this%dec, status)
    call dealloc(this%roll, status)

  end subroutine free_tCorrat

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloc the content of tCorrat type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine init_tCorrat(this, n, status)
    implicit none

    integer, intent(in) :: n
    integer, intent(inout) :: status

    class(tCorrat) :: this

    character(*), parameter :: function = "init_tCorrat"

    call alloc(this%time, n, status)
    call alloc(this%ra, n, status)
    call alloc(this%dec, n, status)
    call alloc(this%roll, n, status)

    this%nelements = n
    this%ccdnr = -1
    this%extname = ""
    this%mjdref = 0_r8

  end subroutine init_tCorrat

end module evtool_corrat_mod
