! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module evtool_deadcor_mod

  use erolib_str_mod
  use erolib_ffits_mod
  use erolib_general_mod
  use erolib_sort_mod
  use evtool_gti_mod
  use erolib_tbintable_mod
  use erolib_math_mod
  implicit none

  private

  character(*), parameter :: deadcor_ext = "DEADCOR"
  character(*), parameter :: time_col = "TIME"
  character(*), parameter :: deadc_col = "DEADC"

  type tDeadcor
    real(r8) :: mjdref = 0.0
    integer :: ccdnr = 0
    integer :: nelements = 0
    character(buffsize) :: extname = ""
    real (r8), allocatable, dimension(:) :: time
    real (r4), allocatable, dimension(:) :: deadc
  contains
    procedure, non_overridable :: read => read_tDeadcor
    procedure, non_overridable :: write => write_tDeadcor
    procedure, non_overridable :: create_extension => create_extension_tDeadcor
    procedure, non_overridable :: sort => sort_tDeadcor
    procedure, non_overridable :: free => free_tDeadcor
    procedure, non_overridable :: init => init_tDeadcor
    procedure, non_overridable :: delete_duplicates => delete_duplicates_tDeadcor
    procedure, non_overridable :: apply_index_cols => apply_index_cols_tDeadcor
    procedure, non_overridable :: apply_gti => apply_gti_tDeadcor
  end type tDeadcor

  public :: tDeadcor, deadcor_ext, alloc_deadcor, dealloc_deadcor, merge_deadcor

contains

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> dealloacte deadcor structures
  !
  !
  !
  !> @param[inout] deadcor
  !> @param[inout] status
  !---------------------------------------------------------------------------
  subroutine dealloc_deadcor(deadcor, status)

    integer :: stat
    integer, intent(inout) :: status

    type(tDeadcor), allocatable, dimension(:) :: deadcor

    character(*), parameter :: function = "dealloc_deadcor"

    if(allocated(deadcor)) then
      deallocate(deadcor, stat=stat)
      call FError(function, stat, "dealloc deadcor type", status)
    end if

  end subroutine dealloc_deadcor

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloacte n deadcor structures
  !
  !
  !
  !> @param[inout] deadcor
  !> @param[in] n
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine alloc_deadcor(deadcor, n, status)
    implicit none

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    type(tDeadcor), allocatable, dimension(:) :: deadcor

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_deadcor"

    call dealloc_deadcor(deadcor, status)
    allocate(deadcor(n), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

  end subroutine alloc_deadcor

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create the merged deadcor extensions for all infiles
  !
  !
  !
  !> @param[in] EvtenFile_out
  !> @param[out] files
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine merge_deadcor(deadcor_in, deadcor_out, status)
    implicit none

    integer, intent(inout) :: status
    integer(i4) :: ii, nelements, a, b

    type(tDeadcor), intent(out) :: deadcor_out
    !        type(tDeadcor), intent(inout), dimension(:) :: deadcor_in
    type(tDeadcor), intent(in), dimension(:) :: deadcor_in

    character(*), parameter :: function = "merge_deadcor"

    if(status > 100 .or. status == s_errRep) return

    nelements = 0
    do ii = 1, size(deadcor_in)
      nelements = nelements + deadcor_in(ii)%nelements
    end do

    ! reallocate memory for columns
    call deadcor_out%init(nelements, status)

    deadcor_out%nelements = nelements
    deadcor_out%ccdnr = deadcor_in(1)%ccdnr
    deadcor_out%extname = deadcor_in(1)%extname
    deadcor_out%mjdref = deadcor_in(1)%mjdref

    if(nelements == 0) return

    ! merge all columns
    a = 1
    do ii = 1, size(deadcor_in)
      b = a + deadcor_in(ii)%nelements - 1
      deadcor_out%time(a:b) = deadcor_in(ii)%time
      deadcor_out%deadc(a:b) = deadcor_in(ii)%deadc
      if(status == s_errRep) return
      a = b + 1
    end do

  end subroutine merge_deadcor



  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort the type entries according to one array
  !
  !
  !
  !> @param[in] this
  !> @param[in] array
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_tDeadcor(this, array, status, ascend)
    implicit none

    integer, intent(inout) :: status
    integer, allocatable, dimension(:) :: order
    integer, optional :: ascend

    class(tDeadcor) :: this


    class(*), intent(in), dimension(:) :: array
    real(r4), allocatable, dimension(:) :: tmparray_r4
    real(r8), allocatable, dimension(:) :: tmparray_r8
    !real(dp), intent(in), dimension(:) :: array
    !real(dp), allocatable, dimension(:) :: tmparray

    character(*), parameter :: function = "sort_tDeadcor"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%time) /= size(array)) then
      call wrerr(function // " tDeadcor type", s_err, "length of sorting array does not conform to length of type")
      status = s_errRep
    end if

    select type (array)
      type is (real(r4))
      call alloc(tmparray_r4, size(array), status, array)
      call sort(tmparray_r4, status, order)
      if(present(ascend)) then
        if(ascend == false) call reverse_array(order)
      end if

      type is (real(r8))
      call alloc(tmparray_r8, size(array), status, array)
      call sort(tmparray_r8, status, order)
      if(present(ascend)) then
        if(ascend == false) call reverse_array(order)
      end if

    end select

    this%time = this%time(order)
    this%deadc = this%deadc(order)

    call dealloc(order, status)
    call dealloc(tmparray_r4, status)
    call dealloc(tmparray_r8, status)

  end subroutine sort_tDeadcor

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the values from the deadcor extension
  !
  !
  !
  !> @param[in] this (tDeadcor)
  !> @param[in] unit, ccdnr
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_tDeadcor(this, unit, ccdnr, success, status, mjdref)
    implicit none

    class(tDeadcor) :: this

    integer :: nelements, rowsize, frow, nrows, lrow, felem
    integer :: time_colnum, deadc_colnum
    integer, intent(in) :: ccdnr
    integer, intent(inout) :: unit, status
    integer, intent(out) :: success

    real(r8), optional, intent(in) :: mjdref

    character(buffsize)  :: telid_str
    character(*), parameter :: function = "read_tDeadcor"

    logical :: anyf

    if(status > 100 .or. status == s_errRep) return

    ! convert int to string
    telid_str = int2str(ccdnr, status)
    if(status == s_errRep) return

    success = false
    status = ignore
    call ffits_movnam_hdu(unit, BINARY_TBL, deadcor_ext // trim(telid_str), status)
    if(status == ignore) then
      status = ok
    else
      status = ok
      return
    end if

    call ffits_get_num_rows(unit, nelements, status)
    call ffits_get_rowsize(unit, rowsize, status)

    call this%init(nelements, status)

    this%ccdnr = ccdnr
    this%nelements = nelements
    this%extname = deadcor_ext // trim(telid_str)

    ! if no mjdref in the extension is found take the optional mjdref
    status = ignore
    call ffits_read_key(unit, "MJDREF", this%mjdref, status)
    if(status == ignore) then
      status = ok
    else
      status = ok
      if(present(mjdref)) this%mjdref = mjdref
    end if

    call ffits_get_colnum(unit, .false., time_col, time_colnum, status)
    call ffits_get_colnum(unit, .false., deadc_col, deadc_colnum, status)

    anyf = .false.
    felem = 1

    if(nelements == 0 .or. status == s_errRep) then
      if(status == ok) success = true
      return
    end if

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1
      call ffits_read_col(unit, time_colnum, frow, felem, nrows, 0, this%time(frow:lrow), status, anyf)
      call ffits_read_col(unit, deadc_colnum, frow, felem, nrows, 0, this%deadc(frow:lrow), status, anyf)
    end do

    if(status == ok) then
      success = true
    else
      success = false
      call this%init(0, status)
    end if

  end subroutine read_tDeadcor

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write in the columns of an deadcor extension
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_tDeadcor(this, unit, status)
    implicit none

    integer :: nelements, rowsize, frow, nrows, lrow, felem
    integer :: time_colnum, deadc_colnum
    integer, intent(inout) :: status
    integer, intent(in) :: unit
    character(buffsize)  :: telid_str

    class(tDeadcor) :: this

    character(*), parameter :: function = "write_tDeadcor"

    if(status > 100 .or. status == s_errRep) return

    if(this%extname == "") then
      telid_str = int2str(this%ccdnr, status)
      this%extname = deadcor_ext // trim(telid_str)
    end if

    ! move to the correct extension
    status = ignore
    call ffits_movnam_hdu(unit, BINARY_TBL, this%extname, status)
    if(status /= ignore) then
      status = ok
      call this%create_extension(unit, status)
    else
      status = ok
    end if

    call ffits_get_rowsize(unit, rowsize, status)

    nelements = this%nelements

    call ffits_get_colnum(unit, .false., time_col, time_colnum, status)
    call ffits_get_colnum(unit, .false., deadc_col, deadc_colnum, status)

    felem = 1

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1
      call ffits_write_col(unit, time_colnum, frow, felem, nrows, this%time(frow:lrow), status)
      call ffits_write_col(unit, deadc_colnum, frow, felem, nrows, this%deadc(frow:lrow), status)
    end do

    call ffits_update_key(unit, "MJDREF", this%mjdref, "[d] updated by evtool", status)

  end subroutine write_tDeadcor

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create a basic deadcor extension in the file
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine create_extension_tDeadcor(this, unit, status)
    implicit none

    integer, intent(inout) :: status
    integer, intent(in) :: unit

    character(*), parameter :: function = "create_extension_tDeadcor"
    character(fcolnamelength), dimension(2) :: ttype
    character(fcolnamelength), dimension(2) :: tform
    character(fcolnamelength), dimension(2) :: tunit

    class(tDeadcor) :: this
    type(tBinTable) :: table

    if(status > 100 .or. status == s_errRep) return

    ttype(1) = time_col
    ttype(2) = deadc_col

    tform = ["D","E"]
    tunit(:) = ""

    call table%init(ttype, tform, tunit, status)
    call table%insertFits(unit, this%extname, status);

  end subroutine create_extension_tDeadcor

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> reduce all entries in the columns to contain only a subset of the complete
  !> by providing an array list with the desired indices
  !
  !
  !> @param[in] this
  !> @param[in] order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_index_cols_tDeadcor(this, order, status)
    implicit none

    integer, intent(inout) :: status

    class(tDeadcor) :: this

    integer, dimension(:), intent(in) :: order
    integer :: n

    character(*), parameter :: function = "apply_index_cols_tDeadcor"

    if(status > 100 .or. status == s_errRep) return

    n = size(order)
    if(n < 2) then
      !            call wrerr(function, warn, trim(this%extname) // " would be reduced to below two entries - but at least two entries in " // trim(this%extname) // &
      !    " are required for further processing - no filter will be applied")
      call wrchatty(function, trim(this%extname) // " would be reduced to below two entries - but at least two entries in " // trim(this%extname) // &
        " are required for further processing - no filter will be applied")
      !      call wrerr(function, warn, trim(this%extname) // " is reduced to zero entries")
      !      call alloc(this%time, n, status)
      !      call alloc(this%deadc, n, status)
      !      this%nelements = n
      return
    end if

    this%time = this%time(order)
    this%deadc = this%deadc(order)
    this%nelements = n

  end subroutine apply_index_cols_tDeadcor

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a GTI filter to the deadcor type
  !
  !
  !> @param[in] this, GTI
  !> @param[in, optional] extend_order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_gti_tDeadcor(this, GTI, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tDeadcor) :: this
    type(tGTI), intent(in) :: GTI
    type(tGTI) :: iGTI

    integer, dimension(this%nelements) :: mask
    integer, dimension(:), allocatable :: order

    character(*), parameter :: function = "apply_gti_tDeadcor"

    if(status > 100 .or. status == s_errRep) return

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]

    iGTI = GTI
    ! internally correct the GTI
    if(iGTI%use_mjdref == true) then

      if(this%mjdref /= iGTI%mjdref) then

        iGTI%start = (iGTI%mjdref - this%mjdref ) * 3600.0_r8 * 24.0_r8 + iGTI%start
        iGTI%stop = ( iGTI%mjdref - this%mjdref ) * 3600.0_r8 * 24.0_r8 + iGTI%stop
        iGTI%mjdref = this%mjdref
      end if

    end if

    mask = get_gti_mask(this%time, iGTI, status)

    order = pack(order, mask == true)

    call this%apply_index_cols(order, status)
    call dealloc(order, status)
    call iGTI%free(status)

  end subroutine apply_gti_tDeadcor

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete duplicate entries in the tDeadcor type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_duplicates_tDeadcor(this, status)
    implicit none

    class(tDeadcor) :: this

    integer, intent(inout) :: status
    integer, dimension(:), allocatable :: order
    integer, dimension(this%nelements) :: mask
    integer :: ii
    character(buffsize) :: nelements_dup_str
    character(*), parameter :: function = "delete_duplicates_tDeadcor"

    if(status > 100 .or. status == s_errRep) return

    call this%sort(this%deadc, status)
    call this%sort(this%time, status)

    if(this%nelements < 2) return

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]

    mask = true
    do ii = 2, this%nelements
      if(abs(this%time(ii) - this%time(ii-1)) < tiny(this%time(1))) then
        if(abs(this%deadc(ii) - this%deadc(ii-1)) < tiny(this%deadc(1))) then
          mask(ii) = false
        end if
      end if
    end do

    order = pack(order, mask == true)
    if(this%nelements > size(order)) then
      nelements_dup_str = int2str(this%nelements - size(order), status)
      call wrerr(function, warn, "Number of duplicate DEADCOR rows removed: " // trim(nelements_dup_str))
      call this%apply_index_cols(order, status)
    end if

    call dealloc(order, status)

  end subroutine delete_duplicates_tDeadcor


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the content of tDeadcor type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_tDeadcor(this, status)
    implicit none

    integer, intent(inout) :: status

    class(tDeadcor) :: this

    character(*), parameter :: function = "free_tDeadcor"

    call dealloc(this%time, status)
    call dealloc(this%deadc, status)

  end subroutine free_tDeadcor

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloc the content of tDeadcor type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine init_tDeadcor(this, n, status)
    implicit none

    integer, intent(in) :: n
    integer, intent(inout) :: status

    class(tDeadcor) :: this

    character(*), parameter :: function = "init_tDeadcor"

    call alloc(this%time, n, status)
    call alloc(this%deadc, n, status)

    this%nelements = n
    this%ccdnr = -1
    this%extname = ""
    this%mjdref = 0_r8

  end subroutine init_tDeadcor

end module evtool_deadcor_mod
