! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module evtool_eventfile_mod

  use param
  use erolib_general_mod
  use erolib_ffits_mod
  use erolib_region_mod
  use erolib_wcs_mod
  use evtool_gti_mod
  use evtool_events_mod
  use evtool_image_mod
  use evtool_corrat_mod
  use evtool_deadcor_mod
  use evtool_badpix_mod
  use evtool_offset_mod
  use evtool_hk_mod
  use evtool_hkx1_mod
  use evtool_hkx2_mod
  use evtool_hkx3_mod
  use evtool_hkx5_mod
  use evtool_rawxy_mod
  use fitsimage, only : stdhdr_type, fread_stdhdr, fwrite_stdhdr, wcs_imahdr_type, fread_wcshdr, fwrite_wcshdr, wcs_refhdr_type, update_stdhdr
  use srctool_generic , only : coord_type, COORDSYS_CODE_ICRS
  use srctool_generic_regions , only : srctool_regions_test_if_coord_in_region, region_type

  implicit none

logical, private, parameter :: doTest=.false.

  type tEventFile
    integer :: unit = 0 ! fitsio unit to access the file
    character(buffsize) :: name="" ! filename

    integer :: ninst = 0
    integer, dimension(:), allocatable :: ccdnr_infile
    character(buffsize), dimension(:), allocatable :: instrum

    integer :: filter_pattern = 0
    integer :: filter_energy = 0
    integer :: filter_telid = 0
    integer :: filter_gti = 0
    integer :: filter_flag = 0
    integer :: filter_region = 0
    logical :: filter_rawxy = .false.

    ! flags if extension is present in file and to be written
    ! by default no extensions are to be written
    integer :: ext_events   = false
    integer :: ext_image    = false
    integer :: ext_corrat   = false
    integer :: ext_deadcor  = false
    integer :: ext_badpix   = false
    integer :: ext_offset   = false
    integer :: ext_hk       = false
    integer :: ext_hkx1     = false
    integer :: ext_hkx2     = false
    integer :: ext_hkx3     = false
    integer :: ext_hkx5     = false
    integer :: ext_regtable = false
    integer :: ext_gti      = false
    integer :: ext_qualgti  = false
    integer :: ext_flaregti = false
    integer :: ext_regiongti = false

    type(stdhdr_type) :: std_header

    type(tGTI), dimension(:), allocatable :: gti
    type(tGTI), dimension(:), allocatable :: qualgti
    type(tGTI), dimension(:), allocatable :: flaregti
    type(tGTI) :: externalgti
    type(tGTI), dimension(:), allocatable :: regiongti

    type(tEvents) :: events
    type(tImage2D_i4) :: image
    type(tCorrat), dimension(:), allocatable :: corrat
    type(tDeadcor), dimension(:), allocatable :: deadcor
    type(tBadpix), dimension(:), allocatable :: badpix
    type(tOffset), dimension(:), allocatable :: offset
    type(tHK), dimension(:), allocatable :: hk
    type(tHKx1), dimension(:), allocatable :: hkx1
    type(tHKx2), dimension(:), allocatable :: hkx2
    type(tHKx3), dimension(:), allocatable :: hkx3
    type(tHKx5), dimension(:), allocatable :: hkx5
    type(tRegTable) :: regtable

    type(tRawXYMask) :: rawxy_mask

  contains

    procedure :: read => read_tEventFile
    procedure :: write => write_tEventFile
    procedure :: free => free_tEventFile
    procedure :: open => open_tEventFile
    procedure :: close => close_tEventFile

    procedure :: read_stdheader => read_stdheader_tEventFile
    procedure :: write_stdheader => write_stdheader_tEventFile

    procedure :: apply_filter_telid => apply_filter_telid_tEventFile
    procedure :: apply_filter_energy => apply_filter_energy_tEventFile
    procedure :: apply_filter_flag => apply_filter_flag_tEventFile
    procedure :: apply_filter_pattern => apply_filter_pattern_tEventFile
    procedure :: apply_filter_gti => apply_filter_gti_tEventFile
    procedure :: apply_filter_qualgti => apply_filter_qualgti_tEventFile
    procedure :: apply_filter_flaregti => apply_filter_flaregti_tEventFile
    procedure :: apply_filter_externalgti => apply_filter_externalgti_tEventFile
    procedure :: apply_filter_region => apply_filter_region_tEventFile
    procedure :: apply_filter_rawxy => apply_filter_rawxy_tEventFile

    procedure :: apply_region_gti => apply_region_gti_tEventFile
    procedure :: apply_stdheader => apply_stdheader_tEventFile

    procedure :: sort_events => sort_events_tEventFile
    procedure :: free_events => free_events_tEventFile
    procedure :: read_events => read_events_tEventFile
    procedure :: write_events => write_events_tEventFile
    procedure :: delete_duplicates_events => delete_duplicates_events_tEventFile

    procedure :: read_image_events => read_image_events_tEventFile
    procedure :: write_image => write_image_tEventFile
    procedure :: free_image => free_image_tEventFile

    procedure :: write_corrat => write_corrat_tEventFile
    procedure :: read_corrat => read_corrat_tEventFile
    procedure :: free_corrat => free_corrat_tEventFile
    procedure :: sort_corrat => sort_corrat_tEventFile
    procedure :: delete_duplicates_corrat => delete_duplicates_corrat_tEventFile
    procedure :: calc_gti_from_corrat_region_filter => calc_gti_from_corrat_region_filter_tEventFile

    procedure :: write_deadcor => write_deadcor_tEventFile
    procedure :: read_deadcor => read_deadcor_tEventFile
    procedure :: free_deadcor => free_deadcor_tEventFile
    procedure :: sort_deadcor => sort_deadcor_tEventFile
    procedure :: delete_duplicates_deadcor => delete_duplicates_deadcor_tEventFile

    procedure :: write_badpix => write_badpix_tEventFile
    procedure :: read_badpix => read_badpix_tEventFile
    procedure :: free_badpix => free_badpix_tEventFile
    procedure :: sort_badpix => sort_badpix_tEventFile
    procedure :: delete_duplicates_badpix => delete_duplicates_badpix_tEventFile

    procedure :: write_offset => write_offset_tEventFile
    procedure :: read_offset => read_offset_tEventFile
    procedure :: free_offset => free_offset_tEventFile
    procedure :: sort_offset => sort_offset_tEventFile
    procedure :: delete_duplicates_offset => delete_duplicates_offset_tEventFile

    procedure :: write_hk => write_hk_tEventFile
    procedure :: read_hk => read_hk_tEventFile
    procedure :: free_hk => free_hk_tEventFile
    procedure :: sort_hk => sort_hk_tEventFile
    procedure :: delete_duplicates_hk => delete_duplicates_hk_tEventFile

    procedure :: write_hkx1 => write_hkx1_tEventFile
    procedure :: read_hkx1 => read_hkx1_tEventFile
    procedure :: free_hkx1 => free_hkx1_tEventFile
    procedure :: sort_hkx1 => sort_hkx1_tEventFile

    procedure :: write_hkx2 => write_hkx2_tEventFile
    procedure :: read_hkx2 => read_hkx2_tEventFile
    procedure :: free_hkx2 => free_hkx2_tEventFile
    procedure :: sort_hkx2 => sort_hkx2_tEventFile

    procedure :: write_hkx3 => write_hkx3_tEventFile
    procedure :: read_hkx3 => read_hkx3_tEventFile
    procedure :: free_hkx3 => free_hkx3_tEventFile
    procedure :: sort_hkx3 => sort_hkx3_tEventFile

    procedure :: write_hkx5 => write_hkx5_tEventFile
    procedure :: read_hkx5 => read_hkx5_tEventFile
    procedure :: free_hkx5 => free_hkx5_tEventFile
    procedure :: sort_hkx5 => sort_hkx5_tEventFile

    procedure :: derive_region_gti => derive_region_gti_tEventFile
    procedure :: write_region => write_region_tEventFile

    procedure :: write_gti => write_gti_tEventFile
    procedure :: write_qualgti => write_qualgti_tEventFile
    procedure :: write_regiongti => write_regiongti_tEventFile
    procedure :: write_flaregti => write_flaregti_tEventFile
    procedure :: read_gti => read_gti_tEventFile
    procedure :: read_qualgti => read_qualgti_tEventFile
    procedure :: read_flaregti => read_flaregti_tEventFile

    procedure :: free_gti => free_gti_tEventFile
    procedure :: free_qualgti => free_qualgti_tEventFile
    procedure :: free_flaregti => free_flaregti_tEventFile

    procedure :: read_rawxy_sel => read_rawxy_sel_tEventFile
    procedure :: record_rawxy_sel => record_rawxy_sel_tEventFile
    procedure :: write_rawxy_sel => write_rawxy_sel_tEventFile

  end type tEventFile

  public :: tEventFile

contains

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> derive region gti based on the corrat extension and the region filter
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine derive_region_gti_tEventFile(this, overlap, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    real(r8), intent(in) :: overlap

    class(tEventFile) :: this

    character(*), parameter :: function = "derive_region_gti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%regtable%Region) == 0) return
    if(this%ext_corrat == false) then
      call wrlog(function, "region gti cannot be derived / no CORRAT ext was found")
      return
    end if

    call alloc_gti(this%regiongti, this%ninst, status)
        
    ! in the derived REGION GTI the overlap is already included
    do ii = 1, this%ninst
      call this%corrat(ii)%derive_region_GTI(this%regtable, this%regiongti(ii), overlap, status)
    end do

    if(status == ok) this%ext_regiongti = true

  end subroutine derive_region_gti_tEventFile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the gti extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] reopen
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_gti_tEventFile(this, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    class(tEventFile) :: this

    integer, dimension(this%ninst) :: success

    character(*), parameter :: function = "read_gti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call alloc_gti(this%gti, this%ninst, status)

    if(present(reopen)) then
      if(reopen == true) call this%open(readonly, status)
    end if

    do ii = 1, this%ninst
      call this%gti(ii)%read(this%unit, gti_ext, success(ii), status, ccdnr=this%ccdnr_infile(ii), mjdref=this%std_header%mjdref)
    end do

    if(present(reopen)) then
      if(reopen == true) call this%close(status)
    end if

    if( (sum(success) == 0) ) then
      this%ext_gti = false
      call alloc_gti(this%gti, 0, status)
    else if (sum(success) == this%ninst) then
      this%ext_gti = true
    else
      call wrerr(function, warn, trim(gti_ext) // " is not present for all instruments")
      this%ext_gti = false
    end if

  end subroutine read_gti_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the qualgti extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] reopen
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_qualgti_tEventFile(this, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    class(tEventFile) :: this

    integer, dimension(this%ninst) :: success

    character(*), parameter :: function = "read_qualgti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call alloc_gti(this%qualgti, this%ninst, status)

    if(present(reopen)) then
      if(reopen == true) call this%open(readonly, status)
    end if

    do ii = 1, this%ninst
      call this%qualgti(ii)%read(this%unit, qualgti_ext, success(ii), status, ccdnr=this%ccdnr_infile(ii), mjdref=this%std_header%mjdref)
    end do

    if(present(reopen)) then
      if(reopen == true) call this%close(status)
    end if

    if (sum(success) == 0) then
      call alloc_gti(this%qualgti, 0, status)
      this%ext_qualgti = false
    else if (sum(success) == this%ninst) then
      this%ext_qualgti = true
    else
      call wrerr(function, warn, trim(qualgti_ext) // " is not present for all instruments")
      this%ext_qualgti = false
    end if

  end subroutine read_qualgti_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the flaregti extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] reopen, convert
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_flaregti_tEventFile(this, status, reopen, convert)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen, convert

    class(tEventFile) :: this

    integer, dimension(this%ninst) :: success

    character(*), parameter :: function = "read_flaregti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call alloc_gti(this%flaregti, this%ninst, status)

    if(present(reopen)) then
      if(reopen == true) call this%open(readonly, status)
    end if

    do ii = 1, this%ninst
      call this%flaregti(ii)%read(this%unit, flaregti_ext, success(ii), status, ccdnr=this%ccdnr_infile(ii), mjdref=this%std_header%mjdref)
    end do

    if( (sum(success) == 0) ) then
      ! try to convert old FLAREGTI to new FLARGTIx
      if(present(convert)) then
        if(convert == true) then
          call this%flaregti(1)%read(this%unit, "FLAREGTI", success(1), status, mjdref=this%std_header%mjdref)
          if(success(1) == true) then
            call wrlog(function, "convert old FLAREGTI to FLARGTIx")
            this%flaregti(1)%ccdnr = this%ccdnr_infile(1)
            do ii = 2, this%ninst
              this%flaregti(ii) = this%flaregti(1)
              this%flaregti(ii)%ccdnr = this%ccdnr_infile(ii)
            end do
          end if
        end if
      end if

      ! if reading the old flaregti still does not work -> flaregti does not exist
      if(sum(success) == 0) then
        this%ext_flaregti = false
        call alloc_gti(this%flaregti, 0, status)
      end if
    else if (sum(success) == this%ninst) then
      this%ext_flaregti = true
    else
      call wrerr(function, warn, trim(flaregti_ext) // " is not present for all instruments")
      this%ext_flaregti = false
    end if

    ! close file if reopen was used
    if(present(reopen)) then
      if(reopen == true) call this%close(status)
    end if

  end subroutine read_flaregti_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create image from input events
  !> TODO RENAME to create image from events
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_image_events_tEventFile(this, resolution, center_position, rebin, status)
    implicit none

    integer, intent(inout) :: status

    integer(i4), dimension(2), intent(in) :: rebin
    integer(i4), dimension(2), intent(in) :: resolution ! naxis
    real(r8), dimension(2), intent(in) :: center_position

    class(tEventFile) :: this

    character(*), parameter :: function = "read_image_events_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call this%image%createFromEvents(this%events, resolution, center_position, rebin, status)

  end subroutine read_image_events_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write image to eventfile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_image_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status

    class(tEventFile) :: this

    character(*), parameter :: function = "write_image_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "writing image extension")
    call this%image%write(this%unit, status)
    call fwrite_stdhdr(this%unit,this%std_header,status)

  end subroutine write_image_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the image in the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_image_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status

    class(tEventFile) :: this

    character(*), parameter :: function = "free_image_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call this%image%free(status)

  end subroutine free_image_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the gti extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_gti_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "write_gti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write gti extension")
    do ii = 1, size(this%gti)
      call this%gti(ii)%write(this%unit, status)
      call this%write_stdheader(status)
    end do

  end subroutine write_gti_tEventFile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the qualgti extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_qualgti_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "write_qualgti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write qualgti extension")
    do ii = 1, size(this%qualgti)
      call this%qualgti(ii)%write(this%unit, status)
      call this%write_stdheader(status)
    end do

  end subroutine write_qualgti_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the regiongti extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_regiongti_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "write_regiongti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write regiongti extension")
    do ii = 1, size(this%regiongti)
      call this%regiongti(ii)%write(this%unit, status)
    end do

  end subroutine write_regiongti_tEventFile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the flaregti extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_flaregti_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "write_flaregti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write flaregti extension")

    do ii = 1, size(this%flaregti)
      call this%flaregti(ii)%write(this%unit, status)
      call this%write_stdheader(status)
    end do

  end subroutine write_flaregti_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the events extensions of the EventFile, by default the wcs keywords will be written as well
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_events_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status

    class(tEventFile) :: this

    character(*), parameter :: function = "write_events_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write events extension")
    call this%events%write(this%unit, status, this%std_header)

  end subroutine write_events_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete the duplicates in the corratt extension
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_duplicates_corrat_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    character(*), parameter :: function = "delete_duplicates_corrat_tEventFile"
    class(tEventFile) :: this

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, this%ninst
      call this%corrat(ii)%delete_duplicates(status)
    end do

  end subroutine delete_duplicates_corrat_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the corrat extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_corrat_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "free_corrat_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%corrat) > 0) then
      do ii = 1, this%ninst
        call this%corrat(ii)%free(status)
      end do

      call dealloc_corrat(this%corrat, status)
    end if

  end subroutine free_corrat_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the corrat extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] reopen
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_corrat_tEventFile(this, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    class(tEventFile) :: this

    integer, dimension(this%ninst) :: success

    character(*), parameter :: function = "read_corrat_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call alloc_corrat(this%corrat, this%ninst, status)

    if(present(reopen)) then
      if(reopen == true) call this%open(readonly, status)
    end if

    do ii = 1, this%ninst
      call this%corrat(ii)%read(this%unit, this%ccdnr_infile(ii), success(ii), status, mjdref = this%std_header%mjdref)
    end do

    if(present(reopen)) then
      if(reopen == true) call this%close(status)
    end if

    if (sum(success) == 0) then
      call alloc_corrat(this%corrat, 0, status)
      this%ext_corrat = false
    else if (sum(success) == this%ninst) then
      this%ext_corrat = true
    else
      call wrerr(function, warn, trim(corratt_ext) // " is not present for all instruments")
      this%ext_corrat = false
    end if

  end subroutine read_corrat_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the corrat extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_corrat_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "write_corrat_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write corratt extension")

    do ii = 1, size(this%corrat)
      call this%corrat(ii)%write(this%unit, status)
      call this%write_stdheader(status)
    end do


  end subroutine write_corrat_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort corrat extensions
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_corrat_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "sort_corrat_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, this%ninst
      call this%corrat(ii)%sort(this%corrat(ii)%roll, status)
      call this%corrat(ii)%sort(this%corrat(ii)%dec, status)
      call this%corrat(ii)%sort(this%corrat(ii)%ra, status)
      call this%corrat(ii)%sort(this%corrat(ii)%time, status)
    end do

  end subroutine sort_corrat_tEventFile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete the duplicates in the deadcor extension
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_duplicates_deadcor_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    character(*), parameter :: function = "delete_duplicates_deadcor_tEventFile"
    class(tEventFile) :: this

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, this%ninst
      call this%deadcor(ii)%delete_duplicates(status)
    end do

  end subroutine delete_duplicates_deadcor_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the deadcor extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_deadcor_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "free_deadcor_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%deadcor) > 0) then
      do ii = 1, this%ninst
        call this%deadcor(ii)%free(status)
      end do

      call dealloc_deadcor(this%deadcor, status)
    end if

  end subroutine free_deadcor_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the deadcor extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] reopen
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_deadcor_tEventFile(this, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    class(tEventFile) :: this

    integer, dimension(this%ninst) :: success

    character(*), parameter :: function = "read_deadcor_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call alloc_deadcor(this%deadcor, this%ninst, status)

    if(present(reopen)) then
      if(reopen == true) call this%open(readonly, status)
    end if

    do ii = 1, this%ninst
      call this%deadcor(ii)%read(this%unit, this%ccdnr_infile(ii), success(ii), status, mjdref = this%std_header%mjdref)
    end do

    if(present(reopen)) then
      if(reopen == true) call this%close(status)
    end if

    if (sum(success) == 0) then
      call alloc_deadcor(this%deadcor, 0, status)
      this%ext_deadcor = false
    else if (sum(success) == this%ninst) then
      this%ext_deadcor = true
    else
      call wrerr(function, warn, trim(deadcor_ext) // " is not present for all instruments")
      this%ext_deadcor = false
    end if

  end subroutine read_deadcor_tEventFile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the events extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] reopen
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_events_tEventFile(this, status, reopen, use_wcs)
    implicit none

    integer, intent(inout) :: status

    class(tEventFile) :: this

    integer, intent(in), optional :: reopen, use_wcs

    integer :: success

    character(*), parameter :: function = "read_events_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(present(reopen)) then
      if(reopen == true) call this%open(readonly, status)
    end if

    if(present(use_wcs)) then
      call this%events%read(this%unit, success, status, mjdref = this%std_header%mjdref, use_wcs=use_wcs)
    else
      call this%events%read(this%unit, success, status, mjdref = this%std_header%mjdref, use_wcs=false)
!      call this%events%read(this%unit, success, status, mjdref = this%std_header%mjdref, use_wcs=true)
    end if

    if(present(reopen)) then
      if(reopen == true) call this%close(status)
    end if

    if(success == true) then
      this%ext_events = true
    else
      this%ext_events = false
      call wrerr(function, s_err, trim(events_ext) // " could not be read")
      status = s_errRep
      return
    end if

  end subroutine read_events_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete the duplicate entries in the events extensions
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_duplicates_events_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status

    class(tEventFile) :: this

    character(*), parameter :: function = "delete_duplicates_events_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call this%events%delete_duplicates(this%ccdnr_infile, status)

  end subroutine delete_duplicates_events_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the events extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_events_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status

    class(tEventFile) :: this

    character(*), parameter :: function = "free_events_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call this%events%free(status)

  end subroutine free_events_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the deadcor extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_deadcor_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "write_deadcor_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write deadcor extension")

    do ii = 1, size(this%deadcor)
      call this%deadcor(ii)%write(this%unit, status)
      call this%write_stdheader(status)
    end do


  end subroutine write_deadcor_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort deadcor extensions
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_deadcor_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "sort_deadcor_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, this%ninst
      call this%deadcor(ii)%sort(this%deadcor(ii)%deadc, status)
      call this%deadcor(ii)%sort(this%deadcor(ii)%time, status)
    end do

  end subroutine sort_deadcor_tEventFile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete the duplicates in the badpix extension
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_duplicates_badpix_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    character(*), parameter :: function = "delete_duplicates_badpix_tEventFile"
    class(tEventFile) :: this

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, this%ninst
!      call this%badpix(ii)%fail_on_duplicates(status)
      call this%badpix(ii)%delete_duplicates(status)
    end do

  end subroutine delete_duplicates_badpix_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the gti extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_gti_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "free_gti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%gti) > 0 ) then
      do ii = 1, this%ninst
        call this%gti(ii)%free(status)
      end do

      call dealloc_gti(this%gti, status)
    end if

  end subroutine free_gti_tEventFile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the qualgti extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_qualgti_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "free_qualgti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%qualgti) > 0) then
      do ii = 1, this%ninst
        call this%qualgti(ii)%free(status)
      end do

      call dealloc_gti(this%qualgti, status)
    end if

  end subroutine free_qualgti_tEventFile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the flaregti extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_flaregti_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "free_flaregti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%flaregti) > 0 ) then
      do ii = 1, this%ninst
        call this%flaregti(ii)%free(status)
      end do

      call dealloc_gti(this%flaregti, status)
    end if

  end subroutine free_flaregti_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the badpix extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_badpix_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "free_badpix_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%badpix) > 0) then
      do ii = 1, this%ninst
        call this%badpix(ii)%free(status)
      end do

      call dealloc_badpix(this%badpix, status)
    end if

  end subroutine free_badpix_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the badpix extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] reopen
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_badpix_tEventFile(this, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, optional, intent(in) :: reopen

    class(tEventFile) :: this

    integer, dimension(this%ninst) :: success

    character(*), parameter :: function = "read_badpix_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call alloc_badpix(this%badpix, this%ninst, status)

    if(present(reopen)) then
      if(reopen == true) call this%open(readonly, status)
    end if

    do ii = 1, this%ninst
      call this%badpix(ii)%read(this%unit, this%ccdnr_infile(ii), success(ii), status, mjdref = this%std_header%mjdref)
    end do
        
    if(present(reopen)) then
      if(reopen == true) call this%close(status)
    end if

    if (sum(success) == 0 ) then
      call alloc_badpix(this%badpix, 0, status)
      this%ext_badpix = false
    else if (sum(success) == this%ninst) then
      this%ext_badpix = true
    else
      call wrerr(function, warn, trim(badpix_ext) // " is not present for all instruments")
      this%ext_badpix = false
    end if

  end subroutine read_badpix_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the badpix extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_badpix_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "write_badpix_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write badpix extension")
    do ii = 1, size(this%badpix)
      call this%badpix(ii)%write(this%unit, status)
      call this%write_stdheader(status)
    end do


  end subroutine write_badpix_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort badpix extensions
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_badpix_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "sort_badpix_tEventFile"

    if(status > 100 .or. status == s_errRep) return
    do ii = 1, this%ninst
      call this%badpix(ii)%sort(this%badpix(ii)%rawx, status)
      call this%badpix(ii)%sort(this%badpix(ii)%rawy, status)
      call this%badpix(ii)%sort(this%badpix(ii)%timemax, status)
      call this%badpix(ii)%sort(this%badpix(ii)%timemin, status)

    end do

  end subroutine sort_badpix_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete the duplicates in the offset extension
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_duplicates_offset_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    character(*), parameter :: function = "delete_duplicates_offset_tEventFile"
    class(tEventFile) :: this

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, this%ninst
      call this%offset(ii)%delete_duplicates(status)
    end do

  end subroutine delete_duplicates_offset_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the offset extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_offset_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "free_offset_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%offset) > 0) then
      do ii = 1, this%ninst
        call this%offset(ii)%free(status)
      end do

      call dealloc_offset(this%offset, status)
    end if

  end subroutine free_offset_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the offset extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] reopen
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_offset_tEventFile(this, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    class(tEventFile) :: this

    integer, dimension(this%ninst) :: success

    character(*), parameter :: function = "read_offset_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call alloc_offset(this%offset, this%ninst, status)

    if(present(reopen)) then
      if(reopen == true) call this%open(readonly, status)
    end if

    do ii = 1, this%ninst
      call this%offset(ii)%read(this%unit, this%ccdnr_infile(ii), success(ii), status, mjdref = this%std_header%mjdref)
    end do

    if(present(reopen)) then
      if(reopen == true) call this%close(status)
    end if

    if (sum(success) == 0) then
      call alloc_offset(this%offset, 0, status)
      this%ext_offset = false
    else if (sum(success) == this%ninst) then
      this%ext_offset = true
    else
      call wrerr(function, warn, trim(offset_ext) // " is not present for all instruments")
      this%ext_offset = false
    end if

  end subroutine read_offset_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the offset extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_offset_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "write_offset_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write offset extension")
    do ii = 1, size(this%offset)
      call this%offset(ii)%write(this%unit, status)
      call this%write_stdheader(status)
    end do


  end subroutine write_offset_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort offset extensions
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_offset_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "sort_offset_tEventFile"

    if(status > 100 .or. status == s_errRep) return
    do ii = 1, this%ninst
      call this%offset(ii)%sort(this%offset(ii)%rawx, status)
      call this%offset(ii)%sort(this%offset(ii)%offsetx, status)
      call this%offset(ii)%sort(this%offset(ii)%time, status)
    end do

  end subroutine sort_offset_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete the duplicates in the hk extension
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_duplicates_hk_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    character(*), parameter :: function = "delete_duplicates_hk_tEventFile"
    class(tEventFile) :: this

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, this%ninst
      call this%hk(ii)%delete_duplicates(status)
    end do

  end subroutine delete_duplicates_hk_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the hk extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_hk_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "free_hk_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%hk) > 0) then
      do ii = 1, this%ninst
        call this%hk(ii)%free(status)
      end do
      call dealloc_hk(this%hk, status)
    end if

  end subroutine free_hk_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the hk extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] reopen
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_hk_tEventFile(this, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    class(tEventFile) :: this

    integer, dimension(this%ninst) :: success

    character(*), parameter :: function = "read_hk_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call alloc_hk(this%hk, this%ninst, status)

    if(present(reopen)) then
      if(reopen == true) call this%open(readonly, status)
    end if

    do ii = 1, this%ninst
      call this%hk(ii)%read(this%unit, this%ccdnr_infile(ii), success(ii), status, mjdref = this%std_header%mjdref)
    end do

    if(present(reopen)) then
      if(reopen == true) call this%close(status)
    end if

    if (sum(success) == 0) then
      call alloc_hk(this%hk, 0, status)
      this%ext_hk = false
    else if (sum(success) == this%ninst) then
      this%ext_hk = true
    else
      call wrerr(function, warn, trim(hk_ext) // " is not present for all instruments")
      this%ext_hk= false
    end if
  end subroutine read_hk_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the hk extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_hk_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "write_hk_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write hk extension")
    do ii = 1, size(this%hk)
      call this%hk(ii)%write(this%unit, status)
      call this%write_stdheader(status)
    end do

  end subroutine write_hk_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort hk extensions
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_hk_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "sort_hk_tEventFile"

    if(status > 100 .or. status == s_errRep) return
    do ii = 1, this%ninst
      call this%hk(ii)%sort(this%hk(ii)%time, status)
    end do

  end subroutine sort_hk_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the hkx1 extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_hkx1_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "free_hkx1_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%hkx1) > 0) then
      do ii = 1, this%ninst
        call this%hkx1(ii)%free(status)
      end do
      call dealloc_hkx1(this%hkx1, status)
    end if

  end subroutine free_hkx1_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the hkx2 extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_hkx2_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "free_hkx2_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%hkx2) > 0) then
      do ii = 1, this%ninst
        call this%hkx2(ii)%free(status)
      end do
      call dealloc_hkx2(this%hkx2, status)
    end if

  end subroutine free_hkx2_tEventFile

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the hkx3 extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_hkx3_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "free_hkx3_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%hkx3) > 0) then
      do ii = 1, this%ninst
        call this%hkx3(ii)%free(status)
      end do
      call dealloc_hkx3(this%hkx3, status)
    end if

  end subroutine free_hkx3_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the hkx5 extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_hkx5_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "free_hkx5_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%hkx5) > 0) then
      do ii = 1, this%ninst
        call this%hkx5(ii)%free(status)
      end do
      call dealloc_hkx5(this%hkx5, status)
    end if

  end subroutine free_hkx5_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the hkx1 extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] reopen
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_hkx1_tEventFile(this, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    class(tEventFile) :: this

    character(*), parameter :: function = "read_hkx1_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call alloc_hkx1(this%hkx1, this%ninst, status)

    if(present(reopen)) then
      if(reopen == true) call this%open(readonly, status)
    end if

    this%ext_hkx1 = true
    do ii = 1, this%ninst
      this%hkx1(ii)%unit = this%unit
      call this%hkx1(ii)%read(status, tel_id = this%ccdnr_infile(ii))
      if(status /= ok) then
        call wrerr(function, warn, hkx1_ext // " is not present for all instruments")
        this%ext_hkx1 = false
        status = ok
        exit
      end if
    end do

    if(present(reopen)) then
      if(reopen == true) call this%close(status)
    end if

  end subroutine read_hkx1_tEventFile

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the hkx2 extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] reopen
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_hkx2_tEventFile(this, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    class(tEventFile) :: this

    character(*), parameter :: function = "read_hkx2_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call alloc_hkx2(this%hkx2, this%ninst, status)

    if(present(reopen)) then
      if(reopen == true) call this%open(readonly, status)
    end if

    this%ext_hkx2 = true
    do ii = 1, this%ninst
      this%hkx2(ii)%unit = this%unit
      call this%hkx2(ii)%read(status, tel_id = this%ccdnr_infile(ii))
      if(status /= ok) then
        call wrerr(function, warn, hkx2_ext // " is not present for all instruments")
        this%ext_hkx2 = false
        status = ok
        exit
      end if
    end do

    if(present(reopen)) then
      if(reopen == true) call this%close(status)
    end if

  end subroutine read_hkx2_tEventFile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the hkx3 extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] reopen
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_hkx3_tEventFile(this, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    class(tEventFile) :: this

    character(*), parameter :: function = "read_hkx3_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call alloc_hkx3(this%hkx3, this%ninst, status)

    if(present(reopen)) then
      if(reopen == true) call this%open(readonly, status)
    end if

    this%ext_hkx3 = true
    do ii = 1, this%ninst
      this%hkx3(ii)%unit = this%unit
      call this%hkx3(ii)%read(status, tel_id = this%ccdnr_infile(ii))
      if(status /= ok) then
        call wrerr(function, warn, hkx3_ext // " is not present for all instruments")
        this%ext_hkx3 = false
        status = ok
        exit
      end if
    end do

    if(present(reopen)) then
      if(reopen == true) call this%close(status)
    end if

  end subroutine read_hkx3_tEventFile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the hkx5 extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] reopen
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_hkx5_tEventFile(this, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    class(tEventFile) :: this

    character(*), parameter :: function = "read_hkx5_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call alloc_hkx5(this%hkx5, this%ninst, status)

    if(present(reopen)) then
      if(reopen == true) call this%open(readonly, status)
    end if

    this%ext_hkx5 = true
    do ii = 1, this%ninst
      this%hkx5(ii)%unit = this%unit
      call this%hkx5(ii)%read(status, tel_id = this%ccdnr_infile(ii))
      if(status /= ok) then
        call wrerr(function, warn, hkx5_ext // " is not present for all instruments")
        this%ext_hkx5 = false
        status = ok
        exit
      end if
    end do

    if(present(reopen)) then
      if(reopen == true) call this%close(status)
    end if

  end subroutine read_hkx5_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the hkx1 extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_hkx1_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "write_hkx1_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write hkx1 extension")
    do ii = 1, size(this%hkx1)
      this%hkx1(ii)%unit = this%unit
      call this%hkx1(ii)%write(status)
      call this%write_stdheader(status)
    end do

  end subroutine write_hkx1_tEventFile

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the hkx2 extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_hkx2_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "write_hkx2_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write hkx2 extension")
    do ii = 1, size(this%hkx2)
      this%hkx2(ii)%unit = this%unit
      call this%hkx2(ii)%write(status)
      call this%write_stdheader(status)
    end do

  end subroutine write_hkx2_tEventFile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the hkx3 extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_hkx3_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "write_hkx3_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write hkx3 extension")
    do ii = 1, size(this%hkx3)
      this%hkx3(ii)%unit = this%unit
      call this%hkx3(ii)%write(status)
      call this%write_stdheader(status)
    end do

  end subroutine write_hkx3_tEventFile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the hkx5 extensions of the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_hkx5_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "write_hkx5_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write hkx5 extension")
    do ii = 1, size(this%hkx5)
      this%hkx5(ii)%unit = this%unit
      call this%hkx5(ii)%write(status)
      call this%write_stdheader(status)
    end do

  end subroutine write_hkx5_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort hkx1 extensions
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_hkx1_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "sort_hkx1_tEventFile"

    if(status > 100 .or. status == s_errRep) return
    do ii = 1, this%ninst
      call this%hkx1(ii)%sort(this%hkx1(ii)%time%value, status)
    end do

  end subroutine sort_hkx1_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort hkx2 extensions
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_hkx2_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "sort_hkx2_tEventFile"

    if(status > 100 .or. status == s_errRep) return
    do ii = 1, this%ninst
      call this%hkx2(ii)%sort(this%hkx2(ii)%time%value, status)
    end do

  end subroutine sort_hkx2_tEventFile

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort hkx3 extensions
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_hkx3_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "sort_hkx3_tEventFile"

    if(status > 100 .or. status == s_errRep) return
    do ii = 1, this%ninst
      call this%hkx3(ii)%sort(this%hkx3(ii)%time%value, status)
    end do

  end subroutine sort_hkx3_tEventFile

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort hkx5 extensions
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_hkx5_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this

    character(*), parameter :: function = "sort_hkx5_tEventFile"

    if(status > 100 .or. status == s_errRep) return
    do ii = 1, this%ninst
      call this%hkx5(ii)%sort(this%hkx5(ii)%time%value, status)
    end do

  end subroutine sort_hkx5_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> open various eventfiles
  !
  !
  !
  !> @param[in] EventFile, filename
  !> @param[out] iotype
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine open_eventfile(EventFile, filename, iotype, status, delete)
    implicit none

    integer, intent(inout) :: status
    integer, intent(in) :: iotype
    integer, intent(in), optional :: delete

    type(tEventFile), intent(inout) :: EventFile

    character(*), intent(in) :: filename
    character(*), parameter :: function = "open_eventfile"

    if(status > 100 .or. status == s_errRep) return

    call ffits_get_iounit(EventFile%unit, status)
    EventFile%name = filename
    if(iotype /= create) then
      call ffits_open_file(EventFile%unit, filename, iotype, status)
    else
      if(present(delete)) then
        if(delete == true) call delete_file_fortran(filename, status)
      end if
      call ffits_create_file(EventFile%unit, filename, fblocksize, status)
    end if


  end subroutine open_eventfile

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> open an eventfile
  !
  !
  !
  !> @param[in] EventFile, filename
  !> @param[out] iotype
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine open_tEventFile(this, iotype, status, filename, delete)
    implicit none

    integer, intent(inout) :: status
    integer, intent(in) :: iotype
    integer, intent(in), optional :: delete

    class(tEventFile) :: this

    character(*), intent(in), optional :: filename
    character(*), parameter :: function = "open_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call ffits_get_iounit(this%unit, status)
    if(present(filename)) this%name = filename

!*** 2019-09-11 IMS: this%name doesn't seem to be initialized unless present(filename). I added an init to "" in the struct definition, but this does not seem to have effect on the output file name, which seems to contain nonsense before being set.
    if(this%name == "") then
      call wrerr(function, s_err, "no filename provided")
      status = s_errRep
    end if

    if(iotype /= create) then
      call ffits_open_file(this%unit, this%name, iotype, status)
    else
      if(present(delete)) then
        if(delete == true) call delete_file_fortran(this%name, status)
      end if
      call ffits_create_file(this%unit, this%name, fblocksize, status)
    end if

  end subroutine open_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> close various eventfiles
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine close_eventfile(EventFile, status)
    implicit none

    integer, intent(inout) :: status

    type(tEventFile), intent(inout) :: EventFile

    character(*), parameter :: function = "close_eventfile"

    if(status > 100 .or. status == s_errRep) return

    call ffits_close_file(EventFile%unit, status)
    call ffits_free_iounit(EventFile%unit, status)

  end subroutine close_eventfile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> close eventfile
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine close_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status

    class(tEventFile) :: this

    character(*), parameter :: function = "close_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call ffits_close_file(this%unit, status)
    call ffits_free_iounit(this%unit, status)

  end subroutine close_tEventFile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> open various eventfiles
  !
  !
  !
  !> @param[in] EventFile, filename
  !> @param[out] iotype
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine open_eventfile_array(EventFile, filename, iotype, status)
    implicit none

    integer, intent(inout) :: status
    integer, intent(in) :: iotype
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), dimension(:), intent(in) :: filename
    character(*), parameter :: function = "open_eventfile_array"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      if(size(EventFile) > 50) then
        call wrlog(function, "number of input event files > 50")
        EventFile(ii)%unit = 100 + ii
      else
        call ffits_get_iounit(EventFile(ii)%unit, status)
      end if
      EventFile(ii)%name = filename(ii)
      if(iotype /= create) then
        call ffits_open_file(EventFile(ii)%unit, filename(ii), iotype, status)
      else
        call ffits_create_file(EventFile(ii)%unit, filename(ii), fblocksize, status)
      end if
    end do


  end subroutine open_eventfile_array

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> close various eventfiles
  !
  !
  !
  !> @param[in] EventFile
  !> @param[out] iotype
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine close_eventfile_array(EventFile, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "close_eventfile_array"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call ffits_close_file(EventFile(ii)%unit, status)
      if(size(EventFile) <= 50) call ffits_free_iounit(EventFile(ii)%unit, status)
    end do

  end subroutine close_eventfile_array

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply stdheader type keywords to eventfile
  !
  !
  !
  !> @param[in] this
  !> @param[out] stdheader
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_stdheader_tEventFile(this, stdheader, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, ldigit

    class(tEventFile) :: this
    type(stdhdr_type), intent(in) :: stdheader

    character(*), parameter :: function = "apply_stdheader_tEventFile"

    if(status > 100 .or. status == s_errRep) return

     ! apply the stdheader to
    this%filter_pattern = false
    this%filter_flag = false
    this%filter_gti = false
    this%filter_energy = false
    this%filter_telid = false
    if(stdheader%pat_sel /= 0)this%filter_pattern = true
    if(stdheader%flagsel /= 0)this%filter_flag = true
    if(stdheader%gti_sel /= "NONE")this%filter_gti = true
    if(stdheader%nenergy /= 0)this%filter_energy = true
    if(stdheader%ninst /= 0)this%filter_telid = true

    this%ninst = stdheader%ninst

    call alloc(this%instrum, this%ninst, status)
    call alloc(this%ccdnr_infile, this%ninst, status)

    if(this%ninst == 0) return

    if(this%ninst > 0) then
      this%instrum = stdheader%instrumx(:this%ninst)
    end if

    do ii = 1, this%ninst
      ldigit = len_trim(this%instrum(ii))
      this%ccdnr_infile(ii) = str2int(this%instrum(ii)(ldigit:ldigit))
    end do

  end subroutine apply_stdheader_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read in events extension keywords
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_events_ext_keywords_eventfile(EventFile, status)
    implicit none

    integer :: ii
    integer, intent(inout) :: status

    type(tEventFile), dimension(:) :: EventFile

    character(*), parameter :: function = "read_events_ext_keywords_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call ffits_movnam_hdu(EventFile(ii)%unit, BINARY_TBL, events_ext, status)
      call EventFile(ii)%read_stdheader(status)
      call EventFile(ii)%events%wcs%read(EventFile(ii)%unit, BINARY_TBL, x_col, y_col, status)
    end do

  end subroutine read_events_ext_keywords_eventfile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> dealloacte eventfile structures
  !
  !
  !
  !> @param[inout] eventfile
  !> @param[inout] status
  !---------------------------------------------------------------------------
  subroutine dealloc_eventfile(eventfile, status)

    integer :: stat
    integer, intent(inout) :: status

    type(tEventFile), allocatable, dimension(:) :: eventfile

    character(*), parameter :: function = "dealloc_eventfile"

    if(allocated(eventfile)) then
      deallocate(eventfile, stat=stat)
      call FError(function, stat, "dealloc eventfile type", status)
    end if

  end subroutine dealloc_eventfile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloacte n eventfile structures
  !
  !
  !
  !> @param[inout] eventfile
  !> @param[in] n
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine alloc_eventfile(eventfile, n, status)
    implicit none

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    type(tEventFile), allocatable, dimension(:) :: eventfile

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_eventfile"

    call dealloc_eventfile(eventfile, status)
    allocate(eventfile(n), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

  end subroutine alloc_eventfile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply the derived gti from the region selection and the corrat extension to all relevant extensions
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_region_gti_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEventFile) :: this
    type(tGTI) :: tmp

    character(*), parameter :: function = "apply_region_gti_tEventFile"

    if(status > 100 .or. status == s_errRep) return
        
    if(this%ext_corrat == false) then
      call wrlog(function, "no region gti will be applied / no CORRAT ext was found")
      return
    end if

    call wrlog(function, "reduction of deadcor extensions - regiongti - temporarily disabled")
    call wrlog(function, "reduction of hk extensions - regiongti - temporarily disabled")
    call wrlog(function, "reduction of offset extensions - regiongti - temporarily disabled")

    ! the overlap is already included in the regiongti
    do ii = 1, this%ninst

      call this%corrat(ii)%apply_gti(this%regiongti(ii), status)
      !            call this%deadcor(ii)%apply_gti(this%regiongti(ii), status)
      !            if(this%ext_hk == true) call this%hk(ii)%apply_gti(this%regiongti(ii), status)
      !            if(this%ext_offset == true) call this%offset(ii)%apply_gti(this%regiongti(ii), status)

      if(this%ext_gti == true) then
        tmp = this%gti(ii)
        call merge_gti(tmp, this%regiongti(ii), this%gti(ii), and, status)
      end if

      if(this%ext_qualgti == true) then
        tmp = this%qualgti(ii)
        call merge_gti(tmp, this%regiongti(ii), this%qualgti(ii), and, status)
      end if

      if(this%ext_flaregti == true) then
        tmp = this%flaregti(ii)
        call merge_gti(tmp, this%regiongti(ii), this%flaregti(ii), and, status)
      end if

    end do

  end subroutine apply_region_gti_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply energy filter to the EventFile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_filter_energy_tEventFile(this, status)
    implicit none
    integer, intent(inout) :: status
    integer :: nenergy

    character(*), parameter :: function = "apply_filter_energy_tEventFile"

    class(tEventFile), intent(inout) :: this

    if(status > 100 .or. status == s_errRep) return

    nenergy = this%std_header%nenergy
    ! only apply energy unit conversion on std_header here, since std_header is keV while energy column are in eV
    call this%events%apply_energy(this%std_header%eminx(:nenergy)*1000.0_r8, this%std_header%emaxx(:nenergy)*1000.0_r8, status)
  !    call this%events%apply_energy(this%std_header%eminx(:nenergy), this%std_header%emaxx(:nenergy), status)

  end subroutine apply_filter_energy_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a filter on the ccdnr column to only contain certain numbers
  !
  !
  !
  !> @param[in] this
  !> @param[out] array
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_filter_telid_tEventFile(this, status)
    implicit none
    integer, intent(inout) :: status

    character(*), parameter :: function = "apply_filter_telid_tEventFile"

    class(tEventFile), intent(inout) :: this

    if(status > 100 .or. status == s_errRep) return

    call this%events%apply_telid(this%ccdnr_infile, status)

  end subroutine apply_filter_telid_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a flag selection filter on the events
  !
  !
  !
  !> @param[in] unit, flagsel
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_filter_flag_tEventFile(this, status)
    implicit none
    integer, intent(inout) :: status

    class(tEventFile) :: this

    character(*), parameter :: function = "apply_filter_flag_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call this%events%apply_flag(this%std_header%flagsel, status)

  end subroutine apply_filter_flag_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a pattern selection filter on the events
  !
  !
  !
  !> @param[in] unit, patternsel
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_filter_pattern_tEventFile(this, status)
    implicit none
    integer, intent(inout) :: status

    class(tEventFile) :: this

    character(*), parameter :: function = "apply_filter_pattern_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call this%events%apply_pattern(this%std_header%pat_sel, status)

  end subroutine apply_filter_pattern_tEventFile

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a gti selection filter on the events
  !
  !
  !
  !> @param[inout] this
  !> @param[in, optional] overlap
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_filter_gti_tEventFile(this, status, overlap)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    real(r8), optional, intent(in) :: overlap

    class(tEventFile) :: this
    type(tGti) :: tmp

    character(*), parameter :: function = "apply_filter_gti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(this%ccdnr_infile)
      call this%events%apply_gti(this%gti(ii), status)

      call wrlog(function, "reduction of deadcor extensions - gti - temporarily disabled")
      call wrlog(function, "reduction of hk extensions - gti - temporarily disabled")
      call wrlog(function, "reduction of offset extensions - gti - temporarily disabled")

      tmp = this%gti(ii)
      ! modify the GTI applied to the auxilliary extensions to be able to interpolate in the upcoming tasks
      if(present(overlap)) then
        tmp%start = tmp%start - overlap
        tmp%stop = tmp%stop + overlap
      end if

      if(this%ext_corrat == true) call this%corrat(ii)%apply_gti(tmp, status)
    !      if(this%ext_deadcor == true) call this%deadcor(ii)%apply_gti(tmp, status)
    !      if(this%ext_hk == true) call this%hk(ii)%apply_gti(tmp, status)
    !      if(this%ext_offset == true) call this%offset(ii)%apply_gti(tmp, status)
    end do

  end subroutine apply_filter_gti_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a qualgti selection filter on the events
  !
  !
  !
  !> @param[inout] this
  !> @param[in, optional] overlap
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_filter_qualgti_tEventFile(this, status, overlap)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    real(r8), intent(in), optional :: overlap

    class(tEventFile) :: this
    type(tGti) :: tmp

    character(*), parameter :: function = "apply_filter_qualgti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(this%ext_qualgti == true) then

      call wrlog(function, "reduction of deadcor extensions - qualgti - temporarily disabled")
      call wrlog(function, "reduction of hk extensions - qualgti - temporarily disabled")
      call wrlog(function, "reduction of offset extensions - qualgti - temporarily disabled")

      do ii = 1, size(this%ccdnr_infile)
        call this%events%apply_gti(this%qualgti(ii), status)
        if(this%ext_gti == true) call this%gti(ii)%apply_gti(this%qualgti(ii), and, status)

        tmp = this%qualgti(ii)
        ! modify the GTI applied to the auxilliary extensions to be able to interpolate in the upcoming tasks
        if(present(overlap)) then
          tmp%start = tmp%start - overlap
          tmp%stop = tmp%stop + overlap
        end if

        if(this%ext_corrat == true) call this%corrat(ii)%apply_gti(tmp, status)
      !        if(this%ext_deadcor == true) call this%deadcor(ii)%apply_gti(tmp, status)
      !        if(this%ext_hk == true) call this%hk(ii)%apply_gti(tmp, status)
      !        if(this%ext_offset == true) call this%offset(ii)%apply_gti(tmp, status)
      end do
    else
      call wrerr(function, s_err, trim(qualgti_ext) // " selected as filter but (almost) all infiles without " // trim(qualgti_ext) // " extension")
      status = s_errRep
    end if

  end subroutine apply_filter_qualgti_tEventFile


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a flaregti selection filter on the events
  !
  !
  !
  !> @param[inout] this
  !> @param[in, optional] overlap
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_filter_flaregti_tEventFile(this, status, overlap)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    real(r8), intent(in), optional :: overlap

    class(tEventFile) :: this
    type(tGTI) :: tmp

    character(*), parameter :: function = "apply_filter_flaregti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(this%ext_flaregti == true) then

      call wrlog(function, "reduction of deadcor extensions - flaregti - temporarily disabled")
      call wrlog(function, "reduction of hk extensions - flaregti - temporarily disabled")
      call wrlog(function, "reduction of offset extensions - flaregti - temporarily disabled")

      do ii = 1, size(this%ccdnr_infile)
        call this%events%apply_gti(this%flaregti(ii), status)
        if(this%ext_gti == true) call this%gti(ii)%apply_gti(this%flaregti(ii), and, status)
        if(this%ext_qualgti == true) call this%qualgti(ii)%apply_gti(this%flaregti(ii), and, status)

        tmp = this%flaregti(ii)
        ! modify the GTI applied to the auxilliary extensions to be able to interpolate in the upcoming tasks
        if(present(overlap)) then
          tmp%start = tmp%start - overlap
          tmp%stop = tmp%stop + overlap
        end if

        if(this%ext_corrat == true) call this%corrat(ii)%apply_gti(tmp, status)
      !        if(this%ext_deadcor == true) call this%deadcor(ii)%apply_gti(tmp, status)
      !        if(this%ext_hk == true) call this%hk(ii)%apply_gti(tmp, status)
      !        if(this%ext_offset == true) call this%offset(ii)%apply_gti(tmp, status)
      end do
    else
      call wrerr(function, s_err, trim(flaregti_ext) // " selected as filter but (almost) all infiles without " // trim(qualgti_ext) // " extension")
      status = s_errRep
    end if

  end subroutine apply_filter_flaregti_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a external gti selection filter on the events
  !
  !
  !
  !> @param[inout] this
  !> @param[in], externalgti
  !> @param[in, optional] overlap
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_filter_externalgti_tEventFile(this, externalgti, status, overlap)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    real(r8), optional, intent(in) :: overlap

    class(tEventFile) :: this
    type(tGTI) :: tmp

    type(tGTI), intent(in) :: externalgti

    character(*), parameter :: function = "apply_filter_externalgti_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call this%events%apply_gti(externalgti, status)
    tmp = externalgti
    ! modify the GTI applied to the auxilliary extensions to be able to interpolate in the upcoming tasks
    if(present(overlap)) then
      tmp%start = tmp%start - overlap
      tmp%stop = tmp%stop + overlap
    end if

    call wrlog(function, "reduction of deadcor extensions - externalgti - temporarily disabled")
    call wrlog(function, "reduction of hk extensions - externalgti - temporarily disabled")
    call wrlog(function, "reduction of offset extensions - externalgti - temporarily disabled")

    do ii = 1, size(this%ccdnr_infile)
      if(this%ext_gti == true) call this%gti(ii)%apply_gti(externalgti, and, status)
      if(this%ext_qualgti == true) call this%qualgti(ii)%apply_gti(externalgti, and, status)
      if(this%ext_flaregti == true) call this%flaregti(ii)%apply_gti(externalgti, and, status)

      if(this%ext_corrat == true) call this%corrat(ii)%apply_gti(tmp, status)
    !      if(this%ext_deadcor == true) call this%deadcor(ii)%apply_gti(tmp, status)
    !      if(this%ext_hk == true) call this%hk(ii)%apply_gti(tmp, status)
    !      if(this%ext_offset == true) call this%offset(ii)%apply_gti(tmp, status)
    end do

  end subroutine apply_filter_externalgti_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply the region filter on the event list
  !
  !
  !
  !> @param[in] this
  !> @param[out] Evtool
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_filter_region_tEventFile(this, regtable, status)
    implicit none

    integer, intent(inout) :: status
    integer(i4), dimension(:), allocatable :: order, mask
    integer :: ii, nelements

    class(tEventFile) :: this
    type(tRegTable) :: regtable

    ! for using srctool spatial region selections
    type(region_type) :: srctool_region
    type(coord_type) :: srctool_coord

    character(*), parameter :: function = "apply_filter_region_tEventFile"

    if(status > 100 .or. status == s_errRep) return
    this%regtable = regtable

    nelements = this%events%nelements
    call alloc(order, nelements, status)
    call alloc(mask, nelements, status)
    order = [(ii,ii=1, nelements,1)]
    mask = false

    call regtable%get_srctool_region(srctool_region, status)

    ! TODO update when srctool can handle the other formats
    ! THIS ASSUMES THAT ONLY ONE TYOE OF COORDSYS in all regions
    do ii = 1, nelements
      srctool_coord%long = this%events%ra(ii)
      srctool_coord%lat = this%events%dec(ii)
      srctool_coord%x = this%events%x(ii)
      srctool_coord%y = this%events%y(ii)
      srctool_coord%rawx = this%events%rawx(ii)
      srctool_coord%rawy = this%events%rawy(ii)
      srctool_coord%coordsys_code = COORDSYS_CODE_ICRS
      ! test
      !      if(srctool_region%pregcomp(1)%coordsys_code == COORDSYS_CODE_PHYSICAL) then
      !        srctool_coord%coordsys_code = COORDSYS_CODE_PHYSICAL
      !      end if

      call srctool_regions_test_if_coord_in_region(srctool_coord, srctool_region, mask(ii), status)
      if(status /= ok) status = s_errRep
    end do

    order = pack(order, mask == true)

    call this%events%apply_index_cols(order, status)

    call dealloc(order, status)
    call dealloc(mask, status)

  end subroutine apply_filter_region_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply the RAWX/Y filter on the event list
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_filter_rawxy_tEventFile(this, rawxy_sel, status)
    implicit none

    integer, intent(inout) :: status
    type(tRawXYSelection), intent(in) :: rawxy_sel
    class(tEventFile) :: this

    character(*), parameter :: function = "apply_filter_rawxy_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call this%events%apply_rawxy(rawxy_sel, status)

  end subroutine apply_filter_rawxy_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the region extension to the eventfile
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_region_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status

    class(tEventFile) :: this

    character(*), parameter :: function = "write_region_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call wrchatty(function, "write region extension")
    call this%regtable%write(this%unit, "REGION", status)
    call this%write_stdheader(status)

  end subroutine write_region_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> reduce extensions (GTI, QUALGTI, FLAREGTI) to contain a subset of the entries based on an
  !> encapsulating circle, extended by the eFoV parameter, for each region and
  !> the CORRAT extension
  !
  !> @param[inout] Evselect
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine calc_gti_from_corrat_region_filter_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    integer, dimension(:), allocatable :: order, mask
    integer :: ii, jj, nrows

    character(*), parameter :: function = "calc_gti_from_corrat_region_filter_tEventFile"

    type(tCorrat) :: iCorrat
    class(tEventFile) :: this
    type(region_type) :: srctool_region
    type(coord_type) :: srctool_coord

    if(status > 100 .or. status == s_errRep) return

    call this%regtable%get_srctool_encap_circle_region(srctool_region, status)

    allocate(this%regiongti(this%ninst))

    do ii = 1, size(this%events%ccdnr)

      ! presort the corrat extension in time
      call this%Corrat(ii)%sort(this%Corrat(ii)%time, status, true)

      ! interpolate the corrat extension
      ! NOTE: interpolation for roll angle is not performed
      ! - only necessary if we want to store a reduced corrat extension
      call this%Corrat(ii)%interpolate(iCorrat, frametime, status)

      nrows = size(iCorrat%ra)

      call alloc(order, nrows, status)
      call alloc(mask, nrows, status)

      order = [(jj,jj=1,nrows,1)]
      mask = false

      do jj = 1, nrows
        srctool_coord%long = iCorrat%ra(jj)
        srctool_coord%lat = iCorrat%dec(jj)
        ! TODO enable again after srctool update
        !           srctool_coord%x = 0.0_dp
        !           srctool_coord%y = 0.0_dp
        !           srctool_coord%rawx = 0.0_dp
        !           srctool_coord%rawy = 0.0_dp
        srctool_coord%coordsys_code = COORDSYS_CODE_ICRS
        call srctool_regions_test_if_coord_in_region(srctool_coord, srctool_region, mask(jj), status)
        if(status /= ok) status = s_errRep
      end do

      order = pack(order, mask == false)

      call create_GTI_from_order(iCorrat%time, order, this%regiongti(ii), status)

    end do

  end subroutine calc_gti_from_corrat_region_filter_tEventFile


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort the events in order : time ,rawx, rawy, pha, ccdnr
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_events_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status

    class(tEventFile) :: this

    character(*), parameter :: function = "sort_events_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call this%events%sort(this%events%ccdnr, status)
    call this%events%sort(this%events%pha, status)
    call this%events%sort(this%events%rawy, status)
    call this%events%sort(this%events%rawx, status)
    call this%events%sort(this%events%time, status)

  end subroutine sort_events_tEventFile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the standard header keys of an EventFile
  !
  !
  !
  !> @param[in] inunit
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_stdheader_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tEventFile) :: this

    integer :: unit

    character(*), parameter :: function = "read_stdheader_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    unit = this%unit

    call fread_stdhdr(unit, this%std_header, status)
    if(status == 1) then
      call wrlog(function, "Missing keywords : reading standard header")
      status = ok
    end if

    call this%apply_stdheader(this%std_header, status)

  end subroutine read_stdheader_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write standard header keys of an EventFile to a new EventFile
  !
  !
  !
  !> @param[in] inunit
  !> @param[in] outunit
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_stdheader_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status

    character(buffsize) :: sass_parint

    class(tEventFile) :: this

    character(*), parameter :: function = "write_stdheader_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call getenv("SASS_PARINT", sass_parint)

    if(trim(sass_parint) == "PIPELINE") then
      call update_stdhdr(this%name, this%std_header, status)
    end if

    ! keywords written by write stdheader should be written with update routine instead of write routine
    call fwrite_stdhdr(this%unit, this%std_header, status)


  end subroutine write_stdheader_tEventFile

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read an eventfile
  !
  !
  !> @param[inout] this, status
  !---------------------------------------------------------------------------

  subroutine read_tEventFile(this, status, use_wcs, filename)
    implicit none

    integer, intent(inout) :: status
    class(tEventFile), intent(inout) :: this
    integer, intent(in), optional :: use_wcs
    character(*), intent(in), optional :: filename

    character(*), parameter :: function = "read_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    if(present(filename)) call this%open(readonly, status, filename=filename)

    call ffits_movnam_hdu(this%unit, BINARY_TBL, events_ext, status)
    call this%read_stdheader(status)

    if(present(use_wcs)) then
      call this%read_events(status, use_wcs=use_wcs)
    else
      call this%read_events(status)
    end if

    call this%read_corrat(status)
    call this%read_deadcor(status)
    call this%read_badpix(status)
    call this%read_gti(status)

    ! optional extensions not necessary
    call this%read_offset(status)
    call this%read_hk(status)
    call this%read_qualgti(status)
    call this%read_flaregti(status)

    call this%read_hkx1(status)
    call this%read_hkx2(status)
    call this%read_hkx3(status)
    call this%read_hkx5(status)

    call this%read_rawxy_sel(status)

    if(present(filename)) call this%close(status)

  end subroutine read_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write an eventfile
  !
  !
  !> @param[inout] this, status
  !---------------------------------------------------------------------------

  subroutine write_tEventFile(this, status, filename)
    implicit none

    class(tEventFile),      intent(inout) :: this
    integer,                intent(inout) :: status
    character(*), optional, intent(in) :: filename

    character(*), parameter :: function = "write_tEventFile"
    character(buffsize) :: sass_parint
    integer :: ierr
    integer, dimension(1) :: dummy_naxes
    dummy_naxes(1) = 0

    if(status > 100 .or. status == s_errRep) return

    if(present(filename)) call this%open(create, status, filename=filename)

    ! update std_header with filename
    call getenv("SASS_PARINT", sass_parint)
    if(trim(sass_parint) == "PIPELINE") then
      call update_stdhdr(this%name, this%std_header, status)
    end if

    if(this%ext_image    == true) then
      call this%write_image(status)
    else
      call FTIIMG(this%unit, 32, 1, dummy_naxes, ierr) ! 20200429 IMS: FITSIO cocks up if we try to write a keyword to the PHDU before creating some sort of image.
      call report_FITS_err(function, ierr, status)
      call fwrite_stdhdr(this%unit,this%std_header,status)
    end if
    if(this%ext_events   == true) call this%write_events(status)
    if(this%ext_corrat   == true) call this%write_corrat(status)
    if(this%ext_deadcor  == true) call this%write_deadcor(status)
    if(this%ext_badpix   == true) call this%write_badpix(status)
    if(this%ext_hk       == true) call this%write_hk(status)
    if(this%ext_hkx1     == true) call this%write_hkx1(status)
    if(this%ext_hkx2     == true) call this%write_hkx2(status)
    if(this%ext_hkx3     == true) call this%write_hkx3(status)
    if(this%ext_hkx5     == true) call this%write_hkx5(status)
    if(this%ext_offset   == true) call this%write_offset(status)
    if(this%ext_gti      == true) call this%write_gti(status)
    if(this%ext_qualgti  == true) call this%write_qualgti(status)
    if(this%ext_flaregti == true) call this%write_flaregti(status)
    if(this%ext_regtable == true) call this%write_region(status)
    if(this%ext_regiongti == true) call this%write_regiongti(status)

    if(this%filter_rawxy) call this%write_rawxy_sel(status)

    if(present(filename)) call this%close(status)

  end subroutine write_tEventFile

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free an eventfile
  !
  !
  !> @param[inout] this, status
  !---------------------------------------------------------------------------

  subroutine free_tEventFile(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tEventFile), intent(inout) :: this

    character(*), parameter :: function = "free_tEventFile"

    if(status > 100 .or. status == s_errRep) return

    call this%free_events(status)
    call this%free_image(status)
    call this%free_corrat(status)
    call this%free_deadcor(status)
    call this%free_badpix(status)
    call this%free_hk(status)
    call this%free_hkx1(status)
    call this%free_offset(status)
    call this%free_gti(status)
    call this%free_qualgti(status)
    call this%free_flaregti(status)
    call this%rawxy_mask%free(status)

  end subroutine free_tEventFile

  !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> ANDs the RAWX/Y selection with the mask in the RAWXY_MASK extension.
  !
  !
  !
  !> @param[in] this
  !> @param[in] rawxy_sel
  !> @param[in, optional] is_new
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine record_rawxy_sel_tEventFile(this, rawxy_sel, status, is_new)
    implicit none

    type(tRawXYSelection), intent(in) :: rawxy_sel
    integer, intent(inout) :: status
    logical, intent(in), optional :: is_new

    class(tEventFile) :: this

    character(*), parameter :: function = "record_rawxy_sel_tEventFile"

    integer :: xiLo,xiHi,yiLo,yiHi,xi,yi
    integer, dimension(2) :: naxis
!*** instead declare it a parameter in erolib_general?

    if(status > 100 .or. status == s_errRep) return

    if(present(is_new)) then
      if(is_new) then ! Can't concatenate the two 'ifs' since with Fortran there is no guarantee that B is not evaluated if .not.A in an expression such as 'A .and. B'.
        naxis(1) = n_ccd_pixels
        naxis(2) = n_ccd_pixels
        call this%rawxy_mask%init(status)
        this%rawxy_mask%extname = rawxy_ext
        this%rawxy_mask%values = this%rawxy_mask%all_bits_set
      end if
    end if

    if(allocated(rawxy_sel%mask%values)) then
      if(rawxy_sel%invert) then
        if(rawxy_sel%ccdnr_defined) then
          where(rawxy_sel%mask%values)&
            this%rawxy_mask%values = ibclr(this%rawxy_mask%values, rawxy_sel%ccdnr-1)
        else ! .not. rawxy_sel%ccdnr_defined. Clear all CCD bits inside the selected rectangle (including the boundary).
          where(rawxy_sel%mask%values)&
            this%rawxy_mask%values = 0
        end if
      else
        if(rawxy_sel%ccdnr_defined) then
          where(.not.rawxy_sel%mask%values)&
            this%rawxy_mask%values = ibclr(this%rawxy_mask%values, rawxy_sel%ccdnr-1)
        else ! .not. rawxy_sel%ccdnr_defined. Clear all CCD bits inside the selected rectangle (including the boundary).
          where(.not.rawxy_sel%mask%values)&
            this%rawxy_mask%values = 0
        end if
      end if

  return
    end if

    if(rawxy_sel%axes(1)%lo%is_defined) then
      xiLo = rawxy_sel%axes(1)%lo%value
    else
      xiLo = 1
    end if
    if(rawxy_sel%axes(1)%hi%is_defined) then
      xiHi = rawxy_sel%axes(1)%hi%value
    else
      xiHi = n_ccd_pixels
    end if
    if(rawxy_sel%axes(2)%lo%is_defined) then
      yiLo = rawxy_sel%axes(2)%lo%value
    else
      yiLo = 1
    end if
    if(rawxy_sel%axes(2)%hi%is_defined) then
      yiHi = rawxy_sel%axes(2)%hi%value
    else
      yiHi = n_ccd_pixels
    end if

!*** Note that the output image array this%image seems to be constructed in evtool_image_mod.create_from_events_tImage2D_i4() with the indices in (x,y) order rather than (y,x). I have followed that convention in constructing this%rawxy_mask%values, but it probably warrants further digging - for example, to see what order the WCS keywords come in: which is 1 and which 2.

    if(rawxy_sel%invert) then
      if(rawxy_sel%ccdnr_defined) then
        this%rawxy_mask%values(xiLo:xiHi,yiLo:yiHi) = ibclr(this%rawxy_mask%values(xiLo:xiHi,yiLo:yiHi), rawxy_sel%ccdnr-1)
      else ! .not. rawxy_sel%ccdnr_defined. Clear all CCD bits inside the selected rectangle (including the boundary).
        this%rawxy_mask%values(xiLo:xiHi,yiLo:yiHi) = 0
      end if
    else
      if(rawxy_sel%ccdnr_defined) then
        this%rawxy_mask%values(:,1:yiLo-1)&
          = ibclr(this%rawxy_mask%values(:,1:yiLo-1),            rawxy_sel%ccdnr-1)
        this%rawxy_mask%values(1:xiLo-1,:)&
          = ibclr(this%rawxy_mask%values(1:xiLo-1,:),            rawxy_sel%ccdnr-1)
        this%rawxy_mask%values(xiHi+1:n_ccd_pixels,:)&
          = ibclr(this%rawxy_mask%values(xiHi+1:n_ccd_pixels,:), rawxy_sel%ccdnr-1)
        this%rawxy_mask%values(:,yiHi+1:n_ccd_pixels)&
          = ibclr(this%rawxy_mask%values(:,yiHi+1:n_ccd_pixels), rawxy_sel%ccdnr-1)

      else ! .not. rawxy_sel%ccdnr_defined. Clear all CCD bits outside the selected rectangle.
        this%rawxy_mask%values(:,1:yiLo-1) = 0
        this%rawxy_mask%values(1:xiLo-1,yiLo:yiHi) = 0
        this%rawxy_mask%values(xiHi+1:n_ccd_pixels,yiLo:yiHi) = 0
        this%rawxy_mask%values(:,yiHi+1:n_ccd_pixels) = 0
      end if
    end if

  end subroutine record_rawxy_sel_tEventFile

    !---------------------------------------------------------------------------
    !> @author
    !> Ian M Stewart
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> Read the RAWX/Y selection image extension.
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_rawxy_sel_tEventFile(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tEventFile) :: this

        character(*), parameter :: function = "read_rawxy_sel_tEventFile"
        logical :: extension_exists=.false.

        if(status > 100 .or. status == s_errRep)&
    return

        this%filter_rawxy = .false. ! default

        call this%rawxy_mask%read(this%unit, rawxy_ext, extension_exists, status)

        if(status == 0 .and. extension_exists) then
          if(size(this%rawxy_mask%values,1) /= n_ccd_pixels .or. &
             size(this%rawxy_mask%values,2) /= n_ccd_pixels) then

              call wrerr(function, s_err, "RAWX/Y selection image has wrong dimensions")
              status = s_errRep
    return
          end if

          this%filter_rawxy = .true.
        end if

    end subroutine read_rawxy_sel_tEventFile

    !---------------------------------------------------------------------------
    !> @author
    !> Ian M Stewart
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> Write the RAWX/Y selection image extension.
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_rawxy_sel_tEventFile(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tEventFile) :: this

        character(*), parameter :: function = "write_rawxy_sel_tEventFile"

        if(status > 100 .or. status == s_errRep) return

        call this%rawxy_mask%write(this%unit, status)

    end subroutine write_rawxy_sel_tEventFile

end module evtool_eventfile_mod
