! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module evtool_events_mod

  use ero_general ! need for flag masks.
  use erolib_str_mod
  use erolib_ffits_mod
  use erolib_general_mod
  use erolib_sort_mod
  use evtool_gti_mod
  use evtool_rawxy_mod
  use erolib_tbintable_mod
  use erolib_region_mod
  use erolib_math_mod
  use erolib_wcs_mod
  use radec2pix_mod
  use fitsimage, only : stdhdr_type, fread_stdhdr, fwrite_stdhdr, wcs_imahdr_type, &
    fread_wcshdr, fwrite_wcshdr, wcs_refhdr_type, update_stdhdr, evscalehdr_type, fwrite_evsclhdr

  !  use, intrinsic :: IEEE_ARITHMETIC

  implicit none

  private

!  integer(i4), parameter :: badFlagMask = ior(FLAG_CORRUPT_EVENT, FLAG_CORRUPT_FRAME)
! 20191230 IMS: moved this to ero_general_mod.f90.

  !  integer(i1), parameter :: tnull_i1 = -huge(tnull_i1)
  integer(i2), parameter :: tnull_i2 = -huge(tnull_i2)
  integer(i4), parameter :: tnull_i4 = -huge(tnull_i4)
  integer(i8), parameter :: tnull_i8 = -huge(tnull_i8)
  !  integer(r4), parameter :: tnull_r4 = -huge(tnull_r4)
  !  integer(r8), parameter :: tnull_r8 = -huge(tnull_r8)

  character(*), parameter :: events_ext = "EVENTS"

  character(*), parameter :: energy_col_old = "ENERGY"
  character(*), parameter :: ccdnr_col_old  = "CCDNR"

  integer, parameter :: numColsDefined = 21

  character(*), parameter :: time_col       = "TIME"
  character(*), parameter :: ra_col         = "RA"
  character(*), parameter :: dec_col        = "DEC"
  character(*), parameter :: x_col          = "X"
  character(*), parameter :: y_col          = "Y"
  character(*), parameter :: energy_col     = "PI"
  character(*), parameter :: ev_weight_col  = "EV_WEIGHT"
  character(*), parameter :: rawx_col       = "RAWX"
  character(*), parameter :: rawy_col       = "RAWY"
  character(*), parameter :: subx_col       = "SUBX"
  character(*), parameter :: suby_col       = "SUBY"
  character(*), parameter :: pha_col        = "PHA"
  character(*), parameter :: pat_typ_col    = "PAT_TYP"
  character(*), parameter :: pat_inf_col    = "PAT_INF"
  character(*), parameter :: ccdnr_col      = "TM_NR"
  character(*), parameter :: flag_col       = "FLAG"
  character(*), parameter :: frametime_col  = "FRAMETIME"
  character(*), parameter :: recordtime_col = "RECORDTIME"
  character(*), parameter :: pat_ind_col    = "PAT_IND"
  character(*), parameter :: amp_gain_col   = "AMP_GAIN"
  character(*), parameter :: amp_cor_col    = "AMP_COR"

  type tRawXYSelection
    type(tRangeI), dimension(2) :: axes
    integer :: ccdnr = 0
    logical :: ccdnr_defined = .false.
    logical :: invert = .false.
    type(tRawXYMask_logical) :: mask
  contains
    procedure, non_overridable :: print_me => print_tRawXYSelection
    procedure, non_overridable :: get_entry_strs => get_entry_strs_tRawXYSelection
  end type tRawXYSelection

  type tEvents
    type(tWcs) :: wcs
    integer :: use_wcs = 0
    integer :: use_tnull = 0

    real(r8) :: mjdref = 0.0
    character(buffsize) :: extname = ""
    integer :: nelements = 0
    real(r8), allocatable, dimension(:) :: time, ra, dec , subx, suby, x, y, frametime, recordtime
    real(r4), allocatable, dimension(:) :: energy,ev_weight,amp_gain,amp_cor
    integer(i4), allocatable, dimension(:) :: flag
    integer(i2), allocatable, dimension(:) :: rawx, rawy, pha, pat_typ
    integer(i1), allocatable, dimension(:) :: pat_inf,ccdnr,pat_ind
    !    real(r8) :: tscal_ra, tzero_ra, tscal_dec, tzero_dec, tscal_subx, tzero_subx, tscal_suby, tzero_suby

    integer(i4), allocatable, dimension(:) :: temp_frame ! Used by evprep to flag events with CORRUPT_FRAME.

    ! to be compatible with old simulation files - energy column name changed from energy to pi
    ! a flag is set in the type indicating which column name was found
    integer :: use_old_energy = 0
    logical :: read_pat_ind=.false., write_pat_ind=.false.
    logical :: read_amp_gain=.false., write_amp_gain=.false.
    logical :: read_amp_cor=.false., write_amp_cor=.false.
    logical :: timecorr_was_performed = .false.

  contains
    procedure, non_overridable :: read => read_tEvents
    procedure, non_overridable :: write => write_tEvents
    procedure, non_overridable :: create_extension => create_extension_tEvents
    procedure, non_overridable :: sort => sort_tEvents
    procedure, non_overridable :: free => free_tEvents
    procedure, non_overridable :: init => init_tEvents
    procedure, non_overridable :: delete_duplicates => delete_duplicates_tEvents
    procedure, non_overridable :: flag_duplicates => flag_duplicates_tEvents
    procedure, non_overridable :: delete_flagged => delete_flagged_tEvents
    procedure, non_overridable :: apply_index_cols => apply_index_cols_tEvents
    procedure, non_overridable :: apply_gti => apply_gti_tEvents
    procedure, non_overridable :: apply_telid => apply_telid_tEvents
    procedure, non_overridable :: apply_energy => apply_energy_tEvents
    procedure, non_overridable :: apply_flag => apply_flag_tEvents
    procedure, non_overridable :: apply_pattern => apply_pattern_tEvents
    procedure, non_overridable :: apply_rawxy => apply_rawxy_tEvents
  end type tEvents

  public :: tRawXYSelection,tEvents, events_ext, x_col, y_col, alloc_events, dealloc_events, merge_events
  public :: tnull_i2,tnull_i4,tnull_i8
!,badFlagMask

contains

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> dealloacte events structures
  !
  !
  !
  !> @param[inout] events
  !> @param[inout] status
  !---------------------------------------------------------------------------
  subroutine dealloc_events(events, status)

    integer :: stat
    integer, intent(inout) :: status

    type(tEvents), allocatable, dimension(:) :: events

    character(*), parameter :: function = "dealloc_events"

    if(allocated(events)) then
      deallocate(events, stat=stat)
      call FError(function, stat, "dealloc events type", status)
    end if

  end subroutine dealloc_events

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloacte n events structures
  !
  !
  !
  !> @param[inout] events
  !> @param[in] n
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine alloc_events(events, n, status)
    implicit none

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    type(tEvents), allocatable, dimension(:) :: events

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_events"

    call dealloc_events(events, status)
    allocate(events(n), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

  end subroutine alloc_events

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create the merged events extensions for all infiles
  !> this will modify the X and Y col if the WCS system differs and the wcs aswell
  !
  !
  !> @param[in] EvtenFile_out
  !> @param[out] files
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine merge_events(events_in, events_out, status)
    implicit none

    integer, intent(inout) :: status
    integer(i4) :: ii, nelements, a, b
    integer :: radec2xy

    real(r8) :: correction

    type(tEvents), intent(inout) :: Events_out
    type(tEvents), intent(inout), dimension(:) :: Events_in

    character(*), parameter :: function = "merge_events"

    if(status > 100 .or. status == s_errRep) return

    nelements = 0
    do ii = 1, size(Events_in)
      nelements = nelements + Events_in(ii)%nelements
    end do
    ! reallocate memory for columns
    call Events_out%init(nelements, status)

    Events_out%nelements = nelements
    Events_out%extname = Events_in(1)%extname
    Events_out%mjdref = Events_in(1)%mjdref
    !    Events_out%tscal_ra = Events_in(1)%tscal_ra
    !    Events_out%tscal_dec = Events_in(1)%tscal_dec
    !    Events_out%tscal_subx = Events_in(1)%tscal_subx
    !    Events_out%tscal_suby = Events_in(1)%tscal_suby
    !    Events_out%tzero_ra = Events_in(1)%tzero_ra
    !    Events_out%tzero_dec = Events_in(1)%tzero_dec
    !    Events_out%tzero_subx = Events_in(1)%tzero_subx
    !    Events_out%tzero_suby = Events_in(1)%tzero_suby

    ! wcs keywords, check if the first EventFile has a WCS else ignore it
    if(Events_in(1)%use_wcs == true) then
      Events_out%wcs = Events_in(1)%wcs
      Events_out%use_wcs = true
    else
      Events_out%use_wcs = false
    end if

    ! use the flag for the old energy name of the first input file
    Events_out%use_old_energy = Events_in(1)%use_old_energy

    if(Events_out%nelements == 0) return

    ! merge all columns
    a = 1
    do ii = 1, size(Events_in)

      ! TODO do the radec2xy only if all wcs types are set
      if(Events_out%use_wcs == true) then
        radec2xy = false
        if(compare(Events_in(ii)%wcs%refpix, Events_in(1)%wcs%refpix) == false) radec2xy = true
        if(compare(Events_in(ii)%wcs%refval, Events_in(1)%wcs%refval) == false) radec2xy = true
        if(compare(Events_in(ii)%wcs%delta, Events_in(1)%wcs%delta) == false) radec2xy = true
        if(radec2xy == true) then
          Events_in(ii)%wcs = Events_in(1)%wcs
          call wrlog(function, "call radec2xy_wrapper")
          call radec2xy_wrapper(Events_in(ii)%ra, Events_in(ii)%dec, Events_in(ii)%x, Events_in(ii)%y, &
            Events_out%wcs%refval(1), Events_out%wcs%refval(2))
        end if
      end if

      b = a + Events_in(ii)%nelements - 1
      correction = 0_r8
      ! correct potential mjdref differences

      if(abs(Events_in(ii)%mjdref - Events_out%mjdref) > tiny(Events_out%mjdref)) then
        call wrlog(function, "mjdref correction applied to events extension")
        correction = (Events_in(ii)%mjdref - Events_out%mjdref) * 3600.0_r8 * 24.0_r8
      end if

      Events_out%time(a:b) = Events_in(ii)%time + correction

      Events_out%x(a:b) = Events_in(ii)%x
      Events_out%y(a:b) = Events_in(ii)%y
      Events_out%ra(a:b) = Events_in(ii)%ra
      Events_out%dec(a:b) = Events_in(ii)%dec
      Events_out%rawx(a:b) = Events_in(ii)%rawx
      Events_out%rawy(a:b) = Events_in(ii)%rawy
      Events_out%flag(a:b) = Events_in(ii)%flag
      Events_out%energy(a:b) = Events_in(ii)%energy
      Events_out%ev_weight(a:b) = Events_in(ii)%ev_weight
      Events_out%pha(a:b) = Events_in(ii)%pha
      Events_out%pat_typ(a:b) = Events_in(ii)%pat_typ
      Events_out%pat_ind(a:b) = Events_in(ii)%pat_ind
      Events_out%subx(a:b) = Events_in(ii)%subx
      Events_out%suby(a:b) = Events_in(ii)%suby
      Events_out%pat_inf(a:b) = Events_in(ii)%pat_inf
      Events_out%ccdnr(a:b) = Events_in(ii)%ccdnr
      Events_out%frametime(a:b) = Events_in(ii)%frametime
      Events_out%recordtime(a:b) = Events_in(ii)%recordtime
      Events_out%temp_frame(a:b) = Events_in(ii)%temp_frame
      Events_out%amp_gain(a:b) = Events_in(ii)%amp_gain
      Events_out%amp_cor(a:b) = Events_in(ii)%amp_cor

      if(status == s_errRep) return
      a = b + 1

    end do

  end subroutine merge_events

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort the type entries according to one array
  !
  !
  !
  !> @param[in] this
  !> @param[in] array
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_tEvents(this, array, status, ascend)
    implicit none

    integer, intent(inout) :: status
    integer, optional :: ascend

    class(tEvents) :: this
    class(*), dimension(:) :: array

    integer, dimension(:), allocatable :: order

    real(r8),allocatable, dimension(:) :: tmparray_r8
    real(r4),allocatable, dimension(:) :: tmparray_r4
    integer(i4),allocatable, dimension(:) :: tmparray_i4
    integer(i2),allocatable, dimension(:) :: tmparray_i2
    integer(i1),allocatable, dimension(:) :: tmparray_i1

    character(*), parameter :: function = "sort_tEvents"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%time) /= size(array)) then
      call wrerr(function // " tEvents type", s_err, "length of sorting array does not conform to length of type")
      status = s_errRep
    end if

    select type(array)
      type is ( real(r4) )
      tmparray_r4 = array
      call sort_real(tmparray_r4 ,status, order)
      type is ( real(r8) )
      tmparray_r8 = array
      call sort_double(tmparray_r8 ,status, order)
      type is ( integer(i4) )
      tmparray_i4 = array
      call sort_i4(tmparray_i4 ,status, order)
      type is ( integer(i2) )
      tmparray_i2 = array
      call sort_i2(tmparray_i2 ,status, order)
      type is ( integer(i1) )
      tmparray_i1 = array
      call sort_i1(tmparray_i1 ,status, order)
    end select

    if(present(ascend)) then
      if(ascend == false) call reverse_array(order)
    end if

    this%time = this%time(order)
    this%frametime = this%frametime(order)
    this%recordtime = this%recordtime(order)
    this%ra = this%ra(order)
    this%dec = this%dec(order)
    this%x = this%x(order)
    this%y = this%y(order)
    this%rawx = this%rawx(order)
    this%rawy = this%rawy(order)
    this%flag = this%flag(order)
    this%energy = this%energy(order)
    this%ev_weight = this%ev_weight(order)
    this%pha = this%pha(order)
    this%pat_typ = this%pat_typ(order)
    this%pat_ind = this%pat_ind(order)
    this%subx = this%subx(order)
    this%suby = this%suby(order)
    this%pat_inf = this%pat_inf(order)
    this%ccdnr = this%ccdnr(order)
    this%temp_frame = this%temp_frame(order)
    this%amp_gain = this%amp_gain(order)
    this%amp_cor = this%amp_cor(order)

    call dealloc(order, status)
    call dealloc(tmparray_i1, status)
    call dealloc(tmparray_i2, status)
    call dealloc(tmparray_i4, status)
    call dealloc(tmparray_r4, status)
    call dealloc(tmparray_r8, status)

  end subroutine sort_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the values from the events extension
  !
  !
  !
  !> @param[in] this (tEvents)
  !> @param[in] unit, ccdnr
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_tEvents(this, unit, success, status, mjdref, ext, use_wcs)
    implicit none

    class(tEvents), intent(inout) :: this

    integer :: nelements, rowsize, frow, nrows, lrow, felem
    integer :: time_colnum, x_colnum, y_colnum, ra_colnum, dec_colnum
    integer :: rawx_colnum, rawy_colnum, flag_colnum, energy_colnum, ev_weight_colnum, pha_colnum
    integer :: pat_typ_colnum, subx_colnum, suby_colnum, pat_inf_colnum, ccdnr_colnum,pat_ind_colnum
    integer :: amp_gain_colnum,amp_cor_colnum
    integer :: frametime_colnum, use_frametime_as_time, recordtime_colnum

    integer, intent(in) :: unit
    integer, intent(inout) :: status
    integer, intent(out) :: success

    integer, intent(in), optional :: use_wcs

    real(r8), optional, intent(in) :: mjdref

    character(*), parameter :: function = "read_tEvents"
    character(*), optional , intent(in) :: ext

    logical :: anyf,not_found

    if(status > 100 .or. status == s_errRep) return

    success = false
    status = ignore

    if(present(ext)) then
      call ffits_movnam_hdu(unit, BINARY_TBL, ext, status)
    else
      call ffits_movnam_hdu(unit, BINARY_TBL, events_ext, status)
    end if

    if(status == ignore) then
      status = ok
    else
      status = ok
      return
    end if

    call ffits_get_num_rows(unit, nelements, status)
    call ffits_get_rowsize(unit, rowsize, status)

    call this%init(nelements, status)


    if(present(use_wcs)) then
      this%use_wcs = use_wcs
    end if

    if(present(ext)) then
      this%extname = ext
    else
      this%extname = events_ext
    end if

    ! if no mjdref in the extension is found take the optional mjdref
    status = ignore
    call ffits_read_key(unit, "MJDREF", this%mjdref, status)
    if(status == ignore) then
      status = ok
    else
      status = ok
      if(present(mjdref)) this%mjdref = mjdref
    end if

    this%nelements = nelements

    not_found = .false.
    call ffits_get_colnum(unit, .false., frametime_col, frametime_colnum, status, not_found)
    call ffits_get_colnum(unit, .false., recordtime_col, recordtime_colnum, status, not_found)
    call ffits_get_colnum(unit, .false., x_col, x_colnum, status, not_found)
    call ffits_get_colnum(unit, .false., y_col, y_colnum, status, not_found)
    call ffits_get_colnum(unit, .false., ra_col, ra_colnum, status, not_found)
    call ffits_get_colnum(unit, .false., dec_col, dec_colnum, status, not_found)
    call ffits_get_colnum(unit, .false., rawx_col, rawx_colnum, status, not_found)
    call ffits_get_colnum(unit, .false., rawy_col, rawy_colnum, status, not_found)
    call ffits_get_colnum(unit, .false., flag_col, flag_colnum, status, not_found)
    call ffits_get_colnum(unit, .false., ev_weight_col, ev_weight_colnum, status, not_found)
    call ffits_get_colnum(unit, .false., pha_col, pha_colnum, status, not_found)
    call ffits_get_colnum(unit, .false., pat_typ_col, pat_typ_colnum, status, not_found)
    if(this%read_pat_ind)&
    call ffits_get_colnum(unit, .false., pat_ind_col, pat_ind_colnum, status, not_found)
    call ffits_get_colnum(unit, .false., subx_col, subx_colnum, status, not_found)
    call ffits_get_colnum(unit, .false., suby_col, suby_colnum, status, not_found)
    call ffits_get_colnum(unit, .false., pat_inf_col, pat_inf_colnum, status, not_found)
    if(this%read_amp_gain)&
    call ffits_get_colnum(unit, .false., amp_gain_col, amp_gain_colnum, status, not_found)
    if(this%read_amp_cor)&
    call ffits_get_colnum(unit, .false., amp_cor_col, amp_cor_colnum, status, not_found)

    if(status == ok) then
      status = ignore
      call ffits_get_colnum(unit, .false., ccdnr_col_old, ccdnr_colnum, status)
      if(status == ignore) then
        status = ok
        call wrerr(function, warn, "input event file has old column names replacing " // trim(ccdnr_col_old) // " with " // trim(ccdnr_col))
      else
        status = ok
        call ffits_get_colnum(unit, .false., ccdnr_col, ccdnr_colnum, status, not_found)
      end if
    end if
    
    use_frametime_as_time = false
    if(status == ok) then
      status = ignore

      call ffits_get_colnum(unit, .false., time_col, time_colnum, status)
      if(status /= ignore) then
        call wrlog(function, "input event file does not contain time - use frametime column instead")
        use_frametime_as_time = true
        status = ok
      end if
    end if

    ! try reading the new energy column - if not - try reading the old one
    status = ignore
    call ffits_get_colnum(unit, .false., energy_col, energy_colnum, status)
    if (status /= ignore) then
      status = ignore
      call ffits_get_colnum(unit, .false., energy_col_old, energy_colnum, status, not_found)
      if(status == ignore) then
        !        call wrlog(function, "old energy column name " // trim(energy_col_old) // " is found and will be used")
        status = ok
        this%use_old_energy = true
      end if
    else
      status = ok
    end if

    if (not_found) then
      call wrerr(function, s_err, "Some columns of the event extension were not found.")
      status = s_errRep
  return
    end if

    anyf = .false.
    felem = 1

    ! reading tscal and tzero keywords
    ! not longer needed every file written by evtool will be set to default scale parameters
    !    call ffits_read_key(unit, "TSCAL" // trim(int2str(ra_colnum)), this%tscal_ra, status)
    !    call ffits_read_key(unit, "TSCAL" // trim(int2str(dec_colnum)), this%tscal_dec, status)
    !    call ffits_read_key(unit, "TSCAL" // trim(int2str(subx_colnum)), this%tscal_subx, status)
    !    call ffits_read_key(unit, "TSCAL" // trim(int2str(suby_colnum)), this%tscal_suby, status)
    !
    !    call ffits_read_key(unit, "TZERO" // trim(int2str(ra_colnum)), this%tzero_ra, status)
    !    call ffits_read_key(unit, "TZERO" // trim(int2str(dec_colnum)), this%tzero_dec, status)
    !    call ffits_read_key(unit, "TZERO" // trim(int2str(subx_colnum)), this%tzero_subx, status)
    !    call ffits_read_key(unit, "TZERO" // trim(int2str(suby_colnum)), this%tzero_suby, status)

    ! read the wcs
    if(this%use_wcs == true) then
      call this%wcs%read(unit, BINARY_TBL, x_col, y_col, status)
    end if

    ! before reading the columns check if there is something to read

    if(nelements == 0 .or. status == s_errRep) then
      if(status == ok) success = true
      return
    end if

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1
      call ffits_read_col(unit, frametime_colnum, frow, felem, nrows, 0, this%frametime(frow:lrow), status, anyf)
      call ffits_read_col(unit, recordtime_colnum, frow, felem, nrows, 0, this%recordtime(frow:lrow), status, anyf)
      call ffits_read_col(unit, x_colnum, frow, felem, nrows, 0, this%x(frow:lrow), status, anyf)
      call ffits_read_col(unit, y_colnum, frow, felem, nrows, 0, this%y(frow:lrow), status, anyf)
      call ffits_read_col(unit, ra_colnum, frow, felem, nrows, 0, this%ra(frow:lrow), status, anyf)
      call ffits_read_col(unit, dec_colnum, frow, felem, nrows, 0, this%dec(frow:lrow), status, anyf)
      call ffits_read_col(unit, rawx_colnum, frow, felem, nrows, 0, this%rawx(frow:lrow), status, anyf)
      call ffits_read_col(unit, rawy_colnum, frow, felem, nrows, 0, this%rawy(frow:lrow), status, anyf)
      call ffits_read_col(unit, flag_colnum, frow, felem, nrows, 0, this%flag(frow:lrow), status, anyf)
      call ffits_read_col(unit, energy_colnum, frow, felem, nrows, 0, this%energy(frow:lrow), status, anyf)
      call ffits_read_col(unit, ev_weight_colnum, frow, felem, nrows, 0, this%ev_weight(frow:lrow), status, anyf)
      call ffits_read_col(unit, pha_colnum, frow, felem, nrows, 0, this%pha(frow:lrow), status, anyf)
      call ffits_read_col(unit, pat_typ_colnum, frow, felem, nrows, 0, this%pat_typ(frow:lrow), status, anyf)
      if(this%read_pat_ind)&
      call ffits_read_col(unit, pat_ind_colnum, frow, felem, nrows, 0, this%pat_ind(frow:lrow), status, anyf)
      call ffits_read_col(unit, subx_colnum, frow, felem, nrows, 0, this%subx(frow:lrow), status, anyf)
      call ffits_read_col(unit, suby_colnum, frow, felem, nrows, 0, this%suby(frow:lrow), status, anyf)
      call ffits_read_col(unit, pat_inf_colnum, frow, felem, nrows, 0, this%pat_inf(frow:lrow), status, anyf)
      call ffits_read_col(unit, ccdnr_colnum, frow, felem, nrows, 0, this%ccdnr(frow:lrow), status, anyf)
      if(use_frametime_as_time == false) then
        call ffits_read_col(unit, time_colnum, frow, felem, nrows, 0, this%time(frow:lrow), status, anyf)
      end if
      if(this%read_amp_gain)&
      call ffits_read_col(unit, amp_gain_colnum, frow, felem, nrows, 0, this%amp_gain(frow:lrow), status, anyf)
      if(this%read_amp_cor)&
      call ffits_read_col(unit, amp_cor_colnum, frow, felem, nrows, 0, this%amp_cor(frow:lrow), status, anyf)
    end do

    if(use_frametime_as_time == true) this%time = this%frametime

    if(status == ok) success = true

  end subroutine read_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write in the columns of an events extension
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_tEvents(this, unit, status, std_header)
    implicit none

    integer, intent(inout) :: status
    integer, intent(in) :: unit
    integer :: nelements, rowsize, frow, nrows, lrow, felem
    integer :: time_colnum, x_colnum, y_colnum, ra_colnum, dec_colnum
    integer :: rawx_colnum, rawy_colnum, flag_colnum, energy_colnum, ev_weight_colnum, pha_colnum
    integer :: pat_typ_colnum, subx_colnum, suby_colnum, pat_inf_colnum, ccdnr_colnum,pat_ind_colnum
    integer :: amp_gain_colnum,amp_cor_colnum
    integer :: frametime_colnum, recordtime_colnum

    integer(i4) :: i4_number

    class(tEvents) :: this
    type(stdhdr_type), optional, intent(in) :: std_header
    type(evscalehdr_type) :: evscale_header
    character(buffsize) :: col_str

    character(*), parameter :: function = "write_tEvents"

    if(status > 100 .or. status == s_errRep) return

    ! move to the correct extension
    status = ignore
    call ffits_movnam_hdu(unit, BINARY_TBL, this%extname, status)
    if(status /= ignore) then
      status = ok
      call wrchatty(function, "Create extensions " // trim(this%extname))
      call this%create_extension(unit, status)
    else
      status = ok
    end if
    
    call ffits_flush(unit, status)

    if(present(std_header)) then
      ! keywords written by write stdheader should be written with update routine instead of write routine
      call fwrite_stdhdr(unit, std_header, status)
      ! also write the event scale header
      call fwrite_evsclhdr(unit, evscale_header, status)
      ! write OGIP HDUCLAS1
      call ffits_update_key_str(unit, "HDUCLAS1", "EVENTS", "", status)
    end if

    if(this%timecorr_was_performed) then
      call ffits_update_key_str(unit, "T_OFFCOR", "T", "", status)
    else
      call ffits_update_key_str(unit, "T_OFFCOR", "F", "", status)
    end if

    call ffits_get_rowsize(unit, rowsize, status)
    nelements = this%nelements
    call ffits_get_colnum(unit, .false., time_col, time_colnum, status)
    call ffits_get_colnum(unit, .false., x_col, x_colnum, status)
    call ffits_get_colnum(unit, .false., y_col, y_colnum, status)
    call ffits_get_colnum(unit, .false., ra_col, ra_colnum, status)
    call ffits_get_colnum(unit, .false., dec_col, dec_colnum, status)
    call ffits_get_colnum(unit, .false., rawx_col, rawx_colnum, status)
    call ffits_get_colnum(unit, .false., rawy_col, rawy_colnum, status)
    call ffits_get_colnum(unit, .false., flag_col, flag_colnum, status)
    call ffits_get_colnum(unit, .false., ev_weight_col, ev_weight_colnum, status)
    call ffits_get_colnum(unit, .false., pha_col, pha_colnum, status)
    call ffits_get_colnum(unit, .false., pat_typ_col, pat_typ_colnum, status)
    if(this%write_pat_ind)&
    call ffits_get_colnum(unit, .false., pat_ind_col, pat_ind_colnum, status)
    call ffits_get_colnum(unit, .false., subx_col, subx_colnum, status)
    call ffits_get_colnum(unit, .false., suby_col, suby_colnum, status)
    call ffits_get_colnum(unit, .false., pat_inf_col, pat_inf_colnum, status)
    call ffits_get_colnum(unit, .false., ccdnr_col, ccdnr_colnum, status)
    call ffits_get_colnum(unit, .false., frametime_col, frametime_colnum, status)
    call ffits_get_colnum(unit, .false., recordtime_col, recordtime_colnum, status)
    if(this%write_amp_gain)&
    call ffits_get_colnum(unit, .false., amp_gain_col, amp_gain_colnum, status)
    if(this%write_amp_cor)&
    call ffits_get_colnum(unit, .false., amp_cor_col, amp_cor_colnum, status)

    if(this%use_old_energy == false) then
      call ffits_get_colnum(unit, .false., energy_col, energy_colnum, status)
    else
      call wrlog(function, "the old energy column name " // trim(energy_col_old) // " will be used")
      call ffits_get_colnum(unit, .false., energy_col_old, energy_colnum, status)
    end if
    felem = 1
    
    !    ! write tscal and tzero keywords must be performed before the column is written to be automatically applied
    !    call ffits_update_key(unit, "TSCAL" // trim(int2str(ra_colnum)), this%tscal_ra, "", status)
    !    call ffits_update_key(unit, "TSCAL" // trim(int2str(dec_colnum)), this%tscal_dec, "", status)
    !    call ffits_update_key(unit, "TSCAL" // trim(int2str(subx_colnum)), this%tscal_subx, "", status)
    !    call ffits_update_key(unit, "TSCAL" // trim(int2str(suby_colnum)), this%tscal_suby, "", status)
    !
    !    call ffits_update_key(unit, "TZERO" // trim(int2str(ra_colnum)), this%tzero_ra, "", status)
    !    call ffits_update_key(unit, "TZERO" // trim(int2str(dec_colnum)), this%tzero_dec, "", status)
    !    call ffits_update_key(unit, "TZERO" // trim(int2str(subx_colnum)), this%tzero_subx, "", status)
    !    call ffits_update_key(unit, "TZERO" // trim(int2str(suby_colnum)), this%tzero_suby, "", status)

    if(this%use_tnull == true) then

      col_str = "TNULL" // int2str(ra_colnum, status)
      call ffits_update_key(unit, col_str, tnull_i4, "NULL VALUE", status)
      col_str = "TNULL" // int2str(dec_colnum, status)
      call ffits_update_key(unit, col_str, tnull_i4, "NULL VALUE", status)
      col_str = "TNULL" // int2str(x_colnum, status)
      call ffits_update_key(unit, col_str, tnull_i4, "NULL VALUE", status)
      col_str = "TNULL" // int2str(y_colnum, status)
      call ffits_update_key(unit, col_str, tnull_i4, "NULL VALUE", status)

      i4_number = int(tnull_i2, kind(i4_number))
      col_str = "TNULL" // int2str(rawx_colnum, status)
      call ffits_update_key(unit, col_str, i4_number, "NULL VALUE", status)
      col_str = "TNULL" // int2str(rawy_colnum, status)
      call ffits_update_key(unit, col_str, i4_number, "NULL VALUE", status)
      col_str = "TNULL" // int2str(pha_colnum, status)
      call ffits_update_key(unit, col_str, i4_number, "NULL VALUE", status)
      col_str = "TNULL" // int2str(pat_typ_colnum, status)
      call ffits_update_key(unit, col_str, i4_number, "NULL VALUE", status)
!      if(this%write_pha_qua) then
!        col_str = "TNULL" // int2str(pha_qua_colnum, status)
!        call ffits_update_key(unit, col_str, i4_number, "NULL VALUE", status)
!      end if

!      if(this%write_pat_ind) then
!        i4_number = int(tnull_i1, kind(i4_number))
!        col_str = "TNULL" // int2str(pat_ind_colnum, status)
!        call ffits_update_key(unit, col_str, i4_number, "NULL VALUE", status)
!      end if

    !      i4_number = int(tnull_i1, kind(i4_number))
    !      col_str = "TNULL" // int2str(pat_inf_colnum, status)
    !      call ffits_update_key(unit, col_str, i4_number, "NULL VALUE", status)
    !      col_str = "TNULL" // int2str(ccdnr_colnum, status)
    !      call ffits_update_key(unit, col_str, i4_number, "NULL VALUE", status)
    end if


    call ffits_flush(unit, status)

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1
      call ffits_write_col(unit, time_colnum, frow, felem, nrows, this%time(frow:lrow), status)
      call ffits_write_col(unit, x_colnum, frow, felem, nrows, this%x(frow:lrow), status)
      call ffits_write_col(unit, y_colnum, frow, felem, nrows, this%y(frow:lrow), status)
      call ffits_write_col(unit, ra_colnum, frow, felem, nrows, this%ra(frow:lrow), status)
      call ffits_write_col(unit, dec_colnum, frow, felem, nrows, this%dec(frow:lrow), status)
      call ffits_write_col(unit, rawx_colnum, frow, felem, nrows, this%rawx(frow:lrow), status)
      call ffits_write_col(unit, rawy_colnum, frow, felem, nrows, this%rawy(frow:lrow), status)
      call ffits_write_col(unit, flag_colnum, frow, felem, nrows, this%flag(frow:lrow), status)
      call ffits_write_col(unit, energy_colnum, frow, felem, nrows, this%energy(frow:lrow), status)
      call ffits_write_col(unit, ev_weight_colnum, frow, felem, nrows, this%ev_weight(frow:lrow), status)
      call ffits_write_col(unit, pha_colnum, frow, felem, nrows, this%pha(frow:lrow), status)
      call ffits_write_col(unit, pat_typ_colnum, frow, felem, nrows, this%pat_typ(frow:lrow), status)
      if(this%write_pat_ind)&
      call ffits_write_col(unit, pat_ind_colnum, frow, felem, nrows, this%pat_ind(frow:lrow), status)

      call ffits_write_col(unit, subx_colnum, frow, felem, nrows, this%subx(frow:lrow), status)

      call ffits_write_col(unit, suby_colnum, frow, felem, nrows, this%suby(frow:lrow), status)
      call ffits_write_col(unit, pat_inf_colnum, frow, felem, nrows, this%pat_inf(frow:lrow), status)
      call ffits_write_col(unit, ccdnr_colnum, frow, felem, nrows, this%ccdnr(frow:lrow), status)
      call ffits_write_col(unit, frametime_colnum, frow, felem, nrows, this%frametime(frow:lrow), status)
      call ffits_write_col(unit, recordtime_colnum, frow, felem, nrows, this%recordtime(frow:lrow), status)

      if(this%write_amp_gain)&
      call ffits_write_col(unit, amp_gain_colnum, frow, felem, nrows, this%amp_gain(frow:lrow), status)
      if(this%write_amp_cor)&
      call ffits_write_col(unit, amp_cor_colnum, frow, felem, nrows, this%amp_cor(frow:lrow), status)

    end do

    if(this%use_wcs == true) then
      call this%wcs%write(unit, BINARY_TBL, status)
    end if

  end subroutine write_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create a basic events extension in the file
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine create_extension_tEvents(this, unit, status)
    implicit none

    integer, intent(inout) :: status
    integer, intent(in) :: unit

    character(*), parameter :: function = "create_extension_tEvents"
    character(fcolnamelength), dimension(numColsDefined) :: ttype,ttype2
    character(fcolnamelength), dimension(numColsDefined) :: tform,tform2
    character(fcolnamelength), dimension(numColsDefined) :: tunit,tunit2
    integer :: actualNumCols=18

    class(tEvents) :: this
    type(tBinTable) :: table

    if(status > 100 .or. status == s_errRep) return

    ttype(1) = time_col		! r8
    ttype(2) = ra_col		! r8
    ttype(3) = dec_col		! r8
    ttype(4) = x_col		! r8
    ttype(5) = y_col		! r8
    ttype(6) = energy_col	! r4
    if(this%use_old_energy == true) ttype(6) = energy_col_old
    ttype(7) = ev_weight_col	! r4
    ttype(8) = rawx_col		! i2
    ttype(9) = rawy_col		! i2
    ttype(10) = subx_col	! r8
    ttype(11) = suby_col	! r8
    ttype(12) = pha_col		! i2
    ttype(13) = pat_typ_col	! i2
    ttype(14) = pat_inf_col	! i1
    ttype(15) = ccdnr_col	! i1
    ttype(16) = flag_col	! i4
    ttype(17) = frametime_col	! r8
    ttype(18) = recordtime_col	! r8
    ttype(19) = pat_ind_col	! i2
    ttype(20) = amp_gain_col	! r4
    ttype(21) = amp_cor_col	! r4

    tform = ["D","D","D","D","D","E","E","I","I","D","D","I","I","B","B","J","D","D","I","E","E"]
    tunit(:) = ""
    tunit(2) = "deg"
    tunit(3) = "deg"
    tunit(1) = "s"
    tunit(17) = "s"

    ttype2(:actualNumCols) = ttype(:actualNumCols)
    tform2(:actualNumCols) = tform(:actualNumCols)
    tunit2(:actualNumCols) = tunit(:actualNumCols)

    if(this%write_pat_ind) then
      actualNumCols = actualNumCols + 1
      ttype2(actualNumCols) = ttype(19)
      tform2(actualNumCols) = tform(19)
      tunit2(actualNumCols) = tunit(19)
    end if

    if(this%write_amp_gain) then
      actualNumCols = actualNumCols + 1
      ttype2(actualNumCols) = ttype(20)
      tform2(actualNumCols) = tform(20)
      tunit2(actualNumCols) = tunit(20)
    end if

    if(this%write_amp_cor) then
      actualNumCols = actualNumCols + 1
      ttype2(actualNumCols) = ttype(21)
      tform2(actualNumCols) = tform(21)
      tunit2(actualNumCols) = tunit(21)
    end if

    call table%init(ttype2(:actualNumCols), tform2(:actualNumCols), tunit2(:actualNumCols), status)
    call table%insertFits(unit, this%extname, status);
!*** remove the ; in master.

  end subroutine create_extension_tEvents

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> reduce all entries in the columns to contain only a subset of the complete
  !> by providing an array list with the desired indices
  !
  !
  !> @param[in] this
  !> @param[in] order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_index_cols_tEvents(this, order, status)
    implicit none

    integer, intent(inout) :: status

    class(tEvents) :: this

    integer, dimension(:), intent(in) :: order
    integer :: n

    character(*), parameter :: function = "apply_index_cols_tEvents"

    if(status > 100 .or. status == s_errRep) return

    n = size(order)
    if(n == 0) then
      !            call wrerr(function, warn, trim(this%extname) // " was reduced to zero entries")
      call wrchatty(function, trim(this%extname) // " was reduced to zero entries")
      call alloc(this%time, n, status)
      call alloc(this%ra, n, status)
      call alloc(this%dec, n, status)
      call alloc(this%x, n, status)
      call alloc(this%y, n, status)
      call alloc(this%rawx, n, status)
      call alloc(this%rawy, n, status)
      call alloc(this%subx, n, status)
      call alloc(this%suby, n, status)
      call alloc(this%energy, n, status)
      call alloc(this%ev_weight, n, status)
      call alloc(this%flag, n, status)
      call alloc(this%pha, n, status)
      call alloc(this%pat_typ, n, status)
      call alloc(this%pat_ind, n, status)
      call alloc(this%pat_inf, n, status)
      call alloc(this%ccdnr, n, status)
      call alloc(this%frametime, n, status)
      call alloc(this%recordtime, n, status)
      call alloc(this%temp_frame, n, status)
      call alloc(this%amp_gain, n, status)
      call alloc(this%amp_cor, n, status)
      this%nelements = n
      return
    end if

    this%time = this%time(order)
    this%ra = this%ra(order)
    this%dec = this%dec(order)
    this%x = this%x(order)
    this%y = this%y(order)
    this%rawx = this%rawx(order)
    this%rawy = this%rawy(order)
    this%subx = this%subx(order)
    this%suby = this%suby(order)
    this%energy = this%energy(order)
    this%ev_weight = this%ev_weight(order)
    this%flag = this%flag(order)
    this%pha = this%pha(order)
    this%pat_typ = this%pat_typ(order)
    this%pat_ind = this%pat_ind(order)
    this%pat_inf = this%pat_inf(order)
    this%ccdnr = this%ccdnr(order)
    this%frametime = this%frametime(order)
    this%recordtime = this%recordtime(order)
    this%temp_frame = this%temp_frame(order)
    this%amp_gain = this%amp_gain(order)
    this%amp_cor = this%amp_cor(order)

    this%nelements = n

  end subroutine apply_index_cols_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a GTI filter to the events type
  !
  !
  !
  !> @param[in] this
  !> @param[in] GTI
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_gti_tEvents(this, GTI, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tEvents) :: this
    type(tGTI), intent(in) :: GTI
    type(tGTI) :: iGTI

    integer, dimension(this%nelements) :: mask, mask_ccdnr
    integer, dimension(:), allocatable :: order

    character(*), parameter :: function = "apply_gti_tEvents"

    if(status > 100 .or. status == s_errRep) return

    iGTI = GTI
    ! internally correct the GTI
    if(iGTI%use_mjdref == true) then

      if(this%mjdref /= iGTI%mjdref) then
        call wrlog(function, "mjdref correction applied")
        iGTI%start = (iGTI%mjdref - this%mjdref ) * 3600.0_r8 * 24.0_r8 + iGTI%start
        iGTI%stop = (iGTI%mjdref - this%mjdref ) * 3600.0_r8 * 24.0_r8 + iGTI%stop
        iGTI%mjdref = this%mjdref
      end if

    end if

    call this%sort(this%time, status)
    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]
    mask = get_gti_mask(this%time, iGTI, status)

    if(iGTI%ccdnr >= 1 .and. iGTI%ccdnr <= 7) then
      mask_ccdnr = get_ccdnr_mask(this%ccdnr, iGTI%ccdnr, status, falseval=2)
      mask = mask * mask_ccdnr
    end if

    order = pack(order, mask > 0)
    call this%apply_index_cols(order, status)
    call dealloc(order, status)
    call iGTI%free(status)

  end subroutine apply_gti_tEvents

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a telid filter to the events type
  !
  !
  !
  !> @param[in] this
  !> @param[in] telid
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_telid_tEvents(this, telid, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, dimension(:), intent(in) :: telid

    class(tEvents) :: this

    integer, dimension(this%nelements) :: mask
    integer, dimension(:), allocatable :: order

    character(*), parameter :: function = "apply_telid_tEvents"

    if(status > 100 .or. status == s_errRep) return

    call alloc(order, this%nelements, status)

    order = [(ii,ii=1,this%nelements,1)]

    mask = false
    do ii = 1, size(telid)
      where (this%ccdnr == telid(ii))
        mask = true
      end where
    end do

    order = pack(order, mask == true)
    call this%apply_index_cols(order, status)
    call dealloc(order, status)

  end subroutine apply_telid_tEvents

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a energy filter to the events type
  !
  !
  !
  !> @param[in] this
  !> @param[in] telid
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_energy_tEvents(this, emin, emax, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    real(r8), dimension(:), intent(in) :: emin, emax

    class(tEvents) :: this

    integer, dimension(this%nelements) :: mask
    integer, dimension(:), allocatable :: order

    character(*), parameter :: function = "apply_energy_tEvents"

    if(status > 100 .or. status == s_errRep) return

    call alloc(order, this%nelements, status)

    order = [(ii,ii=1,this%nelements,1)]

    mask = false
    do ii = 1, size(emin)
      where (this%energy <= emax(ii) .and. this%energy >= emin(ii))
        mask = true
      end where
    end do

    order = pack(order, mask == true)
    call this%apply_index_cols(order, status)
    call dealloc(order, status)

  end subroutine apply_energy_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a flag filter to the events type
  !
  !
  !
  !> @param[in] this
  !> @param[in] flag
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_flag_tEvents(this, flagsel, status, invert)
    implicit none

    class(tEvents) :: this
    integer, intent(in)           :: flagsel
    integer, intent(inout)        :: status
    logical, intent(in), optional :: invert

    integer :: ii
    logical, dimension(this%nelements) :: mask
    integer, dimension(:), allocatable :: order

    character(*), parameter :: function = "apply_flag_tEvents"

    if(status > 100 .or. status == s_errRep) return

    call alloc(order, this%nelements, status)

    order = [(ii,ii=1,this%nelements,1)]

    mask = .false.
    where (iand(this%flag, flagsel) == 0) ! Only events will pass for which the FLAG value has none of the bits set which are 'on' in flagsel, since only for such will the iand() yield zero.
      mask = .true.
    end where

    if(present(invert)) then
      if(invert) mask = .not.mask
    end if

    order = pack(order, mask)
    call this%apply_index_cols(order, status)
    call dealloc(order, status)

  end subroutine apply_flag_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a pattern filter to the events type
  !
  !
  !
  !> @param[in] this
  !> @param[in] flag
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_pattern_tEvents(this, pat_sel, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in) :: pat_sel

    class(tEvents) :: this

    integer(i4), parameter :: zero = 0_i4
    integer, dimension(this%nelements) :: mask, pattern, pat_typ
    integer, dimension(:), allocatable :: order

    character(*), parameter :: function = "apply_pattern_tEvents"

    if(status > 100 .or. status == s_errRep) return

    ! selection == 0 invalid and valid pattern are copied
    if(pat_sel == 0) return

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]
    mask = false

    if(pat_sel < 0) then
      where(this%pat_typ <= 0)
        mask = true
      end where
      order = pack(order, mask == true)
      call this%apply_index_cols(order, status)
      call dealloc(order, status)
      return
    end if

    pattern = this%pat_typ

    ! set all invalid patterns to 0
    where(pattern <= 0)
      pattern = 0
      pat_typ = zero
    end where

    ! transform all patterns to bitmask starting at 0
    where(pattern > 0)
      pat_typ = ibset(zero, pattern - 1)
    end where

    ! set mask to true if pat_typ is found in bitmask pat_sel
    where (iand(pat_typ, pat_sel) /= 0)
      mask = true
    end where

    order = pack(order, mask == true)
    call this%apply_index_cols(order, status)
    call dealloc(order, status)

  end subroutine apply_pattern_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a RAWX filter to the events type
  !
  !
  !
  !> @param[in] this
  !> @param[in] rawx_sel
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_rawxy_tEvents(this, rawxy_sel, status)
    implicit none

    class(tEvents) :: this
    type(tRawXYSelection), intent(in) :: rawxy_sel
    integer, intent(inout) :: status

    integer(i4), dimension(:), allocatable :: order
    logical, dimension(this%nelements) :: mask, ccd_mask
    integer :: ii,jj
    logical :: inside_sel,outside_sel

    character(*), parameter :: function = "apply_rawxy_tEvents"

    if(status > 100 .or. status == s_errRep) return

    if(.not. (rawxy_sel%axes(1)%lo%is_defined .or. rawxy_sel%axes(1)%hi%is_defined &
         .or. rawxy_sel%axes(2)%lo%is_defined .or. rawxy_sel%axes(2)%hi%is_defined &
         .or. allocated(rawxy_sel%mask%values))) return

    if(rawxy_sel%invert) then
      inside_sel  = .false.
    else
      inside_sel  = .true.
    end if
    outside_sel = .not.inside_sel

    if(rawxy_sel%ccdnr_defined) then ! no rawx/y selection done where event CCD /= rawxy_sel%ccdnr.
      mask = .true.
      where (this%ccdnr == rawxy_sel%ccdnr)&
        mask = inside_sel ! default.
    else ! same rawx/y selection will be done regardless of event CCD.
      mask = inside_sel ! default.
    end if

    if(rawxy_sel%mask%containsData) then
      if(rawxy_sel%ccdnr_defined) then
        ccd_mask = .false.
        where (this%ccdnr == rawxy_sel%ccdnr)&
          ccd_mask = .true.

        order = [(ii,ii=1, this%nelements,1)]
        order = pack(order, ccd_mask)

        do ii=1,size(order)
          jj = order(ii)
          if(.not.rawxy_sel%mask%values(this%rawx(jj),this%rawy(jj))) then
            mask(jj) = outside_sel
          end if
        end do

      else
        do jj=1,this%nelements
          if(.not.rawxy_sel%mask%values(this%rawx(jj),this%rawy(jj))) then
            mask(jj) = outside_sel
          end if
        end do
      end if
      
    else
      if(rawxy_sel%ccdnr_defined) then
        if(rawxy_sel%axes(1)%lo%is_defined) then
          where (this%rawx < rawxy_sel%axes(1)%lo%value .and. this%ccdnr == rawxy_sel%ccdnr)
            mask = outside_sel
          end where
        end if
        if(rawxy_sel%axes(1)%hi%is_defined) then
          where (this%rawx > rawxy_sel%axes(1)%hi%value .and. this%ccdnr == rawxy_sel%ccdnr)
            mask = outside_sel
          end where
        end if

        if(rawxy_sel%axes(2)%lo%is_defined) then
          where (this%rawy < rawxy_sel%axes(2)%lo%value .and. this%ccdnr == rawxy_sel%ccdnr)
            mask = outside_sel
          end where
        end if
        if(rawxy_sel%axes(2)%hi%is_defined) then
          where (this%rawy > rawxy_sel%axes(2)%hi%value .and. this%ccdnr == rawxy_sel%ccdnr)
            mask = outside_sel
          end where
        end if

      else ! apply the selection to all CCDs.
        if(rawxy_sel%axes(1)%lo%is_defined) then
          where (this%rawx < rawxy_sel%axes(1)%lo%value)
            mask = outside_sel
          end where
        end if
        if(rawxy_sel%axes(1)%hi%is_defined) then
          where (this%rawx > rawxy_sel%axes(1)%hi%value)
            mask = outside_sel
          end where
        end if

        if(rawxy_sel%axes(2)%lo%is_defined) then
          where (this%rawy < rawxy_sel%axes(2)%lo%value)
            mask = outside_sel
          end where
        end if
        if(rawxy_sel%axes(2)%hi%is_defined) then
          where (this%rawy > rawxy_sel%axes(2)%hi%value)
            mask = outside_sel
          end where
        end if
      end if ! apply to a specific CCD or to all.
    end if ! choice between mask or rectangle.

    order = [(ii,ii=1, this%nelements,1)]
    order = pack(order, mask)

    call this%apply_index_cols(order, status)

    call dealloc(order, status)

  end subroutine apply_rawxy_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete duplicate entries in the tEvents type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_duplicates_tEvents(this, ccdnr_infile, status, report)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, jj, ninst

    class(tEvents) :: this

    logical, dimension(this%nelements) :: mask, mask_ccdnr
    integer, dimension(:), allocatable :: order, order_ccdnr
    integer, dimension(:), intent(in) :: ccdnr_infile
    integer, intent(in), optional :: report
    character(buffsize) :: nelements_dup_str

    character(*), parameter :: function = "delete_duplicates_tEvents"

    if(status > 100 .or. status == s_errRep) return

    ninst = size(ccdnr_infile)

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]

    call this%sort(this%pha, status)
    call this%sort(this%rawy, status)
    call this%sort(this%rawx, status)
    call this%sort(this%time, status)

    mask = .true.
    do jj = 1, ninst
      mask_ccdnr = .false.

      where (this%ccdnr == ccdnr_infile(jj))
        mask_ccdnr = .true.
      end where

      call alloc(order_ccdnr, this%nelements, status)
      order_ccdnr = [(ii,ii=1,this%nelements,1)]
      order_ccdnr = pack(order_ccdnr, mask_ccdnr)

      call wrlog(function, "Events are sorted in order time/rawx/rawy/pha.")
      call wrlog(function, "All events with the same time/rawx/rawy are removed.")
      call wrlog(function, "The first such event is retained but the sign of its PHA is set to negative.")
!      call wrlog(function, "evtool will as such not correctly merging files if duplicates are detected!!!!")
      do ii = 2, size(order_ccdnr)
        if(abs(this%time(order_ccdnr(ii)) - this%time(order_ccdnr(ii-1))) < tiny(this%time(1))) then
          if(this%rawx(order_ccdnr(ii)) == this%rawx(order_ccdnr(ii-1))) then
            if(this%rawy(order_ccdnr(ii)) == this%rawy(order_ccdnr(ii-1)))  then
              ! remove photons more conservately
              !              if(this%pha(order_ccdnr(ii)) == this%pha(order_ccdnr(ii-1))) then
              mask(order_ccdnr(ii)) = .false.
              this%pha(order_ccdnr(ii-1)) = -abs(this%pha(order_ccdnr(ii-1)))
            !              end if
            end if
          end if
        end if
      end do
    end do

    order = pack(order, mask)

    if(present(report)) then
      nelements_dup_str = int2str(this%nelements - size(order), status)
      call wrerr(function, report, "number of duplicate events removed: " // trim(nelements_dup_str))
    end if

    !   call wrlog(function, "delete duplicate disabled")
    call this%apply_index_cols(order, status)

    call dealloc(order, status)
    call dealloc(order_ccdnr, status)

  end subroutine delete_duplicates_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> flags duplicate entries in the tEvents type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine flag_duplicates_tEvents(this, ccdnr_infile, status)
    implicit none

    integer, intent(inout) :: status
    integer, dimension(:), intent(in) :: ccdnr_infile
    class(tEvents) :: this

    logical, dimension(this%nelements) :: mask_ccdnr,mask_bad_frame,mask_flag
    integer, dimension(:), allocatable :: order_ccdnr
    character(buffsize) :: nelements_dup_str,msg1,msg2
    integer :: ii,jj,ninst,start_ii,kk,num_rep,num_cases,total_num_rep,num_weird_frames
    real :: ave_num_rep
    integer(i4) :: start_frame

    character(*), parameter :: function = "flag_duplicates_tEvents"

    if(status > 100 .or. status == s_errRep) return

    ninst = size(ccdnr_infile)

    call this%sort(this%rawy, status)
    call this%sort(this%rawx, status)
    call this%sort(this%time, status)

    mask_flag = .false.
    where (iand(this%flag,badFlagMask) == 0) ! The iand will return non-zero if this%flag has any bits set which are set in badFlagMask. 
      mask_flag = .true.
    end where

    call wrlog(function, "All events in a frame which has some repeated events are flagged with CORRUPT_FRAME.")

    num_cases = 0
    total_num_rep = 0
    do jj = 1, ninst
      mask_ccdnr = .false.

      where (this%ccdnr == ccdnr_infile(jj) .and. mask_flag)
        mask_ccdnr = .true.
      end where

      call alloc(order_ccdnr, this%nelements, status)
      order_ccdnr = [(ii,ii=1,this%nelements,1)]
      order_ccdnr = pack(order_ccdnr, mask_ccdnr)

      ! Search through the ordered times/rawxs/rawys for 'cloned' events.
      !
      mask_bad_frame = .false.
      start_ii = 0
      num_weird_frames = 0
      do ii = 2, size(order_ccdnr)
        ! 20191011 IMS: If we were sure that repeated events would not overlap into
        ! different frames, we could here do:
        !   if(mask_bad_frame(order_ccdnr(ii))) continue

        if(abs(this%time(order_ccdnr(ii)) - this%time(order_ccdnr(ii-1))) < tiny(this%time(1))&
        .and. this%rawx(order_ccdnr(ii)) == this%rawx(order_ccdnr(ii-1))&
        .and. this%rawy(order_ccdnr(ii)) == this%rawy(order_ccdnr(ii-1))) then ! we are in a stretch of clones.
          if(start_ii == 0) start_ii = ii-1

        else
          if(start_ii /= 0) then ! indicates that we are at the first event after a run of clones.
            start_frame = this%temp_frame(order_ccdnr(start_ii))
            do kk = start_ii+1,ii-1
              if(this%temp_frame(order_ccdnr(kk)) /= start_frame) num_weird_frames = num_weird_frames + 1
            end do

            where (this%temp_frame == start_frame)
              mask_bad_frame = .true.
            end where

            num_cases = num_cases + 1
            start_ii = 0 ! indicates we are outside a stretch of cloned events (and have not just concluded one).
          end if
        end if
      end do ! loop over events.

      num_rep = count(mask_bad_frame .and. (iand(this%flag, FLAG_CORRUPT_FRAME)==0))

      if(num_rep > 0) then
        if(num_rep>99999999) then
          msg1 = ">=10^8"
        else
          write (msg1, "(i8)"), num_rep
        end if

        write (msg2, "(a,i1,a)"), "TM=", ccdnr_infile(jj), ": "//trim(msg1)//" additional repeated events found."
        call wrchatty(function, msg2)

        where(mask_bad_frame)
          this%flag = ior(this%flag, FLAG_CORRUPT_FRAME)
        end where

        total_num_rep = total_num_rep + num_rep

        if(num_weird_frames>0) then
          if(num_weird_frames>99999999) then
            msg1 = ">=10^8"
          else
            write (msg1, "(i8)"), num_weird_frames
          end if

          write (msg2, "(a,i1,a)"), "TM=", ccdnr_infile(jj), ": "//trim(msg1)//" repeated events have DIFFERENT frames!"
          call wrchatty(function, msg2)
        end if
      end if ! num_rep > 0
    end do ! loop over insts present.

    if(num_cases>0) then
      ave_num_rep = total_num_rep/real(num_cases)
      write (msg1, "(i10)"), num_cases
      write (msg2, "(a,es8.2)"), "Total additional number of cases of repeated events was "//trim(adjustl(msg1))//". Average num repeats was ", ave_num_rep
    else
      msg2 = "No more repeated events found."
    end if
    call wrlog(function, msg2)

    call dealloc(order_ccdnr, status)

  end subroutine flag_duplicates_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine remove_flagged_events.
  !> @brief
  !> Removes flagged events before deriving GTIs, HK frametimes.
  !
  !
  !> @param
  !> @param[in]
  !> @param[out]
  !> @param[inout] events_pp,status
  !---------------------------------------------------------------------------

  subroutine delete_flagged_tEvents(this, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tEvents) :: this

    integer, allocatable, dimension(:) :: order
    logical, dimension(this%nelements) :: mask
    integer :: ii

    character(*), parameter :: function = "delete_flagged_tEvents"

    if(status > 100 .or. status == s_errRep) return

    order = [(ii,ii=1,this%nelements,1)]
    mask = .false.
    where (iand(this%flag,badFlagMask) == 0) ! The iand will return non-zero if this%flag has any bits set which are set in badFlagMask. 
      mask = .true.
    end where
    order = pack(order, mask)
    call this%apply_index_cols(order, status)
    call dealloc(order, status)

  end subroutine delete_flagged_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the content of tEvents type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_tEvents(this, status)
    implicit none

    integer, intent(inout) :: status

    class(tEvents) :: this

    character(*), parameter :: function = "free_tEvents"

    !        this%nelements = 0
    call dealloc(this%time, status)
    call dealloc(this%x, status)
    call dealloc(this%y, status)
    call dealloc(this%ra, status)
    call dealloc(this%dec, status)
    call dealloc(this%rawx, status)
    call dealloc(this%rawy, status)
    call dealloc(this%flag, status)
    call dealloc(this%energy, status)
    call dealloc(this%ev_weight, status)
    call dealloc(this%pha, status)
    call dealloc(this%pat_typ, status)
    call dealloc(this%pat_ind, status)
    call dealloc(this%subx, status)
    call dealloc(this%suby, status)
    call dealloc(this%pat_inf, status)
    call dealloc(this%ccdnr, status)
    call dealloc(this%frametime, status)
    call dealloc(this%recordtime, status)
    call dealloc(this%temp_frame, status)
    call dealloc(this%amp_gain, status)
    call dealloc(this%amp_cor, status)

  end subroutine free_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloc the content of tEvents type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine init_tEvents(this, n, status, extname, tnull)
    implicit none

    integer, intent(in) :: n
    integer, intent(inout) :: status
    character(*), intent(in), optional :: extname
    integer, intent(in), optional :: tnull

    class(tEvents) :: this

    character(*), parameter :: function = "init_tEvents"

    call alloc(this%time, n, status)
    call alloc(this%x, n, status)
    call alloc(this%y, n, status)
    call alloc(this%ra, n, status)
    call alloc(this%dec, n, status)
    call alloc(this%rawx, n, status)
    call alloc(this%rawy, n, status)
    call alloc(this%flag, n, status)
    call alloc(this%energy, n, status)
    call alloc(this%ev_weight, n, status)
    call alloc(this%pha, n, status)
    call alloc(this%pat_typ, n, status)
    call alloc(this%pat_ind, n, status)
    call alloc(this%subx, n, status)
    call alloc(this%suby, n, status)
    call alloc(this%pat_inf, n, status)
    call alloc(this%ccdnr, n, status)
    call alloc(this%frametime, n, status)
    call alloc(this%recordtime, n, status)
    call alloc(this%temp_frame, n, status)
    call alloc(this%amp_gain, n, status)
    call alloc(this%amp_cor, n, status)

    this%use_tnull = false
    if(present(tnull)) then
      if(tnull == true) then
        this%use_tnull = true

        ! this%time = ieee_value(this%time, ieee_quiet_nan)
        ! this%energy(:) = ieee_value(this%energy, ieee_quiet_nan)
        ! this%ev_weight(:) = ieee_value(this%ev_weight, ieee_quiet_nan)

        this%ra = tnull_i4 * 1.D-6
        this%dec = tnull_i4 * 1.D-6
        this%x = tnull_i4
        this%y = tnull_i4
        !        call wrlog(function, "update tnull for subx/y")
        this%flag = null_i4 ! flag should have a value 0 since no flags have been applied

        this%rawx = tnull_i2
        this%rawy = tnull_i2
        this%pha = tnull_i2
        this%pat_typ = tnull_i2
!        this%pat_ind = tnull_i1

      end if
    end if

    this%nelements = n
    this%extname = events_ext
    if(present(extname)) this%extname = extname
    this%mjdref = 0_r8
    this%use_old_energy = false
    this%use_wcs = true

  end subroutine init_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get a mask for an array (ccdnr) of a specified ccd
  !
  !
  !
  !> @param[in] ccdnr_array, ccdnr
  !> @param[in, optional] falseval
  !> @param[inout] status
  !---------------------------------------------------------------------------

  function get_ccdnr_mask(ccdnr_array, ccdnr, status, falseval)
    implicit none

    integer, intent(inout) :: status
    integer(i4), intent(in) :: ccdnr
    integer(i4), intent(in), optional :: falseval
    integer(i1), dimension(:), intent(in) :: ccdnr_array
    integer, dimension(size(ccdnr_array)) :: get_ccdnr_mask

    character(*), parameter :: function = "get_ccdnr_mask"

    if(status > 100 .or. status == s_errRep) return

    get_ccdnr_mask(:) = false
    if(present(falseval)) get_ccdnr_mask(:) = falseval
    where(ccdnr_array == ccdnr)
      get_ccdnr_mask = true
    end where

  end function get_ccdnr_mask

  !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> Pretty-prints the struct tRawXYSelection.
  !
  !
  !
  !> @param[in] this
  !---------------------------------------------------------------------------

  subroutine print_tRawXYSelection(this)
    class(tRawXYSelection) :: this

    character(buffsize) :: entryStr,ccdnrStr,invertStr

    call this%get_entry_strs(entryStr, ccdnrStr, invertStr)

    print *, "rawxy = "//trim(entryStr)
    print *, "rawxy_telid = "//trim(ccdnrStr)
    print *, "rawxy_invert = "//trim(invertStr)

  end subroutine print_tRawXYSelection

  !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> Obtains print-friendly strings for the parameters rawxy and rawxy_ccdnr
  !
  !
  !
  !> @param[in] this
  !> @param[out] entryStr
  !> @param[out] ccdnrStr
  !---------------------------------------------------------------------------

  subroutine get_entry_strs_tRawXYSelection(this, entryStr, ccdnrStr, invertStr)
    character(*), intent(out) :: entryStr,ccdnrStr,invertStr
    class(tRawXYSelection) :: this

    character(buffsize) :: xLoStr,xHiStr,yLoStr,yHiStr

    if(this%axes(1)%lo%is_defined) then
      write (xLoStr, '(i3)'), this%axes(1)%lo%value
    else
      xLoStr = 'inf'
    end if
    if(this%axes(1)%hi%is_defined) then
      write (xHiStr, '(i3)'), this%axes(1)%hi%value
    else
      xHiStr = 'inf'
    end if
    if(this%axes(2)%lo%is_defined) then
      write (yLoStr, '(i3)'), this%axes(2)%lo%value
    else
      yLoStr = 'inf'
    end if
    if(this%axes(2)%hi%is_defined) then
      write (yHiStr, '(i3)'), this%axes(2)%hi%value
    else
      yHiStr = 'inf'
    end if

    entryStr = trim(xLoStr)//' '//trim(xHiStr)//' '//trim(yLoStr)//' '//trim(yHiStr)

    if(this%ccdnr_defined) then
      write (ccdnrStr, '(i1)'), this%ccdnr
    else
      ccdnrStr = 'undefined'
    end if

    if(this%invert) then
      invertStr = "yes"
    else
      invertStr = "no"
    end if

  end subroutine get_entry_strs_tRawXYSelection

        ! ------------------- imported from radec2xy


  subroutine radec2xy_wrapper(ra, dec, x, y, ra0, dec0)

    !-----------------------------------------------------------------------------
    ! HBrunner 20100430
    !
    ! Program calculates X and Y sky pixel FITS table columns from input RA
    ! and DEC columns, given an RA and DEC projection centre provided on the
    ! command line. Input parameters:
    !
    ! ra, dec     input coordinates
    ! x, y        output coordinates
    ! ra0      RA [deg] of projection centre
    ! dec0     DEC [deg] of projection centre
    !

    implicit none

    ! column buffers
    real(r8), dimension(:), intent(in) :: ra, dec
    real(r8), dimension(:), intent(out) :: x, y

    ! WCS keywords
    real(r8), parameter :: x0=0d0, y0=0d0, xd=-0.05/3600d0, yd=0.05/3600d0
    real(r8), intent(in) :: ra0, dec0

    character(*), parameter :: function = "radec2xy_wrapper"

    call radec2pix(ra, dec, size(ra), ra0, dec0, x0, y0, xd, yd, x, y)

  end subroutine radec2xy_wrapper

end module evtool_events_mod
