! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module evtool_events_pp_mod
  use param
  use ero_general ! need for flag masks.
  use erolib_tbintable_mod
  use erolib_ffits_col_mod
  use evtool_events_mod
  use fitsimage, only : stdhdr_type, fread_stdhdr, fwrite_stdhdr

  implicit none

  ! extnames used in the files
  character(*), parameter, private :: extname_default = "EVENTS"

  integer, private, parameter :: keylength = 32
  integer, private, parameter :: commentlength = 64


  type :: tPreProc_header

    character(keylength) :: date_obs_key = "DATE-OBS"
    character(keylength) :: date_end_key = "DATE-END"
    character(keylength) :: telescop_key = "TELESCOP"
    character(keylength) :: instrume_key = "INSTRUME"
    character(keylength) :: frametime_key = "FRAMETIM"
    character(keylength) :: mjdref_key = "MJDREF"
    character(keylength) :: tel_id_key = "TEL_ID"
    character(keylength) :: ninst_key = "NINST"
    character(keylength) :: app_id_key = "APP_ID"
    character(keylength) :: eroday_key = "ERODAY"

    character(keylength) :: date_obs = ""
    character(keylength) :: date_end = ""
    character(keylength) :: telescop = ""
    character(keylength) :: instrume = ""
    real(r8) :: frametime = 0.0
    real(r8) :: mjdref = 0.0
!    integer(i4) :: tel_id = 0 !
    integer :: tel_id = 0 !
    character(keylength) :: tel_id_str = ""
    character(keylength) :: app_id = ""
    integer(i4) :: eroday = 0

  contains
    procedure :: read => read_tPreProc_header

  end type tPreProc_header

  ! basic type of the catalog used for the cumulative list
  type, public :: tEvents_pp

    integer :: unit = 0
    character(buffsize) :: filename = ""
    character(buffsize) :: extname = extname_default
    type(tBinTable) :: table ! description of the binary table
    type(tPreProc_header) :: pp_header

    integer :: nelements = 0 ! number of elements (rows)

    !https://erosita.mpe.mpg.de/eROdoc/raw/rawFitsFileFormats_2018-08.html

    type(r8_tColumn) :: frametime ! the time of the frame, computed from OTS&frame as defined in CE Manual
    type(r8_tColumn) :: recordtime !(high resolution creation of the record, calculated from EXT_OTS, INT_OTS, SCTime, SubSec)
    type(r8_tColumn) :: time ! same as frametime, but with sporadic time jumps corrected.

    type(i2_tColumn) :: rawx
    type(i2_tColumn) :: rawy
    type(i2_tColumn) :: pha !(INDEF if not present, not present in PMENV2 mode)
    type(i2_tColumn) :: timesrc ! (the source of the time info in the timetag)
    type(i2_tColumn) :: datasrc !(the source of the event data, i.e. science data, testdata etc)
    type(i2_tColumn) :: subsec !(the subsec part of the frame time as defined in CE Manual)
    type(i2_tColumn) :: tmframecounter !(8 Bit counter of the Telemetry frame, unique for every telescope and data type)
    type(i2_tColumn) :: binnedamp !(binned amplitude instead of PHA, INDEF if not present, only present in PMENV2 mode)
    type(i2_tColumn) :: quality !(quality info on the energy reconstruction, INDEF if not present, only present in PMENV2 mode)
    type(i2_tColumn) :: dir1 !(optional column with directional info in compressed event mode)
    type(i2_tColumn) :: pha1 !(optional column with pha info in compressed event mode)
    type(i2_tColumn) :: dir2 !(optional column with directional info in compressed event mode)
    type(i2_tColumn) :: pha2 !(optional column with pha info in compressed event mode)
!    type(i2_tColumn) :: psequence !(ENV info, INDEF if not present, only present in PMENV and PMENV2 mode)

    type(i4_tColumn) :: frame !  the number of  currenct CCD frame from the timetag
    type(i4_tColumn) :: ext_ots
    type(i4_tColumn) :: int_ots
    type(i4_tColumn) :: sctime !(the time info in the frame time in seconds)
    type(i4_tColumn) :: env !(ENV info, INDEF if not present, only present in PMENV and PMENV2 mode)

    integer(i4), allocatable, dimension(:) :: flag
    integer(i1), allocatable, dimension(:) :: procmodes
    integer(i4), allocatable, dimension(:) :: rownum ! introduced to assist in bug finding.

  contains
    procedure :: init => init_tEvents_pp
    procedure :: free => free_tEvents_pp
    procedure :: open => open_tEvents_pp
    procedure :: close => close_tEvents_pp
    procedure :: create_extension => create_extension_tEvents_pp
    procedure :: read => read_tEvents_pp
    procedure :: write => write_tEvents_pp
    procedure :: apply_index => apply_index_tEvents_pp
    procedure :: sort => sort_tEvents_pp
    procedure :: reset_colnum => reset_colnum_tEvents_pp
    procedure :: flag_duplicates => flag_duplicates_tEvents_pp

  end type tEvents_pp

contains

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloacte n events_pp structures
  !
  !
  !
  !> @param[inout] events_pp
  !> @param[in] n
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine alloc_events_pp(events_pp, n, status)
    implicit none

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    class(tEvents_pp), allocatable, dimension(:) :: events_pp

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_events_pp"

    if(status > 100 .or. status == s_errRep) return

    call dealloc_events_pp(events_pp, status)
    allocate(events_pp(n), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

  end subroutine alloc_events_pp

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> dealloacte events_pp structures
  !
  !
  !
  !> @param[inout] events_pp
  !> @param[inout] status
  !---------------------------------------------------------------------------
  subroutine dealloc_events_pp(events_pp, status)

    integer :: stat
    integer, intent(inout) :: status

    class(tEvents_pp), allocatable, dimension(:) :: events_pp

    character(*), parameter :: function = "dealloc_events_pp"

    if(status > 100 .or. status == s_errRep) return

    if(allocated(events_pp)) then
      deallocate(events_pp, stat=stat)
      call FError(function, stat, "dealloc events_pp type", status)
    end if

  end subroutine dealloc_events_pp

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> init the tEvents_pp type
  !> nband = number of different energy bands
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional]
  !---------------------------------------------------------------------------

  subroutine init_tEvents_pp(this, n, status, extname)
    implicit none

    integer, intent(inout) :: status
    class(tEvents_pp) :: this
    integer, intent(in) :: n
    character(*), intent(in), optional :: extname

    character(*), parameter :: function = "init_tEvents_pp"

    if(status > 100 .or. status == s_errRep) return

    this%nelements = n
    this%extname = extname_default
    if(present(extname)) this%extname = extname
      call this%frametime%init("FRAMETIME", n, status)
      call this%time%init("TIME", n, status, is_optional=true) ! not present in the PP file but we need it internally.
      call this%rawx%init("RAWX", n, status, tunit="pixel")
      call this%rawy%init("RAWY", n, status, tunit="pixel")
      call this%pha%init("PHA", n, status, tunit="ADU")

      call this%frame%init("FRAME", n, status)
      call this%ext_ots%init("EXT_OTS", n, status, is_optional=true)
      call this%int_ots%init("INT_OTS", n, status, is_optional=true)
      call this%timesrc%init("TimeSRC", n, status)
      call this%datasrc%init("DataSRC", n, status)
      call this%sctime%init("SCTime", n, status, is_optional=true)
      call this%subsec%init("SubSec", n, status, is_optional=true)
      call this%recordtime%init("RecordTime", n, status, is_optional=true)
      call this%tmframecounter%init("TMFrameCounter", n, status, is_optional=true)
!      call this%psequence%init("PSequence", n, status, is_optional=true)
      call this%env%init("env", n, status)
      call this%binnedamp%init("BinnedAmp", n, status, is_optional=true)
      call this%quality%init("Quality", n, status, is_optional=true)
      call this%dir1%init("dir1", n, status, is_optional=true)
      call this%pha1%init("PHA1", n, status, tunit="ADU", is_optional=true)
      call this%dir2%init("dir2", n, status, is_optional=true)
      call this%pha2%init("PHA2", n, status, tunit="ADU", is_optional=true)

    call alloc(this%flag, n, status)
    call alloc(this%procmodes, n, status)
    call alloc(this%rownum, n, status)

    this%frametime%use_tnull = true
    this%time%use_tnull = true
    this%recordtime%use_tnull = true
    this%rawx%use_tnull = true
    this%rawy%use_tnull = true
    this%pha%use_tnull = true
    this%timesrc%use_tnull = true
    this%datasrc%use_tnull = true
    this%subsec%use_tnull = true
    this%tmframecounter%use_tnull = true
!    this%psequence%use_tnull = true
    this%binnedamp%use_tnull = true
    this%quality%use_tnull = true
    this%dir1%use_tnull = true
    this%pha1%use_tnull = true
    this%dir2%use_tnull = true
    this%pha2%use_tnull = true

    this%frame%use_tnull = true
    this%ext_ots%use_tnull = true
    this%int_ots%use_tnull = true
    this%sctime%use_tnull = true
    this%env%use_tnull = true

  end subroutine init_tEvents_pp

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the tEvents_pp type
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional] nband
  !---------------------------------------------------------------------------

  subroutine free_tEvents_pp(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tEvents_pp) :: this

    character(*), parameter :: function = "free_tEvents_pp"

    if(status > 100 .or. status == s_errRep) return

    this%nelements = 0

    call this%frametime%free(status)
    call this%time%free(status)
    call this%rawx%free(status)
    call this%rawy%free(status)
    call this%pha%free(status)

    call this%frame%free(status)
    call this%ext_ots%free(status)
    call this%int_ots%free(status)
    call this%timesrc%free(status)
    call this%datasrc%free(status)
    call this%sctime%free(status)
    call this%subsec%free(status)
    call this%recordtime%free(status)
    call this%tmframecounter%free(status)
!    call this%psequence%free(status)
    call this%env%free(status)
    call this%binnedamp%free(status)
    call this%quality%free(status)
    call this%dir1%free(status)
    call this%pha1%free(status)
    call this%dir2%free(status)
    call this%pha2%free(status)

  end subroutine free_tEvents_pp

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> routine to open the Events_pp
  !
  !
  !
  !> @param[inout] this
  !> @param[in] filename, rwmode
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine open_tEvents_pp(this, filename, rwmode, status, delete)
    implicit none

    integer, intent(inout) :: status
    class(tEvents_pp) :: this

    character(*), intent(in) :: filename
    integer, intent(in) :: rwmode

    integer :: exists

    integer, intent(in), optional :: delete

    character(*), parameter :: function = "open_tEvents_pp"

    if(status > 100 .or. status == s_errRep) return

    this%filename = filename

    call ffits_get_iounit(this%unit, status)
    if(rwmode /= create) then
      call ffits_open_file(this%unit, filename, rwmode, status)
      if(status == ok) call wrlog(function, trim(this%filename) // " successfully opened")
    else
      call ffits_file_exists(filename, exists, status)
      if(present(delete) .and. exists == 1) then
        if(delete == true) call delete_file_fortran(filename, status)
      end if
      call ffits_create_file(this%unit, filename, fblocksize, status)
      if(status == ok) call wrlog(function, trim(this%filename) // " successfully created")
    end if

  end subroutine open_tEvents_pp

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> close the tEvents_pp file
  !
  !
  !
  !> @param[inout] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine close_tEvents_pp(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tEvents_pp) :: this

    character(*), parameter :: function = "close_tEvents_pp"

    if(status > 100 .or. status == s_errRep) return

    call ffits_close_file(this%unit, status)
    call ffits_free_iounit(this%unit, status)

  end subroutine close_tEvents_pp

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> subroutine of class tEvents_pp
  !> create the extension of this type in a FITS file
  !
  !
  !> @param[inout] this
  !> @param[in, optional] aux
  !> @param[out]
  !> @param[inout]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine create_extension_tEvents_pp(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ncol ! number of columns in the file

    class(tEvents_pp) :: this
    character(*), parameter :: function = "create_extension_tEvents_pp"

    if(status > 100 .or. status == s_errRep) return

    ncol = 20

    call this%table%init(ncol, status)

      call this%table%insert_ffits_col(this%frametime, status)
!*** don't do 'time' because we did not read it from the PP file.
      call this%table%insert_ffits_col(this%rawx, status)
      call this%table%insert_ffits_col(this%rawy, status)
      call this%table%insert_ffits_col(this%pha, status)

      call this%table%insert_ffits_col(this%frame, status)
      call this%table%insert_ffits_col(this%ext_ots, status)
      call this%table%insert_ffits_col(this%int_ots, status)
      call this%table%insert_ffits_col(this%timesrc, status)
      call this%table%insert_ffits_col(this%datasrc, status)
      call this%table%insert_ffits_col(this%sctime, status)
      call this%table%insert_ffits_col(this%subsec, status)
      call this%table%insert_ffits_col(this%recordtime, status)
      call this%table%insert_ffits_col(this%tmframecounter, status)
!      call this%table%insert_ffits_col(this%psequence, status)
      call this%table%insert_ffits_col(this%env, status)
      call this%table%insert_ffits_col(this%binnedamp, status)
      call this%table%insert_ffits_col(this%quality, status)
      call this%table%insert_ffits_col(this%dir1, status)
      call this%table%insert_ffits_col(this%pha1, status)
      call this%table%insert_ffits_col(this%dir2, status)
      call this%table%insert_ffits_col(this%pha2, status)

    call this%table%insertFits(this%unit, this%extname, status)

  end subroutine create_extension_tEvents_pp

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the events_pp type from a FITS file
  !
  !
  !
  !> @param[in] this
  !> @param[out] filename
  !> @param[inout] status
  !> @param[in, optional] extend
  !---------------------------------------------------------------------------

  subroutine read_tEvents_pp(this, status, filename)
    implicit none

    integer, intent(inout) :: status
    integer :: nelements, rowsize, frow, lrow, nrows, ii
    class(tEvents_pp) :: this
    character(*), optional, intent(in) :: filename

    character(fcolnamelength) :: col
    character(*), parameter :: function = "read_tEvents_pp"

    if(status > 100 .or. status == s_errRep) return

    if(present(filename)) call this%open(filename, readonly, status)

    call ffits_movnam_hdu(this%unit, BINARY_TBL, this%extname, status)
    call ffits_get_num_rows(this%unit, nelements, status)
    call ffits_get_rowsize(this%unit, rowsize, status)

    call this%init(nelements, status)

    if(nelements == 0 .or. status == s_errRep) return

    call ffits_read_key(this%unit, "TEL_ID", this%pp_header%tel_id, status)

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1

        call this%frametime%read(this%unit, status, frow=frow, lrow=lrow)
        call this%rawx%read(this%unit, status, frow=frow, lrow=lrow)
        call this%rawy%read(this%unit, status, frow=frow, lrow=lrow)
        call this%pha%read(this%unit, status, frow=frow, lrow=lrow)

        call this%frame%read(this%unit, status, frow=frow, lrow=lrow)
        call this%ext_ots%read(this%unit, status, frow=frow, lrow=lrow)
        call this%int_ots%read(this%unit, status, frow=frow, lrow=lrow)
        call this%timesrc%read(this%unit, status, frow=frow, lrow=lrow)
        call this%datasrc%read(this%unit, status, frow=frow, lrow=lrow)
        call this%sctime%read(this%unit, status, frow=frow, lrow=lrow)
        call this%subsec%read(this%unit, status, frow=frow, lrow=lrow)
        call this%recordtime%read(this%unit, status, frow=frow, lrow=lrow)
        call this%tmframecounter%read(this%unit, status, frow=frow, lrow=lrow)
!        call this%psequence%read(this%unit, status, frow=frow, lrow=lrow)
        call this%env%read(this%unit, status, frow=frow, lrow=lrow)
        call this%binnedamp%read(this%unit, status, frow=frow, lrow=lrow)
        call this%quality%read(this%unit, status, frow=frow, lrow=lrow)
        call this%dir1%read(this%unit, status, frow=frow, lrow=lrow)
        call this%pha1%read(this%unit, status, frow=frow, lrow=lrow)
        call this%dir2%read(this%unit, status, frow=frow, lrow=lrow)
        call this%pha2%read(this%unit, status, frow=frow, lrow=lrow)
    end do

    this%rownum = [(ii,ii=1,nelements,1)]

    call this%pp_header%read(this%unit, status)

    if(present(filename)) call this%close(status)

  end subroutine read_tEvents_pp

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the events_pp type to a FITS file
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] aux
  !> @param[out] filename
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_tEvents_pp(this, status, filename)
    implicit none

    integer, intent(inout) :: status
    class(tEvents_pp) :: this
    integer :: rowsize, ii, frow, nelements, lrow, nrows
    character(*), optional, intent(in) :: filename

    character(*), parameter :: function = "write_tEvents_pp"

    if(status > 100 .or. status == s_errRep) return

    if(present(filename)) then
      call this%open(filename, create, status)
      if(status /= ok) return
    end if

    ! move to the correct extension
    status = ignore
    call ffits_movnam_hdu(this%unit, BINARY_TBL, this%extname, status)
    if(status /= ignore) then
      status = ok
      call this%create_extension(status)
    else
      status = ok
    end if

    call ffits_get_rowsize(this%unit, rowsize, status)
    nelements = this%nelements

    !    call fwrite_stdhdr(this%unit, this%stdhdr, status)

    ! write 1D arrays
    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1

        call this%frametime%write(this%unit, status, frow=frow, lrow=lrow)
! Don't do 'time' because we did not read it from the PP file.
        call this%rawx%write(this%unit, status, frow=frow, lrow=lrow)
        call this%rawy%write(this%unit, status, frow=frow, lrow=lrow)
        call this%pha%write(this%unit, status, frow=frow, lrow=lrow)

        call this%frame%write(this%unit, status, frow=frow, lrow=lrow)
        call this%ext_ots%write(this%unit, status, frow=frow, lrow=lrow)
        call this%int_ots%write(this%unit, status, frow=frow, lrow=lrow)
        call this%timesrc%write(this%unit, status, frow=frow, lrow=lrow)
        call this%datasrc%write(this%unit, status, frow=frow, lrow=lrow)
        call this%sctime%write(this%unit, status, frow=frow, lrow=lrow)
        call this%subsec%write(this%unit, status, frow=frow, lrow=lrow)
        call this%recordtime%write(this%unit, status, frow=frow, lrow=lrow)
        call this%tmframecounter%write(this%unit, status, frow=frow, lrow=lrow)
!        call this%psequence%write(this%unit, status, frow=frow, lrow=lrow)
        call this%env%write(this%unit, status, frow=frow, lrow=lrow)
        call this%binnedamp%write(this%unit, status, frow=frow, lrow=lrow)
        call this%quality%write(this%unit, status, frow=frow, lrow=lrow)
        call this%dir1%write(this%unit, status, frow=frow, lrow=lrow)
        call this%pha1%write(this%unit, status, frow=frow, lrow=lrow)
        call this%dir2%write(this%unit, status, frow=frow, lrow=lrow)
        call this%pha2%write(this%unit, status, frow=frow, lrow=lrow)

    end do

    if(present(filename)) call this%close(status)

  end subroutine write_tEvents_pp

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> reduce all entries in the columns to contain only a subset of the complete
  !> by providing an array list with the desired indices
  !
  !
  !> @param[in] this
  !> @param[in] order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_index_tEvents_pp(this, order, status)
    implicit none

    integer, intent(inout) :: status

    class(tEvents_pp) :: this

    integer, dimension(:), intent(in) :: order

    character(*), parameter :: function = "apply_index_tEvents_pp"

    if(status > 100 .or. status == s_errRep) return

    if(this%nelements == 0) return

    this%nelements = size(order)

      call this%frametime%apply_index(order, status)
      call this%time%apply_index(order, status)
      call this%rawx%apply_index(order, status)
      call this%rawy%apply_index(order, status)
      call this%pha%apply_index(order, status)

      call this%frame%apply_index(order, status)
      call this%ext_ots%apply_index(order, status)
      call this%int_ots%apply_index(order, status)
      call this%timesrc%apply_index(order, status)
      call this%datasrc%apply_index(order, status)
      call this%sctime%apply_index(order, status)
      call this%subsec%apply_index(order, status)
      call this%recordtime%apply_index(order, status)
      call this%tmframecounter%apply_index(order, status)
!      call this%psequence%apply_index(order, status)
      call this%env%apply_index(order, status)
      call this%binnedamp%apply_index(order, status)
      call this%quality%apply_index(order, status)
      call this%dir1%apply_index(order, status)
      call this%pha1%apply_index(order, status)
      call this%dir2%apply_index(order, status)
      call this%pha2%apply_index(order, status)

    this%flag = this%flag(order)
    this%procmodes = this%procmodes(order)
    this%rownum = this%rownum(order)


  end subroutine apply_index_tEvents_pp

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge all entries of the columns with a second tEvents_pp
  !
  !
  !> @param[in] Events_pp_A, Events_pp_B
  !> @param[in] Events_pp_C
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine merge_Events_pp(Events_pp_A, Events_pp_B, Events_pp_C, status)
    implicit none

    integer, intent(inout) :: status

    class(tEvents_pp), intent(in) :: Events_pp_A
    class(tEvents_pp), intent(in) :: Events_pp_B
    class(tEvents_pp), intent(out) :: Events_pp_C

    character(*), parameter :: function = "merge_Events_pp"

    if(status > 100 .or. status == s_errRep) return

    ! copy all important informations
    Events_pp_C%table        = Events_pp_A%table
    !    Events_pp_C%stdhdr        = Events_pp_A%stdhdr

    if(Events_pp_A%nelements == 0 .and. Events_pp_B%nelements == 0) then
      call wrlog(function, "cats to merge have zero elements - nothing to be done here")
      return
    end if

    Events_pp_C%nelements = Events_pp_A%nelements + Events_pp_B%nelements

      call merge_tColumn(Events_pp_A%frametime, Events_pp_B%frametime, Events_pp_C%frametime, status)
      call merge_tColumn(Events_pp_A%time, Events_pp_B%time, Events_pp_C%time, status)
      call merge_tColumn(Events_pp_A%rawx, Events_pp_B%rawx, Events_pp_C%rawx, status)
      call merge_tColumn(Events_pp_A%rawy, Events_pp_B%rawy, Events_pp_C%rawy, status)
      call merge_tColumn(Events_pp_A%pha, Events_pp_B%pha, Events_pp_C%pha, status)

      call merge_tColumn(Events_pp_A%frame, Events_pp_B%frame, Events_pp_C%frame, status)
      call merge_tColumn(Events_pp_A%ext_ots, Events_pp_B%ext_ots, Events_pp_C%ext_ots, status)
      call merge_tColumn(Events_pp_A%int_ots, Events_pp_B%int_ots, Events_pp_C%int_ots, status)
      call merge_tColumn(Events_pp_A%timesrc, Events_pp_B%timesrc, Events_pp_C%timesrc, status)
      call merge_tColumn(Events_pp_A%datasrc, Events_pp_B%datasrc, Events_pp_C%datasrc, status)
      call merge_tColumn(Events_pp_A%sctime, Events_pp_B%sctime, Events_pp_C%sctime, status)
      call merge_tColumn(Events_pp_A%subsec, Events_pp_B%subsec, Events_pp_C%subsec, status)
      call merge_tColumn(Events_pp_A%recordtime, Events_pp_B%recordtime, Events_pp_C%recordtime, status)
      call merge_tColumn(Events_pp_A%tmframecounter, Events_pp_B%tmframecounter, Events_pp_C%tmframecounter, status)
!      call merge_tColumn(Events_pp_A%psequence, Events_pp_B%psequence, Events_pp_C%psequence, status)
      call merge_tColumn(Events_pp_A%env, Events_pp_B%env, Events_pp_C%env, status)
      call merge_tColumn(Events_pp_A%binnedamp, Events_pp_B%binnedamp, Events_pp_C%binnedamp, status)
      call merge_tColumn(Events_pp_A%quality, Events_pp_B%quality, Events_pp_C%quality, status)
      call merge_tColumn(Events_pp_A%dir1, Events_pp_B%dir1, Events_pp_C%dir1, status)
      call merge_tColumn(Events_pp_A%pha1, Events_pp_B%pha1, Events_pp_C%pha1, status)
      call merge_tColumn(Events_pp_A%dir2, Events_pp_B%dir2, Events_pp_C%dir2, status)
      call merge_tColumn(Events_pp_A%pha2, Events_pp_B%pha2, Events_pp_C%pha2, status)

  end subroutine merge_Events_pp


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort the type entries according to one array
  !
  !
  !
  !> @param[in] this
  !> @param[in] array
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_tEvents_pp(this, array, status, ascend)
    implicit none

    integer, intent(inout) :: status
    integer, optional :: ascend

    class(tEvents_pp) :: this
    class(*), dimension(:) :: array

    integer, dimension(:), allocatable :: order

    real(r8),allocatable, dimension(:) :: tmparray_r8
    real(r4),allocatable, dimension(:) :: tmparray_r4
    integer(i4),allocatable, dimension(:) :: tmparray_i4
    integer(i2),allocatable, dimension(:) :: tmparray_i2
    integer(i1),allocatable, dimension(:) :: tmparray_i1

    character(*), parameter :: function = "sort_tEvents_pp"

    if(status > 100 .or. status == s_errRep) return

    if(this%nelements /= size(array)) then
      call wrerr(function // " tEvents_pp(_ext) type", s_err, "length of sorting array does not conform to length of type")
      status = s_errRep
    end if

    select type(array)
      type is ( real(r4) )
      tmparray_r4 = array
      call sort_real(tmparray_r4 ,status, order)
      type is ( real(r8) )
      tmparray_r8 = array
      call sort_double(tmparray_r8 ,status, order)
      type is ( integer(i4) )
      tmparray_i4 = array
      call sort_i4(tmparray_i4 ,status, order)
      type is ( integer(i2) )
      tmparray_i2 = array
      call sort_i2(tmparray_i2 ,status, order)
      type is ( integer(i1) )
      tmparray_i1 = array
      call sort_i1(tmparray_i1 ,status, order)
      class default
      call wrstop(function, "sorting tEvents_pp with the current array type is not supported for tEvents_pp")
    end select

    if(present(ascend)) then
      if(ascend == false) call reverse_array(order)
    end if

    call this%apply_index(order, status)

    call dealloc(order, status)
    call dealloc(tmparray_i1, status)
    call dealloc(tmparray_i2, status)
    call dealloc(tmparray_i4, status)
    call dealloc(tmparray_r4, status)
    call dealloc(tmparray_r8, status)

  end subroutine sort_tEvents_pp

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine read_tPreProc_header.
  !> @brief
  !> read the preproc header
  !
  !
  !> @param this
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_tPreProc_header(this, unit, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tPreProc_header) :: this
    integer, intent(in) :: unit

    !internal param

    character(*), parameter :: function = "read_tPreProc_header"

    if(status > 100 .or. status == s_errRep) return

    call ffits_read_key_str(unit, this%date_obs_key, this%date_obs, status)
    call ffits_read_key_str(unit, this%date_end_key, this%date_end, status)
    call ffits_read_key_str(unit, this%telescop_key, this%telescop, status)
    call ffits_read_key_str(unit, this%instrume_key, this%instrume, status)
    call ffits_read_key_str(unit, this%app_id_key, this%app_id, status)
    call ffits_read_key(unit, this%frametime_key, this%frametime, status)
    call ffits_read_key(unit, this%mjdref_key, this%mjdref, status)
!    call ffits_read_key(unit, tel_id_key, this%tel_id, status)
    call ffits_read_key_str(unit, this%tel_id_key, this%tel_id_str, status)

    call ffits_read_key(unit, this%eroday_key, this%eroday, status)

  end subroutine read_tPreProc_header

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine reset_colnum_tEvents_pp.
  !> @brief
  !> reset column numbers
  !
  !
  !> @param this
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine reset_colnum_tEvents_pp(this, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tEvents_pp) :: this

    !internal param

    character(*), parameter :: function = "reset_colnum_tEvents_pp"

    if(status > 100 .or. status == s_errRep) return

      ! reset column numbering
    ! since in testfiles is a column called psequence
    !      call wrlog(function, "reset column numbers - change when new testfiles are available")
    this%frametime%num = 0
    this%rawx%num = 0
    this%rawy%num = 0
    this%pha%num = 0
    this%frame%num = 0
    this%ext_ots%num = 0
    this%int_ots%num = 0
    this%timesrc%num = 0
    this%datasrc%num = 0
    this%sctime%num = 0
    this%subsec%num = 0
    this%recordtime%num = 0
    this%tmframecounter%num = 0
!    this%psequence%num = 0
    this%env%num = 0

    this%binnedamp%num = 0
    this%quality%num = 0
    this%dir1%num = 0
    this%pha1%num = 0
    this%dir2%num = 0
    this%pha2%num = 0

  end subroutine reset_colnum_tEvents_pp

  !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> flags duplicate entries in the tEvents_pp type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine flag_duplicates_tEvents_pp(this, status)
!***** could probably make this a bit faster if we omitted counting the bad frames.
    implicit none

    integer, intent(inout) :: status
    class(tEvents_pp) :: this

    logical, dimension(this%nelements) :: evmask_bad_frame,evmask_flag
    integer, dimension(:), allocatable :: order
    character(buffsize) :: nelements_dup_str,msg1,msg2,msg3
    integer :: ii,jj,start_ii,kk,num_rep,num_cases,total_num_rep,num_weird_frames
    integer :: num_frames_flagged_corrupt,nframes
    real :: ave_num_rep
    integer(i4) :: start_frame
    integer(i4), allocatable, dimension(:) :: u_frame_indices,frames
    logical, dimension(:), allocatable :: frame_is_bad,frame_mask

    character(*), parameter :: function = "flag_duplicates_tEvents_pp"

    if(status > 100 .or. status == s_errRep) return

    call this%sort(this%rawy%value, status)
    call this%sort(this%rawx%value, status)
    call this%sort(this%frametime%value, status)

    ! Set up a list of the unique frame values.
    !
    u_frame_indices = unique(this%frame%value, status)
    nframes = size(u_frame_indices)
    frames = this%frame%value(u_frame_indices)
    call alloc(frame_is_bad, nframes, status)
    call alloc(frame_mask, nframes, status)
    frame_is_bad = .false.

    evmask_flag = .false.
    where (iand(this%flag,badFlagMask) == 0) ! The iand will return non-zero if this%flag has any bits set which are set in badFlagMask. 
      evmask_flag = .true.
    end where

    call wrlog(function, "All events in a frame which has some repeated events are flagged with CORRUPT_FRAME.")

    frame_mask = .false.
    num_cases = 0
    total_num_rep = 0
      call alloc(order, this%nelements, status)
      order = [(ii,ii=1,this%nelements,1)]
      order = pack(order, evmask_flag)

      ! Search through the ordered times/rawxs/rawys for 'cloned' events.
      !
      evmask_bad_frame = .false.
      start_ii = 0
      num_weird_frames = 0
      do ii = 2, size(order)
        ! 20191011 IMS: If we were sure that repeated events would not overlap into
        ! different frames, we could here do:
        !   if(evmask_bad_frame(order(ii))) continue

        if(abs(this%frametime%value(order(ii)) - this%frametime%value(order(ii-1))) < tiny(this%frametime%value(1))&
        .and. this%rawx%value(order(ii)) == this%rawx%value(order(ii-1))&
        .and. this%rawy%value(order(ii)) == this%rawy%value(order(ii-1))) then ! we are in a stretch of clones.
          if(start_ii == 0) start_ii = ii-1

        else
          if(start_ii /= 0) then ! indicates that we are at the first event after a run of clones.
            start_frame = this%frame%value(order(start_ii))

            frame_mask = .false.
            where(frames==start_frame) frame_mask = .true.
            if(count(frame_mask)<1) then
              write (msg1, "(i8)"), start_frame
              call wrstop(function, "This should not happen. Frame "//trim(adjustl(msg1))//" not found in frame list.")
            else if(count(frame_mask)>1) then
              write (msg1, "(i8)"), start_frame
              call wrstop(function, "This should not happen. Frame "//trim(adjustl(msg1))//" found >once in frame list.")
            else
              where(frame_mask) frame_is_bad = .true.
            end if

            do kk = start_ii+1,ii-1
              if(this%frame%value(order(kk)) /= start_frame) num_weird_frames = num_weird_frames + 1
            end do

            where (this%frame%value == start_frame)
              evmask_bad_frame = .true.
            end where

            num_cases = num_cases + 1
            start_ii = 0 ! indicates we are outside a stretch of cloned events (and have not just concluded one).
          end if
        end if
      end do ! loop over events.

      num_rep = count(evmask_bad_frame .and. (iand(this%flag, FLAG_CORRUPT_FRAME)==0))
      num_frames_flagged_corrupt = count(frame_is_bad)

      if(num_rep > 0 .or. num_frames_flagged_corrupt>0) then
!*** put these in erolib_str_mod.f90???? Oops no, there's already a routine int2str there. Use that.
        if(num_rep>99999999) then
          msg1 = ">=10^8"
        else
          write (msg1, "(i8)"), num_rep
        end if

        if(num_frames_flagged_corrupt>99999999) then
          msg3 = ">=10^8"
        else
          write (msg3, "(i8)"), num_frames_flagged_corrupt
        end if

        write (msg2, "(a,i1,a)"), "TM=", this%pp_header%tel_id, ": "//trim(adjustl(msg1))//" repeated events in "//trim(adjustl(msg3))//" bad frames found."
        call wrchatty(function, msg2)

        where(evmask_bad_frame)
          this%flag = ior(this%flag, FLAG_CORRUPT_FRAME)
        end where

        total_num_rep = total_num_rep + num_rep

        if(num_weird_frames>0) then
          if(num_weird_frames>99999999) then
            msg1 = ">=10^8"
          else
            write (msg1, "(i8)"), num_weird_frames
          end if

          write (msg2, "(a,i1,a)"), "TM=", this%pp_header%tel_id, ": "//trim(msg1)//" repeated events have DIFFERENT frames!"
          call wrchatty(function, msg2)
        end if
      end if ! num_rep > 0

    if(num_cases>0) then
      ave_num_rep = total_num_rep/real(num_cases)
      write (msg1, "(i10)"), num_cases
      write (msg2, "(a,es8.2)"), "Total number of cases of repeated events was "//trim(adjustl(msg1))//". Average num repeats was ", ave_num_rep
    else
      msg2 = "Zero repeated events found."
    end if
    call wrlog(function, msg2)

    call dealloc(order, status)

    call dealloc(u_frame_indices, status)
    call dealloc(frames, status)
    call dealloc(frame_is_bad, status)
    call dealloc(frame_mask, status)

  end subroutine flag_duplicates_tEvents_pp

end module evtool_events_pp_mod
