! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module evtool_gti_mod

    use erolib_sort_mod
    use erolib_ffits_mod
    use erolib_general_mod
    use erolib_tbintable_mod
    implicit none

    private
    character(*), parameter :: start_col = "START"
    character(*), parameter :: stop_col = "STOP"
    character(*), parameter :: gti_ext = "GTI"
    character(*), parameter :: qualgti_ext = "QUALGTI"
    character(*), parameter :: flaregti_ext = "FLAREGTI"
    character(*), parameter :: external_gti_ext = "STDGTI"

    type tGTI
        integer :: ccdnr
        character(buffsize) :: extname
        integer :: nelements
        character(fcolnamelength) :: start_col_name, stop_col_name
        real(r8), dimension(:), allocatable :: start, stop
        real(r8) :: mjdref, timezero, tstart, tstop
        integer :: use_mjdref, use_timezero
    contains
        procedure, non_overridable :: read => read_tGTI
        procedure, non_overridable :: sort => sort_tGTI
        procedure, non_overridable :: replaceInFits => replaceInFits_tGTI
        procedure, non_overridable :: write => writeToFits_tGTI
        procedure, non_overridable :: init => init_tGTI
        procedure, non_overridable :: free => free_tGTI
        procedure, non_overridable :: print => print_tGTI
        procedure, non_overridable :: apply_gti => apply_gti_tGTI
        procedure, non_overridable :: create_extension => create_extension_tGTI
    end type tGTI

    public :: tGTI, get_GTI_mask, get_GTI_mask_logical, get_GTI_mask_extend, gti_ext, qualgti_ext, flaregti_ext, external_gti_ext, &
        alloc_gti, dealloc_gti, merge_gti, create_GTI_from_order

contains

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> dealloacte gti structures
    !
    !
    !
    !> @param[inout] gti
    !> @param[inout] status
    !---------------------------------------------------------------------------
    subroutine dealloc_gti(gti, status)

        integer :: stat
        integer, intent(inout) :: status

        type(tGTI), allocatable, dimension(:) :: gti

        character(*), parameter :: function = "dealloc_gti"

        if(allocated(gti)) then
            deallocate(gti, stat=stat)
            call FError(function, stat, "dealloc gti type", status)
        end if

    end subroutine dealloc_gti

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> alloacte n gti structures
    !
    !
    !
    !> @param[inout] gti
    !> @param[in] n
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine alloc_gti(gti, n, status)
        implicit none

        integer :: stat
        integer, intent(in) :: n
        integer, intent(inout) :: status

        type(tGTI), allocatable, dimension(:) :: gti

        character(buffsize) :: errmsg
        character(*), parameter :: function = "alloc_gti"

        call dealloc_gti(gti, status)
        allocate(gti(n), stat=stat, errmsg=errmsg)
        call FError(function, stat, errmsg, status)

    end subroutine alloc_gti

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> apply a gti to the gti option of 'and' and 'or'
    !
    !
    !
    !> @param[in] this
    !> @param[out] GTI, andor
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine apply_gti_tGTI(this, GTI, andor, status)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) :: andor

        class(tGTI) :: this
        type(tGTI), intent(in) :: GTI
        type(tGTI) :: tmp_gti

        character(*), parameter :: function = "apply_gti_tGTI"

        if(status > 100 .or. status == s_errRep) return

        call merge_gti(this, GTI, tmp_gti, andor, status)

        this%mjdref = tmp_gti%mjdref
        this%timezero = tmp_gti%timezero
        this%extname = tmp_gti%extname
        this%ccdnr = tmp_gti%ccdnr
        this%nelements = tmp_gti%nelements
        this%start_col_name = tmp_gti%start_col_name
        this%stop_col_name = tmp_gti%stop_col_name

        if(this%nelements > 0) then
            this%start = tmp_gti%start
            this%stop = tmp_gti%stop
            this%tstart = tmp_gti%tstart
            this%tstop = tmp_gti%tstop
        end if

        call tmp_gti%free(status)

    end subroutine apply_gti_tGTI

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> merge two GTI types
    !
    !
    !
    !> @param[in] a, b
    !> @param[out] c
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine merge_gti(a, b, c, andor, status)
        implicit none

        type(tGTI), intent(in) :: a
        type(tGTI), intent(in) :: b
        type(tGTI), intent(out) :: c
        type(tGTI) ::tmp_gti, b_tmp

        integer, intent(inout) :: status
        integer, intent(in) :: andor
        integer :: n

        character(*), parameter :: function = "merge_gti"

        if(status > 100 .or. status == s_errRep) return

        ! determine the number of intervals
        n = size(a%start) + size(b%start)

        b_tmp = b
        if(b_tmp%use_mjdref == true) then
            if(a%mjdref /= b_tmp%mjdref) then
                b_tmp%start = (b_tmp%mjdref - a%mjdref ) * 3600.0_r8 * 24.0_r8 + b_tmp%start
                b_tmp%stop = ( b_tmp%mjdref - a%mjdref ) * 3600.0_r8 * 24.0_r8 + b_tmp%stop
                b_tmp%mjdref = a%mjdref
            end if
        end if

        if(b_tmp%use_timezero == true) then
            if(a%timezero /= b%timezero) then
                b_tmp%start = (b_tmp%timezero - a%timezero ) + b_tmp%start
                b_tmp%stop = (b_tmp%timezero - a%timezero ) + b_tmp%stop
                b_tmp%timezero = a%timezero
            end if
        end if

        if(andor == or) then
            call merge_or_gti(a,b_tmp,tmp_gti, status)
        end if

        if(andor == and) then
            call merge_and_gti(a,b_tmp,tmp_gti, status)
        end if

        c = tmp_gti

        call tmp_gti%free(status)
        call b_tmp%free(status)

    end subroutine merge_gti


    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> merge intervals ("or")
    !
    !
    !
    !> @param[in] a_start, a_stop, b_start, b_stop
    !> @param[out] c_start, c_stop
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine merge_or_gti(a, b, c, status)
        implicit none

        integer, intent(inout) :: status
        integer :: nI, ii, n

        type(tGTI), intent(in) :: a, b
        type(tGTI), intent(out) :: c
        type(tGTI) :: tmp

        character(*), parameter :: function = "merge_or_gti"

        if(status > 100 .or. status == s_errRep) return

        call alloc(c%start, a%nelements + b%nelements, status)
        call alloc(c%stop, a%nelements + b%nelements, status)
        
        c%mjdref = a%mjdref
        c%timezero = a%timezero
        c%extname = a%extname
        c%ccdnr = a%ccdnr
        c%nelements = a%nelements + b%nelements
        c%start_col_name = a%start_col_name
        c%stop_col_name = a%stop_col_name

        ! TODO merging to gti with 0 elements yields errors at this point
        ! but nothings wrong therefore return
        if(c%nelements == 0) then
            call alloc(c%start, c%nelements, status)
            call alloc(c%stop, c%nelements, status)
            return
        end if

        if(a%nelements == 0) then
            c%nelements = b%nelements
            c%start = b%start
            c%stop = b%stop
            c%tstart = minVal(b%start)
            c%tstop = maxVal(b%stop)
        end if

        if(b%nelements == 0) then
            c%nelements = a%nelements
            c%start = a%start
            c%stop = a%stop
            c%tstart = minVal(a%start)
            c%tstop = maxVal(a%stop)
        end if
		
        if(a%nelements == 0 .or. b%nelements == 0) then
            call wrerr(function, warn, "merging gti with zero elements")
            return
        end if
		
        tmp = c

        tmp%start = [a%start, b%start]
        tmp%stop = [a%stop, b%stop]

        n = size(tmp%start)
        call tmp%sort(tmp%start, status);

        nI = 1
        c%start(nI) = tmp%start(1)
        c%stop(nI) = tmp%stop(1)

        do ii = 2, n
            if(c%stop(nI) < tmp%start(ii)) then
                nI = nI + 1
                c%start(nI) = tmp%start(ii)
                c%stop(nI) = tmp%stop(ii)
            else
                if(c%stop(nI) >= tmp%stop(ii)) cycle
                ! else c%stop(nI) > tmp%start(ii) .and. c%stop(nI) < tmp%stop(ii)
                c%stop(nI) = tmp%stop(ii)
            end if
        end do

        c%start = c%start(:nI)
        c%stop = c%stop(:nI)

        c%tstart = minVal(c%start)
        c%tstop = maxVal(c%stop)

        c%nelements = size(c%start)

        call tmp%free(status)

    end subroutine merge_or_gti

      !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> merge intervals ("and")
    !
    !
    !
    !> @param[in] a_start, a_stop, b_start, b_stop
    !> @param[out] c_start, c_stop
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine merge_and_gti(a, b, c, status)
        implicit none

        type(tGTI), intent(in) :: a, b
        type(tGTI), intent(out) :: c
        type(tGTI) :: tmp

        integer, intent(inout) :: status
        integer :: nI, ii, n, last
        integer, dimension(a%nelements) :: a0
        integer, dimension(b%nelements) :: b1
        integer, dimension(a%nelements + b%nelements) :: c01
        integer, dimension(:), allocatable :: order

        real(dp) :: last_start, last_stop

        character(*), parameter :: function = "merge_and_gti"

        if(status > 100 .or. status == s_errRep) return

        a0(:) = 0
        b1(:) = 1
        c01 = [a0, b1]

        c%mjdref = a%mjdref
        c%timezero = a%timezero
        c%extname = a%extname
        c%ccdnr = a%ccdnr
        c%nelements = a%nelements + b%nelements
        c%start_col_name = a%start_col_name
        c%stop_col_name = a%stop_col_name

        ! if either of the input GTI for and merging has zero entries so has the final GTI
        if(a%nelements == 0 .or. b%nelements == 0) then
            c%nelements = 0
            call alloc(c%start, c%nelements, status)
            call alloc(c%stop, c%nelements, status)
            return
        end if

        tmp = c

        call alloc(c%start, c%nelements, status)
        call alloc(c%stop, c%nelements, status)

        tmp%start = [a%start, b%start]
        tmp%stop = [a%stop, b%stop]

        n = tmp%nelements

        call tmp%sort(tmp%start, status, optorder = order)

        !        call tmp%sort(tmp%start, status, ascend=false)
        c01 = c01(order)

        nI = 1 ! next merged Interval number
        last_start = tmp%start(1)
        last_stop = tmp%stop(1)
        last = c01(1)

        do ii = 2, n
            ! if the last processed interval is of the same GTI => no overlaps
            ! => update last start / last stop
            if(last == c01(ii)) then
                last_start = tmp%start(ii)
                last_stop = tmp%stop(ii)
                cycle ! check the next interval
            end if

            c%start(nI) = max(last_start, tmp%start(ii))
            c%stop(nI) = min(last_stop, tmp%stop(ii))

            if(c%start(nI) < c%stop(nI)) then
                last_start = c%stop(nI)
                nI = nI + 1
                if(last_stop < tmp%stop(ii)) then
                    last_stop = tmp%stop(ii)
                    last = c01(ii)
                end if
            else ! no intersection
                last_start = tmp%start(ii)
                last_stop = tmp%stop(ii)
                last = c01(ii)
            end if

        end do

        ! changed from FATAL error to warn
        if(nI == 1) then
            call wrerr(function, warn, "no intersecting GTI")
            c%nelements = 0
            !            call dealloc(c%start, status)
            !            call dealloc(c%stop, status)
            call alloc(c%start, 0, status)
            call alloc(c%stop, 0, status)
            c%tstart=0
            c%tstop=0
            return
        end if
        ! since its next Interval number subtract 1
        c%start = c%start(:nI-1)
        c%stop = c%stop(:nI-1)

        c%tstart = minVal(c%start)
        c%tstop = maxVal(c%stop)

        c%nelements = size(c%start)

        call tmp%free(status)

    end subroutine merge_and_gti

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> create a basic deadcor extension in the file
    !
    !
    !
    !> @param[in] this
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine create_extension_tGTI(this, unit, status)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) :: unit

        character(*), parameter :: function = "create_extension_tGTI"
        character(fcolnamelength), dimension(2) :: ttype
        character(fcolnamelength), dimension(2) :: tform
        character(fcolnamelength), dimension(2) :: tunit

        class(tGTI) :: this
        type(tBinTable) :: table

        if(status > 100 .or. status == s_errRep) return

        ttype(1) = this%start_col_name
        ttype(2) = this%stop_col_name

        tform = ["D","D"]
        tunit(:) = ""

        call table%init(ttype, tform, tunit, status)
        call table%insertFits(unit, this%extname, status);

    end subroutine create_extension_tGTI

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> write a GTI extension to an existing fits file
    !
    !
    !
    !> @param[in] this
    !> @param[out] outParam
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine writeToFits_tGTI(this, unit, status)
        implicit none

        class(tGTI) :: this

        integer :: nelements, rowsize, frow, nrows, lrow, felem
        integer :: start_colnum, stop_colnum
        integer, intent(inout) :: status
        integer, intent(in) :: unit

        character(*), parameter :: function = "writeToFits_tGTI"

        if(status > 100 .or. status == s_errRep) return

        ! move to the correct extension
        status = ignore
        call ffits_movnam_hdu(unit, BINARY_TBL, this%extname, status)
        if(status /= ignore) then
            status = ok
            call this%create_extension(unit, status)
        else
            status = ok
        end if

        call ffits_get_rowsize(unit, rowsize, status)

        nelements = this%nelements

        call ffits_get_colnum(unit, .false., this%start_col_name, start_colnum, status)
        call ffits_get_colnum(unit, .false., this%stop_col_name, stop_colnum, status)

        felem = 1

        if(nelements > 0) then
            do frow = 1, nelements, rowsize
                nrows = min(nelements - frow + 1, rowsize)
                lrow = frow + nrows - 1
                call ffits_write_col(unit, start_colnum, frow, felem, nrows, this%start(frow:lrow), status)
                call ffits_write_col(unit, stop_colnum, frow, felem, nrows, this%stop(frow:lrow), status)
            end do
        end if

!        call ffits_update_key (unit, "MJDREF", this%mjdref, "[d] updated by evtool", status)
!        call ffits_update_key (unit, "TSTART", this%tstart, "updated by evtool", status)
!        call ffits_update_key (unit, "TSTOP", this%tstop, "updated by evtool", status)
               !        call ffits_update_key (unit, "TIMEZERO", this%timezero, "updated by evtool", status)

    end subroutine writeToFits_tGTI

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> replace the GTI extension of a file with this
    !
    !
    !
    !> @param[in] this
    !> @param[out] filename, extension, startcol, stopcol
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine replaceInFits_tGTI(this, filename, extname, startcol, stopcol, status, iounit)
        implicit none

        integer, intent(inout) :: status
        integer :: unit, exists, nrows, colnum
        integer, optional, intent(inout) :: iounit

        class(tGTI) :: this

        character(*), intent(in) :: filename, extname, startcol, stopcol
        character(*), parameter :: function = "replaceInFits_tGTI"

        if(status > 100 .or. status == s_errRep) return

        if(.not. present(iounit)) then
            call FTGIOU(unit, status)
            call ffits_file_exists(filename, exists, status);
            if(exists < 1) then
                call wrerr(function, s_err, trim(filename) // " does not exist")
                status = s_errRep
                return
            end if
            call ffits_open_table(unit, filename, readwrite, status)
        else
            unit = iounit
        end if

        ! go to the specified GTI extension
        call ffits_movnam_hdu(unit, BINARY_TBL, extname, status)

        call ffits_get_num_rows(unit, nrows, status)
        call ffits_delete_rows(unit, 1, nrows, status)

        call ffits_get_colnum(unit, .false., startcol, colnum, status)
        call ffits_write_col(unit, colnum, 1, 1, size(this%start), this%start, status)

        call ffits_get_colnum(unit, .false., stopcol, colnum, status)
        call ffits_write_col(unit, colnum, 1, 1, size(this%stop), this%stop, status)

        call ffits_update_key (unit, "MJDREF", this%mjdref, "recalculated MJDREF", status)
        call ffits_update_key (unit, "TIMEZERO", this%timezero, "recalculated TIMEZERO", status)

        call ffits_update_key (unit, "TSTART", minval(this%start), "recalculated TSTART", status)
        call ffits_update_key (unit, "TSTOP", maxval(this%stop), "recalculated TSTOP", status)

        ! if a unit was provided return before closing
        if(present(iounit)) return

        call ffits_close_file(unit, status)
        call FTFIOU(unit, status)


    end subroutine replaceInFits_tGTI

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> read in the GTI content of a specified extension
    !> colStart, colStop (name of the columns)
    !
    !> @param[in] filename, colStart, colStop
    !> @param[out] iounit, extension
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_tGTI(this, unit, extname, success, status, ccdnr, colStart, colStop, mjdref, timezero)

        integer, intent(inout) :: status
        integer, intent(inout) :: unit
        integer :: nelements, rowsize, frow, nrows, lrow, felem
        integer :: start_colnum, stop_colnum
        integer, optional, intent(in) :: ccdnr
        integer, intent(inout) :: success

        character(*), parameter :: function = "read_tGTI"
        character(*), intent(in) :: extname
        character(*), intent(in), optional :: colStart, colStop
        character(buffsize) :: telid_str

        real(r8), optional, intent(in) :: mjdref, timezero

        class(tGTI) :: this

        logical :: anyf

        if(status > 100 .or. status == s_errRep) return

        success = false
        this%start_col_name = start_col
        this%start_col_name = stop_col

        ! overwrite some settings
        if(present(colStart)) this%start_col_name = colStart
        if(present(colStop)) this%start_col_name = colStop
        if(present(ccdnr)) then
            telid_str = trim(int2str(ccdnr, status))
            if(status == s_errRep) return
        else
            telid_str = ""
        end if

        this%extname = extname // trim(telid_str)
        status = ignore
        call ffits_movnam_hdu(unit, BINARY_TBL, this%extname, status)
        if(status == ignore) then
            status = ok
        else
            status = ok
            return
        end if

        ! determine the number of rows for memory allocation
        call ffits_get_num_rows(unit, nelements, status)

        call ffits_get_rowsize(unit, rowsize, status)

        call this%init(nelements, status)

        if(present(ccdnr)) this%ccdnr = ccdnr
        this%nelements = nelements
        this%tstart = 0
        this%tstop = 0
        this%extname = extname // trim(telid_str)

        ! if no mjdref in the extension is found take the optional mjdref
        status = ignore
        call ffits_read_key(unit, "MJDREF", this%mjdref, status)
        if(status == ignore) then
            status = ok
        else
            status = ok
            if(present(mjdref)) this%mjdref = mjdref
        end if

        ! if no timezero in the extension is found take the optional mjdref
        status = ignore
        call ffits_read_key(unit, "TIMEZERO", this%timezero, status)
        if(status == ignore) then
            status = ok
        else
            status = ok
            if(present(timezero)) this%timezero = timezero
        end if

        if(present(ccdnr)) then
            this%ccdnr = ccdnr
        else
            this%ccdnr = -1
        end if

        call ffits_get_colnum(unit, .false., this%start_col_name, start_colnum, status)
        call ffits_get_colnum(unit, .false., this%stop_col_name, stop_colnum, status)

        anyf = .false.
        felem = 1

        if(nelements == 0 .or. status == s_errRep) then
            success = true
            return
        end if

        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1
            call ffits_read_col(unit, start_colnum, frow, felem, nrows, 0, this%start(frow:lrow), status, anyf)
            call ffits_read_col(unit, stop_colnum, frow, felem, nrows, 0, this%stop(frow:lrow), status, anyf)
        end do

        this%tstart = minval(this%start)
        this%tstop = maxval(this%stop)

        if(status == ok) success = true

    end subroutine read_tGTI

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> create a GTI type from an input time array and a potentially "continues order"
    !> the order is continues for individual GTI blocks
    !
    !
    !
    !> @param[in] inParam
    !> @param[out] outParam
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine create_GTI_from_order(time, order, gti, status)
        implicit none

        integer, intent(inout), dimension(:) :: order
        integer, dimension(:), allocatable :: neworder
        integer, intent(inout) :: status
        integer :: n, ii

        real(dp), dimension(:), intent(in) :: time

        type(tGTI), intent(out) :: gti

        character(*), parameter :: function = "create_GTI_from_order"

        if(status > 100 .or. status == s_errRep) return

        ! get the index numbers of the intervals (continues)
        call get_cont_intervals_int(order, neworder, status)

        n = size(neworder)

        call gti%init(n, status)

        ! implcit do loop to fill in the correct numbers
        ! start and stop (alternating)
        gti%start = [(time(neworder(ii)), ii=1, n-1, 2)]
        gti%stop = [(time(neworder(ii)), ii=2, n, 2)]

        gti%tstart = minval(gti%start)
        gti%tstop = maxval(gti%stop)
        gti%nelements = size(gti%start)

    end subroutine create_GTI_from_order

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> sort the type entries according to one array
    !
    !
    !
    !> @param[in] this
    !> @param[in] array
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine sort_tGTI(this, array, status, ascend, optorder)
        implicit none

        integer, intent(inout) :: status
        integer, allocatable, dimension(:) :: order
        integer, allocatable, dimension(:), optional, intent(out) :: optorder
        integer, optional :: ascend

        class(tGTI) :: this

        real(dp), intent(in), dimension(:) :: array
        real(dp), allocatable, dimension(:) :: tmparray

        character(*), parameter :: function = "sort_tGTI"

        if(status > 100 .or. status == s_errRep) return

        if(this%nelements /= size(array)) then
            call wrerr(function // " tGTI type", s_err, "length of sorting array does not conform to length of type")
            status = s_errRep
        end if

        tmparray = array
        call sort(tmparray, status, order)
        if(present(ascend)) then
            if(ascend == false) call reverse_array(order)
        end if

        this%start = this%start(order)
        this%stop = this%stop(order)

        if(present(optorder)) optorder = order

        call dealloc(order, status)
        call dealloc(tmparray, status)

    end subroutine sort_tGTI

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> free the content of tGTI type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tGTI(this, status)
        implicit none

        integer, intent(inout) :: status

        class(tGTI) :: this

        character(*), parameter :: function = "free_tGTI"

        call dealloc(this%start, status)
        call dealloc(this%stop, status)

    end subroutine free_tGTI

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> alloc the content of tGTI type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine init_tGTI(this, n, status)
        implicit none

        integer, intent(in) :: n
        integer, intent(inout) :: status

        class(tGTI) :: this

        character(*), parameter :: function = "init_tGTI"

        call alloc(this%start, n, status)
        call alloc(this%stop, n, status)

        this%nelements = n
        this%start_col_name = start_col
        this%stop_col_name = stop_col
        this%ccdnr = -1
        this%extname = ""
        this%mjdref = 0_r8
        this%timezero = 0_r8
        this%use_mjdref = true
        this%use_timezero = true

    end subroutine init_tGTI

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> print the keywords of the gti type
    !
    !
    !
    !> @param[in] this
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine print_tGTI(this, status)
        implicit none

        integer, intent(inout) :: status

        class(tGTI) :: this

        character(*), parameter :: function = "print_tGTI"

        if(status > 100 .or. status == s_errRep) return

        call wrlog(function, "extname: " // trim(this%extname))
        call wrlog(function, "ccdnr: " // trim(int2str(this%ccdnr)))
        call wrlog(function, "mjdref: " // trim(double2str(this%mjdref)))
        call wrlog(function, "timezero: " // trim(double2str(this%timezero)))
        call wrlog(function, "start col name: " // trim(this%start_col_name))
        call wrlog(function, "stop col name: " // trim(this%stop_col_name))
        call wrlog(function, "tstart: " // trim(double2str(this%tstart)))
        call wrlog(function, "tstop: " // trim(double2str(this%tstop)))
        call wrlog(function, "nelements: " // trim(int2str(this%nelements)))
        call wrlog(function, "size(start col): " // trim(int2str(size(this%start))))
        call wrlog(function, "size(stop col): " // trim(int2str(size(this%stop))))

    end subroutine print_tGTI

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> function to count the number of continues intervals in an array
    !> the array must not include 0
    !
    !
    !> @param[in] array
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine get_cont_intervals_int(array, newarray, status)
        implicit none

        integer, intent(inout) :: status
        integer, intent(inout), dimension(:) :: array
        integer, dimension(size(array)) :: tmparray
        integer, dimension(:), allocatable :: newarray
        integer :: n, ii

        if(status > 100 .or. status == s_errRep) return

        n = size(array)
        tmparray = array
        do ii = 2, n-1
            ! check for continues array
            ! element before and after are -/+1
            if(array(ii) == array(ii-1) + 1 .and. array(ii) == array(ii+1) -1) then
                tmparray(ii) = 0
            end if
        end do

        n = size(pack(tmparray, tmparray /= 0))
        call alloc(newarray, n, status)
        newarray = pack(tmparray, tmparray /= 0)

    end subroutine get_cont_intervals_int

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> create a mask to flag times outside the GTI
    !
    !
    !
    !> @param[in] time, GTI
    !> @param[inout] status
    !---------------------------------------------------------------------------

    function get_GTI_mask(time, GTI, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii, n

        real(r8), dimension(:), intent(in) :: time

        integer, dimension(size(time)) :: get_GTI_mask

        type(tGTI), intent(in) :: GTI

        character(*), parameter :: function = "get_GTI_mask"

        if(status > 100 .or. status == s_errRep) return

        get_GTI_mask(:) = false
        n = size(GTI%start)
        do ii = 1, n

            where((time > GTI%start(ii)) .and. (time < GTI%stop(ii)))
                get_GTI_mask = true
            end where

            ! account for comparing floating point arithmetic
            where(time > GTI%start(ii) .and. abs(time - GTI%stop(ii)) < epsilon(GTI%stop(ii)))
                get_GTI_mask = true
            end where

            ! account for comparing floating point arithmetic
            where(abs(time - GTI%start(ii)) < epsilon(GTI%start(ii)) .and. time < GTI%stop(ii))
                get_GTI_mask = true
            end where

        end do

    end function get_GTI_mask

    !---------------------------------------------------------------------------
    !> @author
    !> Ian M Stewart
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> create a mask to flag times outside the GTI
    !
    !
    !
    !> @param[in] time, GTI
    !> @param[inout] status
    !---------------------------------------------------------------------------

    function get_GTI_mask_logical(time, GTI, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii, n

        real(r8), dimension(:), intent(in) :: time

        logical, dimension(size(time)) :: get_GTI_mask_logical

        type(tGTI), intent(in) :: GTI

        character(*), parameter :: function = "get_GTI_mask_logical"

        if(status > 100 .or. status == s_errRep) return

        get_GTI_mask_logical(:) = .false.
        n = size(GTI%start)
        do ii = 1, n

            where((time > GTI%start(ii)) .and. (time < GTI%stop(ii)))
                get_GTI_mask_logical = .true.
            end where

            ! account for comparing floating point arithmetic
            where(time > GTI%start(ii) .and. abs(time - GTI%stop(ii)) < epsilon(GTI%stop(ii)))
                get_GTI_mask_logical = .true.
            end where

            ! account for comparing floating point arithmetic
            where(abs(time - GTI%start(ii)) < epsilon(GTI%start(ii)) .and. time < GTI%stop(ii))
                get_GTI_mask_logical = .true.
            end where

        end do

    end function get_GTI_mask_logical

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> create a mask to flag times outside the GTI, but extend the masked range by N (before and after the intervall)
    !> the time column must be in ascending order
    !
    !
    !> @param[in] time, GTI
    !> @param[inout] status
    !---------------------------------------------------------------------------

    function get_GTI_mask_extend(time, GTI, extend, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii, jj, n
        integer, intent(in) :: extend

        real(r8), dimension(:), intent(in) :: time

        integer, dimension(size(time)) :: get_GTI_mask_extend

        type(tGTI), intent(in) :: GTI

        character(*), parameter :: function = "get_GTI_mask_extend"

        if(status > 100 .or. status == s_errRep) return

        get_GTI_mask_extend(:) = false
        n = size(GTI%start)
        do ii = 1, n
            do jj = 1, size(time)
                if( time(jj) > GTI%start(ii)) then
                    if(time(jj) < GTI%stop(ii)) then
                        get_GTI_mask_extend(jj) = true
                        if(jj - extend < 1) then
                            get_GTI_mask_extend(1:jj) = true
                        else
                            get_GTI_mask_extend(jj - extend:jj) = true
                        end if

                        if(jj + extend > size(time)) then
                            get_GTI_mask_extend(jj : size(time)) = true
                        else
                            get_GTI_mask_extend(jj : jj + extend) = true
                        end if
                    else
                        ! mask entries as true if GTI
                        if(extend > 0) then
                            if(jj - extend < 1) then
                                get_GTI_mask_extend(1 : jj) = true
                            else
                                get_GTI_mask_extend(jj - extend : jj) = true
                            end if

                            if(jj + extend - 1 > size(time)) then
                                get_GTI_mask_extend(jj : size(time)) = true
                            else
                                get_GTI_mask_extend(jj : jj + extend - 1) = true
                            end if
                        end if
                        ! if time is above the upper GTI limit you can ignore the rest since the provided time is ascending order
                        ! exit the jj loop
                        exit
                    ! end if GTI stop
                    end if
                ! end if GTI start
                end if
                ! end of jj loop
            end do
            ! end of ii loop
        end do

    end function get_GTI_mask_extend



end module evtool_gti_mod
