! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module evtool_hk_mod

  use erolib_str_mod
  use erolib_ffits_mod
  use erolib_general_mod
  use erolib_sort_mod
  use evtool_gti_mod
  use erolib_tbintable_mod
  use erolib_math_mod
  implicit none

  private

  character(*), parameter :: hk_ext = "HK"
  character(*), parameter :: time_col = "TIME"
  character(*), parameter :: fwheel_col = "FWHEEL"
  character(*), parameter :: detstat_col = "DETSTAT"

  type tHK
    real(r8) :: mjdref = 0.0
    character(buffsize) :: extname = ""
    integer :: nelements = 0
    integer :: ccdnr = 0
    real (r8), allocatable, dimension(:) :: time
    integer (i2), allocatable, dimension(:) :: fwheel
    integer (r8), allocatable, dimension(:) :: detstat
  contains
    procedure, non_overridable :: read => read_tHK
    procedure, non_overridable :: write => write_tHK
    procedure, non_overridable :: create_extension => create_extension_tHK
    procedure, non_overridable :: sort => sort_tHK
    procedure, non_overridable :: free => free_tHK
    procedure, non_overridable :: init => init_tHK
    procedure, non_overridable :: delete_duplicates => delete_duplicates_tHK
    procedure, non_overridable :: apply_index_cols => apply_index_cols_tHK
    procedure, non_overridable :: apply_gti => apply_gti_tHK
  end type tHK

  public :: tHK, hk_ext, alloc_hk, dealloc_hk, merge_hk, sortTimes

contains

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> dealloacte hk structures
  !
  !
  !
  !> @param[inout] hk
  !> @param[inout] status
  !---------------------------------------------------------------------------
  subroutine dealloc_hk(hk, status)

    integer :: stat
    integer, intent(inout) :: status

    type(tHK), allocatable, dimension(:) :: hk

    character(*), parameter :: function = "dealloc_hk"

    if(allocated(hk)) then
      deallocate(hk, stat=stat)
      call FError(function, stat, "dealloc hk type", status)
    end if

  end subroutine dealloc_hk

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloacte n hk structures
  !
  !
  !
  !> @param[inout] hk
  !> @param[in] n
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine alloc_hk(hk, n, status)
    implicit none

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    type(tHK), allocatable, dimension(:) :: hk

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_hk"

    call dealloc_hk(hk, status)
    allocate(hk(n), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

  end subroutine alloc_hk

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create the merged hk extensions for all infiles
  !
  !
  !
  !> @param[in] EvtenFile_out
  !> @param[out] files
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine merge_hk(hk_in, hk_out, status)
    implicit none

    integer, intent(inout) :: status
    integer(i4) :: ii, nelements, a, b

    type(tHK), intent(out) :: hk_out
    type(tHK), intent(in), dimension(:) :: hk_in

    character(*), parameter :: function = "merge_hk"

    if(status > 100 .or. status == s_errRep) return

    nelements = 0
    do ii = 1, size(hk_in)
      nelements = nelements + hk_in(ii)%nelements
    end do

    ! reallocate memory for columns
    call hk_out%init(nelements, status)

    hk_out%nelements = nelements
    hk_out%ccdnr = hk_in(1)%ccdnr
    hk_out%extname = hk_in(1)%extname
    hk_out%mjdref = hk_in(1)%mjdref

    if(nelements == 0) return

    ! merge all columns
    a = 1
    do ii = 1, size(hk_in)
      b = a + hk_in(ii)%nelements - 1
      hk_out%time(a:b) = hk_in(ii)%time
      hk_out%fwheel(a:b) = hk_in(ii)%fwheel
      hk_out%detstat(a:b) = hk_in(ii)%detstat
      if(status == s_errRep) return
      a = b + 1
    end do

  end subroutine merge_hk


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort the type entries according to one array
  !
  !
  !
  !> @param[in] this
  !> @param[in] array
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_tHK(this, array, status, ascend)
    implicit none

    integer, intent(inout) :: status
    integer, allocatable, dimension(:) :: order
    integer, optional :: ascend

    class(tHK) :: this

    class(*), intent(in), dimension(:) :: array
    real(r8), allocatable, dimension(:) :: tmparray_r8

    character(*), parameter :: function = "sort_tHK"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%time) /= size(array)) then
      call wrerr(function // " tHK type", s_err, "length of sorting array does not conform to length of type")
      status = s_errRep
    end if

    select type (array)
      type is (real(r8))
      call alloc(tmparray_r8, size(array), status, array)
      call sort(tmparray_r8, status, order)
      if(present(ascend)) then
        if(ascend == false) call reverse_array(order)
      end if

    end select

    this%time = this%time(order)
    this%fwheel = this%fwheel(order)
    this%detstat = this%detstat(order)

    call dealloc(order, status)
    call dealloc(tmparray_r8, status)

  end subroutine sort_tHK

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the values from the hk extension
  !
  !
  !
  !> @param[in] this (tHK)
  !> @param[in] unit, ccdnr
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_tHK(this, unit, ccdnr, success, status, mjdref)
    implicit none

    class(tHK) :: this

    integer :: nelements, rowsize, frow, nrows, lrow, felem
    integer :: time_colnum, fwheel_colnum, detstat_colnum
    integer, intent(in) :: ccdnr
    integer, intent(inout) :: unit, status
    integer, intent(out) :: success
    real(r8), optional, intent(in) :: mjdref

    character(buffsize)  :: telid_str
    character(*), parameter :: function = "read_tHK"

    logical :: anyf

    if(status > 100 .or. status == s_errRep) return

    ! convert int to string
    telid_str = int2str(ccdnr, status)
    if(status == s_errRep) return

    success = false

    status = ignore
    call ffits_movnam_hdu(unit, BINARY_TBL, hk_ext // trim(telid_str), status)
    if(status == ignore) then
      status = ok
    else
      status = ok
      return
    end if

    call ffits_get_num_rows(unit, nelements, status)
    call ffits_get_rowsize(unit, rowsize, status)

    call this%init(nelements, status)

    this%ccdnr = ccdnr
    this%nelements = nelements
    this%extname = hk_ext // trim(telid_str)

    ! if no mjdref in the extension is found take the optional mjdref
    status = ignore
    call ffits_read_key(unit, "MJDREF", this%mjdref, status)
    if(status == ignore) then
      status = ok
    else
      status = ok
      if(present(mjdref)) this%mjdref = mjdref
    end if

    call ffits_get_colnum(unit, .false., time_col, time_colnum, status)
    call ffits_get_colnum(unit, .false., fwheel_col, fwheel_colnum, status)
    call ffits_get_colnum(unit, .false., detstat_col, detstat_colnum, status)

    anyf = .false.
    felem = 1

    if(nelements == 0 .or. status == s_errRep) then
      if(status == ok) success = true
      return
    end if

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1
      call ffits_read_col(unit, time_colnum, frow, felem, nrows, 0, this%time(frow:lrow), status, anyf)
      call ffits_read_col(unit, fwheel_colnum, frow, felem, nrows, 0, this%fwheel(frow:lrow), status, anyf)
      call ffits_read_col(unit, detstat_colnum, frow, felem, nrows, 0, this%detstat(frow:lrow), status, anyf)
    end do

    if(status == ok) success = true

  end subroutine read_tHK

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write in the columns of an hk extension
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_tHK(this, unit, status)
    implicit none

    integer :: nelements, rowsize, frow, nrows, lrow, felem
    integer :: time_colnum, fwheel_colnum, detstat_colnum
    integer, intent(inout) :: status
    integer, intent(in) :: unit

    class(tHK) :: this

    character(*), parameter :: function = "write_tHK"

    if(status > 100 .or. status == s_errRep) return

    if(this%ccdnr == -1 .or. this%extname == "") return

    ! move to the correct extension
    status = ignore
    call ffits_movnam_hdu(unit, BINARY_TBL, this%extname, status)
    if(status /= ignore) then
      status = ok
      call this%create_extension(unit, status)
    else
      status = ok
    end if

    call ffits_get_rowsize(unit, rowsize, status)

    nelements = this%nelements

    call ffits_get_colnum(unit, .false., time_col, time_colnum, status)
    call ffits_get_colnum(unit, .false., fwheel_col, fwheel_colnum, status)
    call ffits_get_colnum(unit, .false., detstat_col, detstat_colnum, status)

    felem = 1

    if(nelements > 0) then
      do frow = 1, nelements, rowsize
        nrows = min(nelements - frow + 1, rowsize)
        lrow = frow + nrows - 1
        call ffits_write_col(unit, time_colnum, frow, felem, nrows, this%time(frow:lrow), status)
        call ffits_write_col(unit, fwheel_colnum, frow, felem, nrows, this%fwheel(frow:lrow), status)
        call ffits_write_col(unit, detstat_colnum, frow, felem, nrows, this%detstat(frow:lrow), status)
      end do
    end if

    call ffits_update_key(unit, "MJDREF", this%mjdref, "[d] updated by evtool", status)

  end subroutine write_tHK

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create a basic hk extension in the file
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine create_extension_tHK(this, unit, status)
    implicit none

    integer, intent(inout) :: status
    integer, intent(in) :: unit

    character(*), parameter :: function = "create_extension_tHK"
    character(fcolnamelength), dimension(3) :: ttype
    character(fcolnamelength), dimension(3) :: tform
    character(fcolnamelength), dimension(3) :: tunit

    class(tHK) :: this
    type(tBinTable) :: table

    if(status > 100 .or. status == s_errRep) return

    ttype(1) = time_col
    ttype(2) = fwheel_col
    ttype(3) = detstat_col

    tform = ["D", "I", "I"]
    tunit(:) = ""

    call table%init(ttype, tform, tunit, status)
    call table%insertFits(unit, this%extname, status);

  end subroutine create_extension_tHK

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> reduce all entries in the columns to contain only a subset of the complete
  !> by providing an array list with the desired indices
  !
  !
  !> @param[in] this
  !> @param[in] order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_index_cols_tHK(this, order, status)
    implicit none

    integer, intent(inout) :: status

    class(tHK) :: this

    integer, dimension(:), intent(in) :: order
    integer :: n

    character(*), parameter :: function = "apply_index_cols_tHK"

    if(status > 100 .or. status == s_errRep) return

    n = size(order)
    if(n < 2) then
      !      call wrerr(function, warn, trim(this%extname) // " would be reduced to below two entries - but at least two entries in " // trim(this%extname) // &
      !    " are required for further processing - no filter will be applied")
      call wrchatty(function, trim(this%extname) // " would be reduced to below two entries - but at least two entries in " // trim(this%extname) // &
        " are required for further processing - no filter will be applied")
      !      call wrerr(function, warn, trim(this%extname) // " is reduced to zero entries")
      !      call alloc(this%time, n, status)
      !      call alloc(this%deadc, n, status)
      !      this%nelements = n
      return
    end if

    this%time = this%time(order)
    this%fwheel = this%fwheel(order)
    this%detstat = this%detstat(order)
    this%nelements = n

  end subroutine apply_index_cols_tHK

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !
  !
  !> @param[in] this, GTI
  !> @param[in, optional] extend_order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_gti_tHK(this, GTI, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tHK) :: this
    type(tGTI), intent(in) :: GTI
    type(tGTI) :: iGTI

    integer, dimension(this%nelements) :: mask
    integer, dimension(:), allocatable :: order

    character(*), parameter :: function = "apply_gti_tHK"

    if(status > 100 .or. status == s_errRep) return

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]

    iGTI = GTI
    ! internally correct the GTI
    if(iGTI%use_mjdref == true) then

      if(this%mjdref /= iGTI%mjdref) then

        iGTI%start = (iGTI%mjdref - this%mjdref ) * 3600.0_r8 * 24.0_r8 + iGTI%start
        iGTI%stop = ( iGTI%mjdref - this%mjdref ) * 3600.0_r8 * 24.0_r8 + iGTI%stop
        iGTI%mjdref = this%mjdref
      end if

    end if

    mask = get_gti_mask(this%time, iGTI, status)
!    mask = get_gti_mask_extend(this%time, iGTI, 1, status)

    order = pack(order, mask == true)

    call this%apply_index_cols(order, status)
    call dealloc(order, status)
    call iGTI%free(status)


  end subroutine apply_gti_tHK

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete duplicate entries in the tHK type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_duplicates_tHK(this, status)
    implicit none

    class(tHK) :: this

    integer, intent(inout) :: status
    integer, dimension(:), allocatable :: order
    integer, dimension(this%nelements) :: mask
    integer :: ii

    character(*), parameter :: function = "delete_duplicates_tHK"

    if(status > 100 .or. status == s_errRep) return

    call this%sort(this%time, status)

    if(this%nelements < 2) return

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]

    mask = true
    do ii = 2, this%nelements
      if(abs(this%time(ii) - this%time(ii-1)) < tiny(this%time(1))) then
        mask(ii) = false
      end if
    end do

    order = pack(order, mask == true)
    call this%apply_index_cols(order, status)
    call dealloc(order, status)

  end subroutine delete_duplicates_tHK

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the content of tHK type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_tHK(this, status)
    implicit none

    integer, intent(inout) :: status

    class(tHK) :: this

    character(*), parameter :: function = "free_tHK"

    call dealloc(this%time, status)
    call dealloc(this%fwheel, status)
    call dealloc(this%detstat, status)

  end subroutine free_tHK

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloc the content of tHK type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine init_tHK(this, n, status)
    implicit none

    integer, intent(in) :: n
    integer, intent(inout) :: status

    class(tHK) :: this

    character(*), parameter :: function = "init_tHK"

    call alloc(this%time, n, status)
    call alloc(this%fwheel, n, status)
    call alloc(this%detstat, n, status)

    this%nelements = n
    this%ccdnr = -1
    this%extname = ""
    this%mjdref = 0_r8

  end subroutine init_tHK

  subroutine sortTimes(nTimes, frameTimes, recordTimes)
    implicit none

    integer,                     intent(in)    :: nTimes
    real(r8), dimension(nTimes), intent(inout) :: frameTimes, recordTimes

    real(r8), dimension(nTimes) :: tmparray_r8
!    integer, dimension(nTimes) :: order
    integer, dimension(:), allocatable :: order
    integer :: status
    character(*), parameter :: function = "sortTimes"

    tmparray_r8 = frameTimes
    call sort_double(tmparray_r8, status, order)
    frameTimes  = frameTimes( order)
    recordTimes = recordTimes(order)

    tmparray_r8 = recordTimes
    call sort_double(tmparray_r8, status, order)
    frameTimes  = frameTimes( order)
    recordTimes = recordTimes(order)

    call dealloc(order, status)

  end subroutine sortTimes

end module evtool_hk_mod
