! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module evtool_hkx1_mod
  use param
  use erolib_tbintable_mod
  use erolib_ffits_col_mod
  use evtool_events_pp_mod
  use evtool_gti_mod
  use fitsimage, only : stdhdr_type, fread_stdhdr, fwrite_stdhdr

  implicit none

logical, private, parameter :: doTest=.false.

  ! extnames used in the files

  character(*), parameter, public :: hkx1_ext = "HKx1"
  character(*), parameter, private :: extname_default = "HKx1"

  ! basic type of the catalog used for the cumulative list
  type, public :: tHKx1

    integer :: unit = 0
    integer :: tel_id = -1
    character(1) :: tel_id_str = ''
    integer :: nelements = 0 ! number of elements (rows)
    character(buffsize) :: filename = ""
    character(buffsize) :: extname = extname_default
    type(tBinTable) :: table ! description of the binary table
    !    type(stdhdr_type) :: stdhdr

    type(r8_tColumn) :: time
    type(r8_tColumn) :: recordtime
    type(i1_tColumn) :: pxprocmode
    type(i4_tColumn) :: pxprocopts
    type(r4_tColumn) :: pxtmframes
    type(r4_tColumn) :: pxccdfrms
    type(r4_tColumn) :: pxevtctr
    type(r4_tColumn) :: pxmipctr
    type(r4_tColumn) :: pxfwstepct
    type(i1_tColumn) :: pxfwstat
!    type(i1_tColumn) :: pxfw
    type(r4_tColumn) :: pxchopper

    ! requested by KD
    type(r4_tColumn) :: pxsplt2noi
    type(r4_tColumn) :: pxevtqlimt

    ! derivative hk columns
    type(i2_tColumn) :: dxprocmode
    type(i2_tColumn) :: dxprocopts
    type(i2_tColumn) :: dxfwstat

    ! frametime from record information
    type(r8_tColumn) :: rframetime
    type(i2_tColumn) :: q_rframetime ! quality of the information used to derive the frametime

    ! columns required by telgti/telselect
    !    type(i2_tColumn) :: fwheel
    type(i2_tColumn) :: detstat


  contains
    procedure :: init => init_tHKx1
    procedure :: free => free_tHKx1
    procedure :: open => open_tHKx1
    procedure :: close => close_tHKx1
    procedure :: create_extension => create_extension_tHKx1
    procedure :: read => read_tHKx1
    procedure :: write => write_tHKx1
    procedure :: apply_index => apply_index_tHKx1
    procedure :: sort => sort_tHKx1
    procedure :: set_tel_id => set_tel_id_tHKx1
    procedure :: reset_tel_id => reset_tel_id_tHKx1
    procedure :: reset_colnum => reset_colnum_tHKx1

    procedure :: derive_columns => derive_columns_tHKx1
    procedure :: derive_frametime => derive_frametime_tHKx1
    procedure :: delete_duplicates => delete_duplicates_tHKx1

  end type tHKx1

contains

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloacte n hkx1 structures
  !
  !
  !
  !> @param[inout] hkx1
  !> @param[in] n
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine alloc_hkx1(hkx1, n, status)
    implicit none

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    type(tHKx1), allocatable, dimension(:) :: hkx1

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_hkx1"

    call dealloc_hkx1(hkx1, status)
    allocate(hkx1(n), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

  end subroutine alloc_hkx1

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> dealloacte hkx1 structures
  !
  !
  !
  !> @param[inout] hkx1
  !> @param[inout] status
  !---------------------------------------------------------------------------
  subroutine dealloc_hkx1(hkx1, status)

    integer :: stat
    integer, intent(inout) :: status

    type(tHKx1), allocatable, dimension(:) :: hkx1

    character(*), parameter :: function = "dealloc_hkx1"

    if(status > 100 .or. status == s_errRep) return

    if(allocated(hkx1)) then
      deallocate(hkx1, stat=stat)
      call FError(function, stat, "dealloc hkx1 type", status)
    end if

  end subroutine dealloc_hkx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> init the tHKx1 type
  !> nband = number of different energy bands
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional]
  !---------------------------------------------------------------------------

  subroutine init_tHKx1(this, n, status, extname)
    implicit none

    integer, intent(inout) :: status
    class(tHKx1) :: this
    integer, intent(in) :: n
    character(*), intent(in), optional :: extname

    character(*), parameter :: function = "init_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    this%nelements = n
    this%extname = extname_default
    if(present(extname)) this%extname = extname

    call this%time%init("TIME", n, status, tunit="s", is_optional=true)
    call this%recordtime%init("RecordTime", n, status, tunit="s")
    call this%pxprocmode%init("PxPROCMODE", n, status)
    call this%pxprocopts%init("PxPROCOPTS", n, status)
    call this%pxtmframes%init("PxTMFRAMES", n, status)
    call this%pxccdfrms%init("PxCCDFRMS", n, status)
    call this%pxevtctr%init("PxEVTCTR", n, status)
    call this%pxmipctr%init("PxMIPCTR", n, status)
    call this%pxfwstepct%init("PxFWSTEPCT", n, status)
    call this%pxfwstat%init("PxFWSTAT", n, status)
    call this%pxchopper%init("PxCHOPPER", n, status)
    call this%pxsplt2noi%init("PxSPLT2NOI", n, status)
    call this%pxevtqlimt%init("PxEVTQLIMT", n, status)
    call this%dxprocmode%init("DxPROCMODE", n, status, is_optional=true)
    call this%dxprocopts%init("DxPROCOPTS", n, status, is_optional=true)
    call this%dxfwstat%init("DxFWSTAT", n, status, is_optional=true)
    call this%rframetime%init("RFRAMETIME", n, status, is_optional=true, tunit="s")
    call this%q_rframetime%init("QRFRAMETIME", n, status, is_optional=true)

    !    call this%fwheel%init("FWHEEL", n, status, is_optional=true)
    call this%detstat%init("DETSTAT", n, status, is_optional=true)

  end subroutine init_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the tHKx1 type
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional] nband
  !---------------------------------------------------------------------------

  subroutine free_tHKx1(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tHKx1) :: this

    character(*), parameter :: function = "free_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    this%nelements = 0

    call this%time%free(status)
    call this%recordtime%free(status)
    call this%pxprocmode%free(status)
    call this%pxprocopts%free(status)
    call this%pxtmframes%free(status)
    call this%pxccdfrms%free(status)
    call this%pxevtctr%free(status)
    call this%pxmipctr%free(status)
    call this%pxfwstepct%free(status)
    call this%pxfwstat%free(status)
    call this%pxchopper%free(status)

    call this%pxsplt2noi%free(status)
    call this%pxevtqlimt%free(status)

    call this%rframetime%free(status)
    call this%q_rframetime%free(status)

    call this%dxprocmode%free(status)
    call this%dxprocopts%free(status)
    call this%dxfwstat%free(status)

    !    call this%fwheel%free(status)
    call this%detstat%free(status)

  end subroutine free_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> routine to open the HKx1
  !
  !
  !
  !> @param[inout] this
  !> @param[in] filename, rwmode
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine open_tHKx1(this, filename, rwmode, status, delete)
    implicit none

    integer, intent(inout) :: status
    class(tHKx1) :: this

    character(*), intent(in) :: filename
    integer, intent(in) :: rwmode

    integer :: exists

    integer, intent(in), optional :: delete

    character(*), parameter :: function = "open_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    this%filename = filename

    call ffits_get_iounit(this%unit, status)
    if(rwmode /= create) then
      call ffits_open_file(this%unit, filename, rwmode, status)
      if(status == ok) call wrlog(function, trim(this%filename) // " successfully opened")
    else
      call ffits_file_exists(filename, exists, status)
      if(present(delete) .and. exists == 1) then
        if(delete == true) call delete_file_fortran(filename, status)
      end if
      call ffits_create_file(this%unit, filename, fblocksize, status)
      if(status == ok) call wrlog(function, trim(this%filename) // " successfully created")
    end if

  end subroutine open_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> close the tHKx1 file
  !
  !
  !
  !> @param[inout] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine close_tHKx1(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tHKx1) :: this

    character(*), parameter :: function = "close_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    call ffits_close_file(this%unit, status)
    call ffits_free_iounit(this%unit, status)

  end subroutine close_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> subroutine of class tHKx1
  !> create the extension of this type in a FITS file
  !
  !
  !> @param[inout] this
  !> @param[in, optional] aux
  !> @param[out]
  !> @param[inout]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine create_extension_tHKx1(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ncol ! number of columns in the file

    class(tHKx1) :: this
    character(*), parameter :: function = "create_extension_tHKx1"

    if(status > 100 .or. status == s_errRep) return

!    ncol = 20
    ncol = 19

    this%time%num = 0
    this%recordtime%num = 0
    this%pxprocmode%num = 0
    this%pxprocopts%num = 0
    this%pxtmframes%num = 0
    this%pxccdfrms%num = 0
    this%pxevtctr%num = 0
    this%pxmipctr%num = 0
    this%pxfwstepct%num = 0
    this%pxfwstat%num = 0
    this%pxchopper%num = 0
    this%pxsplt2noi%num = 0
    this%pxevtqlimt%num = 0
    this%dxprocmode%num = 0
    this%dxprocopts%num = 0
    this%dxfwstat%num = 0
    this%rframetime%num = 0
    this%q_rframetime%num = 0
    !    this%fwheel%num = 0
    this%detstat%num = 0

    call this%table%init(ncol, status)

    call this%table%insert_ffits_col(this%time, status)
    call this%table%insert_ffits_col(this%pxprocmode, status)
    call this%table%insert_ffits_col(this%pxprocopts, status)
    call this%table%insert_ffits_col(this%pxtmframes, status)
    call this%table%insert_ffits_col(this%pxccdfrms, status)
    call this%table%insert_ffits_col(this%pxevtctr, status)
    call this%table%insert_ffits_col(this%pxmipctr, status)
    call this%table%insert_ffits_col(this%pxfwstepct, status)
    call this%table%insert_ffits_col(this%pxfwstat, status)
    call this%table%insert_ffits_col(this%pxchopper, status)
    call this%table%insert_ffits_col(this%recordtime, status)

    call this%table%insert_ffits_col(this%pxsplt2noi, status)
    call this%table%insert_ffits_col(this%pxevtqlimt, status)

    call this%table%insert_ffits_col(this%rframetime, status)
    call this%table%insert_ffits_col(this%q_rframetime, status)

    call this%table%insert_ffits_col(this%dxprocmode, status)
    call this%table%insert_ffits_col(this%dxprocopts, status)
    call this%table%insert_ffits_col(this%dxfwstat, status)

    call this%table%insert_ffits_col(this%detstat, status)
    !    call this%table%insert_ffits_col(this%fwheel, status)

    call this%table%insertFits(this%unit, this%extname, status)

  end subroutine create_extension_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the hkx1 type from a FITS file
  !
  !
  !
  !> @param[in] this
  !> @param[out] filename
  !> @param[inout] status
  !> @param[in, optional] extend
  !---------------------------------------------------------------------------

  subroutine read_tHKx1(this, status, filename, tel_id, set_tel_id)
    implicit none

    integer, intent(inout) :: status
    integer :: nelements, rowsize, frow, lrow, nrows, ii
    class(tHKx1) :: this
    character(*), optional, intent(in) :: filename
    integer, optional, intent(in) :: tel_id
    integer :: telescope_id
    character(1) :: tel_id_str
    character(buffsize) :: extname
    logical, optional :: set_tel_id

    character(fcolnamelength) :: col
    character(*), parameter :: function = "read_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    if(present(filename)) call this%open(filename, readonly, status)

    telescope_id = this%tel_id

    if(present(tel_id)) then
      this%extname = extname_default
      tel_id_str = int2str(tel_id, status)
      this%extname(3:3) = tel_id_str
      telescope_id = tel_id
    end if

    status = ignore
    call ffits_movnam_hdu(this%unit, BINARY_TBL, this%extname, status)
    if(status == ignore) then
      status = ok
    else
!      print *, "CHECK HKX1"
!      print *, trim(this%extname)
      ! extension does not exist
      status = warn
      return
    end if

    call ffits_get_num_rows(this%unit, nelements, status)
    call ffits_get_rowsize(this%unit, rowsize, status)

    extname = this%extname
    call this%init(nelements, status)
    this%tel_id = telescope_id
    this%extname = extname

    if(present(set_tel_id)) then
      if(set_tel_id .eqv. .true.) call this%set_tel_id(telescope_id, status)
    end if

    if(nelements == 0 .or. status == s_errRep) return

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1
      ! try reading time if not available read recordtime
      call this%time%read(this%unit, status, frow=frow, lrow=lrow)
      call this%recordtime%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxprocmode%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxprocopts%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxtmframes%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxccdfrms%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxevtctr%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxmipctr%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxfwstepct%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxfwstat%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxchopper%read(this%unit, status, frow=frow, lrow=lrow)

      call this%pxsplt2noi%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxevtqlimt%read(this%unit, status, frow=frow, lrow=lrow)

      call this%rframetime%read(this%unit, status, frow=frow, lrow=lrow)
      call this%q_rframetime%read(this%unit, status, frow=frow, lrow=lrow)

      call this%dxprocmode%read(this%unit, status, frow=frow, lrow=lrow)
      call this%dxprocopts%read(this%unit, status, frow=frow, lrow=lrow)
      call this%dxfwstat%read(this%unit, status, frow=frow, lrow=lrow)

      !      call this%fwheel%read(this%unit, status, frow=frow, lrow=lrow)
      call this%detstat%read(this%unit, status, frow=frow, lrow=lrow)

    end do

    if(present(filename)) call this%close(status)

  end subroutine read_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the hkx1 type to a FITS file
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] aux
  !> @param[out] filename
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_tHKx1(this, status, filename)
    implicit none

    integer, intent(inout) :: status
    class(tHKx1) :: this
    integer :: rowsize, ii, frow, nelements, lrow, nrows
    character(*), optional, intent(in) :: filename

    character(*), parameter :: function = "write_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    if(present(filename)) then
      call this%open(filename, create, status)
      if(status /= ok) return
    end if

    ! move to the correct extension
    status = ignore
    call ffits_movnam_hdu(this%unit, BINARY_TBL, this%extname, status)
    if(status /= ignore) then
      status = ok
      call this%create_extension(status)
    else
      status = ok
    end if

    call ffits_get_rowsize(this%unit, rowsize, status)
    nelements = this%nelements

    ! write 1D arrays
    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1

      call this%time%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxprocmode%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxprocopts%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxtmframes%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxccdfrms%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxevtctr%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxmipctr%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxfwstepct%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxfwstat%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxchopper%write(this%unit, status, frow=frow, lrow=lrow)
      call this%recordtime%write(this%unit, status, frow=frow, lrow=lrow)

      call this%pxsplt2noi%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxevtqlimt%write(this%unit, status, frow=frow, lrow=lrow)

      call this%rframetime%write(this%unit, status, frow=frow, lrow=lrow)
      call this%q_rframetime%write(this%unit, status, frow=frow, lrow=lrow)

      call this%dxprocmode%write(this%unit, status, frow=frow, lrow=lrow)
      call this%dxprocopts%write(this%unit, status, frow=frow, lrow=lrow)
      call this%dxfwstat%write(this%unit, status, frow=frow, lrow=lrow)

      !      call this%fwheel%write(this%unit, status, frow=frow, lrow=lrow)
      call this%detstat%write(this%unit, status, frow=frow, lrow=lrow)

    end do

    if(present(filename)) call this%close(status)

  end subroutine write_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> reduce all entries in the columns to contain only a subset of the complete
  !> by providing an array list with the desired indices
  !
  !
  !> @param[in] this
  !> @param[in] order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_index_tHKx1(this, order, status)
    implicit none

    integer, intent(inout) :: status

    class(tHKx1) :: this

    integer, dimension(:), intent(in) :: order

    character(*), parameter :: function = "apply_index_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    if(this%nelements == 0) return

    this%nelements = size(order)

    call this%time%apply_index(order, status)
    call this%recordtime%apply_index(order, status)
    call this%pxprocmode%apply_index(order, status)
    call this%pxprocopts%apply_index(order, status)
    call this%pxtmframes%apply_index(order, status)
    call this%pxccdfrms%apply_index(order, status)
    call this%pxevtctr%apply_index(order, status)
    call this%pxmipctr%apply_index(order, status)
    call this%pxfwstepct%apply_index(order, status)
    call this%pxfwstat%apply_index(order, status)
    call this%pxchopper%apply_index(order, status)

    call this%pxsplt2noi%apply_index(order, status)
    call this%pxevtqlimt%apply_index(order, status)

    call this%rframetime%apply_index(order, status)
    call this%q_rframetime%apply_index(order, status)

    call this%dxprocmode%apply_index(order, status)
    call this%dxprocopts%apply_index(order, status)
    call this%dxfwstat%apply_index(order, status)

    !    call this%fwheel%apply_index(order, status)
    call this%detstat%apply_index(order, status)

  end subroutine apply_index_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge all entries of the columns with a second tHKx1
  !
  !
  !> @param[in] HKx1_A, HKx1_B
  !> @param[in] HKx1_C
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine merge_HKx1(HKx1_A, HKx1_B, HKx1_C, status)
    implicit none

    integer, intent(inout) :: status

    class(tHKx1), intent(in) :: HKx1_A
    class(tHKx1), intent(in) :: HKx1_B
    class(tHKx1), intent(out) :: HKx1_C

    character(*), parameter :: function = "merge_HKx1"

    if(status > 100 .or. status == s_errRep) return

    ! copy all important informations
    HKx1_C%table        = HKx1_A%table
    !    HKx1_C%stdhdr        = HKx1_A%stdhdr

    if(HKx1_A%nelements == 0 .and. HKx1_B%nelements == 0) then
      call wrlog(function, "cats to merge have zero elements - nothing to be done here")
  return
    end if

    HKx1_C%nelements = HKx1_A%nelements + HKx1_B%nelements
    HKx1_C%extname = HKx1_A%extname

    call merge_tColumn(HKx1_A%time, HKx1_B%time, HKx1_C%time, status)
    call merge_tColumn(HKx1_A%recordtime, HKx1_B%recordtime, HKx1_C%recordtime, status)
    call merge_tColumn(HKx1_A%pxprocmode, HKx1_B%pxprocmode, HKx1_C%pxprocmode, status)
    call merge_tColumn(HKx1_A%pxprocopts, HKx1_B%pxprocopts, HKx1_C%pxprocopts, status)
    call merge_tColumn(HKx1_A%pxtmframes, HKx1_B%pxtmframes, HKx1_C%pxtmframes, status)
    call merge_tColumn(HKx1_A%pxccdfrms, HKx1_B%pxccdfrms, HKx1_C%pxccdfrms, status)
    call merge_tColumn(HKx1_A%pxevtctr, HKx1_B%pxevtctr, HKx1_C%pxevtctr, status)
    call merge_tColumn(HKx1_A%pxmipctr, HKx1_B%pxmipctr, HKx1_C%pxmipctr, status)
    call merge_tColumn(HKx1_A%pxfwstepct, HKx1_B%pxfwstepct, HKx1_C%pxfwstepct, status)
    call merge_tColumn(HKx1_A%pxfwstat, HKx1_B%pxfwstat, HKx1_C%pxfwstat, status)
    call merge_tColumn(HKx1_A%pxchopper, HKx1_B%pxchopper, HKx1_C%pxchopper, status)

    call merge_tColumn(HKx1_A%pxsplt2noi, HKx1_B%pxsplt2noi, HKx1_C%pxsplt2noi, status)
    call merge_tColumn(HKx1_A%pxevtqlimt, HKx1_B%pxevtqlimt, HKx1_C%pxevtqlimt, status)

    call merge_tColumn(HKx1_A%rframetime, HKx1_B%rframetime, HKx1_C%rframetime, status)
    call merge_tColumn(HKx1_A%q_rframetime, HKx1_B%q_rframetime, HKx1_C%q_rframetime, status)

    call merge_tColumn(HKx1_A%dxprocmode, HKx1_B%dxprocmode, HKx1_C%dxprocmode, status)
    call merge_tColumn(HKx1_A%dxprocopts, HKx1_B%dxprocopts, HKx1_C%dxprocopts, status)
    call merge_tColumn(HKx1_A%dxfwstat, HKx1_B%dxfwstat, HKx1_C%dxfwstat, status)

    !    call merge_tColumn(HKx1_A%fwheel, HKx1_B%fwheel, HKx1_C%fwheel, status)
    call merge_tColumn(HKx1_A%detstat, HKx1_B%detstat, HKx1_C%detstat, status)

  end subroutine merge_HKx1


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort the type entries according to one array
  !
  !
  !
  !> @param[in] this
  !> @param[in] array
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_tHKx1(this, array, status, ascend)
    implicit none

    integer, intent(inout) :: status
    integer, optional :: ascend

    class(tHKx1) :: this
    class(*), dimension(:) :: array

    integer, dimension(:), allocatable :: order

    real(r8),allocatable, dimension(:) :: tmparray_r8
    real(r4),allocatable, dimension(:) :: tmparray_r4
    integer(i4),allocatable, dimension(:) :: tmparray_i4
    integer(i2),allocatable, dimension(:) :: tmparray_i2
    integer(i1),allocatable, dimension(:) :: tmparray_i1

    character(*), parameter :: function = "sort_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    if(this%nelements /= size(array)) then
      call wrerr(function // " tHKx1(_ext) type", s_err, "length of sorting array does not conform to length of type")
      status = s_errRep
    end if

    select type(array)
      type is ( real(r4) )
      tmparray_r4 = array
      call sort_real(tmparray_r4 ,status, order)
      type is ( real(r8) )
      tmparray_r8 = array
      call sort_double(tmparray_r8 ,status, order)
      type is ( integer(i4) )
      tmparray_i4 = array
      call sort_i4(tmparray_i4 ,status, order)
      type is ( integer(i2) )
      tmparray_i2 = array
      call sort_i2(tmparray_i2 ,status, order)
      type is ( integer(i1) )
      tmparray_i1 = array
      call sort_i1(tmparray_i1 ,status, order)
      class default
      call wrstop(function, "sorting tHKx1 with the current array type is not supported for tHKx1")
    end select

    if(present(ascend)) then
      if(ascend == false) call reverse_array(order)
    end if

    call this%apply_index(order, status)

    call dealloc(order, status)
    call dealloc(tmparray_i1, status)
    call dealloc(tmparray_i2, status)
    call dealloc(tmparray_i4, status)
    call dealloc(tmparray_r4, status)
    call dealloc(tmparray_r8, status)

  end subroutine sort_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine set_tel_id_tHKx1.
  !> @brief
  !> set the telesope id for this extension / modify extanme
  !
  !
  !> @param this, tel_id
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine set_tel_id_tHKx1(this, tel_id, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    integer, intent(in) :: tel_id
    class(tHKx1) :: this

    character(*), parameter :: function = "set_tel_id_tHKx1"

    if(status > 100 .or. status == s_errRep) return
    this%tel_id = tel_id
    this%tel_id_str =  int2str(tel_id, status)
    this%extname = extname_default
    this%extname(3:3) = this%tel_id_str

    ! replace x with TelID except time
    !    this%time%name(2:2) = this%tel_id_str
    this%pxprocmode%name(2:2) = this%tel_id_str
    this%pxprocopts%name(2:2) = this%tel_id_str
    this%pxtmframes%name(2:2) = this%tel_id_str
    this%pxccdfrms%name(2:2) = this%tel_id_str
    this%pxevtctr%name(2:2) = this%tel_id_str
    this%pxmipctr%name(2:2) = this%tel_id_str
    this%pxfwstepct%name(2:2) = this%tel_id_str
    this%pxfwstat%name(2:2) = this%tel_id_str
    this%pxchopper%name(2:2) = this%tel_id_str

    this%pxsplt2noi%name(2:2) = this%tel_id_str
    this%pxevtqlimt%name(2:2) = this%tel_id_str

  !   this%rframetime%name(2:2) = this%tel_id_str

  !    this%dxprocmode%name(2:2) = this%tel_id_str
  !    this%dxprocopts%name(2:2) = this%tel_id_str
  !    this%dxfwstat%name(2:2) = this%tel_id_str

  end subroutine set_tel_id_tHKx1

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine reset_tel_id_tHKx1.
  !> @brief
  !> reset the telesope id for this extension / modify extanme
  !
  !
  !> @param this
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine reset_tel_id_tHKx1(this, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx1) :: this

    character(*), parameter :: function = "reset_tel_id_tHKx1"

    if(status > 100 .or. status == s_errRep) return


    ! replace x with TelID except time
    !    this%time%name(2:2) = this%tel_id_str
    this%pxprocmode%name(2:2) = "x"
    this%pxprocopts%name(2:2) = "x"
    this%pxtmframes%name(2:2) = "x"
    this%pxccdfrms%name(2:2) = "x"
    this%pxevtctr%name(2:2) = "x"
    this%pxmipctr%name(2:2) = "x"
    this%pxfwstepct%name(2:2) = "x"
    this%pxfwstat%name(2:2) = "x"
    this%pxchopper%name(2:2) = "x"

    this%pxsplt2noi%name(2:2) = "x"
    this%pxevtqlimt%name(2:2) = "x"


  end subroutine reset_tel_id_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine reset_colnum_tHKx1.
  !> @brief
  !> reset the colnum of all columns in tHKx1
  !
  !
  !> @param this
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine reset_colnum_tHKx1(this, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx1) :: this

    !internal param

    character(*), parameter :: function = "reset_colnum_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    this%time%num = 0
    this%recordtime%num = 0
    this%pxprocmode%num = 0
    this%pxprocopts%num = 0
    this%pxtmframes%num = 0
    this%pxccdfrms%num = 0
    this%pxevtctr%num = 0
    this%pxmipctr%num = 0
    this%pxfwstepct%num = 0
    this%pxfwstat%num = 0
    this%pxchopper%num = 0

    this%pxsplt2noi%num = 0
    this%pxevtqlimt%num = 0

    this%rframetime%num = 0
    this%q_rframetime%num = 0

    this%dxprocmode%num = 0
    this%dxprocopts%num = 0
    this%dxfwstat%num = 0

    !    this%fwheel%num = 0
    this%detstat%num = 0

  end subroutine reset_colnum_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine derive_columns_tHKx1.
  !> @brief
  !> derive additional columns representing the changes of a column
  !
  !
  !> @param this
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine derive_columns_tHKx1(this, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx1) :: this

    !internal param
    integer :: ii
    character(*), parameter :: function = "derive_columns_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    if(this%nelements < 1) return

    ! ----------- 1 Element -------------------
    this%dxprocopts%value(1) = 0
    this%dxprocmode%value(1) = 0
    this%dxfwstat%value(1) = 0

    if(this%time%exists == false) then
      this%time%value = this%recordtime%value
      this%time%exists = true
    end if

    this%dxprocopts%exists = true
    this%dxprocmode%exists = true
    this%dxfwstat%exists = true

    !    this%fwheel%exists = true
    this%detstat%exists = true

    ! calculate fhweel
    !    where(this%pxfwstat%value == 14) this%fwheel%value = 0
    !    where(this%pxfwstat%value == 12) this%fwheel%value = 1
    !    where(this%pxfwstat%value == 08) this%fwheel%value = 2
    !    where(this%pxfwstat%value == 13) this%fwheel%value = 3
    !    where(this%pxfwstat%value == 15) this%fwheel%value = 4

    ! calculate detstat
    where(this%pxprocmode%value ==  1) this%detstat%value = 1
    where(this%pxprocmode%value ==  4) this%detstat%value = 1
    where(this%pxprocmode%value ==  9) this%detstat%value = 1
    where(this%pxprocmode%value == 10) this%detstat%value = 1

    !call wrlog(function, "fake detstat and filter")
    ! this%fwheel%value = 2
    ! this%detstat%value = 1

    if(this%nelements < 2) return

    ! -------------- 2 or more Elements ------------------
    do ii = 2, this%nelements

      if(this%pxprocopts%value(ii) == this%pxprocopts%value(ii-1)) this%dxprocopts%value(ii) = 0
      if(this%pxprocmode%value(ii) == this%pxprocmode%value(ii-1)) this%dxprocmode%value(ii) = 0
      if(this%pxfwstat%value(ii) == this%pxfwstat%value(ii-1)) this%dxfwstat%value(ii) = 0

      if(this%pxprocopts%value(ii) > this%pxprocopts%value(ii-1)) this%dxprocopts%value(ii) = 1
      if(this%pxprocmode%value(ii) > this%pxprocmode%value(ii-1)) this%dxprocmode%value(ii) = 1
      if(this%pxfwstat%value(ii) > this%pxfwstat%value(ii-1)) this%dxfwstat%value(ii) = 1

      if(this%pxprocopts%value(ii) < this%pxprocopts%value(ii-1)) this%dxprocopts%value(ii) = -1
      if(this%pxprocmode%value(ii) < this%pxprocmode%value(ii-1)) this%dxprocmode%value(ii) = -1
      if(this%pxfwstat%value(ii) < this%pxfwstat%value(ii-1)) this%dxfwstat%value(ii) = -1

    end do


  end subroutine derive_columns_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine derive_frametime_tHKx1.
  !> @brief
  !> description
  !
  !
  !> @param parameter
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine derive_frametime_tHKx1(this, events_pp, rGTI, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx1) :: this
    type(tEvents_pp), intent(inout) :: events_pp
    type(tGTI), intent(in) :: rGTI

    !internal param
    integer :: ii, jj, kk, n_hkx, n_uni_rtime, n_evt, n_evt_complete, counter
    integer, allocatable, dimension(:) :: order, order_ftime
    real(r8) :: ftime_duration, rtime_duration, ratio, time_scaling
    real(r8), allocatable, dimension(:) :: rtime, uni_rtime, ftime_event, ftime
    integer(i2), allocatable, dimension(:) :: type      ! 1 = hkx
    integer(i4), allocatable, dimension(:) :: order_hkx ! index of hk record
    integer(i4), allocatable, dimension(:) :: within_gti, evt_within_gti

    character(*), parameter :: function = "derive_frametime_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    ! alternative to loop which is commented below
    ! sort events in frametime and afterwards in recordtime
!    call events_pp%sort(events_pp%frametime%value, status)
    call events_pp%sort(events_pp%time%value, status)
    call events_pp%sort(events_pp%recordtime%value, status)

    uni_rtime = events_pp%recordtime%value(unique(events_pp%recordtime%value, status))

    n_hkx = this%nelements
    n_uni_rtime = size(uni_rtime)

    ! create combined array of recordtimes
    call alloc(type, n_uni_rtime + n_hkx, status)
    call alloc(rtime, n_uni_rtime + n_hkx, status)
    call alloc(order_hkx, n_uni_rtime + n_hkx, status)
    call alloc(ftime, n_uni_rtime + n_hkx, status)

    type(n_uni_rtime + 1: n_uni_rtime + n_hkx) = 1
    order_hkx(n_uni_rtime + 1 : n_uni_rtime + n_hkx) = [(ii,ii=1,n_hkx,1)]
    rtime = [uni_rtime, this%recordtime%value]

    counter = 1
    do ii = 1, events_pp%nelements
      if( abs(events_pp%recordtime%value(ii) - uni_rtime(counter)) < tiny(uni_rtime(counter))) then
!        ftime(counter) = events_pp%frametime%value(ii)
        ftime(counter) = events_pp%time%value(ii)
        cycle
      else
        counter = counter + 1
!        ftime(counter) = events_pp%frametime%value(ii)
        ftime(counter) = events_pp%time%value(ii)
      end if
    end do
    ! write last frametime to last recordtime - this might not be correct
!    ftime(n_uni_rtime) = events_pp%frametime%value(events_pp%nelements)
    ftime(n_uni_rtime) = events_pp%time%value(events_pp%nelements)

    call sort(rtime, status, order)
    type = type(order)
    order_hkx = order_hkx(order)
    ftime = ftime(order)

    ! all entries from events_pp + hk extension is checked
    within_gti = get_GTI_mask(rtime, rGTI, status)

    ! only check for the recordtimes in the events extension
    ! to get the maximum number of unique recordtimes as an indicator on how to calculate the frametime
    evt_within_gti = get_GTI_mask(uni_rtime, rGTI, status)
    n_evt_complete = sum(evt_within_gti)
    call dealloc(evt_within_gti, status)

    ! number of processed evts with complete records (all frametimes are available)
    ! complete record are always within gti
    n_evt = 0
    do ii = 1, size(rtime)

      ! check if its a record and within the recordtime gti
      if(type(ii) == 0) then
        if(within_gti(ii) == true) n_evt = n_evt + 1
      end if

      ! check if record is hkx
      if(type(ii) == 1) then
        ! no complete record yet -> get the next two complete record
        if(n_evt == 0) then
          call get_next_complete_event_record(rtime, type, within_gti, ii, jj, status)
          call get_next_complete_event_record(rtime, type, within_gti, jj, kk, status)
          time_scaling = (ftime(jj) - ftime(kk)) / (rtime(jj) - rtime(kk))
          rtime_duration = rtime(jj) - rtime(ii)
          ftime_duration = rtime_duration * time_scaling
          ftime(ii) = ftime(jj) - ftime_duration
          this%rframetime%value(order_hkx(ii)) = ftime(ii)
          this%q_rframetime%value(order_hkx(ii)) = 3
        ! all complete records have been processed -> get the last two complete records
        else if(n_evt == n_evt_complete) then
          call get_last_complete_event_record(rtime, type, within_gti, ii, kk, status)
          call get_last_complete_event_record(rtime, type, within_gti, kk, jj, status)
          time_scaling = (ftime(jj) - ftime(kk)) / (rtime(jj) - rtime(kk))
          rtime_duration = rtime(jj) - rtime(ii)
          ftime_duration = rtime_duration * time_scaling
          ftime(ii) = ftime(jj) - ftime_duration
          this%rframetime%value(order_hkx(ii)) = ftime(ii)
          this%q_rframetime%value(order_hkx(ii)) = 4
        ! position between two complete event records
        else
          call get_next_complete_event_record(rtime, type, within_gti, ii, kk, status)
          call get_last_complete_event_record(rtime, type, within_gti, ii, jj, status)
          time_scaling = (ftime(jj) - ftime(kk)) / (rtime(jj) - rtime(kk))
          rtime_duration = rtime(jj) - rtime(ii)
          ftime_duration = rtime_duration * time_scaling
          ftime(ii) = ftime(jj) - ftime_duration
          this%rframetime%value(order_hkx(ii)) = ftime(ii)
          ! if a recordtime of HK is within the rGTI
          ! since a frametime / recordtime exists for the start / stop of a rGTI
          ! not much interpolation is needed and thus the quality should be higher
          if(within_gti(ii) == true) then
            this%q_rframetime%value(order_hkx(ii)) = 1
          else
            this%q_rframetime%value(order_hkx(ii)) = 2
          end if
        end if

      end if

    end do

    this%rframetime%exists = true
    this%q_rframetime%exists = true

    call dealloc(rtime, status)
    call dealloc(uni_rtime, status)
    call dealloc(ftime_event, status)
    call dealloc(ftime, status)
    call dealloc(order, status)
    call dealloc(order_ftime, status)
    call dealloc(type, status)
    call dealloc(order_hkx, status)
    call dealloc(within_gti, status)

  end subroutine derive_frametime_tHKx1

      !---------------------------------------------------------------------------
      !> @author
      !> Christoph Grossberger
      !
      ! DESCRIPTION:
      !> Brief description of routine get_next_complete_event_record.
      !> @brief
      !> return an index with the next complete event record - recordtime with all frametimes
      !
      !
      !> @param rtime, type, within_gti, ii, jj, status
      !> @param[in]
      !> @param[out]
      !> @param[inout] status
      !---------------------------------------------------------------------------

  subroutine get_next_complete_event_record(rtime, type, within_gti, ii, jj, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    real(r8), dimension(:), intent(in) :: rtime
    integer(i2), dimension(:), intent(in) :: type
    integer(i4), dimension(:), intent(in) :: within_gti
    integer, intent(in) :: ii
    integer, intent(out) :: jj

    !internal param
    integer :: kk

    character(*), parameter :: function = "get_next_complete_event_record"

    if(status > 100 .or. status == s_errRep) return
    do kk = ii+1, size(type)
      ! check if type is events
!      if(type(kk) == 0 .and. rtime(kk) > rtime(ii)) then
      if(type(kk) == 0) then
        ! check if recordtime is within gti
        if(within_gti(kk) == true) then
          jj = kk
          return
        end if
      end if
    end do

    call wrerr(function, warn, "contact developer - algorithm should never reach this point")

  end subroutine get_next_complete_event_record

      !---------------------------------------------------------------------------
      !> @author
      !> Christoph Grossberger
      !
      ! DESCRIPTION:
      !> Brief description of routine get_last_complete_event_record.
      !> @brief
      !> return an index with the last complete event record - recordtime with all frametimes
      !
      !
      !> @param rtime, type, within_gti, ii, jj, status
      !> @param[in]
      !> @param[out]
      !> @param[inout] status
      !---------------------------------------------------------------------------

  subroutine get_last_complete_event_record(rtime, type, within_gti, ii, jj, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    real(r8), dimension(:), intent(in) :: rtime
    integer(i2), dimension(:), intent(in) :: type
    integer(i4), dimension(:), intent(in) :: within_gti
    integer, intent(in) :: ii
    integer, intent(out) :: jj

    !internal param
    integer :: kk

    character(*), parameter :: function = "get_last_complete_event_record"

    if(status > 100 .or. status == s_errRep) return
    do kk = ii-1, 1, -1
      ! check if type is events
!      if(type(kk) == 0 .and. rtime(kk) < rtime(ii)) then
      if(type(kk) == 0) then
        ! check if recordtime is within gti
        if(within_gti(kk) == true) then
          jj = kk
          return
        end if
      end if
    end do

    call wrerr(function, warn, "contact developer - algorithm should never reach this point")

  end subroutine get_last_complete_event_record

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine delete_duplicates_tHKx1.
  !> @brief
  !> delete duplicate entries based on time
  !
  !
  !> @param this
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_duplicates_tHKx1(this, status, report)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx1) :: this
    integer, optional, intent(in) :: report

    !internal param
    integer :: ii,num_dup
    integer, dimension(:), allocatable :: order
    logical, dimension(this%nelements) :: mask
    character(buffsize) :: nelements_dup_str

    character(*), parameter :: function = "delete_duplicates_tHKx1"

    if(status > 100 .or. status == s_errRep) return
    if(this%nelements == 1) return

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]
    call this%sort(this%recordtime%value, status)
    mask = .true.
    do ii = 2, this%nelements
      if(abs(this%recordtime%value(ii) - this%recordtime%value(ii-1)) < epsilon(this%recordtime%value(1))) then
        mask(ii) = .false.
      end if
    end do
    order = pack(order, mask)

    if(present(report)) then
      num_dup = this%nelements - size(order)
      if(num_dup > 0) then
        nelements_dup_str = int2str(num_dup, status)
        call wrerr(function, report, "number of duplicate records in HKx1 removed: " // trim(nelements_dup_str))
      end if
    end if

    call this%apply_index(order, status)

    call dealloc(order, status)

  end subroutine delete_duplicates_tHKx1

end module evtool_hkx1_mod
