! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module evtool_hkx3_mod
  use param
  use evtool_hkx1_mod
  use erolib_tbintable_mod
  use erolib_ffits_col_mod
  use fitsimage, only : stdhdr_type, fread_stdhdr, fwrite_stdhdr

  implicit none

  ! extnames used in the files

  character(*), parameter, public :: hkx3_ext = "HKx3"
  character(*), parameter, private :: extname_default = "HKx3"

  ! basic type of the catalog used for the cumulative list
  type, public :: tHKx3

    integer :: unit = 0
    integer :: tel_id = -1
    character(1) :: tel_id_str = ''
    integer :: nelements = 0 ! number of elements (rows)
    character(buffsize) :: filename = ""
    character(buffsize) :: extname = extname_default
    type(tBinTable) :: table ! description of the binary table
    !    type(stdhdr_type) :: stdhdr

    type(r8_tColumn) :: time
    type(r8_tColumn) :: recordtime
    type(r4_tColumn) :: pxfwdelay
    type(r4_tColumn) :: pxenvtab

    ! derivative columns
    type(i2_tColumn) :: dxenvtab

    ! frametime information
    type(r8_tColumn) :: rframetime
    type(i2_tColumn) :: q_rframetime

  contains
    procedure :: init => init_tHKx3
    procedure :: free => free_tHKx3
    procedure :: open => open_tHKx3
    procedure :: close => close_tHKx3
    procedure :: create_extension => create_extension_tHKx3
    procedure :: read => read_tHKx3
    procedure :: write => write_tHKx3
    procedure :: apply_index => apply_index_tHKx3
    procedure :: sort => sort_tHKx3
    procedure :: set_tel_id => set_tel_id_tHKx3
    procedure :: reset_tel_id => reset_tel_id_tHKx3
    procedure :: reset_colnum => reset_colnum_tHKx3

    procedure :: derive_columns => derive_columns_tHKx3
    procedure :: derive_frametime => derive_frametime_tHKx3
    procedure :: delete_duplicates => delete_duplicates_tHKx3


  end type tHKx3

contains

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloacte n hkx3 structures
  !
  !
  !
  !> @param[inout] hkx3
  !> @param[in] n
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine alloc_hkx3(hkx3, n, status)
    implicit none

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    type(tHKx3), allocatable, dimension(:) :: hkx3

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_hkx3"

    call dealloc_hkx3(hkx3, status)
    allocate(hkx3(n), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

  end subroutine alloc_hkx3

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> dealloacte hkx3 structures
  !
  !
  !
  !> @param[inout] hkx3
  !> @param[inout] status
  !---------------------------------------------------------------------------
  subroutine dealloc_hkx3(hkx3, status)

    integer :: stat
    integer, intent(inout) :: status

    type(tHKx3), allocatable, dimension(:) :: hkx3

    character(*), parameter :: function = "dealloc_hkx3"

    if(status > 100 .or. status == s_errRep) return

    if(allocated(hkx3)) then
      deallocate(hkx3, stat=stat)
      call FError(function, stat, "dealloc hkx3 type", status)
    end if

  end subroutine dealloc_hkx3

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> init the tHKx3 type
  !> nband = number of different energy bands
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional]
  !---------------------------------------------------------------------------

  subroutine init_tHKx3(this, n, status, extname)
    implicit none

    integer, intent(inout) :: status
    class(tHKx3) :: this
    integer, intent(in) :: n
    character(*), intent(in), optional :: extname

    character(*), parameter :: function = "init_tHKx3"

    if(status > 100 .or. status == s_errRep) return

    this%nelements = n
    this%extname = extname_default
    if(present(extname)) this%extname = extname

    call this%time%init("TIME", n, status, tunit="s", is_optional=true)
    call this%recordtime%init("RecordTime", n, status, tunit="s")
    call this%pxfwdelay%init("PxFWDELAY", n, status)
    call this%pxenvtab%init("PxENVTAB", n, status)

    call this%dxenvtab%init("DxENVTAB", n, status, is_optional=true)

    call this%rframetime%init("RFRAMETIME", n, status, tunit="s", is_optional=true)
    call this%q_rframetime%init("QRFRAMETIME", n, status, is_optional=true)

  end subroutine init_tHKx3

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the tHKx3 type
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional] nband
  !---------------------------------------------------------------------------

  subroutine free_tHKx3(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tHKx3) :: this

    character(*), parameter :: function = "free_tHKx3"

    if(status > 100 .or. status == s_errRep) return

    this%nelements = 0

    call this%time%free(status)
    call this%recordtime%free(status)
    call this%pxfwdelay%free(status)
    call this%pxenvtab%free(status)
    call this%dxenvtab%free(status)

    call this%rframetime%free(status)
    call this%q_rframetime%free(status)

  end subroutine free_tHKx3

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> routine to open the HKx3
  !
  !
  !
  !> @param[inout] this
  !> @param[in] filename, rwmode
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine open_tHKx3(this, filename, rwmode, status, delete)
    implicit none

    integer, intent(inout) :: status
    class(tHKx3) :: this

    character(*), intent(in) :: filename
    integer, intent(in) :: rwmode

    integer :: exists

    integer, intent(in), optional :: delete

    character(*), parameter :: function = "open_tHKx3"

    if(status > 100 .or. status == s_errRep) return

    this%filename = filename

    call ffits_get_iounit(this%unit, status)
    if(rwmode /= create) then
      call ffits_open_file(this%unit, filename, rwmode, status)
      if(status == ok) call wrlog(function, trim(this%filename) // " successfully opened")
    else
      call ffits_file_exists(filename, exists, status)
      if(present(delete) .and. exists == 1) then
        if(delete == true) call delete_file_fortran(filename, status)
      end if
      call ffits_create_file(this%unit, filename, fblocksize, status)
      if(status == ok) call wrlog(function, trim(this%filename) // " successfully created")
    end if

  end subroutine open_tHKx3

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> close the tHKx3 file
  !
  !
  !
  !> @param[inout] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine close_tHKx3(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tHKx3) :: this

    character(*), parameter :: function = "close_tHKx3"

    if(status > 100 .or. status == s_errRep) return

    call ffits_close_file(this%unit, status)
    call ffits_free_iounit(this%unit, status)

  end subroutine close_tHKx3

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> subroutine of class tHKx3
  !> create the extension of this type in a FITS file
  !
  !
  !> @param[inout] this
  !> @param[in, optional] aux
  !> @param[out]
  !> @param[inout]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine create_extension_tHKx3(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ncol ! number of columns in the file

    class(tHKx3) :: this
    character(*), parameter :: function = "create_extension_tHKx3"

    if(status > 100 .or. status == s_errRep) return

    ncol = 3

    call this%table%init(ncol, status)

    call this%table%insert_ffits_col(this%time, status)
    call this%table%insert_ffits_col(this%pxfwdelay, status)
    call this%table%insert_ffits_col(this%pxenvtab, status)
    call this%table%insert_ffits_col(this%dxenvtab, status)

    call this%table%insert_ffits_col(this%recordtime, status)
    call this%table%insert_ffits_col(this%rframetime, status)
    call this%table%insert_ffits_col(this%q_rframetime, status)

    call this%table%insertFits(this%unit, this%extname, status)

  end subroutine create_extension_tHKx3

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the hkx3 type from a FITS file
  !
  !
  !
  !> @param[in] this
  !> @param[out] filename
  !> @param[inout] status
  !> @param[in, optional] extend
  !---------------------------------------------------------------------------

  subroutine read_tHKx3(this, status, filename, tel_id, set_tel_id)
    implicit none

    integer, intent(inout) :: status
    integer :: nelements, rowsize, frow, lrow, nrows, ii
    class(tHKx3) :: this
    character(*), optional, intent(in) :: filename
    integer, optional, intent(in) :: tel_id
    integer :: telescope_id
    character(1) :: tel_id_str
    logical, optional :: set_tel_id
    character(buffsize) :: extname

    character(fcolnamelength) :: col
    character(*), parameter :: function = "read_tHKx3"

    if(status > 100 .or. status == s_errRep) return

    if(present(filename)) call this%open(filename, readonly, status)

    telescope_id = this%tel_id

    if(present(tel_id)) then
      this%extname = extname_default
      tel_id_str = int2str(tel_id, status)
      this%extname(3:3) = tel_id_str
      telescope_id = tel_id
    end if

    status = ignore
    call ffits_movnam_hdu(this%unit, BINARY_TBL, this%extname, status)
    if(status == ignore) then
      status = ok
    else
      ! extension does not exist
      status = warn
      return
    end if

    call ffits_get_num_rows(this%unit, nelements, status)
    call ffits_get_rowsize(this%unit, rowsize, status)

    extname = this%extname
    call this%init(nelements, status)
    this%tel_id = telescope_id
    this%extname = extname

    if(present(set_tel_id)) then
      if(set_tel_id .eqv. .true.) call this%set_tel_id(telescope_id, status)
    end if

    !    call fread_stdhdr(this%unit, this%stdhdr, status)

    if(nelements == 0 .or. status == s_errRep) return

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1

      ! try reading time if not available read recordtime
      call this%time%read(this%unit, status, frow=frow, lrow=lrow)
      call this%recordtime%read(this%unit, status, frow=frow, lrow=lrow)
      call this%rframetime%read(this%unit, status, frow=frow, lrow=lrow)
      call this%q_rframetime%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxfwdelay%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxenvtab%read(this%unit, status, frow=frow, lrow=lrow)
      call this%dxenvtab%read(this%unit, status, frow=frow, lrow=lrow)

    end do


    if(present(filename)) call this%close(status)

  end subroutine read_tHKx3

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the hkx3 type to a FITS file
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] aux
  !> @param[out] filename
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_tHKx3(this, status, filename)
    implicit none

    integer, intent(inout) :: status
    class(tHKx3) :: this
    integer :: rowsize, ii, frow, nelements, lrow, nrows
    character(*), optional, intent(in) :: filename

    character(*), parameter :: function = "write_tHKx3"

    if(status > 100 .or. status == s_errRep) return

    if(present(filename)) then
      call this%open(filename, create, status)
      if(status /= ok) return
    end if

    ! move to the correct extension
    status = ignore
    call ffits_movnam_hdu(this%unit, BINARY_TBL, this%extname, status)
    if(status /= ignore) then
      status = ok
      call this%create_extension(status)
    else
      status = ok
    end if

    call ffits_get_rowsize(this%unit, rowsize, status)
    nelements = this%nelements

    !    call fwrite_stdhdr(this%unit, this%stdhdr, status)

    ! write 1D arrays
    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1
      call this%time%write(this%unit, status, frow=frow, lrow=lrow)
      call this%recordtime%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxfwdelay%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxenvtab%write(this%unit, status, frow=frow, lrow=lrow)
      call this%dxenvtab%write(this%unit, status, frow=frow, lrow=lrow)
      call this%rframetime%write(this%unit, status, frow=frow, lrow=lrow)
      call this%q_rframetime%write(this%unit, status, frow=frow, lrow=lrow)
    end do

    if(present(filename)) call this%close(status)

  end subroutine write_tHKx3

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> reduce all entries in the columns to contain only a subset of the complete
  !> by providing an array list with the desired indices
  !
  !
  !> @param[in] this
  !> @param[in] order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_index_tHKx3(this, order, status)
    implicit none

    integer, intent(inout) :: status

    class(tHKx3) :: this

    integer, dimension(:), intent(in) :: order

    character(*), parameter :: function = "apply_index_tHKx3"

    if(status > 100 .or. status == s_errRep) return

    if(this%nelements == 0) return

    this%nelements = size(order)

    call this%time%apply_index(order, status)
    call this%recordtime%apply_index(order, status)
    call this%pxfwdelay%apply_index(order, status)
    call this%pxenvtab%apply_index(order, status)
    call this%dxenvtab%apply_index(order, status)
    call this%rframetime%apply_index(order, status)
    call this%q_rframetime%apply_index(order, status)

  end subroutine apply_index_tHKx3

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge all entries of the columns with a second tHKx3
  !
  !
  !> @param[in] HKx3_A, HKx3_B
  !> @param[in] HKx3_C
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine merge_HKx3(HKx3_A, HKx3_B, HKx3_C, status)
    implicit none

    integer, intent(inout) :: status

    class(tHKx3), intent(in) :: HKx3_A
    class(tHKx3), intent(in) :: HKx3_B
    class(tHKx3), intent(out) :: HKx3_C

    character(*), parameter :: function = "merge_HKx3"

    if(status > 100 .or. status == s_errRep) return

    ! copy all important informations
    HKx3_C%table        = HKx3_A%table
    !    HKx3_C%stdhdr        = HKx3_A%stdhdr

    if(HKx3_A%nelements == 0 .and. HKx3_B%nelements == 0) then
      call wrlog(function, "cats to merge have zero elements - nothing to be done here")
      return
    end if

    HKx3_C%nelements = HKx3_A%nelements + HKx3_B%nelements
    HKx3_C%extname = HKx3_A%extname

    call merge_tColumn(HKx3_A%time, HKx3_B%time, HKx3_C%time, status)
    call merge_tColumn(HKx3_A%recordtime, HKx3_B%recordtime, HKx3_C%recordtime, status)
    call merge_tColumn(HKx3_A%pxfwdelay, HKx3_B%pxfwdelay, HKx3_C%pxfwdelay, status)
    call merge_tColumn(HKx3_A%pxenvtab, HKx3_B%pxenvtab, HKx3_C%pxenvtab, status)
    call merge_tColumn(HKx3_A%dxenvtab, HKx3_B%dxenvtab, HKx3_C%dxenvtab, status)
    call merge_tColumn(HKx3_A%rframetime, HKx3_B%rframetime, HKx3_C%rframetime, status)
    call merge_tColumn(HKx3_A%q_rframetime, HKx3_B%q_rframetime, HKx3_C%q_rframetime, status)

  end subroutine merge_HKx3


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort the type entries according to one array
  !
  !
  !
  !> @param[in] this
  !> @param[in] array
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_tHKx3(this, array, status, ascend)
    implicit none

    integer, intent(inout) :: status
    integer, optional :: ascend

    class(tHKx3) :: this
    class(*), dimension(:) :: array

    integer, dimension(:), allocatable :: order

    real(r8),allocatable, dimension(:) :: tmparray_r8
    real(r4),allocatable, dimension(:) :: tmparray_r4
    integer(i4),allocatable, dimension(:) :: tmparray_i4
    integer(i2),allocatable, dimension(:) :: tmparray_i2
    integer(i1),allocatable, dimension(:) :: tmparray_i1

    character(*), parameter :: function = "sort_tHKx3"

    if(status > 100 .or. status == s_errRep) return

    if(this%nelements /= size(array)) then
      call wrerr(function // " tHKx3(_ext) type", s_err, "length of sorting array does not conform to length of type")
      status = s_errRep
    end if

    select type(array)
      type is ( real(r4) )
      tmparray_r4 = array
      call sort_real(tmparray_r4 ,status, order)
      type is ( real(r8) )
      tmparray_r8 = array
      call sort_double(tmparray_r8 ,status, order)
      type is ( integer(i4) )
      tmparray_i4 = array
      call sort_i4(tmparray_i4 ,status, order)
      type is ( integer(i2) )
      tmparray_i2 = array
      call sort_i2(tmparray_i2 ,status, order)
      type is ( integer(i1) )
      tmparray_i1 = array
      call sort_i1(tmparray_i1 ,status, order)
      class default
      call wrstop(function, "sorting tHKx3 with the current array type is not supported for tHKx3")
    end select

    if(present(ascend)) then
      if(ascend == false) call reverse_array(order)
    end if

    call this%apply_index(order, status)

    call dealloc(order, status)
    call dealloc(tmparray_i1, status)
    call dealloc(tmparray_i2, status)
    call dealloc(tmparray_i4, status)
    call dealloc(tmparray_r4, status)
    call dealloc(tmparray_r8, status)

  end subroutine sort_tHKx3

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine set_tel_id_tHKx3.
  !> @brief
  !> set the telesope id for this extension / modify extanme
  !
  !
  !> @param this, tel_id
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine set_tel_id_tHKx3(this, tel_id, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    integer, intent(in) :: tel_id
    class(tHKx3) :: this

    character(*), parameter :: function = "set_tel_id_tHKx3"

    if(status > 100 .or. status == s_errRep) return
    this%tel_id = tel_id
    this%tel_id_str =  int2str(tel_id, status)
    this%extname = extname_default
    this%extname(3:3) = this%tel_id_str

    ! replace x with TelID except time
    this%pxfwdelay%name(2:2) = this%tel_id_str
    this%pxenvtab%name(2:2) = this%tel_id_str

  end subroutine set_tel_id_tHKx3

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine reset_tel_id_tHKx3.
  !> @brief
  !> set the telesope id for this extension / modify extanme
  !
  !
  !> @param this
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine reset_tel_id_tHKx3(this, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx3) :: this

    character(*), parameter :: function = "reset_tel_id_tHKx3"

    if(status > 100 .or. status == s_errRep) return

    ! replace x with TelID except time
    this%pxfwdelay%name(2:2) = "x"
    this%pxenvtab%name(2:2) = "x"

  end subroutine reset_tel_id_tHKx3

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine reset_colnum_tHKx3.
  !> @brief
  !> reset the colnum of all columns in tHKx3
  !
  !
  !> @param this
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine reset_colnum_tHKx3(this, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx3) :: this

    !internal param

    character(*), parameter :: function = "reset_colnum_tHKx3"

    if(status > 100 .or. status == s_errRep) return

    this%time%num = 0
    this%recordtime%num = 0
    this%pxfwdelay%num = 0
    this%pxenvtab%num = 0
    this%dxenvtab%num = 0

    this%rframetime%num = 0
    this%q_rframetime%num = 0

  end subroutine reset_colnum_tHKx3

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine derive_columns_tHKx3.
  !> @brief
  !> derive additional columns representing the changes of a column
  !
  !
  !> @param this
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine derive_columns_tHKx3(this, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx3) :: this

    !internal param
    integer :: ii
    character(*), parameter :: function = "derive_columns_tHKx3"

    if(status > 100 .or. status == s_errRep) return

    if(this%nelements < 1) return

    this%dxenvtab%value(1) = 0
    if(this%time%exists == false) then
      this%time%value = this%recordtime%value
      this%time%exists = true
    end if

    if(this%nelements < 2) return

    do ii = 2, this%nelements

      if(this%pxenvtab%value(ii) == this%pxenvtab%value(ii-1)) this%dxenvtab%value(ii) = 0

      if(this%pxenvtab%value(ii) > this%pxenvtab%value(ii-1)) this%dxenvtab%value(ii) = 1

      if(this%pxenvtab%value(ii) < this%pxenvtab%value(ii-1)) this%dxenvtab%value(ii) = -1

    end do

    this%dxenvtab%exists = true

  end subroutine derive_columns_tHKx3

 !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine derive_frametime_tHKx3.
  !> @brief
  !> description
  !
  !
  !> @param parameter
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine derive_frametime_tHKx3(this, events, rGTI, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx3) :: this
    type(tEvents_pp), intent(inout) :: events
    type(tGTI), intent(in) :: rGTI

    !internal param
    integer :: ii, jj, kk, n_hkx, n_uni_rtime, n_evt, n_evt_complete, counter
    integer, allocatable, dimension(:) :: order, order_ftime
    real(r8) :: ftime_duration, rtime_duration, ratio, time_scaling
    real(r8), allocatable, dimension(:) :: rtime, uni_rtime, ftime_event, ftime
    integer(i2), allocatable, dimension(:) :: type      ! 1 = hkx
    integer(i4), allocatable, dimension(:) :: order_hkx ! index of hk record
    integer(i4), allocatable, dimension(:) :: within_gti, evt_within_gti

    character(*), parameter :: function = "derive_frametime_tHKx3"

    if(status > 100 .or. status == s_errRep) return

    ! alternative to loop which is commented below
    ! sort events in frametime and afterwards in recordtime
!    call events%sort(events%frametime%value, status)
    call events%sort(events%time%value, status)
    call events%sort(events%recordtime%value, status)

    uni_rtime = events%recordtime%value(unique(events%recordtime%value, status))

    n_hkx = this%nelements
    n_uni_rtime = size(uni_rtime)

    ! create combined array of recordtimes
    call alloc(type, n_uni_rtime + n_hkx, status)
    call alloc(rtime, n_uni_rtime + n_hkx, status)
    call alloc(order_hkx, n_uni_rtime + n_hkx, status)
    call alloc(ftime, n_uni_rtime + n_hkx, status)

    type(n_uni_rtime + 1: n_uni_rtime + n_hkx) = 1
    order_hkx(n_uni_rtime + 1 : n_uni_rtime + n_hkx) = [(ii,ii=1,n_hkx,1)]
    rtime = [uni_rtime, this%recordtime%value]

    counter = 1
    do ii = 1, events%nelements
      if( abs(events%recordtime%value(ii) - uni_rtime(counter)) < tiny(uni_rtime(counter))) then
!        ftime(counter) = events%frametime%value(ii)
        ftime(counter) = events%time%value(ii)
        cycle
      else
        counter = counter + 1
!        ftime(counter) = events%frametime%value(ii)
        ftime(counter) = events%time%value(ii)
      end if
    end do
    ! write last frametime to last recordtime - this might not be correct
!    ftime(n_uni_rtime) = events%frametime%value(events%nelements)
    ftime(n_uni_rtime) = events%time%value(events%nelements)

    call sort(rtime, status, order)
    type = type(order)
    order_hkx = order_hkx(order)
    ftime = ftime(order)

    ! all entries from events + hk extension is checked
    within_gti = get_GTI_mask(rtime, rGTI, status)

    ! only check for the recordtimes in the events extension
    ! to get the maximum number of unique recordtimes as an indicator on how to calculate the frametime
    evt_within_gti = get_GTI_mask(uni_rtime, rGTI, status)
    n_evt_complete = sum(evt_within_gti)
    call dealloc(evt_within_gti, status)

    ! number of processed evts with complete records (all frametimes are available)
    ! complete record are always within gti
    n_evt = 0
    do ii = 1, size(rtime)

      ! check if its a record and within the recordtime gti
      if(type(ii) == 0) then
        if(within_gti(ii) == true) n_evt = n_evt + 1
      end if

      ! check if record is hkx
      if(type(ii) == 1) then
        ! no complete record yet -> get the next two complete record
        if(n_evt == 0) then
          call get_next_complete_event_record(rtime, type, within_gti, ii, jj, status)
          call get_next_complete_event_record(rtime, type, within_gti, jj, kk, status)
          time_scaling = (ftime(jj) - ftime(kk)) / (rtime(jj) - rtime(kk))
          rtime_duration = rtime(jj) - rtime(ii)
          ftime_duration = rtime_duration * time_scaling
          ftime(ii) = ftime(jj) - ftime_duration
          this%rframetime%value(order_hkx(ii)) = ftime(ii)
          this%q_rframetime%value(order_hkx(ii)) = 3
        ! all complete records have been processed -> get the last two complete records
        else if(n_evt == n_evt_complete) then
          call get_last_complete_event_record(rtime, type, within_gti, ii, kk, status)
          call get_last_complete_event_record(rtime, type, within_gti, kk, jj, status)
          time_scaling = (ftime(jj) - ftime(kk)) / (rtime(jj) - rtime(kk))
          rtime_duration = rtime(jj) - rtime(ii)
          ftime_duration = rtime_duration * time_scaling
          ftime(ii) = ftime(jj) - ftime_duration
          this%rframetime%value(order_hkx(ii)) = ftime(ii)
          this%q_rframetime%value(order_hkx(ii)) = 4
        ! position between two complete event records
        else
          call get_next_complete_event_record(rtime, type, within_gti, ii, kk, status)
          call get_last_complete_event_record(rtime, type, within_gti, ii, jj, status)
          time_scaling = (ftime(jj) - ftime(kk)) / (rtime(jj) - rtime(kk))
          rtime_duration = rtime(jj) - rtime(ii)
          ftime_duration = rtime_duration * time_scaling
          ftime(ii) = ftime(jj) - ftime_duration
          this%rframetime%value(order_hkx(ii)) = ftime(ii)
          ! if a recordtime of HK is within the rGTI
          ! since a frametime / recordtime exists for the start / stop of a rGTI
          ! not much interpolation is needed and thus the quality should be higher
          if(within_gti(ii) == true) then
            this%q_rframetime%value(order_hkx(ii)) = 1
          else
            this%q_rframetime%value(order_hkx(ii)) = 2
          end if
        end if

      end if

    end do

    this%rframetime%exists = true
    this%q_rframetime%exists = true

    call dealloc(rtime, status)
    call dealloc(uni_rtime, status)
    call dealloc(ftime_event, status)
    call dealloc(ftime, status)
    call dealloc(order, status)
    call dealloc(order_ftime, status)
    call dealloc(type, status)
    call dealloc(order_hkx, status)
    call dealloc(within_gti, status)

  end subroutine derive_frametime_tHKx3
    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine delete_duplicates_tHKx3.
  !> @brief
  !> delete duplicate entries based on time
  !
  !
  !> @param this
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_duplicates_tHKx3(this, status, report)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx3) :: this
    integer, optional, intent(in) :: report

    !internal param
    integer :: ii,num_dup
    integer, dimension(:), allocatable :: order
    logical, dimension(this%nelements) :: mask
    character(buffsize) :: nelements_dup_str

    character(*), parameter :: function = "delete_duplicates_tHKx3"

    if(status > 100 .or. status == s_errRep) return
    if(this%nelements == 1) return

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]
    call this%sort(this%recordtime%value, status)

    mask = .true.
    do ii = 2, this%nelements
      if(abs(this%recordtime%value(ii) - this%recordtime%value(ii-1)) < epsilon(this%recordtime%value(1))) then
        mask(ii) = .false.
      end if
    end do

    order = pack(order, mask)
    if(present(report)) then
      num_dup = this%nelements - size(order)
      if(num_dup > 0) then
        nelements_dup_str = int2str(num_dup, status)
        call wrerr(function, report, "number of duplicate records in HKx3 removed: " // trim(nelements_dup_str))
      end if
    end if
    call this%apply_index(order, status)

    call dealloc(order, status)

  end subroutine delete_duplicates_tHKx3


end module evtool_hkx3_mod
