! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module evtool_hkx5_mod
  use param
  use evtool_hkx1_mod
  use erolib_tbintable_mod
  use erolib_ffits_col_mod
  use fitsimage, only : stdhdr_type, fread_stdhdr, fwrite_stdhdr

  implicit none

  ! extnames used in the files

  character(*), parameter, public :: hkx5_ext = "HKx5"
  character(*), parameter, private :: extname_default = "HKx5"

  ! basic type of the catalog used for the cumulative list
  type, public :: tHKx5

    integer :: unit = 0
    integer :: tel_id = -1
    character(1) :: tel_id_str = ''
    integer :: nelements = 0 ! number of elements (rows)
    character(buffsize) :: filename = ""
    character(buffsize) :: extname = extname_default
    type(tBinTable) :: table ! description of the binary table
    !    type(stdhdr_type) :: stdhdr

    type(r8_tColumn) :: time
    type(r8_tColumn) :: recordtime
    type(r4_tColumn) :: pxmipthr
    type(r4_tColumn) :: pxprimthr
    type(r4_tColumn) :: pxfwsteps
    type(r4_tColumn) :: pxspltct ! split event counter?
    type(r4_tColumn) :: pxevtbmax
    type(r4_tColumn) :: pxoutbfree
    type(r4_tColumn) :: pxmipctrpt

    ! derivative columns
    type(i2_tColumn) :: dxmipthr ! mip threshold
    type(i2_tColumn) :: dxprimthr ! event threshold
    type(i2_tColumn) :: dxmipctrpt ! mip counter wiederholungsrate (#ccd frames)

    ! frametime information
    type(r8_tColumn) :: rframetime
    type(i2_tColumn) :: q_rframetime

  contains
    procedure :: init => init_tHKx5
    procedure :: free => free_tHKx5
    procedure :: open => open_tHKx5
    procedure :: close => close_tHKx5
    procedure :: create_extension => create_extension_tHKx5
    procedure :: read => read_tHKx5
    procedure :: write => write_tHKx5
    procedure :: apply_index => apply_index_tHKx5
    procedure :: sort => sort_tHKx5
    procedure :: set_tel_id => set_tel_id_tHKx5
    procedure :: reset_tel_id => reset_tel_id_tHKx5
    procedure :: reset_colnum => reset_colnum_tHKx5

    procedure :: derive_columns => derive_columns_tHKx5
    procedure :: derive_frametime => derive_frametime_tHKx5
    procedure :: delete_duplicates => delete_duplicates_tHKx5

  end type tHKx5

contains

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloacte n hkx5 structures
  !
  !
  !
  !> @param[inout] hkx5
  !> @param[in] n
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine alloc_hkx5(hkx5, n, status)
    implicit none

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    type(tHKx5), allocatable, dimension(:) :: hkx5

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_hkx5"

    call dealloc_hkx5(hkx5, status)
    allocate(hkx5(n), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

  end subroutine alloc_hkx5

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> dealloacte hkx5 structures
  !
  !
  !
  !> @param[inout] hkx5
  !> @param[inout] status
  !---------------------------------------------------------------------------
  subroutine dealloc_hkx5(hkx5, status)

    integer :: stat
    integer, intent(inout) :: status

    type(tHKx5), allocatable, dimension(:) :: hkx5

    character(*), parameter :: function = "dealloc_hkx5"

    if(status > 100 .or. status == s_errRep) return

    if(allocated(hkx5)) then
      deallocate(hkx5, stat=stat)
      call FError(function, stat, "dealloc hkx5 type", status)
    end if

  end subroutine dealloc_hkx5

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> init the tHKx5 type
  !> nband = number of different energy bands
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional]
  !---------------------------------------------------------------------------

  subroutine init_tHKx5(this, n, status, extname)
    implicit none

    integer, intent(inout) :: status
    class(tHKx5) :: this
    integer, intent(in) :: n
    character(*), intent(in), optional :: extname

    character(*), parameter :: function = "init_tHKx5"

    if(status > 100 .or. status == s_errRep) return

    this%nelements = n
    this%extname = extname_default
    if(present(extname)) this%extname = extname

    call this%time%init("TIME", n, status, tunit="s", is_optional=true)
    call this%recordtime%init("RecordTime", n, status, tunit="s")
    call this%pxmipthr%init("PxMIPTHR", n, status)
    call this%pxprimthr%init("PxPRIMTHR", n, status)
    call this%pxfwsteps%init("PxFWSTEPS", n, status)
    call this%pxspltct%init("PxSPLTCT", n, status)
    call this%pxevtbmax%init("PxEVTBMAX", n, status)
    call this%pxoutbfree%init("PxOUTBFREE", n, status)
    call this%pxmipctrpt%init("PxMIPCTRPT", n, status)

    call this%dxmipthr%init("DxMIPTHR", n, status, is_optional=true)
    call this%dxprimthr%init("DxPRIMTHR", n, status, is_optional=true)
    call this%dxmipctrpt%init("DxMIPCTRPT", n, status, is_optional=true)

    call this%rframetime%init("RFRAMETIME", n, status, tunit="s", is_optional=true)
    call this%q_rframetime%init("QRFRAMETIME", n, status, is_optional=true)

  end subroutine init_tHKx5

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the tHKx5 type
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional] nband
  !---------------------------------------------------------------------------

  subroutine free_tHKx5(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tHKx5) :: this

    character(*), parameter :: function = "free_tHKx5"

    if(status > 100 .or. status == s_errRep) return

    this%nelements = 0

    call this%time%free(status)
    call this%recordtime%free(status)
    call this%pxmipthr%free(status)
    call this%pxprimthr%free(status)
    call this%pxfwsteps%free(status)
    call this%pxspltct%free(status)
    call this%pxevtbmax%free(status)
    call this%pxoutbfree%free(status)
    call this%pxmipctrpt%free(status)

    call this%dxmipthr%free(status)
    call this%dxprimthr%free(status)
    call this%dxmipctrpt%free(status)

    call this%rframetime%free(status)
    call this%q_rframetime%free(status)

  end subroutine free_tHKx5

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> routine to open the HKx5
  !
  !
  !
  !> @param[inout] this
  !> @param[in] filename, rwmode
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine open_tHKx5(this, filename, rwmode, status, delete)
    implicit none

    integer, intent(inout) :: status
    class(tHKx5) :: this

    character(*), intent(in) :: filename
    integer, intent(in) :: rwmode

    integer :: exists

    integer, intent(in), optional :: delete

    character(*), parameter :: function = "open_tHKx5"

    if(status > 100 .or. status == s_errRep) return

    this%filename = filename

    call ffits_get_iounit(this%unit, status)
    if(rwmode /= create) then
      call ffits_open_file(this%unit, filename, rwmode, status)
      if(status == ok) call wrlog(function, trim(this%filename) // " successfully opened")
    else
      call ffits_file_exists(filename, exists, status)
      if(present(delete) .and. exists == 1) then
        if(delete == true) call delete_file_fortran(filename, status)
      end if
      call ffits_create_file(this%unit, filename, fblocksize, status)
      if(status == ok) call wrlog(function, trim(this%filename) // " successfully created")
    end if

  end subroutine open_tHKx5

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> close the tHKx5 file
  !
  !
  !
  !> @param[inout] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine close_tHKx5(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tHKx5) :: this

    character(*), parameter :: function = "close_tHKx5"

    if(status > 100 .or. status == s_errRep) return

    call ffits_close_file(this%unit, status)
    call ffits_free_iounit(this%unit, status)

  end subroutine close_tHKx5

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> subroutine of class tHKx5
  !> create the extension of this type in a FITS file
  !
  !
  !> @param[inout] this
  !> @param[in, optional] aux
  !> @param[out]
  !> @param[inout]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine create_extension_tHKx5(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ncol ! number of columns in the file

    class(tHKx5) :: this
    character(*), parameter :: function = "create_extension_tHKx5"

    if(status > 100 .or. status == s_errRep) return

    ncol = 8

    call this%table%init(ncol, status)

    call this%table%insert_ffits_col(this%time, status)
    call this%table%insert_ffits_col(this%pxmipthr, status)
    call this%table%insert_ffits_col(this%pxprimthr, status)
    call this%table%insert_ffits_col(this%pxfwsteps, status)
    call this%table%insert_ffits_col(this%pxspltct, status)
    call this%table%insert_ffits_col(this%pxevtbmax, status)
    call this%table%insert_ffits_col(this%pxoutbfree, status)
    call this%table%insert_ffits_col(this%recordtime, status)
    call this%table%insert_ffits_col(this%pxmipctrpt, status)

    call this%table%insert_ffits_col(this%dxmipthr, status)
    call this%table%insert_ffits_col(this%dxprimthr, status)
    call this%table%insert_ffits_col(this%dxmipctrpt, status)

    call this%table%insert_ffits_col(this%rframetime, status)
    call this%table%insert_ffits_col(this%q_rframetime, status)

    call this%table%insertFits(this%unit, this%extname, status)

  end subroutine create_extension_tHKx5

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the hkx5 type from a FITS file
  !
  !
  !
  !> @param[in] this
  !> @param[out] filename
  !> @param[inout] status
  !> @param[in, optional] extend
  !---------------------------------------------------------------------------

  subroutine read_tHKx5(this, status, filename, tel_id, set_tel_id)
    implicit none

    integer, intent(inout) :: status
    integer :: nelements, rowsize, frow, lrow, nrows, ii
    class(tHKx5) :: this
    character(*), optional, intent(in) :: filename
    integer, optional, intent(in) :: tel_id
    integer :: telescope_id
    character(1) :: tel_id_str
    logical, optional :: set_tel_id
    character(buffsize) :: extname

    character(fcolnamelength) :: col
    character(*), parameter :: function = "read_tHKx5"

    if(status > 100 .or. status == s_errRep) return

    if(present(filename)) call this%open(filename, readonly, status)

    telescope_id = this%tel_id

    if(present(tel_id)) then
      this%extname = extname_default
      tel_id_str = int2str(tel_id, status)
      this%extname(3:3) = tel_id_str
      telescope_id = tel_id
    end if

    status = ignore
    call ffits_movnam_hdu(this%unit, BINARY_TBL, this%extname, status)
    if(status == ignore) then
      status = ok
    else
      ! extension does not exist
      status = warn
      return
    end if

    call ffits_get_num_rows(this%unit, nelements, status)
    call ffits_get_rowsize(this%unit, rowsize, status)

    extname = this%extname
    call this%init(nelements, status)
    this%tel_id = telescope_id
    this%extname = extname

    if(present(set_tel_id)) then
      if(set_tel_id .eqv. .true.) call this%set_tel_id(telescope_id, status)
    end if

    !    call fread_stdhdr(this%unit, this%stdhdr, status)

    if(nelements == 0 .or. status == s_errRep) return

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1

      call this%time%read(this%unit, status, frow=frow, lrow=lrow)
      call this%recordtime%read(this%unit, status, frow=frow, lrow=lrow)

      call this%pxmipthr%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxprimthr%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxfwsteps%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxspltct%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxevtbmax%read(this%unit, status, frow=frow, lrow=lrow)
      call this%pxoutbfree%read(this%unit, status, frow=frow, lrow=lrow)

      call this%pxmipctrpt%read(this%unit, status, frow=frow, lrow=lrow)

      call this%dxmipthr%read(this%unit, status, frow=frow, lrow=lrow)
      call this%dxprimthr%read(this%unit, status, frow=frow, lrow=lrow)
      call this%dxmipctrpt%read(this%unit, status, frow=frow, lrow=lrow)

      call this%rframetime%read(this%unit, status, frow=frow, lrow=lrow)
      call this%q_rframetime%read(this%unit, status, frow=frow, lrow=lrow)

    end do

    if(present(filename)) call this%close(status)

  end subroutine read_tHKx5

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the hkx5 type to a FITS file
  !
  !
  !
  !> @param[in] this
  !> @param[in, optional] aux
  !> @param[out] filename
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_tHKx5(this, status, filename)
    implicit none

    integer, intent(inout) :: status
    class(tHKx5) :: this
    integer :: rowsize, ii, frow, nelements, lrow, nrows
    character(*), optional, intent(in) :: filename

    character(*), parameter :: function = "write_tHKx5"

    if(status > 100 .or. status == s_errRep) return

    if(present(filename)) then
      call this%open(filename, create, status)
      if(status /= ok) return
    end if

    ! move to the correct extension
    status = ignore
    call ffits_movnam_hdu(this%unit, BINARY_TBL, this%extname, status)
    if(status /= ignore) then
      status = ok
      call this%create_extension(status)
    else
      status = ok
    end if

    call ffits_get_rowsize(this%unit, rowsize, status)
    nelements = this%nelements

    !    call fwrite_stdhdr(this%unit, this%stdhdr, status)


    ! write 1D arrays
    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1

      call this%time%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxmipthr%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxprimthr%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxfwsteps%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxspltct%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxevtbmax%write(this%unit, status, frow=frow, lrow=lrow)
      call this%pxoutbfree%write(this%unit, status, frow=frow, lrow=lrow)
      call this%recordtime%write(this%unit, status, frow=frow, lrow=lrow)

      call this%pxmipctrpt%write(this%unit, status, frow=frow, lrow=lrow)

      call this%dxmipthr%write(this%unit, status, frow=frow, lrow=lrow)
      call this%dxprimthr%write(this%unit, status, frow=frow, lrow=lrow)
      call this%dxmipctrpt%write(this%unit, status, frow=frow, lrow=lrow)

      call this%rframetime%write(this%unit, status, frow=frow, lrow=lrow)
      call this%q_rframetime%write(this%unit, status, frow=frow, lrow=lrow)
    end do

    if(present(filename)) call this%close(status)

  end subroutine write_tHKx5

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> reduce all entries in the columns to contain only a subset of the complete
  !> by providing an array list with the desired indices
  !
  !
  !> @param[in] this
  !> @param[in] order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_index_tHKx5(this, order, status)
    implicit none

    integer, intent(inout) :: status

    class(tHKx5) :: this

    integer, dimension(:), intent(in) :: order

    character(*), parameter :: function = "apply_index_tHKx5"

    if(status > 100 .or. status == s_errRep) return

    if(this%nelements == 0) return

    this%nelements = size(order)

    call this%time%apply_index(order, status)
    call this%recordtime%apply_index(order, status)
    call this%pxmipthr%apply_index(order, status)
    call this%pxprimthr%apply_index(order, status)
    call this%pxfwsteps%apply_index(order, status)
    call this%pxspltct%apply_index(order, status)
    call this%pxevtbmax%apply_index(order, status)
    call this%pxoutbfree%apply_index(order, status)
    call this%pxmipctrpt%apply_index(order, status)

    call this%dxmipthr%apply_index(order, status)
    call this%dxprimthr%apply_index(order, status)
    call this%dxmipctrpt%apply_index(order, status)

    call this%rframetime%apply_index(order, status)
    call this%q_rframetime%apply_index(order, status)

  end subroutine apply_index_tHKx5

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge all entries of the columns with a second tHKx5
  !
  !
  !> @param[in] HKx5_A, HKx5_B
  !> @param[in] HKx5_C
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine merge_HKx5(HKx5_A, HKx5_B, HKx5_C, status)
    implicit none

    integer, intent(inout) :: status

    class(tHKx5), intent(in) :: HKx5_A
    class(tHKx5), intent(in) :: HKx5_B
    class(tHKx5), intent(out) :: HKx5_C

    character(*), parameter :: function = "merge_HKx5"

    if(status > 100 .or. status == s_errRep) return

    ! copy all important informations
    HKx5_C%table        = HKx5_A%table
    !    HKx5_C%stdhdr        = HKx5_A%stdhdr

    if(HKx5_A%nelements == 0 .and. HKx5_B%nelements == 0) then
      call wrlog(function, "cats to merge have zero elements - nothing to be done here")
      return
    end if

    HKx5_C%nelements = HKx5_A%nelements + HKx5_B%nelements
    HKx5_C%extname = HKx5_A%extname

    call merge_tColumn(HKx5_A%time, HKx5_B%time, HKx5_C%time, status)
    call merge_tColumn(HKx5_A%recordtime, HKx5_B%recordtime, HKx5_C%recordtime, status)
    call merge_tColumn(HKx5_A%pxmipthr, HKx5_B%pxmipthr, HKx5_C%pxmipthr, status)
    call merge_tColumn(HKx5_A%pxprimthr, HKx5_B%pxprimthr, HKx5_C%pxprimthr, status)
    call merge_tColumn(HKx5_A%pxfwsteps, HKx5_B%pxfwsteps, HKx5_C%pxfwsteps, status)
    call merge_tColumn(HKx5_A%pxspltct, HKx5_B%pxspltct, HKx5_C%pxspltct, status)
    call merge_tColumn(HKx5_A%pxevtbmax, HKx5_B%pxevtbmax, HKx5_C%pxevtbmax, status)
    call merge_tColumn(HKx5_A%pxoutbfree, HKx5_B%pxoutbfree, HKx5_C%pxoutbfree, status)
    call merge_tColumn(HKx5_A%pxmipctrpt, HKx5_B%pxmipctrpt, HKx5_C%pxmipctrpt, status)

    call merge_tColumn(HKx5_A%dxmipthr, HKx5_B%dxmipthr, HKx5_C%dxmipthr, status)
    call merge_tColumn(HKx5_A%dxprimthr, HKx5_B%dxprimthr, HKx5_C%dxprimthr, status)
    call merge_tColumn(HKx5_A%dxmipctrpt, HKx5_B%dxmipctrpt, HKx5_C%dxmipctrpt, status)

    call merge_tColumn(HKx5_A%rframetime, HKx5_B%rframetime, HKx5_C%rframetime, status)
    call merge_tColumn(HKx5_A%q_rframetime, HKx5_B%q_rframetime, HKx5_C%q_rframetime, status)

  end subroutine merge_HKx5


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort the type entries according to one array
  !
  !
  !
  !> @param[in] this
  !> @param[in] array
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_tHKx5(this, array, status, ascend)
    implicit none

    integer, intent(inout) :: status
    integer, optional :: ascend

    class(tHKx5) :: this
    class(*), dimension(:) :: array

    integer, dimension(:), allocatable :: order

    real(r8),allocatable, dimension(:) :: tmparray_r8
    real(r4),allocatable, dimension(:) :: tmparray_r4
    integer(i4),allocatable, dimension(:) :: tmparray_i4
    integer(i2),allocatable, dimension(:) :: tmparray_i2
    integer(i1),allocatable, dimension(:) :: tmparray_i1

    character(*), parameter :: function = "sort_tHKx5"

    if(status > 100 .or. status == s_errRep) return

    if(this%nelements /= size(array)) then
      call wrerr(function // " tHKx5(_ext) type", s_err, "length of sorting array does not conform to length of type")
      status = s_errRep
    end if

    select type(array)
      type is ( real(r4) )
      tmparray_r4 = array
      call sort_real(tmparray_r4 ,status, order)
      type is ( real(r8) )
      tmparray_r8 = array
      call sort_double(tmparray_r8 ,status, order)
      type is ( integer(i4) )
      tmparray_i4 = array
      call sort_i4(tmparray_i4 ,status, order)
      type is ( integer(i2) )
      tmparray_i2 = array
      call sort_i2(tmparray_i2 ,status, order)
      type is ( integer(i1) )
      tmparray_i1 = array
      call sort_i1(tmparray_i1 ,status, order)
      class default
      call wrstop(function, "sorting tHKx5 with the current array type is not supported for tHKx5")
    end select

    if(present(ascend)) then
      if(ascend == false) call reverse_array(order)
    end if

    call this%apply_index(order, status)

    call dealloc(order, status)
    call dealloc(tmparray_i1, status)
    call dealloc(tmparray_i2, status)
    call dealloc(tmparray_i4, status)
    call dealloc(tmparray_r4, status)
    call dealloc(tmparray_r8, status)

  end subroutine sort_tHKx5

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine set_tel_id_tHKx5.
  !> @brief
  !> set the telesope id for this extension / modify extanme
  !
  !
  !> @param this, tel_id
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine set_tel_id_tHKx5(this, tel_id, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    integer, intent(in) :: tel_id
    class(tHKx5) :: this

    character(*), parameter :: function = "set_tel_id_tHKx5"

    if(status > 100 .or. status == s_errRep) return
    this%tel_id = tel_id
    this%tel_id_str =  int2str(tel_id, status)
    this%extname = extname_default
    this%extname(3:3) = this%tel_id_str

    ! replace x with TelID except time
    !    this%time%name(2:2) = this%tel_id_str
    this%pxmipthr%name(2:2) = this%tel_id_str
    this%pxprimthr%name(2:2) = this%tel_id_str
    this%pxfwsteps%name(2:2) = this%tel_id_str
    this%pxspltct%name(2:2) = this%tel_id_str
    this%pxevtbmax%name(2:2) = this%tel_id_str
    this%pxoutbfree%name(2:2) = this%tel_id_str
    this%pxmipctrpt%name(2:2) = this%tel_id_str

  end subroutine set_tel_id_tHKx5

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine reset_tel_id_tHKx5.
  !> @brief
  !> set the telesope id for this extension / modify extanme
  !
  !
  !> @param this
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine reset_tel_id_tHKx5(this, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx5) :: this

    character(*), parameter :: function = "reset_tel_id_tHKx5"

    if(status > 100 .or. status == s_errRep) return

    ! replace x with TelID except time
    !    this%time%name(2:2) = this%tel_id_str
    this%pxmipthr%name(2:2) = "x"
    this%pxprimthr%name(2:2) = "x"
    this%pxfwsteps%name(2:2) = "x"
    this%pxspltct%name(2:2) = "x"
    this%pxevtbmax%name(2:2) = "x"
    this%pxoutbfree%name(2:2) = "x"
    this%pxmipctrpt%name(2:2) = "x"

  end subroutine reset_tel_id_tHKx5

        !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine reset_colnum_tHKx5.
  !> @brief
  !> reset the colnum of all columns in tHKx5
  !
  !
  !> @param this
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine reset_colnum_tHKx5(this, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx5) :: this

    !internal param

    character(*), parameter :: function = "reset_colnum_tHKx5"

    if(status > 100 .or. status == s_errRep) return

    this%time%num = 0
    this%recordtime%num = 0
    this%pxmipthr%num = 0
    this%pxprimthr%num = 0
    this%pxfwsteps%num = 0
    this%pxspltct%num = 0
    this%pxevtbmax%num = 0
    this%pxoutbfree%num = 0
    this%pxmipctrpt%num = 0

    this%dxmipthr%num = 0
    this%dxprimthr%num = 0
    this%dxmipctrpt%num = 0

    this%rframetime%num = 0
    this%q_rframetime%num = 0

  end subroutine reset_colnum_tHKx5

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine derive_columns_tHKx5.
  !> @brief
  !> derive additional columns representing the changes of a column
  !
  !
  !> @param this
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine derive_columns_tHKx5(this, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx5) :: this

    !internal param
    integer :: ii
    character(*), parameter :: function = "derive_columns_tHKx5"

    if(status > 100 .or. status == s_errRep) return

    if(this%nelements < 1) return

    this%dxmipthr%value(1) = 0
    this%dxprimthr%value(1) = 0
    this%dxmipctrpt%value(1) = 0

    if(this%time%exists == false) then
      this%time%value = this%recordtime%value
      this%time%exists = true
    end if

    if(this%nelements < 2) return

    do ii = 2, this%nelements

      if(this%pxmipthr%value(ii) == this%pxmipthr%value(ii-1)) this%dxmipthr%value(ii) = 0
      if(this%pxprimthr%value(ii) == this%pxprimthr%value(ii-1)) this%dxprimthr%value(ii) = 0
      if(this%pxmipctrpt%value(ii) == this%pxmipctrpt%value(ii-1)) this%dxmipctrpt%value(ii) = 0

      if(this%pxmipthr%value(ii) > this%pxmipthr%value(ii-1)) this%dxmipthr%value(ii) = 1
      if(this%pxprimthr%value(ii) > this%pxprimthr%value(ii-1)) this%dxprimthr%value(ii) = 1
      if(this%pxmipctrpt%value(ii) > this%pxmipctrpt%value(ii-1)) this%dxmipctrpt%value(ii) = 1

      if(this%pxmipthr%value(ii) < this%pxmipthr%value(ii-1)) this%dxmipthr%value(ii) = -1
      if(this%pxprimthr%value(ii) < this%pxprimthr%value(ii-1)) this%dxprimthr%value(ii) = -1
      if(this%pxmipctrpt%value(ii) < this%pxmipctrpt%value(ii-1)) this%dxmipctrpt%value(ii) = -1

    end do

    this%dxmipthr%exists = true
    this%dxprimthr%exists = true
    this%dxmipctrpt%exists = true

  end subroutine derive_columns_tHKx5

 !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine derive_frametime_tHKx5.
  !> @brief
  !> description
  !
  !
  !> @param parameter
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine derive_frametime_tHKx5(this, events, rGTI, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx5) :: this
    type(tEvents_pp), intent(inout) :: events
    type(tGTI), intent(in) :: rGTI

    !internal param
    integer :: ii, jj, kk, n_hkx, n_uni_rtime, n_evt, n_evt_complete, counter
    integer, allocatable, dimension(:) :: order, order_ftime
    real(r8) :: ftime_duration, rtime_duration, ratio, time_scaling
    real(r8), allocatable, dimension(:) :: rtime, uni_rtime, ftime_event, ftime
    integer(i2), allocatable, dimension(:) :: type      ! 1 = hkx
    integer(i4), allocatable, dimension(:) :: order_hkx ! index of hk record
    integer(i4), allocatable, dimension(:) :: within_gti, evt_within_gti

    character(*), parameter :: function = "derive_frametime_tHKx5"

    if(status > 100 .or. status == s_errRep) return

    ! alternative to loop which is commented below
    ! sort events in frametime and afterwards in recordtime
!    call events%sort(events%frametime%value, status)
    call events%sort(events%time%value, status)
    call events%sort(events%recordtime%value, status)

    uni_rtime = events%recordtime%value(unique(events%recordtime%value, status))

    n_hkx = this%nelements
    n_uni_rtime = size(uni_rtime)

    ! create combined array of recordtimes
    call alloc(type, n_uni_rtime + n_hkx, status)
    call alloc(rtime, n_uni_rtime + n_hkx, status)
    call alloc(order_hkx, n_uni_rtime + n_hkx, status)
    call alloc(ftime, n_uni_rtime + n_hkx, status)

    type(n_uni_rtime + 1: n_uni_rtime + n_hkx) = 1
    order_hkx(n_uni_rtime + 1 : n_uni_rtime + n_hkx) = [(ii,ii=1,n_hkx,1)]
    rtime = [uni_rtime, this%recordtime%value]

    counter = 1
    do ii = 1, events%nelements
      if( abs(events%recordtime%value(ii) - uni_rtime(counter)) < tiny(uni_rtime(counter))) then
!        ftime(counter) = events%frametime%value(ii)
        ftime(counter) = events%time%value(ii)
        cycle
      else
        counter = counter + 1
!        ftime(counter) = events%frametime%value(ii)
        ftime(counter) = events%time%value(ii)
      end if
    end do
    ! write last frametime to last recordtime - this might not be correct
!    ftime(n_uni_rtime) = events%frametime%value(events%nelements)
    ftime(n_uni_rtime) = events%time%value(events%nelements)

    call sort(rtime, status, order)
    type = type(order)
    order_hkx = order_hkx(order)
    ftime = ftime(order)

    ! all entries from events + hk extension is checked
    within_gti = get_GTI_mask(rtime, rGTI, status)

    ! only check for the recordtimes in the events extension
    ! to get the maximum number of unique recordtimes as an indicator on how to calculate the frametime
    evt_within_gti = get_GTI_mask(uni_rtime, rGTI, status)
    n_evt_complete = sum(evt_within_gti)
    call dealloc(evt_within_gti, status)

    ! number of processed evts with complete records (all frametimes are available)
    ! complete record are always within gti
    n_evt = 0
    do ii = 1, size(rtime)

      ! check if its a record and within the recordtime gti
      if(type(ii) == 0) then
        if(within_gti(ii) == true) n_evt = n_evt + 1
      end if

      ! check if record is hkx
      if(type(ii) == 1) then
        ! no complete record yet -> get the next two complete record
        if(n_evt == 0) then
          call get_next_complete_event_record(rtime, type, within_gti, ii, jj, status)
          call get_next_complete_event_record(rtime, type, within_gti, jj, kk, status)
          time_scaling = (ftime(jj) - ftime(kk)) / (rtime(jj) - rtime(kk))
          rtime_duration = rtime(jj) - rtime(ii)
          ftime_duration = rtime_duration * time_scaling
          ftime(ii) = ftime(jj) - ftime_duration
          this%rframetime%value(order_hkx(ii)) = ftime(ii)
          this%q_rframetime%value(order_hkx(ii)) = 3
        ! all complete records have been processed -> get the last two complete records
        else if(n_evt == n_evt_complete) then
          call get_last_complete_event_record(rtime, type, within_gti, ii, kk, status)
          call get_last_complete_event_record(rtime, type, within_gti, kk, jj, status)
          time_scaling = (ftime(jj) - ftime(kk)) / (rtime(jj) - rtime(kk))
          rtime_duration = rtime(jj) - rtime(ii)
          ftime_duration = rtime_duration * time_scaling
          ftime(ii) = ftime(jj) - ftime_duration
          this%rframetime%value(order_hkx(ii)) = ftime(ii)
          this%q_rframetime%value(order_hkx(ii)) = 4
        ! position between two complete event records
        else
          call get_next_complete_event_record(rtime, type, within_gti, ii, kk, status)
          call get_last_complete_event_record(rtime, type, within_gti, ii, jj, status)
          time_scaling = (ftime(jj) - ftime(kk)) / (rtime(jj) - rtime(kk))
          rtime_duration = rtime(jj) - rtime(ii)
          ftime_duration = rtime_duration * time_scaling
          ftime(ii) = ftime(jj) - ftime_duration
          this%rframetime%value(order_hkx(ii)) = ftime(ii)
          ! if a recordtime of HK is within the rGTI
          ! since a frametime / recordtime exists for the start / stop of a rGTI
          ! not much interpolation is needed and thus the quality should be higher
          if(within_gti(ii) == true) then
            this%q_rframetime%value(order_hkx(ii)) = 1
          else
            this%q_rframetime%value(order_hkx(ii)) = 2
          end if
        end if

      end if

    end do

    this%rframetime%exists = true
    this%q_rframetime%exists = true

    call dealloc(rtime, status)
    call dealloc(uni_rtime, status)
    call dealloc(ftime_event, status)
    call dealloc(ftime, status)
    call dealloc(order, status)
    call dealloc(order_ftime, status)
    call dealloc(type, status)
    call dealloc(order_hkx, status)
    call dealloc(within_gti, status)

  end subroutine derive_frametime_tHKx5
    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine delete_duplicates_tHKx5.
  !> @brief
  !> delete duplicate entries based on time
  !
  !
  !> @param this
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_duplicates_tHKx5(this, status, report)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHKx5) :: this
    integer, optional, intent(in) :: report

    !internal param
    integer :: ii,num_dup
    integer, dimension(:), allocatable :: order
    logical, dimension(this%nelements) :: mask
    character(buffsize) :: nelements_dup_str

    character(*), parameter :: function = "delete_duplicates_tHKx5"

    if(status > 100 .or. status == s_errRep) return
    if(this%nelements == 1) return

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]
    call this%sort(this%recordtime%value, status)

    mask = .true.
    do ii = 2, this%nelements
      if(abs(this%recordtime%value(ii) - this%recordtime%value(ii-1)) < epsilon(this%recordtime%value(1))) then
        mask(ii) = .false.
      end if
    end do

    order = pack(order, mask)
    if(present(report)) then
      num_dup = this%nelements - size(order)
      if(num_dup > 0) then
        nelements_dup_str = int2str(num_dup, status)
        call wrerr(function, report, "number of duplicate records in HKx5 removed: " // trim(nelements_dup_str))
      end if
    end if
    call this%apply_index(order, status)

    call dealloc(order, status)

  end subroutine delete_duplicates_tHKx5


end module evtool_hkx5_mod
