! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module evtool_image_mod
    use erolib_general_mod
    use erolib_ffits_mod
    use evtool_events_mod
    use erolib_wcs_mod
    implicit none

    private

        character(*), parameter :: image_ext = "IMAGE"

    type tImage2D_i4
        type(tWcs) :: wcs
        integer(i4) :: nelements = 0
        integer(i4), dimension(2) :: rebin
        integer(i4), dimension(2) :: resolution ! naxis
        real(r8), dimension(2) :: center_position
        integer(i4), dimension(:,:), allocatable :: value
        character(buffsize) :: extname = ""
    contains
        !        procedure, non_overridable :: readFromFits => read_tImage2D_i4
        procedure, non_overridable :: write => write_tImage2D_i4
        procedure, non_overridable :: create_extension => create_extension_tImage2D_i4
        procedure, non_overridable :: free => free_tImage2D_i4
        procedure, non_overridable :: init => init_tImage2D_i4
        procedure, non_overridable :: createFromEvents => create_from_events_tImage2D_i4
    end type tImage2D_i4

    public :: tImage2D_i4

contains

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> write the image to a fits file
    !
    !
    !
    !> @param[in] this
    !> @param[out] unit
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_tImage2D_i4(this, unit, status)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) :: unit

        class(tImage2D_i4) :: this

        character(*), parameter :: function = "write_tImage2D_i4"

        if(status > 100 .or. status == s_errRep) return

        ! move to the correct extension
        status = ignore
        call ffits_movnam_hdu(unit, IMAGE_HDU, this%extname, status)
        if(status /= ignore) then
            status = ok
            call this%create_extension(unit, status)
        else
            status = ok
        end if

        call ffits_write_img_2D(unit, this%resolution(1), this%resolution(1), this%resolution(2), this%value, status)

        call this%wcs%write(unit, IMAGE_HDU, status)

    end subroutine write_tImage2D_i4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> create an image extension
    !
    !
    !
    !> @param[in] this
    !> @param[out] unit
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine create_extension_tImage2D_i4(this, unit, status, abshdu, extname)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) :: unit
        integer, intent(in), optional :: abshdu
        integer :: moved, hdutype

        class(tImage2D_i4) :: this

        character(*), parameter :: function = "create_extension_tImage2D_i4"
        character(*), optional, intent(in) :: extname

        if(status > 100 .or. status == s_errRep) return

        moved = false

        if(present(extname) .and. present(abshdu)) then
            call wrerr(function, warn, "abshdu and extname provided, abshdu has the priority")
            call ffits_movabs_hdu(unit, abshdu, hdutype, status)
            moved = true
        end if

        if(present(abshdu) .and. moved == false) then
            call ffits_movabs_hdu(unit, abshdu, hdutype, status)
            moved = true
        end if

        if(present(extname) .and. moved == false) then
            call ffits_movnam_hdu(unit, BINARY_TBL, extname, status)
!**** 2019-09-12 IMS: shouldn't this be IMAGE_HDU not BINARY_TBL?
            moved = true
        end if

        ! go to the first primary extension by default
        if(moved == false) call ffits_movabs_hdu(unit, 1, hdutype, status)

        ! create img extension
        call ffits_create_img(unit, 32, 2, this%resolution, status)
        call ffits_flush(unit, status)

    end subroutine create_extension_tImage2D_i4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> alloc the content of tImage2D_i4 type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine init_tImage2D_i4(this, n, status)
        implicit none

        integer, dimension(2), intent(in) :: n
        integer, intent(inout) :: status

        class(tImage2D_i4) :: this

        character(*), parameter :: function = "init_tImage2D_i4"

        call alloc(this%value, n(1), n(2), status)
        this%value = 0
        this%nelements = n(1) * n(2)
        this%resolution = n
        this%extname = ""

    end subroutine init_tImage2D_i4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> free the content of tImage2D_i4 type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tImage2D_i4(this, status)
        implicit none

        integer, intent(inout) :: status

        class(tImage2D_i4) :: this

        character(*), parameter :: function = "free_tImage2D_i4"

        call dealloc(this%value, status)

    end subroutine free_tImage2D_i4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> create an image from an input event type (optional wcs keywords)
    !
    !
    !
    !> @param[in] this
    !> @param[out] events
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine create_from_events_tImage2D_i4(this, events, resolution, center_position, rebin, status)
        implicit none

        class(tImage2D_i4) :: this
        type(tEvents), intent(in) :: events
        integer(i4), intent(in), dimension(2) :: rebin, resolution
        real(r8), intent(in), dimension(2) :: center_position
        integer, intent(inout) :: status

        character(*), parameter :: function = "create_from_events_tImage2D_i4"
        integer :: ii,numGoodEvents
        logical, dimension(events%nelements) :: mask
        real(r8), dimension(:), allocatable :: xs, ys
        integer, dimension(:), allocatable :: xis, yis
        real(r8) :: meanX,meanY,xPixelsLo,xPixelsHi,yPixelsLo,yPixelsHi
        real(r4) :: halfImageXSize,halfImageYSize
        character(buffsize) :: meanXStr,meanYStr,ctrXStr,ctrYStr,num0Str,num1Str
        character(buffsize) :: xPixelsSpanStr,yPixelsSpanStr,xDimStr,yDimStr

        if(status > 100 .or. status == s_errRep) return

        mask = .true.
        where(events%ra==0.0 .and. events%dec==0.0) mask = .false.

        numGoodEvents = count(mask)
        if(events%nelements > numGoodEvents) then
          num0Str = int2str(events%nelements-numGoodEvents)
          call wrerr(function, warn, trim(num0Str)//" events found with RA==0 and DEC==0. These should be flagged or set to NULL!")
        end if

        xs = pack(events%x, mask)
        ys = pack(events%y, mask)
        allocate(xis(size(xs)))
        allocate(yis(size(ys)))

        ! Report comparison between mean and range of X and Y values on the one
        ! hand, and what the user has specified for the image centre and bounds
        ! on the other.
        !
        meanX = sum(xs/float(numGoodEvents))
        meanY = sum(ys/float(numGoodEvents))
        write(meanXStr, *) meanX
        write(meanYStr, *) meanY
        write(ctrXStr, *) center_position(1)
        write(ctrYStr, *) center_position(2)

        xPixelsLo = minval(xs)/rebin(1)
        xPixelsHi = maxval(xs)/rebin(1)
        yPixelsLo = minval(ys)/rebin(2)
        yPixelsHi = maxval(ys)/rebin(2)
        write(xPixelsSpanStr, *) xPixelsHi-xPixelsLo
        write(yPixelsSpanStr, *) yPixelsHi-yPixelsLo
        write(xDimStr, *) resolution(1)
        write(yDimStr, *) resolution(2)

        call wrchatty(function, "User has specified centre ("//trim(adjustl(ctrXStr))&
          //","//trim(adjustl(ctrYStr))//") whereas mean X/Y are ("&
          //trim(adjustl(meanXStr))//","//trim(adjustl(meanYStr))//").")

        call wrchatty(function, "User has specified image pixel sizes ("&
          //trim(adjustl(xDimStr))//","//trim(adjustl(yDimStr))&
          //") whereas event X/Y pixellised extents are ("&
          //trim(adjustl(xPixelsSpanStr))//","//trim(adjustl(yPixelsSpanStr))//").")

        this%rebin = rebin
        call this%init(resolution, status)

        ! Apply shift due to center position provided
        xis = nint(xs - center_position(1))
        yis = nint(ys - center_position(2))

        ! Check that at least some events are within the image bounds. If so, then check for spillover. If not, then warn, with a suggestion for a better image centre.
        !
        halfImageXSize = rebin(1)*resolution(1) / 2
        halfImageYSize = rebin(2)*resolution(2) / 2
        if(all(abs(xis)>halfImageXSize) .and. all(abs(yis)>halfImageYSize)) then
          call wrerr(function, warn, "Zero events fall within the image bounds.")
        else if (any(abs(xis)>halfImageXSize .or. abs(yis)>halfImageYSize)) then
          num0Str = int2str(count(mask))
          num1Str = int2str(count(abs(xis)>halfImageXSize .or. abs(yis)>halfImageYSize))
          call wrerr(function, warn, trim(num1Str)//" out of "//trim(num0Str)//" possible events fall outside the image bounds.")
        end if

        ! shift by half the resolution * factor
        xis = xis + halfImageXSize
        yis = yis + halfImageYSize

        ! calculate the new pixel position
        do ii = 1, numGoodEvents

            xis(ii) = xis(ii) / rebin(1)
            yis(ii) = yis(ii) / rebin(2)

            ! increase the number of photons / events
            if(xis(ii) > 0 .and. xis(ii) <= resolution(1) .and. &
               yis(ii) > 0 .and. yis(ii) <= resolution(2)) then
                this%value(xis(ii), yis(ii)) = this%value(xis(ii), yis(ii)) + 1
            end if

        end do

        this%extname = image_ext

        this%wcs = events%wcs
        call this%wcs%adjustForImage(rebin, resolution, center_position, status)

        deallocate(xs)
        deallocate(ys)
        deallocate(xis)
        deallocate(yis)

    end subroutine create_from_events_tImage2D_i4

end module evtool_image_mod
