! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

program evtool_main

    use param
    use erolib_ffits_mod
    use erolib_str_mod
    use erolib_general_mod
    use evtool2_mod
    use erolib_eventfile_mod

    implicit none

    integer :: status, ii, ierr

    type(tEvtool) :: task
    type(tEventFile), allocatable, dimension(:) :: eventfiles
    type(tEventFile) :: eventfile_out
    character(buffsize) :: msg,fileName

    character(*), parameter :: function = "evtool_main"

    status = 0

    ! read/open the .par file and the logs
    call opast(status)
    if(status /= ok) call wrstop(function, "error read par file")

    call wrchatty(function, "Read params.")
    call task%init(status)
    call task%read_params(status)
    !call task%print_params(status)
    call task%validate(status)

    call alloc_eventfiles(eventfiles, size(task%infile), status)

    if(task%memset) then
      call wrlog(function, "memset==.true., so memory setting for small files is used.")

      call wrlog(function, "Reading infiles.")
      do ii = 1, size(task%infile)
        write(msg,'(a,i2,a)'), "File ", ii, ":"
        call wrchatty(function, msg)
        call wrchatty(function, "  Opening the file.")
        call eventfiles(ii)%open(readonly, status, filename=task%infile(ii))
        call eventfiles(ii)%setup_for_read(status) ! This initializes and sets useage for all extensions.
        call wrchatty(function, "  Reading the file.")
        call eventfiles(ii)%read(status, want_wcs=.true.)
        call wrchatty(function, "  Closing the file.")
        call eventfiles(ii)%close(status)
        if(status /= ok) then
          call wrerr(function, s_err, "ERROR with file: " // trim(task%infile(ii)))
        end if
      end do

    else ! memset==.false., the current default
      call wrlog(function, "memset==.false., so memory setting for large files is used.")

      ! read in the informations for the stdheader and wcs
      do ii = 1, size(task%infile)
        write(msg,'(a,i2,a)'), "File ", ii, ":"
        call wrchatty(function, msg)
        call wrchatty(function, "  Opening the file.")
        call eventfiles(ii)%open(readonly, status, filename=task%infile(ii))
        call eventfiles(ii)%setup_for_read(status) ! This reads the std_header then initializes and sets useage for all extensions.
        call wrchatty(function, "  Moving to the EVENTS extension.")
        call FTMNHD(eventfiles(ii)%info%unitNo, BINARY_TBL, "EVENTS", 0, ierr)
        if(ierr==301) then
          call eventfiles(ii)%info%getFileName(fileName)
call wrstop(function, "EVENTS extension not found in file "//trim(fileName))
        else
          call report_FITS_err(function, ierr, s_err, "- extension EVENTS")
        end if

!        call wrchatty(function, "  Reading the standard header.")
!        call eventfiles(ii)%read_stdheader(status) ! sets eventfiles(ii)%ccdnr_infile.
        !      print *, eventfiles(ii)%std_header%pat_sel
        !      call eventfiles(ii)%events%wcs%readFromFits(eventfiles(ii)%unit, BINARY_TBL, x_col, y_col, status)
        call wrchatty(function, "  Reading the rawx/y selection information.")
        call eventfiles(ii)%read_rawxy_sel(status)
        call wrchatty(function, "  Closing the file.")
        call eventfiles(ii)%close(status)
!*** destroy eventfiles(ii) extensions, to clear memory?
        if(status /= ok) then
          call wrerr(function, s_err, "ERROR with file: " // trim(task%infile(ii)))
        end if
      end do
    end if ! memset

    if(task%filter_gti==true .and. task%filter_type_gti==externalgti_type)&
      call task%externalgti%header%keywords(r8GH_MJDREF)%setValue(eventfile_out%std_header%mjdref, status)

    if(size(eventfiles) > 1) call wrlog(function, "Merging infiles.")
    call compare_stdheader_eventfile(eventfiles, status)
    call wrchatty(function, "Merging headers.")
    call merge_stdheader_eventfile(eventfiles, eventfile_out, status) ! sets eventfile_out%ccdnr_infile etc.
    call eventfile_out%setup_for_write(status) ! note that this needs eventfile_out%ccdnr_infile to be filled first.
    call update_filter_settings_eventfile(task, eventfile_out, status)

    call wrchatty(function, "Merging extensions.")
    call merge_extensions(task%memset, eventfiles, eventfile_out, status) ! also destroys each of the input extensions after use.

    call dealloc_eventfiles(eventfiles, status)

    call wrlog(function, "Applying filters.")
    call apply_filter_eventfile(task, eventfile_out, status) ! also creates the image if this was requested.

    ! if the events extension should not be written set it here
    ! since events might be used before to create images and such
    if(task%write_events) then
      eventfile_out%write_events = .true.
    else
      eventfile_out%write_events = .false.
    end if

    if(.not.eventfile_out%events%header%keywords(strEvH_T_OFFCOR)%containsData)&
      call eventfile_out%events%header%keywords(strEvH_T_OFFCOR)%setValue("F", status)

    ! create outfile
    call wrchatty(function, "Create outfile.")
    call eventfile_out%open(create, status, filename=task%outfile, clobber=(task%delete_outfile>0))

    ! use temporary external skyfield
    if(task%external_skyfield /= "") eventfile_out%std_header%skyfield = task%external_skyfield

    call wrlog(function, "Writing outfile.")
    call eventfile_out%write(status)
    !  call eventfile_out%free(status)

    call wrchatty(function, "Close outfile")
    call eventfile_out%close(status)

!*** 2019-09-11 IMS: eventfile_out never seems to be freed?

    call task%free(status)

    call cpast (status, ierr)

end program evtool_main
