! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

!------------------------------------------------------------------------------
! eROSITA software team
!------------------------------------------------------------------------------
!
! MODULE: evtool_mod
!
!> @author
!> Christoph Grossberger
!
! DESCRIPTION:
!> functions and subroutines used by the evtool tool
!
!------------------------------------------------------------------------------

! Non-method subroutines in this module:
!    subroutine write_eventfile_out(EventFile, Evtool, status)
!    subroutine apply_filter_eventfile(EventFile, Evtool, status)
!    subroutine update_filter_settings_eventfile(Evtool, EventFile, status)
!    subroutine update_filter_gti_settings(Evtool, EventFile, status)
!    subroutine update_filter_telid_settings(Evtool, EventFile, status)
!    subroutine update_filter_energy_settings(Evtool, EventFile, status)
!    subroutine update_filter_flag_settings(Evtool, EventFile, status)
!    subroutine update_filter_pattern_settings(Evtool, EventFile, status)
!    subroutine update_filter_rawxy_settings(Evtool, EventFile, status)
!    subroutine extractRawXYBounds(rawxy_str, rawxy_sel, status)
!    subroutine validateEnergyInt(emin, emax, status)
!    subroutine interprete_str_gti (str, gti, status)

module evtool_mod

    use erolib_general_mod
    use erolib_sort_mod
    use erolib_str_mod
    use erolib_ffits_mod
    use erolib_region_mod
    use erolib_file_mod
    use evtool_deadcor_mod
    use evtool_corrat_mod
    use evtool_gti_mod
    use evtool_rawxy_mod
    use evtool_events_mod
    use evtool_eventfile_mod
    use time_mod
    use param

    implicit none

logical, private, parameter :: doTest=.false.

    ! list of private functions/subroutines
    private :: read_filelist

    type tEvtool
        integer :: nfile = 0
        character(buffsize), dimension(:), allocatable :: infile
        integer :: delete_outfile = 0
        character(buffsize) :: outfile = ""
        logical :: clobber = .false.
        integer :: memset = 0 ! the memory setting to read in all files at once (0 = false, 1 = true)
        real(r8) :: overlap = 0.0

        logical :: write_events = .false.
        logical :: write_image = .false.

        integer :: nenergy = 0
        integer :: filter_energy = 0
        real(dp), dimension(:), allocatable :: emin, emax
        integer :: filter_telid = 0
        integer :: ninst = 0
        integer, dimension(:), allocatable :: ccdnr
        character(buffsize), dimension(:), allocatable :: instrum
        integer :: filter_region = 0
        character(buffsize) :: region = ""
        type(tRegTable) :: regtable
        integer :: filter_gti = 0
        integer :: filter_type_gti = 0
        character(buffsize) :: gti_str = ""
        integer :: filter_flag = 0
        character(buffsize) :: flag = ""
        integer :: flag_hex = 0
        integer :: filter_pattern = 0
        integer :: pattern = 0
        integer :: filter_image = 0

        integer, dimension(:), allocatable :: resolution, rebin
        real(dp), dimension(:), allocatable :: center_position

        character(buffsize) :: rawxy_str = ""
        type(tRawXYSelection) :: rawxy_sel
        logical :: filter_rawxy = .false.

        type(tGTI) :: externalgti
        logical :: write_externalgti = .false.

        ! temporary
        character(buffsize) :: external_skyfield = ""

    contains
        procedure :: read_params => read_params_tEvtool
        procedure :: print_params => print_params_tEvtool
        procedure :: validate => validate_tEvtool
        procedure :: free => free_tEvtool
        procedure :: init => init_tEvtool
    end type tEvtool

contains

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> init the Evtool type structure with preset of allocated arrays
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine init_tEvtool(this, status, nfile, ninst, nenergy)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in), optional :: nfile, ninst, nenergy
        class(tEvtool) :: this

        character(*), parameter :: function = "init_tEvtool"

        if(status > 100 .or. status == s_errRep) return

        if(present(nfile)) then
            this%nfile = nfile
        else
            this%nfile = 100
        end if

        if(present(ninst)) then
            this%ninst = ninst
        else
            this%ninst = 7
        end if

        if(present(nfile)) then
            this%nenergy = nenergy
        else
            this%nenergy = 9
        end if

        call alloc(this%infile, this%nfile, status)
        call alloc(this%emin, this%nenergy, status)
        call alloc(this%emax, this%nenergy, status)
        call alloc(this%ccdnr, this%ninst, status)
        call alloc(this%resolution, 2, status)
        call alloc(this%center_position, 2, status)
        call alloc(this%rebin, 2, status)

    end subroutine init_tEvtool

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> free all allocated arrays in the Evtool type
    !
    !
    !
    !> @param[in] this
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tEvtool(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tEvtool) :: this

        character(*), parameter :: function = "free_tEvtool"

        if(status > 100 .or. status == s_errRep) return

        call dealloc(this%resolution, status)
        call dealloc(this%rebin, status)
        call dealloc(this%center_position, status)
        call dealloc(this%instrum, status)
        call dealloc(this%ccdnr, status)
        call dealloc(this%emin, status)
        call dealloc(this%emax, status)
        call dealloc(this%infile, status)

        call this%rawxy_sel%mask%free(status)

    end subroutine free_tEvtool

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> read the parameters of the Evtool
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_params_tEvtool(this, status)
        implicit none

        class(tEvtool) :: this
        integer, intent(inout) :: status

        integer :: icount,ierr
        integer :: ninfiles = 100
        character(buffsize) :: filelist_file
        integer, dimension(4) :: tempRawXYSel
        integer :: ii

        character(*), parameter :: function = "read_params_tEvtool"

        if(status > 100 .or. status == s_errRep) return

        call rgeta ("eventfiles", ninfiles, this%infile, icount, ierr)
        if(ierr /= 0) call wrstop(function, "reading infile parameter")
        this%infile = this%infile(:icount)

        ! check whether infile is actually a file containing the infiles
        if(icount == 1) then
            if(this%infile(1)(1:1) == "@") then
                filelist_file = this%infile(1)(2:)
                call read_filelist(filelist_file, this%infile, status)
            end if
        end if

        call rget ("outfile", this%outfile, ierr)
        if(ierr /= 0) call wrstop(function, "reading outfile parameter")

        call rget ("clobber", this%clobber, ierr)
        if(ierr /= 0) call wrstop(function, "reading clobber parameter")

        call rget("events", this%write_events, ierr)
        if(ierr /= 0) call wrstop(function, "reading events parameter")

        call rget("image", this%write_image, ierr)
        if(ierr /= 0) call wrstop(function, "reading image parameter")

        call rgeta("size", 2, this%resolution, icount, ierr)
        if(ierr /= 0) call wrstop(function, "reading image resolution parameter")
        if(icount == 1) this%resolution(2) = this%resolution(1)
        if(icount == 0) this%resolution = this%resolution(:icount)

        call rgeta("center_position", 2, this%center_position, icount, ierr)
        if(ierr /= 0) call wrstop(function, "reading image center_position parameter")
        if(icount == 1) this%center_position(2) = this%center_position(1)
        if(icount == 0) this%center_position = this%center_position(:icount)

        call rgeta("rebin", 2, this%rebin, icount, ierr)
        if(ierr /= 0) call wrstop(function, "reading image rebin parameter")
        if(icount == 1) this%rebin(2) = this%rebin(1)
        if(icount == 0) this%rebin = this%rebin(:icount)

        ! max number of nenergy will be set to 9 due to std_header settings
        ! should be updated if std_header will be updated
        !    call rgeti ("nenergy", this%nenergy, ierr)
        !    if(ierr /= 0) call wrstop(function, "reading nenergy parameter")

        call rgeta ("emin", this%nenergy, this%emin, icount, ierr)
        if(ierr == 1) ierr = 0
        if(ierr /= 0) call wrstop(function, "reading emin parameter")
        this%emin = this%emin(:icount) ! changed conversion to be done only in the events extension conversion
        !    this%emin = this%emin(:icount) * 1000.0_r8 ! keV to eV conversion

        call rgeta ("emax", this%nenergy, this%emax, icount, ierr)
        if(ierr == 1) ierr = 0
        if(ierr /= 0) call wrstop(function, "reading emax parameter")
        this%emax = this%emax(:icount) ! changed conversion to be done only in the events extension conversion
        !    this%emax = this%emax(:icount) * 1000.0_r8 ! keV to eV conversion

        ! update the nenergy keyword
        this%nenergy = icount

        call rgeta("telid", 7, this%ccdnr, icount, ierr)
        if(ierr == 1) ierr = 0
        if(ierr /= 0) call wrstop(function, "reading telid parameter")
        this%ccdnr = this%ccdnr(:icount)

        call rget ("region", this%region, ierr)
        if(ierr == 1) then
            this%region = ""
        else if (ierr /= 0) then
            call wrstop(function, "reading region parameter")
        end if

        call rget ("gti", this%gti_str, ierr)
        if(ierr == 1) then
            this%gti_str = ""
        else if (ierr /= 0) then
            call wrstop(function, "reading gti parameter")
        end if

        call rget ("flag", this%flag, ierr)
        if(ierr == 1) then
            this%flag = ""
        else if (ierr /= 0) then
            call wrstop(function, "reading flag parameter")
        end if

        call rget ("pattern", this%pattern, ierr)
        if(ierr /= 0) call wrstop(function, "reading pattern parameter")

        call rget ("rawxy", this%rawxy_str, ierr)
        if(ierr == 1) then
            this%rawxy_str = ""
        else if(ierr /= 0) then
          call wrstop(function, "reading rawxy parameter")
        end if

        this%rawxy_sel%ccdnr_defined = .true. ! default
        call rget ("rawxy_telid", this%rawxy_sel%ccdnr, ierr)
        if(this%rawxy_sel%ccdnr < 1 .or. this%rawxy_sel%ccdnr > ntel) then
            this%rawxy_sel%ccdnr = 0
            this%rawxy_sel%ccdnr_defined = .false.
        else if (ierr /= 0) then
            call wrstop(function, "reading rawxy_ccdnr parameter")
        end if

        call rget ("rawxy_invert", this%rawxy_sel%invert, ierr)
        if(ierr /= 0) call wrstop(function, "reading rawxy_invert parameter")

        call rget ("memset", this%memset, ierr)
        if(ierr /= 0) call wrstop(function, "reading memset parameter")

        call rget ("overlap", this%overlap, ierr)
        if(ierr /= 0) call wrstop(function, "reading overlap parameter")

        call rget ("skyfield", this%external_skyfield, ierr)
        if(ierr == 1) then
            this%external_skyfield = ""
        else if (ierr /= 0) then
            call wrstop(function, "reading skyfield parameter")
        end if

        ! adjust all read parameters
        this%infile  = adjustl(this%infile)
        this%outfile = adjustl(this%outfile)
        this%region  = adjustl(this%region)
        this%gti_str = adjustl(this%gti_str)
        this%rawxy_str = adjustl(this%rawxy_str)
!*** check other str params.

    end subroutine read_params_tEvtool

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> print the parameters of the Evtool
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine print_params_tEvtool(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tEvtool) :: this
        integer :: ii

        character(*), parameter :: function = "print_params_tEvtool"
        character(buffsize) :: entryStr, ccdnrStr, invertStr

        if(status > 100 .or. status == s_errRep) return

        call wrlog(function, "printing parameters")
        print *, "EVENTFILES: - @FILELISTS are already read in"
        do ii = 1, size(this%infile)
            print *, "*" // trim(this%infile(ii)) // "*"
        end do

        print *, "OUTFILE:"
        print *, "*" // trim(this%outfile) // "*"

        print *, "CLOBBER:"
        print * , this%clobber

        print *, "WRITE EVENTS:"
        print * , this%write_events

        print *, "WRITE IMAGE:"
        print * , this%write_image

        print *, "RESOLUTION"
        print *, this%resolution

        print *, "CENTER POSITION"
        print *, this%center_position

        print *, "REBIN"
        print *, this%rebin

        print *, "EMIN"
        print *, this%emin

        print *, "EMAX"
        print *, this%emax

        print *, "TELID"
        print *, this%ccdnr

        print *, "REGION"
        print *, "*" // trim(this%region) // "*"

        print *, "GTI"
        print *, "*" // trim(this%gti_str) // "*"

        print *, "FLAG"
        print *, "*" // trim(this%flag) // "*"

        print *, "PATTERN"
        print *, this%pattern

        print *, "RAWXY"
        call this%rawxy_sel%get_entry_strs(entryStr, ccdnrStr, invertStr)
        print trim(entryStr)

        print *, "RAWXY_TELID"
        print *, trim(ccdnrStr)

        print *, "RAWXY_INVERT"
        print *, trim(invertStr)

    end subroutine print_params_tEvtool


    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> write the eventfile extensions to the outfile
    !
    !
    !
    !> @param[in] EventFile
    !> @param[out] Evtool
    !> @param[inout] status
    !---------------------------------------------------------------------------

!*** 2019-09-11 IMS: doesn't seem to be used. Superseded by tEventFile%write()?

    subroutine write_eventfile_out(EventFile, Evtool, status)
        implicit none

        integer, intent(inout) :: status
        type(tEventFile), intent(in) :: EventFile
        type(tEvtool), intent(in) :: Evtool

        character(*), parameter :: function = "write_eventfile_out"

        if(status > 100 .or. status == s_errRep) return

        if(Evtool%write_image .eqv. .true.) call EventFile%write_image(status)
        call EventFile%free_image(status)

        if(Evtool%write_events .eqv. .true.) call EventFile%write_events(status)
        call EventFile%free_events(status)

        call EventFile%write_corrat(status)
        call EventFile%free_corrat(status)

        call EventFile%write_deadcor(status)
        call EventFile%free_deadcor(status)

        call EventFile%write_badpix(status)
        call EventFile%free_badpix(status)

        call EventFile%write_hk(status)
        call EventFile%free_hk(status)

        call EventFile%write_offset(status)
        call EventFile%free_offset(status)

        call EventFile%write_gti(status)
        call EventFile%free_gti(status)

        call EventFile%write_qualgti(status)
        call EventFile%free_qualgti(status)

        call EventFile%write_flaregti(status)
        call EventFile%free_flaregti(status)

        !        call EventFile%write_regiongti(status)

        !        if(Evtool%write_external .eqv. .true.) call EventFile%write_externalgti(status)
        !        call EventFile%free_externalgti(status)

        call EventFile%write_region(status)
        !    call EventFile%free_region(status)


    end subroutine write_eventfile_out

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> apply filters of EVTOOL to all extensions if appropriate
    !
    !
    !
    !> @param[in] EventFile
    !> @param[out] Evtool
    !> @param[in, optional] extend_order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine apply_filter_eventfile(EventFile, Evtool, status)
        implicit none

        integer, intent(inout) :: status

        type(tEvtool), intent(in) :: Evtool
        type(tEventFile), intent(inout) :: EventFile

        character(*), parameter :: function = "apply_filter_eventfile"

        if(status > 100 .or. status == s_errRep) return

        call EventFile%sort_events(status)

        if(Evtool%filter_telid == true .and. status == ok) then
            call wrchatty(function, "telid filter")
            call EventFile%apply_filter_telid(status)
        end if

        if(Evtool%filter_energy == true .and. status == ok) then
            call wrchatty(function, "energy filter")
            call EventFile%apply_filter_energy (status)
        end if

        if(Evtool%filter_flag == true .and. status == ok) then
            call wrchatty(function, "flag filter")
            call EventFile%apply_filter_flag (status)
        end if

        if(Evtool%filter_pattern == true .and. status == ok) then
            call wrchatty(function, "pattern filter")
            call EventFile%apply_filter_pattern (status)
        end if

        if(Evtool%filter_region == true .and. status == ok) then
            call wrchatty(function, "region filter")
            EventFile%regtable = Evtool%regtable
            EventFile%ext_regtable = true
            call EventFile%apply_filter_region(Evtool%regtable, status)
            ! if its a "physical filter a region gti cannot be derived -> no conversion from rawx/y to ra/dec
            ! it is enough to check first region, since only one type of wcs is allowed
            if(EventFile%regtable%region(1)%wcs /= "physical") then
                call EventFile%derive_region_gti(Evtool%overlap, status)
                call EventFile%apply_region_gti(status)
            end if
        end if

        if(Evtool%filter_gti == true .and. status == ok) then
            call wrchatty(function, "gti filter")
            if(Evtool%filter_type_gti == gti_type) call EventFile%apply_filter_gti(status, Evtool%overlap)
            if(Evtool%filter_type_gti == qualgti_type) call EventFile%apply_filter_qualgti(status, Evtool%overlap)
            if(Evtool%filter_type_gti == flaregti_type) call EventFile%apply_filter_flaregti(status, Evtool%overlap)
            if(Evtool%filter_type_gti == externalgti_type) call EventFile%apply_filter_externalgti(Evtool%externalgti, status, Evtool%overlap)

            call Evtool%externalgti%free(status)
            ! use new gti tstart and tstop values for the std_header as well
            if(EventFile%ext_gti == true) then
                EventFile%std_header%tstart = EventFile%gti(1)%tstart
                EventFile%std_header%tstop = EventFile%gti(1)%tstop
            else if(EventFile%ext_qualgti == true) then
                EventFile%std_header%tstart = EventFile%qualgti(1)%tstart
                EventFile%std_header%tstop = EventFile%qualgti(1)%tstop
            else if(EventFile%ext_flaregti == true) then
                EventFile%std_header%tstart = EventFile%flaregti(1)%tstart
                EventFile%std_header%tstop = EventFile%flaregti(1)%tstop
            else ! external gti
                EventFile%std_header%tstart = Evtool%externalgti%tstart
                EventFile%std_header%tstop = Evtool%externalgti%tstop
            end if

            call sec2date(EventFile%std_header%tstart, EventFile%std_header%mjdref, EventFile%std_header%date_obs, status)
            call sec2date(EventFile%std_header%tstop, EventFile%std_header%mjdref, EventFile%std_header%date_end, status)
        end if

        if(Evtool%filter_rawxy .and. status == ok) then
            call wrchatty(function, "RAWX/Y filter")
            call EventFile%apply_filter_rawxy(Evtool%rawxy_sel, status)
        end if

        if(Evtool%filter_image == true .and. status == ok) then
            if(EventFile%events%use_wcs==0) then
              call wrerr(function, c_err, "No X/Y WCS was read from the events header, can't make image.")
    return
            end if

            call wrchatty(function, "create image")
            call EventFile%read_image_events(Evtool%resolution, Evtool%center_position, Evtool%rebin, status)
            EventFile%ext_image = true
        end if

    end subroutine apply_filter_eventfile

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> update / merge the EventFile settings with the Evtool settings
    !
    !
    !
    !> @param[inout] Evtool
    !> @param[inout] EventFile
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine update_filter_settings_eventfile(Evtool, EventFile, status)
        implicit none

        integer, intent(inout) :: status

        type(tEventFile), intent(inout) :: EventFile
        type(tEvtool), intent(in) :: Evtool

        character(*), parameter :: function = "update_filter_settings_eventfile"

        if(status > 100 .or. status == s_errRep) return

        call update_filter_telid_settings(Evtool, EventFile, status)
        call update_filter_energy_settings(Evtool, EventFile, status)
        call update_filter_flag_settings(Evtool, EventFile, status)
        call update_filter_pattern_settings(Evtool, EventFile, status)
        call update_filter_rawxy_settings(Evtool, EventFile, status)

        !        call update_filter_region_settings(Evtool, EventFile, status)
        call update_filter_gti_settings(Evtool, EventFile, status)

    end subroutine update_filter_settings_eventfile

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> update the gti settings of the EventFile with Evtool settings
    !
    !
    !
    !> @param[in] Evtool
    !> @param[inout] EventFile
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine update_filter_gti_settings(Evtool, EventFile, status)
        implicit none

        integer, intent(inout) :: status
        type(tEvtool), intent(in) :: Evtool
        type(tEventFile), intent(inout) :: EventFile

        character(*), parameter :: function = "update_filter_gti_settings"

        if(status > 100 .or. status == s_errRep) return

        if(Evtool%filter_gti == true .and. EventFile%filter_gti == false) then
            EventFile%filter_gti = true
            if(Evtool%filter_type_gti == gti_type) EventFile%std_header%gti_sel = gti_ext
            if(Evtool%filter_type_gti == qualgti_type) EventFile%std_header%gti_sel = qualgti_ext
            if(Evtool%filter_type_gti == flaregti_type) EventFile%std_header%gti_sel = flaregti_ext
            if(Evtool%filter_type_gti == externalgti_type) then
              EventFile%std_header%gti_sel = external_gti_ext
            end if
        end if

    end subroutine update_filter_gti_settings

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> update the telid settings of the EventFile with Evtool
    !
    !
    !
    !> @param[in] Evtool
    !> @param[out] EventFile
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine update_filter_telid_settings(Evtool, EventFile, status)
        implicit none

        integer, intent(inout) :: status
        integer, dimension(:), allocatable :: intersect_array, complement_array
        integer :: ii

        type(tEvtool), intent(in) :: Evtool
        type(tEventFile), intent(inout) :: EventFile

        character(*), parameter :: function = "update_filter_telid_settings"

        if(status > 100 .or. status == s_errRep) return

        if(Evtool%filter_telid == true .and. EventFile%filter_telid == false) then
            call wrerr(function, s_err, "EventFile out has no input ccdnr, this should never happen")
            status = s_errRep
            return
        end if

        if(Evtool%filter_telid == true .and. EventFile%filter_telid == true) then

            intersect_array = intersect_int_order(EventFile%ccdnr_infile, Evtool%ccdnr, status)
            complement_array = complement_int_order(Evtool%ccdnr, EventFile%ccdnr_infile, status)
            EventFile%ccdnr_infile = EventFile%ccdnr_infile(intersect_array)

            if(EventFile%ninst < size(Evtool%ccdnr)) then
                do ii = 1, size(complement_array)
                    call wrerr(function, warn, "Selected TELID " // trim(int2str(Evtool%ccdnr(complement_array(ii)), status)) // &
                        " is not present in the infile/s")
                end do
            end if

            EventFile%instrum = EventFile%instrum(intersect_array)
            EventFile%ninst = size(intersect_array)

            !update stdheader
            EventFile%std_header%ninst = EventFile%ninst
            EventFile%std_header%instrumx(:EventFile%ninst) = EventFile%instrum

            if(EventFile%std_header%ninst == 1) then
                EventFile%std_header%instrume = EventFile%std_header%instrumx(1)
            else
                EventFile%std_header%instrume = "merged"
            end if

            call dealloc(intersect_array, status)
            call dealloc(complement_array, status)

        end if


    end subroutine update_filter_telid_settings

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> update the energy filter settings of the EventFile with the choices in Evtool
    !
    !
    !
    !> @param[in] Evtool
    !> @param[out] EventFile
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine update_filter_energy_settings(Evtool, EventFile, status)
        implicit none

        integer, intent(inout) :: status

        type(tEvtool), intent(in) :: Evtool
        type(tEventFile), intent(inout) :: EventFile

        real(r8), dimension(:), allocatable :: emin, emax

        character(*), parameter :: function = "update_filter_energy_settings"

        if(status > 100 .or. status == s_errRep) return

        if(Evtool%filter_energy == true .and. EventFile%filter_energy == false) then
            EventFile%filter_energy = true
            EventFile%std_header%nenergy = Evtool%nenergy
            EventFile%std_header%eminx(:Evtool%nenergy) = Evtool%emin
            EventFile%std_header%emaxx(:Evtool%nenergy) = Evtool%emax
            return
        end if

        if(Evtool%filter_energy == true .and. EventFile%filter_energy == true) then
            call merge_and_r8(Evtool%emin, Evtool%emax, Evtool%nenergy, &
                EventFile%std_header%eminx, EventFile%std_header%emaxx, EventFile%std_header%nenergy, emin, emax, status)
            if(size(emin) > 9) then
                call wrerr(function, s_err, "the number of energy intervals after user applied energy settings exceed the max of 9 energy intervals")
                status = s_errRep
                return
            end if
            EventFile%std_header%nenergy = size(emin)
            EventFile%std_header%eminx(:size(emin)) = emin
            EventFile%std_header%emaxx(:size(emin)) = emax
            call dealloc(emin, status)
            call dealloc(emax, status)
        end if

    end subroutine update_filter_energy_settings

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> update the flag filter settings of the EventFile with the choices in Evtool
    !
    !
    !
    !> @param[in] Evtool
    !> @param[out] EventFile
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine update_filter_flag_settings(Evtool, EventFile, status)
        implicit none

        integer, intent(inout) :: status
        integer :: flag

        type(tEvtool), intent(in) :: Evtool
        type(tEventFile), intent(inout) :: EventFile

        character(*), parameter :: function = "update_filter_flag_settings"

        if(status > 100 .or. status == s_errRep) return

        if(Evtool%filter_flag == true .and. EventFile%filter_flag == false) then
            EventFile%filter_flag = true
            EventFile%std_header%flagsel = Evtool%flag_hex
            return
        end if

        if(Evtool%filter_flag == true .and. EventFile%filter_flag == true) then
            flag = ior(EventFile%std_header%flagsel, Evtool%flag_hex)
            EventFile%std_header%flagsel = flag
        end if

    end subroutine update_filter_flag_settings



    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> update the pattern filter settings of the EventFile with the choices in Evtool
    !
    !
    !
    !> @param[in] Evtool
    !> @param[out] EventFile
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine update_filter_pattern_settings(Evtool, EventFile, status)
        implicit none

        integer, intent(inout) :: status
        integer :: pattern

        type(tEvtool), intent(in) :: Evtool
        type(tEventFile), intent(inout) :: EventFile

        character(*), parameter :: function = "update_filter_pattern_settings"

        if(status > 100 .or. status == s_errRep) return

        if(Evtool%filter_pattern == true .and. EventFile%filter_pattern == false) then
            EventFile%filter_pattern = true
            EventFile%std_header%pat_sel = Evtool%pattern
            return
        end if

        if(Evtool%filter_pattern == true .and. EventFile%filter_pattern == true) then
            pattern = iand(EventFile%std_header%pat_sel, Evtool%pattern)
            EventFile%std_header%pat_sel = pattern
        end if

    end subroutine update_filter_pattern_settings

    !---------------------------------------------------------------------------
    !> @author
    !> Ian M Stewart
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> update the rawxy filter settings of the EventFile with the choices in Evtool
    !
    !
    !
    !> @param[in] Evtool
    !> @param[out] EventFile
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine update_filter_rawxy_settings(Evtool, EventFile, status)
        implicit none

        integer, intent(inout) :: status
        integer :: pattern

        type(tEvtool), intent(in) :: Evtool
        type(tEventFile), intent(inout) :: EventFile

        character(*), parameter :: function = "update_filter_rawxy_settings"

        if(status > 100 .or. status == s_errRep) return

        if(Evtool%filter_rawxy .and. .not. EventFile%filter_rawxy) then
          EventFile%filter_rawxy = .true.
          call EventFile%record_rawxy_sel(Evtool%rawxy_sel, status, .true.)
          return
        end if

        if(Evtool%filter_rawxy .and. EventFile%filter_rawxy) then
          call EventFile%record_rawxy_sel(Evtool%rawxy_sel, status)
        end if

    end subroutine update_filter_rawxy_settings

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> validate the input parameters of Evtool
    !
    !
    !
    !> @param[in] this
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine validate_tEvtool(this, status)
        implicit none

        class(tEvtool) :: this
        integer, intent(inout) :: status

        integer :: exists, ii, tmpunit

        character(*), parameter :: function = "validate_tEvtool"

        if(status > 100 .or. status == s_errRep) return

        do ii = 1, size(this%infile)
            if(trim(this%infile(ii)) == "") call wrerr(function, s_err, "one of the infiles is blank")

            call ffits_file_exists(this%infile(ii), exists, status)
            if (exists == 0) then
                call wrerr(function, s_err, "infile " // trim(this%infile(ii)) // " does not exist")
                status = s_errRep
    return
            end if
        end do

        if(trim(this%outfile) /= "") then
            this%delete_outfile = false
            call ffits_file_exists(this%outfile, exists, status)
            if (exists /= 0) then
                if(this%clobber .eqv. .false.) then
                    call wrerr(function, s_err, "outfile " // trim(this%outfile) // " already exists")
                    status = s_errRep
    return
                else
                    this%delete_outfile = true
                end if
            end if
        else
            call wrerr(function, s_err, "outfile is blank")
        end if

        this%filter_image = false
        if(this%write_image .eqv. .true.) then

            if(size(this%resolution) /= 2 .or. size(this%center_position) /= 2 .or. size(this%rebin) /= 2) then
                call wrerr(function, s_err, "insufficient information about output image (resolution, center_position, rebin)")
                status = s_errRep
    return
            end if

            if(this%resolution(1) < 1 .or. this%resolution(2) < 1) then
                call wrerr(function, s_err, "zero or negative resolution selected")
                status = s_errRep
    return
            end if

            if(this%rebin(1) < 1 .or. this%rebin(2) < 1) then
                call wrerr(function, s_err, "zero or negative rebin selected")
                status = s_errRep
    return
            end if

            this%filter_image = true
        end if

        if(this%nenergy < 0 .and. this%nenergy > 9) then
            call wrerr(function, s_err, "max number of energy intervals cannot be less than zero or more than 9")
            status = s_errRep
    return
        end if

        if(size(this%emin) /= size(this%emax)) then
            call wrerr(function, s_err, "number of given energy boundaries (emin / emax) is inconsistent")
            status = s_errRep
        end if

        if(size(this%emin) > this%nenergy .or. size(this%emax) > this%nenergy) then
            call wrerr(function, s_err, "max number of energy intervals is lower than the max number of energy bands provided")
            status = s_errRep
        end if

        call validateEnergyInt(this%emin, this%emax, status)

        this%filter_energy = false
        if(size(this%emin) > 0) this%filter_energy = true
!*** 20190915 IMS: seems never to be preset to false, so I added this just above as the default.

        ! FIX for HB evtool call

        if(size(this%ccdnr) > 0) then
            this%filter_telid = true

            if(minval(this%ccdnr) < 1) then
                call wrerr(function, s_err, "telescope id is illegal, cannot be less than 1")
                status = s_errRep
            end if

            if(maxval(this%ccdnr) > 7) then
                call wrerr(function, s_err, "telescope id is illegal, cannot be greater than 7")
                status = s_errRep
            end if

        else if (size(this%ccdnr) > 7) then
            call wrerr(function, s_err, "more telescope ids selected than max number of instruments (7)")
            status = s_errRep
        else
            this%filter_telid = false
        end if

        if(trim(this%region) /= "") then
            call getRegionTable(this%region, this%regtable, status)
            if(status == ok) this%filter_region = true
        else
            call alloc_reg(this%regtable%Region, 0, status)
        end if

        this%filter_gti = false ! default
        if(trim(this%gti_str) /= "") then
            ! determine the input format of the gti
            ! type is per telescope
            if(this%gti_str == gti_ext) then
                this%filter_type_gti = gti_type
            else if(this%gti_str == qualgti_ext) then
                this%filter_type_gti = qualgti_type
            ! type is global
            else if(this%gti_str == flaregti_ext) then
                this%filter_type_gti = flaregti_type
            else
                this%filter_type_gti = externalgti_type
                ! assume type is a FITS file
                call ffits_file_exists(this%gti_str, exists, status)
                if (exists == 1) then
                    call ffits_get_iounit(tmpunit, status)
                    call ffits_open_file(tmpunit, this%gti_str, READONLY, status)
                    call this%externalgti%read(tmpunit, external_gti_ext, exists, status)
                    if(exists == false) then
                        call wrerr(function, s_err, external_gti_ext // " in file " // trim(this%gti_str) // " could not be read")
                        status = s_errRep
    return
                    end if
                    call ffits_close_file(tmpunit, status)
                    call ffits_free_iounit(tmpunit, status)
                else
                    ! interprete string as two values indicating tstart and tstop
                    call interprete_str_gti(this%gti_str, this%externalgti, status)

                end if

            end if
            if(status == ok) this%filter_gti = true
        end if

        if(trim(this%flag) /= "") then
          this%flag_hex = str2hex(this%flag, status)
          if(status == ok) this%filter_flag = true
        end if

        ! single = 1, double = 2, triple = 4, quadruple = 8
        this%filter_pattern = false ! default
        if(this%pattern > 0 .and. this%pattern < 16) then
          this%filter_pattern = true
        end if

        this%filter_rawxy = .false. ! default
        if(trim(this%rawxy_str) /= "") then
          ! Determine which of the two acceptable strings it is - either (mask) file name or 'xLo xHi yLo yHi'.
          !
          call ffits_file_exists(this%rawxy_str, exists, status)
          if (exists == 1) then
            this%rawxy_sel%mask%filename = this%rawxy_str
            call this%rawxy_sel%mask%open(readonly, status)
            call this%rawxy_sel%mask%read(status)
            call this%rawxy_sel%mask%close(status)
            this%rawxy_sel%mask%containsData = .true.
          else
            call extractRawXYBounds(this%rawxy_str, this%rawxy_sel, status)
          end if

          if(status == ok) this%filter_rawxy = .true.
        end if

        if(this%rawxy_sel%ccdnr_defined) then
          if(this%rawxy_sel%ccdnr < 1 .or. this%rawxy_sel%ccdnr > ntel) then
            call wrerr(function, s_err, "RAWXY ccd number is out of range.")
            status = s_errRep
    return
          end if
        end if

    end subroutine validate_tEvtool

    !---------------------------------------------------------------------------
    !> @author
    !> Ian M Stewart
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> blah
    !
    !
    !
    !> @param[in] emin
    !> @param[out] emax
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine extractRawXYBounds(rawxy_str, rawxy_sel, status)
        implicit none

        character(*),             intent(in)    :: rawxy_str
        type(tRawXYSelection),    intent(inout) :: rawxy_sel
        integer,                  intent(inout) :: status

        integer :: ii
        character(buffsize), allocatable, dimension(:) :: listOfStrs
        integer :: icount,readstat

        character(*), parameter :: function = "extractRawXYBounds"

        if(status > 100 .or. status == s_errRep) return

        ! Divide the rawxy string into its space-separated substrings:
        !
        call strtok(rawxy_str, ' ', listOfStrs)
        icount = size(listOfStrs)

        if(icount == 4) then
          ! Read the 4 values into the relevant struct elements:
          !
          ii = 1
          read(listOfStrs(ii),*,iostat=readstat) rawxy_sel%axes(1)%lo%value
          if (readstat>0) then
            call wrerr(function, s_err, "Cannot read " // trim(listOfStrs(ii)))
          status = s_errRep
    return
          end if

          ii = 2
          read(listOfStrs(ii),*,iostat=readstat) rawxy_sel%axes(1)%hi%value
          if (readstat>0) then
            call wrerr(function, s_err, "Cannot read " // trim(listOfStrs(ii)))
          status = s_errRep
    return
          end if

          ii = 3
          read(listOfStrs(ii),*,iostat=readstat) rawxy_sel%axes(2)%lo%value
          if (readstat>0) then
            call wrerr(function, s_err, "Cannot read " // trim(listOfStrs(ii)))
          status = s_errRep
    return
          end if

          ii = 4
          read(listOfStrs(ii),*,iostat=readstat) rawxy_sel%axes(2)%hi%value
          if (readstat>0) then
            call wrerr(function, s_err, "Cannot read " // trim(listOfStrs(ii)))
          status = s_errRep
    return
          end if

          do ii = 1, 2
            if(rawxy_sel%axes(ii)%lo%value < 1)&
              rawxy_sel%axes(ii)%lo%value = 1
            if(rawxy_sel%axes(ii)%hi%value > n_ccd_pixels)&
              rawxy_sel%axes(ii)%hi%value = n_ccd_pixels

            rawxy_sel%axes(ii)%lo%is_defined = .true.
            rawxy_sel%axes(ii)%hi%is_defined = .true.
          end do

          ! Fill the mask.
          !
          allocate(rawxy_sel%mask%values(n_ccd_pixels,n_ccd_pixels))
          rawxy_sel%mask%values = .false.
          rawxy_sel%mask%values(rawxy_sel%axes(1)%lo%value:rawxy_sel%axes(1)%hi%value&
                           ,rawxy_sel%axes(2)%lo%value:rawxy_sel%axes(2)%hi%value) = .true.

        else
            call wrerr(function, warn, "Invalid rawxy argument "//trim(rawxy_str)//". It is neither the name of an existing file or a list of 4 substrings.")
            do ii = 1, 2
              rawxy_sel%axes(ii)%lo%is_defined = .false.
              rawxy_sel%axes(ii)%lo%value = 0.0
              rawxy_sel%axes(ii)%hi%is_defined = .false.
              rawxy_sel%axes(ii)%hi%value = 0.0
            end do
        end if

        call dealloc(listOfStrs, status)

        do ii=1, 2
          if(rawxy_sel%axes(ii)%lo%is_defined .and. rawxy_sel%axes(ii)%hi%is_defined) then
            if(rawxy_sel%axes(ii)%lo%value > rawxy_sel%axes(ii)%hi%value) then
              call wrerr(function, s_err, "RAWXY filter: lo may not be greater than hi.")
              status = s_errRep
    return
            end if
          end if
        end do

    end subroutine extractRawXYBounds

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> validate energy intervalls
    !
    !
    !
    !> @param[in] emin
    !> @param[out] emax
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine validateEnergyInt(emin, emax, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii

        real(dp), dimension(:), intent(in) :: emin, emax

        character(*), parameter :: function = "validateEnergyInt"

        if(status > 100 .or. status == s_errRep) return

        if(size(emin) /= size(emax)) then
            call wrerr(function, s_err, "number of min/max energy entries is different")
            status = s_errRep
            return
        end if

        do ii = 1, size(emin)
            if(emin(ii) > emax(ii)) then
                call wrerr(function, s_err, "selected min energy > max energy")
                status = s_errRep
                return
            end if
            if(emin(ii) < 0.0_dp .or. emax(ii) < 0.0_dp) then
                call wrerr(function, s_err, "selected min and/or max energy is smaller than 0 eV")
                status = s_errRep
                return
            end if
        end do

    end subroutine validateEnergyInt


    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> interprete string as gti
    !
    !
    !
    !> @param[in] file
    !> @param[out]
    !> @param[inout] status, filelist
    !---------------------------------------------------------------------------

    subroutine interprete_str_gti (str, gti, status)
        implicit none

        integer, intent(inout) :: status
        character(*), intent(in) :: str
        type(tGTI), intent(out) :: gti

        character(buffsize), dimension(:), allocatable :: gti_str

        character(*), parameter :: function = "interprete_str_gti"

        if(status > 100 .or. status == s_errRep) return

        call strtok(str, " ", gti_str)
        if(size(gti_str) /= 2) then
            call wrerr(function, s_err, "could not interprete " // trim(str) // " as external gti")
            status = s_errRep
            return
        end if

        call gti%init(1, status)
        gti%start(1) = str2double(gti_str(1), status)
        gti%stop(1) = str2double(gti_str(2), status)
        gti%tstart = gti%start(1)
        gti%tstop = gti%stop(1)
        ! when set to negative mjdref, the mjdref value is ignored when gti is applied
        gti%use_mjdref = false
        gti%use_timezero = false

    end subroutine interprete_str_gti


end module evtool_mod
