! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module evtool_offset_mod

    use erolib_str_mod
    use erolib_ffits_mod
    use erolib_general_mod
    use erolib_sort_mod
    use erolib_tbintable_mod
    use erolib_math_mod
    use evtool_gti_mod
    implicit none


    private
    character(*), parameter :: offset_ext = "OFFSET"

    character(*), parameter :: time_col = "TIME"
    character(*), parameter :: rawx_col = "RAWX"
    character(*), parameter :: offsetx_col = "OFFSETX"

    type tOffset
        real(r8) :: mjdref = 0.0
        character(buffsize) :: extname = ""
        integer :: nelements = 0
        integer :: ccdnr = 0
        real (r8), allocatable, dimension(:) :: time
        integer (i2), allocatable, dimension(:) :: rawx, offsetx
    contains
        procedure, non_overridable :: read => read_tOffset
        procedure, non_overridable :: write => write_tOffset
        procedure, non_overridable :: create_extension => create_extension_tOffset
        procedure, non_overridable :: sort => sort_tOffset
        procedure, non_overridable :: free => free_tOffset
        procedure, non_overridable :: init => init_tOffset
        procedure, non_overridable :: delete_duplicates => delete_duplicates_tOffset
        procedure, non_overridable :: apply_index_cols => apply_index_cols_tOffset
        procedure, non_overridable :: apply_gti => apply_gti_tOffset
    end type tOffset

    public :: tOffset, offset_ext, alloc_offset, dealloc_offset, merge_offset

contains

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> dealloacte offset structures
    !
    !
    !
    !> @param[inout] offset
    !> @param[inout] status
    !---------------------------------------------------------------------------
    subroutine dealloc_offset(offset, status)

        integer :: stat
        integer, intent(inout) :: status

        type(tOffset), allocatable, dimension(:) :: offset

        character(*), parameter :: function = "dealloc_offset"

        if(allocated(offset)) then
            deallocate(offset, stat=stat)
            call FError(function, stat, "dealloc offset type", status)
        end if

    end subroutine dealloc_offset

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> alloacte n offset structures
    !
    !
    !
    !> @param[inout] offset
    !> @param[in] n
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine alloc_offset(offset, n, status)
        implicit none

        integer :: stat
        integer, intent(in) :: n
        integer, intent(inout) :: status

        type(tOffset), allocatable, dimension(:) :: offset

        character(buffsize) :: errmsg
        character(*), parameter :: function = "alloc_offset"

        call dealloc_offset(offset, status)
        allocate(offset(n), stat=stat, errmsg=errmsg)
        call FError(function, stat, errmsg, status)

    end subroutine alloc_offset

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> create the merged offset extensions for all infiles
    !
    !
    !
    !> @param[in] EvtenFile_out
    !> @param[out] files
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine merge_offset(offset_in, offset_out, status)
        implicit none

        integer, intent(inout) :: status
        integer(i4) :: ii, nelements, a, b

        type(tOffset), intent(out) :: offset_out
        !        type(tOffset), intent(inout), dimension(:) :: offset_in
        type(tOffset), intent(in), dimension(:) :: offset_in

        character(*), parameter :: function = "merge_offset"

        if(status > 100 .or. status == s_errRep) return

        nelements = 0
        do ii = 1, size(offset_in)
            nelements = nelements + offset_in(ii)%nelements
        end do

        offset_out%nelements = nelements
        offset_out%ccdnr = offset_in(1)%ccdnr
        offset_out%extname = offset_in(1)%extname
        offset_out%mjdref = offset_in(1)%mjdref
        ! reallocate memory for columns
        call offset_out%init(nelements, status)

        if(nelements == 0) return

        ! merge all columns
        a = 1
        do ii = 1, size(offset_in)
            b = a + offset_in(ii)%nelements - 1
            offset_out%time(a:b) = offset_in(ii)%time
            offset_out%rawx(a:b) = offset_in(ii)%rawx
            offset_out%offsetx(a:b) = offset_in(ii)%offsetx
            if(status == s_errRep) return
            a = b + 1
        end do

    end subroutine merge_offset


    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> sort the type entries according to one array
    !
    !
    !
    !> @param[in] this
    !> @param[in] array
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine sort_tOffset(this, array, status, ascend)
        implicit none

        integer, intent(inout) :: status
        integer, allocatable, dimension(:) :: order
        integer, optional :: ascend

        class(tOffset) :: this


        class(*), intent(in), dimension(:) :: array
        integer(i2), allocatable, dimension(:) :: tmparray_i2
        real(r8), allocatable, dimension(:) :: tmparray_r8

        character(*), parameter :: function = "sort_tOffset"

        if(status > 100 .or. status == s_errRep) return

        if(size(this%time) /= size(array)) then
            call wrerr(function // " tOffset type", s_err, "length of sorting array does not conform to length of type")
            status = s_errRep
        end if

        select type (array)
            type is (integer(i2))
            call alloc(tmparray_i2, size(array), status, array)
            call sort(tmparray_i2, status, order)
            if(present(ascend)) then
                if(ascend == false) call reverse_array(order)
            end if

            type is (real(r8))
            call alloc(tmparray_r8, size(array), status, array)
            call sort(tmparray_r8, status, order)
            if(present(ascend)) then
                if(ascend == false) call reverse_array(order)
            end if

        end select

        this%time = this%time(order)
        this%rawx = this%rawx(order)
        this%offsetx = this%offsetx(order)

        call dealloc(order, status)
        call dealloc(tmparray_i2, status)
        call dealloc(tmparray_r8, status)

    end subroutine sort_tOffset

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> read the values from the offset extension
    !
    !
    !
    !> @param[in] this (tOffset)
    !> @param[in] unit, ccdnr
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_tOffset(this, unit, ccdnr, success, status, mjdref)
        implicit none

        class(tOffset) :: this

        integer :: nelements, rowsize, frow, nrows, lrow, felem
        integer :: time_colnum, rawx_colnum, offsetx_colnum
        integer, intent(in) :: ccdnr
        integer, intent(inout) :: unit, status
        integer, intent(out) :: success

        real(r8), optional, intent(in) :: mjdref

        character(buffsize)  :: telid_str
        character(*), parameter :: function = "read_tOffset"

        logical :: anyf

        if(status > 100 .or. status == s_errRep) return

        ! convert int to string
        telid_str = int2str(ccdnr, status)
        if(status == s_errRep) return

        success = false
        status = ignore
        call ffits_movnam_hdu(unit, BINARY_TBL, offset_ext // trim(telid_str), status)
        if(status == ignore) then
            status = ok
        else
            status = ok
            return
        end if

        call ffits_get_num_rows(unit, nelements, status)
        call ffits_get_rowsize(unit, rowsize, status)

        call this%init(nelements, status)

        this%ccdnr = ccdnr
        this%nelements = nelements
        this%extname = offset_ext // trim(telid_str)

        ! if no mjdref in the extension is found take the optional mjdref
        status = ignore
        call ffits_read_key(unit, "MJDREF", this%mjdref, status)
        if(status == ignore) then
            status = ok
        else
            status = ok
            if(present(mjdref)) this%mjdref = mjdref
        end if

        call ffits_get_colnum(unit, .false., time_col, time_colnum, status)
        call ffits_get_colnum(unit, .false., rawx_col, rawx_colnum, status)
        call ffits_get_colnum(unit, .false., offsetx_col, offsetx_colnum, status)

        anyf = .false.
        felem = 1

        if(nelements == 0 .or. status == s_errRep) then
            if(status == ok) success = true
            return
        end if


        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1
            call ffits_read_col(unit, time_colnum, frow, felem, nrows, 0, this%time(frow:lrow), status, anyf)
            call ffits_read_col(unit, rawx_colnum, frow, felem, nrows, 0, this%rawx(frow:lrow), status, anyf)
            call ffits_read_col(unit, offsetx_colnum, frow, felem, nrows, 0, this%offsetx(frow:lrow), status, anyf)
        end do

        if(status == ok) success = true

    end subroutine read_tOffset

            !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> write in the columns of an offset extension
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_tOffset(this, unit, status)
        implicit none

        integer :: nelements, rowsize, frow, nrows, lrow, felem
        integer :: time_colnum, rawx_colnum, offsetx_colnum
        integer, intent(inout) :: status
        integer, intent(in) :: unit

        class(tOffset) :: this

        character(*), parameter :: function = "write_tOffset"

        if(status > 100 .or. status == s_errRep) return

        ! move to the correct extension
        status = ignore
        call ffits_movnam_hdu(unit, BINARY_TBL, this%extname, status)
        if(status /= ignore) then
            status = ok
            call this%create_extension(unit, status)
        else
            status = ok
        end if

        call ffits_get_rowsize(unit, rowsize, status)

        nelements = this%nelements

        call ffits_get_colnum(unit, .false., time_col, time_colnum, status)
        call ffits_get_colnum(unit, .false., rawx_col, rawx_colnum, status)
        call ffits_get_colnum(unit, .false., offsetx_col, offsetx_colnum, status)

        felem = 1

        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1
            call ffits_write_col(unit, time_colnum, frow, felem, nrows, this%time(frow:lrow), status)
            call ffits_write_col(unit, rawx_colnum, frow, felem, nrows, this%rawx(frow:lrow), status)
            call ffits_write_col(unit, offsetx_colnum, frow, felem, nrows, this%offsetx(frow:lrow), status)
        end do

        call ffits_update_key(unit, "MJDREF", this%mjdref, "[d] updated by evtool", status)

    end subroutine write_tOffset

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> create a basic offset extension in the file
    !
    !
    !
    !> @param[in] this
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine create_extension_tOffset(this, unit, status)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) :: unit

        character(*), parameter :: function = "create_extension_tOffset"
        character(fcolnamelength), dimension(3) :: ttype
        character(fcolnamelength), dimension(3) :: tform
        character(fcolnamelength), dimension(3) :: tunit

        class(tOffset) :: this
        type(tBinTable) :: table

        if(status > 100 .or. status == s_errRep) return

        ttype(1) = time_col
        ttype(2) = rawx_col
        ttype(3) = offsetx_col

        tform = ["D","I","I"]
        tunit(:) = ""

        call table%init(ttype, tform, tunit, status)
        call table%insertFits(unit, this%extname, status);

    end subroutine create_extension_tOffset

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> reduce all entries in the columns to contain only a subset of the complete
    !> by providing an array list with the desired indices
    !
    !
    !> @param[in] this
    !> @param[in] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine apply_index_cols_tOffset(this, order, status)
        implicit none

        integer, intent(inout) :: status

        class(tOffset) :: this

        integer, dimension(:), intent(in) :: order
        integer :: n

        character(*), parameter :: function = "apply_index_cols_tOffset"

        if(status > 100 .or. status == s_errRep) return

        n = size(order)
        if(n == 0) then
!            call wrerr(function, warn, trim(this%extname) // " was reduced to zero entries")
            call wrchatty(function, trim(this%extname) // " was reduced to zero entries")
            call alloc(this%time, n, status)
            call alloc(this%rawx, n, status)
            call alloc(this%offsetx, n, status)
            return
        end if

        this%time = this%time(order)
        this%rawx = this%rawx(order)
        this%offsetx = this%offsetx(order)
        this%nelements = n

    end subroutine apply_index_cols_tOffset

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> apply a GTI filter to the offset type
    !
    !
    !> @param[in] this, GTI
    !> @param[in, optional] extend_order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine apply_gti_tOffset(this, GTI, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii

        class(tOffset) :: this
        type(tGTI), intent(in) :: GTI
        type(tGTI) :: iGTI

        integer, dimension(this%nelements) :: mask
        integer, dimension(:), allocatable :: order

        character(*), parameter :: function = "apply_gti_tOffset"

        if(status > 100 .or. status == s_errRep) return

        call alloc(order, this%nelements, status)
        order = [(ii,ii=1,this%nelements,1)]

        iGTI = GTI
        ! internally correct the GTI
        if(iGTI%use_mjdref == true) then

            if(this%mjdref /= iGTI%mjdref) then

                iGTI%start = (iGTI%mjdref - this%mjdref ) * 3600.0_r8 * 24.0_r8 + iGTI%start
                iGTI%stop = ( iGTI%mjdref - this%mjdref ) * 3600.0_r8 * 24.0_r8 + iGTI%stop
                iGTI%mjdref = this%mjdref
            end if

        end if

        mask = get_gti_mask(this%time, iGTI, status)

        order = pack(order, mask == true)

        call this%apply_index_cols(order, status)
        call dealloc(order, status)
        call iGTI%free(status)


    end subroutine apply_gti_tOffset

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> delete duplicate entries in the tOffset type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine delete_duplicates_tOffset(this, status)
        implicit none

        class(tOffset) :: this

        integer, intent(inout) :: status
        integer, dimension(:), allocatable :: order
        integer, dimension(this%nelements) :: mask
        integer :: ii

        character(*), parameter :: function = "delete_duplicates_tOffset"

        if(status > 100 .or. status == s_errRep) return

        call this%sort(this%offsetx, status)
        call this%sort(this%rawx, status)
        call this%sort(this%time, status)

        if(this%nelements < 2) return

        call alloc(order, this%nelements, status)
        order = [(ii,ii=1,this%nelements,1)]

        mask = true
        do ii = 2, this%nelements
            if(abs(this%time(ii) - this%time(ii-1)) < tiny(this%time(1))) then
                if(this%rawx(ii) == this%rawx(ii-1)) then
                    if(this%offsetx(ii) == this%offsetx(ii-1)) then
                        mask(ii) = false
                    end if
                end if
            end if
        end do

        order = pack(order, mask == true)
        call this%apply_index_cols(order, status)
        call dealloc(order, status)

    end subroutine delete_duplicates_tOffset

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> free the content of tOffset type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tOffset(this, status)
        implicit none

        integer, intent(inout) :: status

        class(tOffset) :: this

        character(*), parameter :: function = "free_tOffset"

        call dealloc(this%time, status)
        call dealloc(this%rawx, status)
        call dealloc(this%offsetx, status)

    end subroutine free_tOffset

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> alloc the content of tOffset type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine init_tOffset(this, n, status)
        implicit none

        integer, intent(in) :: n
        integer, intent(inout) :: status

        class(tOffset) :: this

        character(*), parameter :: function = "init_tOffset"

        call alloc(this%time, n, status)
        call alloc(this%rawx, n, status)
        call alloc(this%offsetx, n, status)

        this%nelements = n
        this%ccdnr = -1
        this%extname = ""
        this%mjdref = 0_r8

    end subroutine init_tOffset

end module evtool_offset_mod
