! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module evtool_hkx1_mod

  use erolib_str_mod
  use erolib_ffits_mod
  use erolib_general_mod
  use erolib_sort_mod
  use evtool_gti_mod
  use erolib_tbintable_mod
  use erolib_math_mod
  implicit none

  private

  character(*), parameter :: hkx1_ext = "HKx1"
  character(*), parameter :: time_col = "TIME"
  character(*), parameter :: pxprcmode_col = "PxPROCMODE"
  character(*), parameter :: pxprocopts_col = "PxPROCOPTS"
  character(*), parameter :: pxtmframes_col = "PxTMFRAMES"
  character(*), parameter :: pxccdfrms_col = "PxCCDFRMS"
  character(*), parameter :: pxevtctr_col = "PxEVTCTR"
  character(*), parameter :: pxmipctr_col = "PxMIPCTR"
  character(*), parameter :: pxfwstepct_col = "PxFWSTEPCT"
  character(*), parameter :: pxfwstat_col = "PxFWSTAT"
  character(*), parameter :: pxchopper_col = "PxCHOPPER"

  type tHKx1
    real(r8) :: mjdref = 0.0
    character(buffsize) :: extname = hkx1_ext
    integer :: nelements = 0
    integer :: ccdnr = 0
    real (r8), allocatable, dimension(:) :: time

    integer, allocatable, dimension(:) :: pxprcmode
    integer, allocatable, dimension(:) :: pxprocopts
    integer, allocatable, dimension(:) :: pxtmframes
    integer, allocatable, dimension(:) :: pxccdfrms
    integer, allocatable, dimension(:) :: pxevtctr
    integer, allocatable, dimension(:) :: pxmipctr
    integer, allocatable, dimension(:) :: pxfwstepct
    integer, allocatable, dimension(:) :: pxfwstat
    integer, allocatable, dimension(:) :: pxchopper

  contains
    procedure, non_overridable :: read => read_tHKx1
    procedure, non_overridable :: write => write_tHKx1
    procedure, non_overridable :: create_extension => create_extension_tHKx1
    procedure, non_overridable :: sort => sort_tHKx1
    procedure, non_overridable :: free => free_tHKx1
    procedure, non_overridable :: init => init_tHKx1
    procedure, non_overridable :: delete_duplicates => delete_duplicates_tHKx1
    procedure, non_overridable :: apply_index_cols => apply_index_cols_tHKx1
    procedure, non_overridable :: apply_gti => apply_gti_tHKx1
  end type tHKx1

  public :: tHKx1, hkx1_ext, alloc_hkx1, dealloc_hkx1, merge_hkx1

contains

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> dealloacte hkx1 structures
  !
  !
  !
  !> @param[inout] hkx1
  !> @param[inout] status
  !---------------------------------------------------------------------------
  subroutine dealloc_hkx1(hkx1, status)

    integer :: stat
    integer, intent(inout) :: status

    type(tHKx1), allocatable, dimension(:) :: hkx1

    character(*), parameter :: function = "dealloc_hkx1"

    if(allocated(hkx1)) then
      deallocate(hkx1, stat=stat)
      call FError(function, stat, "dealloc hkx1 type", status)
    end if

  end subroutine dealloc_hkx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloacte n hkx1 structures
  !
  !
  !
  !> @param[inout] hkx1
  !> @param[in] n
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine alloc_hkx1(hkx1, n, status)
    implicit none

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    type(tHKx1), allocatable, dimension(:) :: hkx1

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_hkx1"

    call dealloc_hkx1(hkx1, status)
    allocate(hkx1(n), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

  end subroutine alloc_hkx1

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create the merged hkx1 extensions for all infiles
  !
  !
  !
  !> @param[in] EvtenFile_out
  !> @param[out] files
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine merge_hkx1(hkx1_in, hkx1_out, status)
    implicit none

    integer, intent(inout) :: status
    integer(i4) :: ii, nelements, a, b

    type(tHKx1), intent(out) :: hkx1_out
    type(tHKx1), intent(in), dimension(:) :: hkx1_in

    character(*), parameter :: function = "merge_hkx1"

    if(status > 100 .or. status == s_errRep) return

    nelements = 0
    do ii = 1, size(hkx1_in)
      nelements = nelements + hkx1_in(ii)%nelements
    end do

    ! reallocate memory for columns
    call hkx1_out%init(nelements, status)

    hkx1_out%nelements = nelements
    hkx1_out%ccdnr = hkx1_in(1)%ccdnr
    hkx1_out%extname = hkx1_in(1)%extname
    hkx1_out%mjdref = hkx1_in(1)%mjdref

    if(nelements == 0) return

    ! merge all columns
    a = 1
    do ii = 1, size(hkx1_in)
      b = a + hkx1_in(ii)%nelements - 1
      hkx1_out%time(a:b) = hkx1_in(ii)%time
      if(status == s_errRep) return
      a = b + 1
    end do

  end subroutine merge_hkx1


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> sort the type entries according to one array
  !
  !
  !
  !> @param[in] this
  !> @param[in] array
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine sort_tHKx1(this, array, status, ascend)
    implicit none

    integer, intent(inout) :: status
    integer, allocatable, dimension(:) :: order
    integer, optional :: ascend

    class(tHKx1) :: this

    class(*), intent(in), dimension(:) :: array
    real(r8), allocatable, dimension(:) :: tmparray_r8

    character(*), parameter :: function = "sort_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    if(size(this%time) /= size(array)) then
      call wrerr(function // " tHKx1 type", s_err, "length of sorting array does not conform to length of type")
      status = s_errRep
    end if

    select type (array)
      type is (real(r8))
      call alloc(tmparray_r8, size(array), status, array)
      call sort(tmparray_r8, order, status)
      if(present(ascend)) then
        if(ascend == false) call reverse_array(order)
      end if

    end select

    this%time = this%time(order)

    call dealloc(order, status)
    call dealloc(tmparray_r8, status)

  end subroutine sort_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the values from the hkx1 extension
  !
  !
  !
  !> @param[in] this (tHKx1)
  !> @param[in] unit, ccdnr
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_tHKx1(this, unit, ccdnr, success, status, mjdref)
    implicit none

    class(tHKx1) :: this

    integer :: nelements, rowsize, frow, nrows, lrow, felem
    integer :: time_colnum
    integer :: pxprcmode_colnum
    integer :: pxprocopts_colnum
    integer :: pxtmframes_colnum
    integer :: pxccdfrms_colnum
    integer :: pxevtctr_colnum
    integer :: pxmipctr_colnum
    integer :: pxfwstepct_colnum
    integer :: pxfwstat_colnum
    integer :: pxchopper_colnum

    integer, intent(in) :: ccdnr
    integer, intent(inout) :: unit, status
    integer, intent(out) :: success
    real(r8), optional, intent(in) :: mjdref

    character(buffsize)  :: telid_str, extname, col
    character(*), parameter :: function = "read_tHKx1"

    logical :: anyf

    if(status > 100 .or. status == s_errRep) return

    ! convert int to string
    telid_str = int2str(ccdnr, status)
    if(status == s_errRep) return

    success = false

    extname = hkx1_ext
    extname(3:3) = trim(telid_str)

    status = ignore
    call ffits_movnam_hdu(unit, BINARY_TBL, extname, status)
    if(status == ignore) then
      status = ok
    else
      status = ok
      return
    end if

    call ffits_get_num_rows(unit, nelements, status)
    call ffits_get_rowsize(unit, rowsize, status)

    call this%init(nelements, status)

    this%ccdnr = ccdnr
    this%nelements = nelements
    this%extname(3:3) = trim(telid_str)

    call ffits_get_colnum(unit, .false., time_col, time_colnum, status)

    col = pxprcmode_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxprcmode_colnum, status)
    col = pxprocopts_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxprocopts_colnum, status)
    col = pxtmframes_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxtmframes_colnum, status)
    col = pxccdfrms_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxccdfrms_colnum, status)
    col = pxevtctr_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxevtctr_colnum, status)
    col = pxmipctr_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxmipctr_colnum, status)
    col = pxfwstepct_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxfwstepct_colnum, status)
    col = pxfwstat_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxfwstat_colnum, status)
    col = pxchopper_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxchopper_colnum, status)

    anyf = .false.
    felem = 1

    if(nelements == 0 .or. status == s_errRep) then
      if(status == ok) success = true
      return
    end if

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1
      call ffits_read_col(unit, time_colnum, frow, felem, nrows, 0, this%time(frow:lrow), status, anyf)
      call ffits_read_col(unit, pxprcmode_colnum, frow, felem, nrows, 0, this%pxprcmode(frow:lrow), status, anyf)
      call ffits_read_col(unit, pxprocopts_colnum, frow, felem, nrows, 0, this%pxprocopts(frow:lrow), status, anyf)
      call ffits_read_col(unit, pxtmframes_colnum, frow, felem, nrows, 0, this%pxtmframes(frow:lrow), status, anyf)
      call ffits_read_col(unit, pxccdfrms_colnum, frow, felem, nrows, 0, this%pxccdfrms(frow:lrow), status, anyf)
      call ffits_read_col(unit, pxevtctr_colnum, frow, felem, nrows, 0, this%pxevtctr(frow:lrow), status, anyf)
      call ffits_read_col(unit, pxmipctr_colnum, frow, felem, nrows, 0, this%pxmipctr(frow:lrow), status, anyf)
      call ffits_read_col(unit, pxfwstepct_colnum, frow, felem, nrows, 0, this%pxfwstepct(frow:lrow), status, anyf)
      call ffits_read_col(unit, pxfwstat_colnum, frow, felem, nrows, 0, this%pxfwstat(frow:lrow), status, anyf)
      call ffits_read_col(unit, pxchopper_colnum, frow, felem, nrows, 0, this%pxchopper(frow:lrow), status, anyf)
    end do

    if(status == ok) success = true

  end subroutine read_tHKx1

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write in the columns of an hkx1 extension
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_tHKx1(this, unit, status)
    implicit none

    integer :: nelements, rowsize, frow, nrows, lrow, felem
    integer :: time_colnum
    integer, intent(inout) :: status
    integer, intent(in) :: unit

    character(buffsize)  :: telid_str, col

    class(tHKx1) :: this

    character(*), parameter :: function = "write_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    if(this%ccdnr == -1 .or. this%extname == "") return

    ! convert int to string
    telid_str = int2str(ccdnr, status)
    if(status == s_errRep) return

    ! move to the correct extension
    status = ignore
    call ffits_movnam_hdu(unit, BINARY_TBL, this%extname, status)
    if(status /= ignore) then
      status = ok
      call this%create_extension(unit, status)
    else
      status = ok
    end if

    call ffits_get_rowsize(unit, rowsize, status)

    nelements = this%nelements

    call ffits_get_colnum(unit, .false., time_col, time_colnum, status)

    col = pxprcmode_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxprcmode_colnum, status)
    col = pxprocopts_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxprocopts_colnum, status)
    col = pxtmframes_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxtmframes_colnum, status)
    col = pxccdfrms_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxccdfrms_colnum, status)
    col = pxevtctr_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxevtctr_colnum, status)
    col = pxmipctr_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxmipctr_colnum, status)
    col = pxfwstepct_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxfwstepct_colnum, status)
    col = pxfwstat_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxfwstat_colnum, status)
    col = pxchopper_col
    col(3:3) = trim(telid_str)
    call ffits_get_colnum(unit, .false., col, pxchopper_colnum, status)

    felem = 1

    if(nelements > 0) then
      do frow = 1, nelements, rowsize
        nrows = min(nelements - frow + 1, rowsize)
        lrow = frow + nrows - 1
        call ffits_write_col(unit, time_colnum, frow, felem, nrows, this%time(frow:lrow), status)
        call ffits_read_col(unit, pxprcmode_colnum, frow, felem, nrows, this%pxprcmode(frow:lrow), status)
        call ffits_read_col(unit, pxprocopts_colnum, frow, felem, nrows, this%pxprocopts(frow:lrow), status)
        call ffits_read_col(unit, pxtmframes_colnum, frow, felem, nrows, this%pxtmframes(frow:lrow), status)
        call ffits_read_col(unit, pxccdfrms_colnum, frow, felem, nrows, this%pxccdfrms(frow:lrow), status)
        call ffits_read_col(unit, pxevtctr_colnum, frow, felem, nrows, this%pxevtctr(frow:lrow), status)
        call ffits_read_col(unit, pxmipctr_colnum, frow, felem, nrows, this%pxmipctr(frow:lrow), status)
        call ffits_read_col(unit, pxfwstepct_colnum, frow, felem, nrows, this%pxfwstepct(frow:lrow), status)
        call ffits_read_col(unit, pxfwstat_colnum, frow, felem, nrows, this%pxfwstat(frow:lrow), status)
        call ffits_read_col(unit, pxchopper_colnum, frow, felem, nrows, this%pxchopper(frow:lrow), status)

      end do
    end if

  end subroutine write_tHKx1

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create a basic hkx1 extension in the file
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine create_extension_tHKx1(this, unit, status)
    implicit none

    integer, intent(inout) :: status
    integer, intent(in) :: unit

    character(*), parameter :: function = "create_extension_tHKx1"
    character(fcolnamelength), dimension(1) :: ttype
    character(fcolnamelength), dimension(1) :: tform
    character(fcolnamelength), dimension(1) :: tunit

    class(tHKx1) :: this
    type(tBinTable) :: table

    if(status > 100 .or. status == s_errRep) return

    ttype(1) = time_col

    tform = ["D"]
    tunit(:) = ""

    call table%init(ttype, tform, tunit, status)
    call table%insertFits(unit, this%extname, status);

  end subroutine create_extension_tHKx1

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> reduce all entries in the columns to contain only a subset of the complete
  !> by providing an array list with the desired indices
  !
  !
  !> @param[in] this
  !> @param[in] order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_index_cols_tHKx1(this, order, status)
    implicit none

    integer, intent(inout) :: status

    class(tHKx1) :: this

    integer, dimension(:), intent(in) :: order
    integer :: n

    character(*), parameter :: function = "apply_index_cols_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    n = size(order)
    if(n < 2) then
      !            call wrerr(function, warn, trim(this%extname) // " would be reduced to below two entries - but at least two entries in " // trim(this%extname) // &
      !    " are required for further processing - no filter will be applied")
      call wrchatty(function, trim(this%extname) // " would be reduced to below two entries - but at least two entries in " // trim(this%extname) // &
        " are required for further processing - no filter will be applied")
      !      call wrerr(function, warn, trim(this%extname) // " is reduced to zero entries")
      !      call alloc(this%time, n, status)
      !      call alloc(this%deadc, n, status)
      !      this%nelements = n
      return
    end if

    this%time = this%time(order)
    this%nelements = n

  end subroutine apply_index_cols_tHKx1

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !
  !
  !> @param[in] this, GTI
  !> @param[in, optional] extend_order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_gti_tHKx1(this, GTI, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    class(tHKx1) :: this
    type(tGTI), intent(in) :: GTI
    type(tGTI) :: iGTI

    integer, dimension(this%nelements) :: mask
    integer, dimension(:), allocatable :: order

    character(*), parameter :: function = "apply_gti_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]

    iGTI = GTI
    ! internally correct the GTI
    if(iGTI%use_mjdref == true) then

      if(this%mjdref /= iGTI%mjdref) then

        iGTI%start = (iGTI%mjdref - this%mjdref ) * 3600.0_r8 * 24.0_r8 + iGTI%start
        iGTI%stop = ( iGTI%mjdref - this%mjdref ) * 3600.0_r8 * 24.0_r8 + iGTI%stop
        iGTI%mjdref = this%mjdref
      end if

    end if

    mask = get_gti_mask(this%time, iGTI, status)

    order = pack(order, mask == true)

    call this%apply_index_cols(order, status)
    call dealloc(order, status)
    call iGTI%free(status)


  end subroutine apply_gti_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete duplicate entries in the tHKx1 type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_duplicates_tHKx1(this, status)
    implicit none

    class(tHKx1) :: this

    integer, intent(inout) :: status
    integer, dimension(:), allocatable :: order
    integer, dimension(this%nelements) :: mask
    integer :: ii

    character(*), parameter :: function = "delete_duplicates_tHKx1"

    if(status > 100 .or. status == s_errRep) return

    call this%sort(this%time, status)

    if(this%nelements < 2) return

    call alloc(order, this%nelements, status)
    order = [(ii,ii=1,this%nelements,1)]

    mask = true
    do ii = 2, this%nelements
      if(abs(this%time(ii) - this%time(ii-1)) < tiny(this%time(1))) then
        mask(ii) = false
      end if
    end do

    order = pack(order, mask == true)
    call this%apply_index_cols(order, status)
    call dealloc(order, status)

  end subroutine delete_duplicates_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the content of tHKx1 type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_tHKx1(this, status)
    implicit none

    integer, intent(inout) :: status

    class(tHKx1) :: this

    character(*), parameter :: function = "free_tHKx1"

    call dealloc(this%time, status)
    call dealloc(this%pxprcmode, status)
    call dealloc(this%pxprocopts, status)
    call dealloc(this%pxtmframes, status)
    call dealloc(this%pxccdfrms, status)
    call dealloc(this%pxevtctr, status)
    call dealloc(this%pxmipctr, status)
    call dealloc(this%pxfwstepct, status)
    call dealloc(this%pxfwstat, status)
    call dealloc(this%pxchopper, status)

  end subroutine free_tHKx1

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> alloc the content of tHKx1 type
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine init_tHKx1(this, n, status, ccdnr)
    implicit none

    integer, intent(in) :: n
    integer, intent(inout) :: status
    integer, intent(in), optional :: ccdnr

    class(tHKx1) :: this

    character(*), parameter :: function = "init_tHKx1"

    call alloc(this%time, n, status)

    this%nelements = n
    this%ccdnr = -1
    if(present(ccdnr)) this%ccdnr = ccdnr
    this%extname = ""
    this%mjdref = 0_r8

    call alloc(this%time, n, status)
    call alloc(this%pxprcmode, n, status)
    call alloc(this%pxprocopts, n, status)
    call alloc(this%pxtmframes, n, status)
    call alloc(this%pxccdfrms, n, status)
    call alloc(this%pxevtctr, n, status)
    call alloc(this%pxmipctr, n, status)
    call alloc(this%pxfwstepct, n, status)
    call alloc(this%pxfwstat, n, status)
    call alloc(this%pxchopper, n, status)

  end subroutine init_tHKx1

end module evtool_hkx1_mod
